/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrWrapLib
// @module  Folder Wrapper | This module contains the implementation of
//          the CECOrbFolder wrapper class.  This class simply wraps around the
//          CORBA interface CyrStFolder and hides the CORBA code.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrWrapLib/CECOrbFolder.h>
#include <CyrWrapLib/CECOrbFile.h>
#include <Factory/ParamList_i.h>


////////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor is for use by derived Folder types only.  It simply
//          prepares the underlying object for linking and does not retrieve any
//          values from the OpenSTARep database.  It is intended for items such as
//          TestRuns where the name is not known beforehand and specialised
//          construction is required.
// @parm    Pointer to the parent folder object.
CECOrbFolder::CECOrbFolder( CECOrbFolder* parent )
        : CECOrbLink( parent->GetHelper(), parent->GetFactory() ),m_cacheName("")
{
    m_project = NULL;
    m_folder = parent;
    m_hidden = parent->GetHiddenServer();
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor retrieves a root Folder for the specified Project.
//          The root Folder can then be used to walk the hierarchy to other
//          Folders, Files or OpenSTARep objects that are stored in Folders.
// @parm    The Project that this root Folder belongs to.
CECOrbFolder::CECOrbFolder( CECOrbProject* project )
        : CECOrbLink( project->GetHelper(), project->GetFactory() ),m_cacheName("")
{
    // Error information
    const char* where = "CECOrbFolder::CECOrbFolder";

    try
    {
        // Retrieve underlying CyrStServer reference from wrapper class
        m_hidden = project->GetHiddenServer();
        m_project = project;
        m_folder = NULL;

        // Create a new parameter list for Folder.  Remember that all returned
        // strings must be released so we copy them int CORBA::String_var
        // variables before passing them as parameters.  This will cause them
        // to be released as soon as the varaible goes out of scope.
        FactoryParamList* params =
            new FactoryParamList( CyrStFolder::CLASS_NAME, project->GetName().c_str(),
                                  DEFAULT_PERIOD );
        PortableServer::ObjectId* id = m_corba->ActivateObject( params );
        string paramProject = project->GetRegisteredName();
        params->AddValue( CyrStFolder::PARAM_PROJECT, paramProject.c_str() );

        // Create CORBA object in Factory
        CORBA::Object_ptr obj = m_factory->New( params->_this() );
        m_servant = CyrStFolder::_narrow( obj );

        // Register name so that object can be borrowed
        CORBA::String_var regName = m_servant->GetRegisteredName();
        Register( regName, CyrStFolder::CLASS_NAME );

        // Destroy parameter object
        m_corba->DeActivateObject(*id);
        params->_remove_ref();
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor retrieves a specified Folder from the OpenSTARep
//          database, based on its name and its parent folder.
// @parm    Pointer to the parent Folder.
// @parm    Name of the Folder to retrieve.
CECOrbFolder::CECOrbFolder( CECOrbFolder* parent, const char* name )
        : CECOrbLink( parent->GetHelper(), parent->GetFactory() ),m_cacheName("")
{
    // Error information
    const char* where = "CECOrbFolder::CECOrbFolder";

    try
    {
        // Retreive underlying CyrStFolder and CyrStServer references
        // from parent Folder wrapper class
        CyrStFolder_var folder = parent->GetServant();
        m_hidden = parent->GetHiddenServer();
        m_project = NULL;
        m_folder = parent;

        // Create a new parameter list for Folder
        FactoryParamList* params =
            new FactoryParamList( CyrStFolder::CLASS_NAME,
                                  name, DEFAULT_PERIOD );
        PortableServer::ObjectId* id = m_corba->ActivateObject( params );
        CORBA::String_var paramParent = folder->GetRegisteredName();
        params->AddValue( CyrStFolder::PARAM_PARENT, paramParent );

        // Create CORBA object in Factory
        CORBA::Object_ptr obj = m_factory->New( params->_this() );
        m_servant = CyrStFolder::_narrow( obj );

        // Register name so that object can be borrowed
        CORBA::String_var regName = m_servant->GetRegisteredName();
        Register( regName, CyrStFolder::CLASS_NAME );

        // Destroy parameter object
        m_corba->DeActivateObject(*id);
        params->_remove_ref();
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a new Folder in the OpenSTARep database,
//          based on its name, description and its parent folder.
// @parm    Pointer to the parent Folder.
// @parm    Name of the Folder to retrieve.
// @parm    Description of Folder.
CECOrbFolder::CECOrbFolder( CECOrbFolder* parent, const char* name, const char* desc )
        : CECOrbLink( parent->GetHelper(), parent->GetFactory() ),m_cacheName("")
{
    // Error information
    const char* where = "CECOrbFolder::CECOrbFolder";

    try
    {
        // Retrieve underlying CyrStFolder and CyrStServer references
        // from parent Folder wrapper class
        CyrStFolder_var folder = parent->GetServant();
        m_hidden = parent->GetHiddenServer();
        m_project = NULL;
        m_folder = parent;

        // Create a new parameter list for Folder
        FactoryParamList* params =
            new FactoryParamList( CyrStFolder::CLASS_NAME, name, DEFAULT_PERIOD );
        PortableServer::ObjectId* id = m_corba->ActivateObject( params );
        CORBA::String_var paramParent = folder->GetRegisteredName();
        params->AddValue( CyrStFolder::PARAM_PARENT, paramParent );
        params->AddValue( CyrStFolder::PARAM_DESC, desc );

        // Create CORBA object in Factory
        CORBA::Object_ptr obj = m_factory->New( params->_this() );
        m_servant = CyrStFolder::_narrow( obj );

        // Register name so that object can be borrowed
        CORBA::String_var regName = m_servant->GetRegisteredName();
        Register( regName, CyrStFolder::CLASS_NAME );

        // Destroy parameter object
        m_corba->DeActivateObject(*id);
        params->_remove_ref();
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the OpenSTARep Folder.  This is the
//          same name that was passed into the constructors, but is repeated
//          here for completeness.
// @rdesc   The name string.
string CECOrbFolder::GetName()
{
    // Call underlying CORBA object's method, but store result in
    // a CORBA string so that memory is handled correctly.
    if(m_cacheName.empty())
    {
        CORBA::String_var s = m_servant->GetName();
        m_cacheName=s;
    }
    // Convert the CORBA string into something else so that the CORBA
    // mechanism remains hidden.
    //string ret( s );
    return m_cacheName;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a copy of the underlying CORBA servant object
//          to the calling code.  This will increment the reference count and
//          so will need to be released.
// @rdesc   The reference to the servant object.
CyrStFolder_ptr CECOrbFolder::GetServant()
{
    return CyrStFolder::_duplicate( m_servant );
}

CECOrbServer* CECOrbFolder::GetParentServer()
{
    return GetParentProject()->GetParentServer();
}

CECOrbProject* CECOrbFolder::GetParentProject()
{
    if ( m_project != NULL )
        return m_project;
    else
        return GetParentFolder()->GetParentProject();
}

CECOrbFolder* CECOrbFolder::GetParentFolder()
{
    return m_folder;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves the list of child Folders known about by this
//          Folder.  The list is returned as a string that contains a list
//          of names separated by a '\|' character and terminated by two '\|'
//          characters.
// @rdesc   The separated list of Folder names.
// @parm    Will contain the number of Folders found.
string CECOrbFolder::GetFolders( long& count ) const
{
    // Call underlying CORBA object's method, but store result in
    // a CORBA string so that memory is handled correctly.
    CORBA::String_var s = m_servant->GetFolders( count );

    // Convert the CORBA string into something else so that the CORBA
    // mechanism remains hidden.
    string ret( s );
    return ret;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method retrieves the list of Files known about by this
//          Folder.  The list is returned as a string that contains a list
//          of names separated by a '\|' character and terminated by two '\|'
//          characters.
// @rdesc   The separated list of File names.
// @parm    Will contain the number of Files found.
string CECOrbFolder::GetFiles( long& count ) const
{
    // Call underlying CORBA object's method, but store result in
    // a CORBA string so that memory is handled correctly.
    CORBA::String_var s = m_servant->GetFiles( count );

    // Convert the CORBA string into something else so that the CORBA
    // mechanism remains hidden.
    string ret( s );
    return ret;
}

void CECOrbFolder::Refresh()
{
    m_servant->Refresh();
}
