/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrWrapLib
// @module  Project Wrapper | This module contains the implementation of
//          the CECOrbProject wrapper class.  This class simply wraps around the
//          CORBA interface CyrStProject and hides the CORBA code.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include <CyrWrapLib/CECOrbProject.h>
#include <CyrWrapLib/CECOrbFolder.h>
#include <Factory/ParamList_i.h>
///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor retrieves an existing Project from the OpenSTARep
//          Server.  All that is required to do this is the registered name of
//          the parent Server and the name of the Project.
// @parm    Pointer to an instance of the parent Server wrapper class.
// @parm    The name of the Project to retrieve.
CECOrbProject::CECOrbProject( CECOrbServer* server, const char* name )
        : CECOrbBorrow( server->GetHelper(), server->GetFactory() ),
        m_cacheName(std::string(name).c_str())
{
    // Error information
    const char* where = "CECOrbProject::CECOrbProject";

    try
    {
        // Retrieve underlying CyrStServer reference from wrapper class
        m_hidden = server->GetServant();
        m_parent = server;

        // Create a new parameter list for Project
        FactoryParamList* params =
            new FactoryParamList( CyrStProject::CLASS_NAME,
                                  name, DEFAULT_PERIOD );
        PortableServer::ObjectId* id = m_corba->ActivateObject( params );
        CORBA::String_var paramServer = m_hidden->GetRegisteredName();
        params->AddValue( CyrStProject::PARAM_SERVER, paramServer );

        // Create CORBA object in Factory
        CORBA::Object_ptr obj = m_factory->New( params->_this() );
        m_servant = CyrStProject::_narrow( obj );

        // Register name so that object can be borrowed
        CORBA::String_var regName = m_servant->GetRegisteredName();
        Register( regName, CyrStProject::CLASS_NAME );

        // Destroy parameter object
        m_corba->DeActivateObject(*id);
        params->_remove_ref();
    }
    catch( CyrCORBAException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
    }
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the OpenSTARep Project.  This is the
//          same name that was passed into the constructors, but is repeated
//          here for completeness.
// @rdesc   The name string.
string CECOrbProject::GetName()const
{
    // Call underlying CORBA object's method, but store result in
    // a CORBA string so that memory is handled correctly.
    //CORBA::String_var s = m_servant->GetName();

    // Convert the CORBA string into something else so that the CORBA
    // mechanism remains hidden.
    //string ret( s );
    //return ret;
    // return cached name from constructor rather than use method calls
    return m_cacheName;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a copy of the underlying CORBA servant object
//          to the calling code.  This will increment the reference count and
//          so will need to be released.
// @rdesc   The reference to the servant object.
CyrStProject_ptr CECOrbProject::GetServant()
{
    return CyrStProject::_duplicate( m_servant );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a copy of the OpenSTARep CORBA Server interface.
//          This will increment the reference count and so will need to be
//          released.
// @devnote I wish there were a neater way to do this, but the only way to get
//          user details is from the Server interface.  This means I have to
//          pass and copy this reference around.  Bleurgh!!  I thought about
//          adding this functionality into the base class.  I didn't because it
//          firstly didn't logically belong there, and secondly the CECServer
//          class that instantiates the reference does so in its constructor -
//          after the CECBorrow object has been created.  It would have been an
//          even messier solution.
// @rdesc   Reference to the Server object.
CyrStServer_ptr CECOrbProject::GetHiddenServer()
{
    return CyrStServer::_duplicate( m_hidden );
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns a pointer to the CECServer wrapper that created
//          this Project object.
// @rdesc   The pointer to the CECServer object.
CECOrbServer* CECOrbProject::GetParentServer()
{
    return m_parent;
}

