/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CECFFApi
// @module  File Implementation | This module contains the implementation of
//          the CECFFFile class.  This class is simply a C++ wrapper around the
//          OpenSTARep COM interface IStFile.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include "CECFFFile.h"
#include <Repository/CyrStHelper.inl>
///////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor attempts to retrieve the requested File from its
//          parent Folder.  It does this by looping through all the known
//          Files in the Folder and trying to match the name.
// @parm    A pointer to the parent Folder object.
// @parm    The name of the File to search for.
CECFFFile::CECFFFile( CECFFFolder* pParent, const char* name )
{
    // Error information
    const char* where = "CECFFFile::CECFFFile";
    long status = CYR_FILE_NOT_FOUND;

    // Attempt to retrieve file
    if ( RetrieveFile( pParent, name ) == false )
        throw CyrRepositoryException( "Cannot find file", where, status );
    m_name=name;

}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method attempts to locate the named file within the OpenSTARep
//          database.  It does this by retrieving all file objects within the
//          specified folder and seeing if any matches the specified name.
// @rdesc   True if the file is found, else false.
// @parm    Pointer to the parent folder of the file.
// @parm    Name of the file to search for.
bool CECFFFile::RetrieveFile( CECFFFolder* pParent, const char* name )
{
    // Assign member variables
    m_parent = pParent;
    m_name=name;
    return true;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method attempts to create a new File reference in a specified
//          Folder.  This method does NOT copy the File, but simply creates a
//          reference in OpenSTARep.
// @parm    A pointer to the parent Folder object.
// @parm    The requested name of the new File.
// @parm    The description of the new File.
CECFFFile::CECFFFile( CECFFFolder* pParent, const char* name, const char* desc )
{
    // Error information
    const char* where = "CECFFFile::CECFFFile";
    long status = CYR_FILE_ADD_FAILED;
    // See if file already exists
    RetrieveFile( pParent, name );
    m_name=name;
    m_desc=desc;

}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   The destructor destroys the underlying COM object as part of its
//          tidying up.
CECFFFile::~CECFFFile()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the name of the File as recognised by
//          OpenSTARep .  It should be the same name as the one passed into the
//          constructor and therefore should not be needed.  However, it has
//          been added here for completeness.
// @rdesc   The name of the File.

string CECFFFile::GetName() const
{
    return m_name;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the current description for this File.
// @rdesc   The current File description.
string CECFFFile::GetDescription() const
{
    return m_desc;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the full file path for this File.
// @rdesc   The full file path.
string CECFFFile::GetFullPath() const
{
    // Get base directory
    string fullpath = GetRootDirectory().c_str();

    fullpath += m_parent->GetFolderHierarchy();

    // Add filename
    fullpath += GetName();
    return fullpath;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method sets a new description for this File.
// @parm    The new description.
void CECFFFile::SetDescription( const char* desc )
{
    m_desc=desc;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method returns the current comment for this File.
// @rdesc   The current File comment.
string CECFFFile::GetComment() const
{
    return "";
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method sets a new comment for this File.
// @parm    The new comment.
void CECFFFile::SetComment( const char* comment )
{}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method locks the remote File so that no other user can modify
//          it.  This method should be called before updating any changes to
//          the File.
// @rdesc   Returns true if the lock request was successful, otherwise false.

bool CECFFFile::Lock()
{
    return true;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method removes a lock placed previously by this user.
void CECFFFile::Unlock()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Unless this method is called, all changes to the File object will
//          be discarded.  This method persists the changes into the OpenSTARep
//          database.  This method must be called before CheckIn otherwise
//          edited details will be lost.

void CECFFFile::Update()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method removes the File entry from the OpenSTARep database.
//          Note that this action is irreversible and any checking or
//          confirmations should be done by the client before this method is
//          invoked.

void CECFFFile::Remove()
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method transfers the entire contents of the File object to
//          the local machine.
// @rdesc   Returns true if the file was successfully checked out and locked by
//          the current user.  If the file is not locked by the current user
//          then false is returned.
// @parm    Determines if a file lock is requested.

bool CECFFFile::CheckOut( bool lock )
{
    return true;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method transfers the entire contents of a file to the OpenSTARep
//          database.  Using this method will overwrite the contents of any
//          File that is currently in place - therefore the Lock method should
//          be used prior to calling this method.
// @rdesc   Returns true if the file was successfully checked in.
// @parm    Determines if file should be unlocked by this program.
bool CECFFFile::CheckIn( bool release )
{
    return true;
}

CECFFServer* CECFFFile::GetParentServer()
{
    return GetParentProject()->GetParentServer();
}

CECFFProject* CECFFFile::GetParentProject()
{
    return GetParentFolder()->GetParentProject();
}

CECFFFolder* CECFFFile::GetParentFolder()
{
    return m_parent;
}
