/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

///////////////////////////////////////////////////////////////////////////////
// @doc     CyrFFApi
// @module  SCL File Wrapper | This module contains the implementation of
//          the CECFFNodeFile wrapper class.  This class simply extends the
//          CECFFFile wrapper class.
//
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
// Cyrano includes
#include "CECFFNodeFile.h"
#include "CECFFImpactFolders.h"
#include "CECFFImpactProject.h"

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor retrieves a specified source file from the
//          OpenSTARep repository.
// @parm    Pointer to the parent Folder object.
// @parm    The name of the File to retrieve.
CECFFNodeFile::CECFFNodeFile( CECFFFolder* parent, const char* name )
        : CECFFFile( parent, name )
{}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This constructor creates a new File object in the specified Folder.
// @parm    Pointer to the parent Folder.
// @parm    The name of the File to create.
// @parm    The description of the File to create.

CECFFNodeFile::CECFFNodeFile( CECFFFolder* parent, const char* name, const char* desc )
        : CECFFFile( parent, name, desc )
{}


///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method reads the node definitions from a remote file.

void CECFFNodeFile::GetNode (string& PhysicalNode
                             ,string& Username
                             ,string& Password)
{
    //
    //    Get root folder object and from there get the nodes folder
    //
    CECFFFolder* rootFolder = new CECFFFolder( GetParentProject() );
    CECFFNodeFolder NodeFolder (rootFolder);

    //
    //    Recover the name of the node definition file.
    //
    string Nodefile = GetName();
    CECFFFile* Definition = NULL;

    try
    {
        // Get latest node definition file from the node definitions folder
        Definition = new CECFFFile( &NodeFolder, Nodefile.c_str() );
        Definition->CheckOut( false );
        char ReadBuff[256];

        // Open file and read bytes of file as appropriate
        std::ifstream* f = new std::ifstream( Definition->GetFullPath().c_str());
        f->read(ReadBuff,sizeof(ReadBuff));
        PhysicalNode = ReadBuff;
        /*no longer used ???
        f->read(ReadBuff);
        Username = ReadBuff;
        f->read(ReadBuff);
        Password = ReadBuff;
        */
        delete f;
    }
    catch( ... )
    {}

    // Tidy memory

    if ( Definition != NULL )
        delete Definition;
    if ( rootFolder != NULL )
        delete rootFolder;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   This method writes the Node definition settings to a file and then checks
//          that file in.
//
void CECFFNodeFile::SetNode     (string& PhysicalNode
                                 ,string& Username
                                 ,string& Password)
{
    // Error information
    const char* where = "CECFFImpactProject::SetNode";

    //
    //    Get root folder object and from there get the nodes folder
    //
    CECFFFolder* rootFolder = new CECFFFolder( GetParentProject() );
    CECFFNodeFolder NodeFolder (rootFolder);

    //
    //    Recover the name of the node definition file.
    //
    string Nodefile = GetName();
    CECFFFile* Definition = NULL;

    try
    {

        // Open file and write contents to it
        std::ofstream* f = new std::ofstream( Nodefile.c_str());
        f->write( PhysicalNode.c_str(),PhysicalNode.length());
        delete f;

        try
        {
            // Try to update existing entry
            Definition = new CECFFFile( &NodeFolder, Nodefile.c_str() );
            Definition->Update();
        }
        catch( CyrRepositoryException ex )
        {
            if ( ex.status() != CYR_FILE_NOT_FOUND )
                throw CyrRepositoryException( ex.what().c_str(), where, ex.status() );
            else
            {
                // Entry does not exist so Create it
                CECFFFile* Definition = new CECFFFile( &NodeFolder, Nodefile.c_str(), Nodefile.c_str() );
                delete Definition;
            }
        }
    }
    catch( CyrRepositoryException ex )
    {
        throw CyrRepositoryException( ex.what().c_str(), ex.where().c_str(), ex.status() );
    }

    // Tidy memory
    delete rootFolder;
}


