/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: LIST_ROUTINES               */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains the routines required for implementing      */
/*  the standard listing functions.                      */
/*                                       */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard ANSI header files.
 */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL routines */
#include "list_intrtndef.h"     /* Internal listing routines */

/*
 *  Structure Definitions.
 */
#include <shrmsg.h>         /* PSL messages */
#include <psldate_prmdef.h>
/* Date/Time Routine Symbols */
#include <pslerr_prmdef.h>
#include <psldiag_prmdef.h>
#include <list_prmdef.h>

/*
 *  Global data required for communication between the listing routines.
 */
#include <list_ctxdef.h>

/*
 *  Definitions for the location of fields on the main heading line
 */
#define HEADING_DATE_END 70
#define HEADING_PAGE_END 80


/*
 */
/************************************************************************/
/*                                  */
/*    LIST_ALLOCATE - Allocate Listing Context Structure        */
/*    =============                         */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_ALLOCATE (diag_ptr, &context_ptr);          */
/*                                  */
/*    Inputs:                               */
/*                                  */
/*  diag_ptr        = Pointer to an allocated diagnostics stack */
/*            Must be pointer to Structure DIAGSTK.     */
/*                                  */
/*    Outputs:                              */
/*  context_ptr = Address of a variable to hold a pointer to    */
/*            a context block               */
/*                                  */
/*    Returns:                              */
/*  Unsigned long PSL routine status                */
/*                                  */
/*    This routine is responsible for allocating the memory structure   */
/*    required to hold the listing context block which is used to   */
/*    enable a utility to have multiple listing files open concurrently */
/*                                  */
/************************************************************************/

unsigned long   LIST_ALLOCATE (struct diagstk *diag_ptr,
                               struct list_ctx **context_ptr)
{
    /*
     *    Local data
     */
    unsigned long iret;         /* Status variable      */
    struct   list_ctx *temp_ptr;        /* Temporary pointer        */

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status     */

    /*
     *  Allocate sufficient memory to hold the context block and return the
     *  base address of this structure to the user.
     */
    temp_ptr = (struct list_ctx *) calloc (sizeof (struct list_ctx), 1);
    /* Allocate a context block */
    if (temp_ptr == NULL)       /* ...if allocation failed  */
    {

        /*
         *  Memory allocation failed
         *  ------------------------
         *
         *  Report that the listing has been aborted due to lack of memory
         */
        iret = PSL_LISTABORT;

        PSL_PUTDIAG (diag_ptr, 2, iret, PSL_INSVIRMEM);
        PSL_SIGNAL_ERROR (diag_ptr, PSLERRCLS_LIST, iret);
    }
    else
    {

        /*
         *  Memory allocated O.K. 
         *  ---------------------
         *
         *  Setup the initial listing structure.
         */
        (*temp_ptr).diag_ptrps = diag_ptr;

        /*
         *  Setup the context block validity checking fields.
         */
        (void) memcpy ((*temp_ptr).ftypc4, "LIST", 4);

        (*temp_ptr).fvrnl      = LIST_FVRN;
        (*temp_ptr).last_errl  = ERR_NO;
        (*temp_ptr).lstvalc256[0] = '\0';

        (*context_ptr) = temp_ptr;

        /*
         *  Create a Cardfile entry
         */
        PSL_CARDFILE_ADD (temp_ptr, "LIST--> ");
    }

    /*
     *  Exit with the return status.
     */
    return iret;
}


/************************************************************************/
/*                                  */
/*    LIST_DEALLOCATE - Deallocate Listing Context Structure        */
/*    ===============                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  LIST_DEALLOCATE (context_ptr);                  */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Address of a variable to hold a pointer to    */
/*            a context block               */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  None                                            */
/*                                  */
/*    This routine is responsible for deallocating the memory structure */
/*    required to hold the listing context block which was previously   */
/*    allocated via LIST_ALLOCATE.                  */
/*                                  */
/************************************************************************/

void    LIST_DEALLOCATE (struct list_ctx *context_ptr)
{
    /*
     *    Local data
     */
    unsigned long iret;         /* General status return */

    /*
     *  Check the validity of the context block.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (psl_test_status (iret))
    {

        /*
         *  Invalidate the structure so that the user cannot re-use it.
         */
        (void) memset ((*context_ptr).ftypc4,
                       '\0', sizeof ((*context_ptr).ftypc4));
        (*context_ptr).fvrnl = 0;

        /*
         *  De-allocate the structure.
         */
        free (context_ptr);
    }

    return;
}

/************************************************************************/
/*                                  */
/*    LIST_SETUP - Set up listing related qualifiers, manually      */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_SETUP (list_value_ptr, spool_qual, context_ptr) */
/*                                  */
/*    Inputs:                               */
/*                                  */
/*  list_value_ptr  = Pointer to a string containing the        */
/*                primary file specification.           */
/*                                  */
/*  spool_qual      = Boolean spool required?           */
/*              = PFALSE - Disable spooling         */
/*              = PTRUE  - Enable spooling          */
/*                                          */
/*  context_ptr = Pointer to listing context block              */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success) or PSL_LISTABORT) */
/*                                  */
/*    This routine is responsible for setting up the primary file   */
/*    specification and Spool flag, in situations where this        */
/*    information cannot be taken from the command line via the     */
/*    LIST_PARSE routine.                       */
/*                                  */
/************************************************************************/

unsigned long   LIST_SETUP (char *list_value_ptr,
                            Boolean spool_qual,
                            struct list_ctx *context_ptr)

{
    /*
     *    Local data
     */
    unsigned long string_len;       /* Length of string     */
    unsigned long iret;         /* General status return    */

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status     */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *  See if the (alternative) /LIST values were specified and, store them.
     */
    if (list_value_ptr != NULL)
    {

        /*
         *  Save the listing file value (up to 256 characters).
         */
        string_len = strlen (list_value_ptr);   /* Get length of /LIST value */
        if (string_len >= sizeof ((*context_ptr).lstvalc256))
        {
            string_len = sizeof ((*context_ptr).lstvalc256) - 1;
        }

        (void) memcpy ((*context_ptr).lstvalc256, list_value_ptr, string_len);
        (*context_ptr).lstvalc256[string_len] = '\0';   /* Null Terminate */
    }

    /*
     *  Store the (alternative) /SPOOL value.
     */
    (*context_ptr).spool_qualn = spool_qual;

    /*
     *  Exit with the defined status.
     */
    return iret;
}


/************************************************************************/
/*                                  */
/*    LIST_INIT - Intialize data structures and open Listing        */
/*    =========                             */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_INIT (context_ptr, utility_ptr, def_file_ptr,   */
/*            rel_file_ptr, subhdg1_ptr,            */
/*            subhdg2_ptr, version_ptr)         */
/*                                  */
/*    Inputs:                               */
/*  context_ptr     = Pointer to listing context block      */
/*  utility_ptr     = Pointer to null terminated utility name   */
/*  def_file_ptr    = Pointer to default file specification     */
/*  rel_file_ptr    = Pointer to related file specification     */
/*  subhdg1_ptr = Pointer to first sub-heading line     */
/*  subhdg2_ptr = Pointer to second sub-heading line        */
/*  version_ptr = Pointer to the version string.        */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success) or PSL_LISTABORT) */
/*                                  */
/*  This routine is responsible for initialising the data strutures */
/*  and opening the Listing file.                   */
/*                                  */
/************************************************************************/

unsigned long LIST_INIT (      struct list_ctx *context_ptr,
                               const        char     *utility_ptr,
                               const        char     *def_file_ptr,
                               const        char     *rel_file_ptr,
                               const        char     *subhdg1_ptr,
                               const        char     *subhdg2_ptr,
                               const        char     *version_ptr)

{
    /*
     *  Local Data.
     */
    char     date_time[26];    /* Current date/time        */
    char     *line_ptr;        /* Pointer to heading line  */
    size_t   date_time_len;    /* "Length" of time string */
    unsigned long     iret;         /* General status return    */
    unsigned long     string_len;       /* Length of character string   */
    unsigned long     psl_date;     /* PSL Date */
    unsigned long     psl_time;     /* PSL Date */

#ifdef UNIX
    unsigned long     loop;         /* Loop Control */
#endif

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status     */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *  Calculate the Maximum lines per page as follows:
     *
     *      Max. lines per page = page size - 6;
     *
     *  where '6' is to allow a margin of 3 lines, top and bottom.
     *
     */
    (*context_ptr).max_pag_sizl = (LIST__GET_PAGESIZE () - 6);

    /*
     *  Initialise the current page, the lines on the current page and the
     *  line reference number (for numbered lines).
     */
    (*context_ptr).page_nol = 0;    /* No current page - yet    */
    (*context_ptr).ln_on_pagl = (*context_ptr).max_pag_sizl + 1;
    /* initiate a page throw    */
    (*context_ptr).ln_refnol = 0;   /* No numbered lines - yet  */

    /*
     *  Set LIST_OUTPUT_ERROR_BUFFER: specific flag to "enable" trapping of errors
     *  (for output in the listing).
     */
    (*context_ptr).err_trappingl = ERT_ENABLED;
    /* Error trapping is ENABLED    */

    /*
     *    Ensure that the main heading line is space filled initially and null
     *    terminated at its end
     */
    (void) memset ((*context_ptr).headingc84, ' ',
                   sizeof((*context_ptr).headingc84));
    /* SPACE fill the time string   */
    (*context_ptr).headingc84 [81] = 0; /* ...and NULL terminate it.    */

    /*
     *  Get the current date and time and format them into a text buffer
     */
    PSL_GET_DATTIM (&psl_date, &psl_time);
    /* Get Date and Time */
    iret = PSL_FMT_DATTIM (psl_date,    /* Format Date */
                           psl_time,    /* ... and Time */
                           date_time,   /* Into this buffer */
                           sizeof (date_time),
                           /* ... which is this big */
                           PSDTYP_SYSTEM,
                           /* System Date Type */
                           (PSDFMT_DATE | PSDFMT_TIME | PSDFMT_CENTURY));
    /* Flags            */

    if (!(psl_test_status (iret)))  /* Exit if Error */
    {
        PSL_PUTDIAG ((*context_ptr).diag_ptrps , 1, iret);
        PSL_SIGNAL_ERROR ((*context_ptr).diag_ptrps , PSLERRCLS_LIST, iret);
        iret = PSL_LISTABORT;
        goto end;           /* Return primary status    */
    }
    date_time_len = strlen (date_time); /* ...and its size      */

    /*
     *  Copy the Utility and Utility version number strings into the common
     *  data structures.
     */
    (*context_ptr).utlnam_lenl = strlen (utility_ptr);
    /* Length of utility name   */
    if ((*context_ptr).utlnam_lenl > 6) (*context_ptr).utlnam_lenl = 6;
    /* restrict length to 6 chars   */
    (void) memcpy (         /* ...and copy the util name    */
        (*context_ptr).utlnamc8,
        utility_ptr,
        (*context_ptr).utlnam_lenl);
    /*
     *  On UNIX, convert the name to lowercase.
     */
#ifdef UNIX
    for (loop = 0; loop < (*context_ptr).utlnam_lenl; loop++)
    {
        (*context_ptr).utlnamc8[loop] = tolower ((*context_ptr).utlnamc8[loop]);
    }
#endif

    /*
     *    Now do the same for the version
     */
    (*context_ptr).lstver_lenl = strlen (version_ptr);
    /* Length of version        */
    if ((*context_ptr).lstver_lenl > 4) (*context_ptr).lstver_lenl = 4;
    /* restrict length to 4 chars   */
    (void) memcpy (         /* ...and copy the version  */
        (*context_ptr).lstverc8,
        version_ptr,
        (*context_ptr).lstver_lenl);
    /*
     *  On UNIX, convert the version string to lowercase.
     */
#ifdef UNIX
    for (loop = 0; loop < (*context_ptr).lstver_lenl; loop++)
    {
        (*context_ptr).lstverc8[loop] = tolower ((*context_ptr).lstverc8[loop]);
    }
#endif

    /*
     *    Format the main heading line so that we have the utility and version
     *    at the start of the line with the date/time string ending at column
     *    70 followed by a space then followed by 'Page'. The page number itself
     *    will be added at a later date.
     */
    line_ptr = (*context_ptr).headingc84;
    /* point to start of heading    */
    (void) memcpy (line_ptr,
                   (*context_ptr).utlnamc8, (*context_ptr).utlnam_lenl);
    /* copy utility name        */
    line_ptr = line_ptr + (*context_ptr).utlnam_lenl + 1;
    /* Point 1 chars past utility   */
    *line_ptr = '-';            /* Put a hyphen there       */
    line_ptr = line_ptr + 2;        /* Point 2 chars past hyphen    */
    (void) memcpy (line_ptr,
                   (*context_ptr).lstverc8, (*context_ptr).lstver_lenl);
    /* copy version string      */

    /*
     *    We do not know where the date/time string will start, only that it must
     *    end at column 70. Determine the start column and copy it in.
     */
    line_ptr = (*context_ptr).headingc84;
    /* point to start of heading    */
    line_ptr += HEADING_DATE_END - date_time_len;
    /* Get pointer to start of date */
    (void) memcpy (line_ptr, date_time, date_time_len);
    /* ...and add the date/time here */

    /*
     *    ...the date/time string is followed by the text ' Page'. Add this to 
     *    the heading line. The actual page number is added in LIST_NEWPAGE as 
     *    this is where it the heading line is output.
     */
    line_ptr += date_time_len;      /* Point past date/time string  */
    (void) memcpy (line_ptr, " Page", 5);
    /* ...and add page text     */

    /*
     *  Copy "fixed" subheading line 1 into the common data structures.
     */
    string_len = strlen (subhdg1_ptr);  /* Length of sub heading 1      */
    if (string_len > 255) string_len = 255;
    /* restrict to 255 characters   */
    (void) memcpy ((*context_ptr).subhdgc256_6[0],
                   subhdg1_ptr, string_len);   /* copy the subheading and ...  */
    (*context_ptr).subhdgc256_6 [0] [string_len] = '\0';
    /* ...null terminate it     */

    /*
     *    If the subheading was a single space then assume no subheading !
     */
    if (((*context_ptr).subhdgc256_6[0] [0] == ' ')
            && (string_len == 1))      /* If it was a single " ",  */
    {
        (*context_ptr).subhdgc256_6 [0] [0] = '\0';
        /* ...then erase the space  */
    }

    /*
     *  Copy "fixed" subheading line 2 into the common data structures.
     */
    string_len = strlen (subhdg2_ptr);  /* Length of sub heading 2      */
    if (string_len > 255) string_len = 255;
    /* restrict to 255 characters   */
    (void) memcpy ((*context_ptr).subhdgc256_6 [1], subhdg2_ptr, string_len);
    /* copy the subheading and ...  */
    (*context_ptr).subhdgc256_6 [1] [string_len] = '\0';
    /* ...null terminate it     */

    /*
     *    If the subheading was a single space then assume no subheading !
     */
    if (((*context_ptr).subhdgc256_6[1] [0] == ' ')
            && (string_len == 1))      /* If it was a single " ",  */
    {
        (*context_ptr).subhdgc256_6 [1] [0] = '\0';
        /* ...then erase the space  */
    }

    /*
     *    Subheadings 3-6 are always ommited. They must be specified via the 
     *    LIST_SUBHEADINGS routine, so clear them initially.
     */
    (*context_ptr).subhdgc256_6 [2] [0] = '\0';
    /* Clear subheading #3      */
    (*context_ptr).subhdgc256_6 [3] [0] = '\0';
    /* Clear subheading #4      */
    (*context_ptr).subhdgc256_6 [4] [0] = '\0';
    /* Clear subheading #5      */
    (*context_ptr).subhdgc256_6 [5] [0] = '\0';
    /* Clear subheading #6      */

    /*
     *  Now open the print file.
     */
    iret = LIST__OPEN_LISTING (context_ptr,
                               def_file_ptr,
                               rel_file_ptr);

    /*
     *    Check for an error status being returned and report as necessary.
     */
    if (!(psl_test_status (iret)))  /* Error? - then report it  */
    {

        /*
         *    Error opening the listing file - Report it.
         */
        LIST__FILE_ERROR ((*context_ptr).diag_ptrps,
                          PSL_OPENOUT,
                          (*context_ptr).prtfilc256);
        /* report the error     */
        (*context_ptr).list_initn = PFALSE;
        /* Say Listing NOT initialised*/
        iret = PSL_LISTABORT;       /* Set failure return status    */
    }
    else
    {
        (*context_ptr).list_initn = PTRUE;
        /* Show listing initialized */
    }

    /*
     *  Exit, returning status to the caller.
     */
end:
    return iret;
}


/************************************************************************/
/*                                  */
/*    LIST_CLOSE - Close the listing file, and spool if required    */
/*    ==========                            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_CLOSE (context_ptr)                 */
/*                                  */
/*    Inputs:                               */
/*  list_ctx    = Pointer to the listing context block as returned  */
/*            by LIST_INIT                      */
/*  context_ptr = Pointer to listing context block          */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success), PSL_LISTABORT    */
/*            or PSL_LISTNOLINES.               */
/*                                  */
/*    This routine is responsible for closing the Listing file,     */
/*    and spooling it to the print queue if requested (VMS ONLY).   */
/*                                  */
/************************************************************************/

unsigned long LIST_CLOSE (struct list_ctx *context_ptr)
{

    /*
     *  Local Data.
     */
    unsigned long    iret;          /* General status return    */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;

    /*
     *    ...now check to see if Listing was ever initialized and, if not, exit
     *    at once.
     */
    if (!(*context_ptr).list_initn) /* If not Initialized */
    {
        goto list_close_end;        /* Exit at once, with success   */
    }

    /*
     *  See if there has been any output and, if not, output the first
     *  heading and a diagnostic message.
     */
    if  ((*context_ptr).page_nol == 0)  /* If nothing output        */
    {
        (void) LIST_NEWPAGE (context_ptr);
        /* Output new page      */
        iret = PSL_LISTNOLINES;     /* Set up message status    */
        PSL_PUTDIAG ((*context_ptr).diag_ptrps, 1, iret);
        PSL_OUTPUT_DIAG ((*context_ptr).diag_ptrps ,0,DIOUFL_LSTERR);
    }

    /*
     *  Close Listing File, Spooling it if /SPOOL specified.
     */
    if  ((*context_ptr).spool_qualn)    /* If SPOOL requested */
    {
        iret = LIST__CLOSE_PRINT (context_ptr);
        /* Close listing file and spool it */
        if (!(psl_test_status (iret)))  /* If error then report it!     */
        {
            goto list_close_error;
        }
    }

    else
    {
        iret = LIST__CLOSE_KEEP (context_ptr);
        /* Just close the listing file */
        if (!(psl_test_status (iret)))  /* If error then report it!     */
        {
            goto list_close_error;
        }
    }

    /*
     *  Reset Unit Number to show we're not Initialized.
     */
    (*context_ptr).list_initn = PFALSE;

    /*
     *  Exit, returning status to the caller.
     */
list_close_end:
    return iret;

    /*
     *  Error Closing Listing File - Output Error and Return to Caller.
     */
list_close_error:
    LIST__FILE_ERROR ((*context_ptr).diag_ptrps,
                      PSL_CLOSEOUT,
                      (*context_ptr).prtfilc256);
    /* ...and report the error  */
    iret = PSL_LISTABORT;       /* ...then set failure status   */
    goto list_close_end;        /* ...and return it     */
}

/************************************************************************/
/*                                  */
/*    LIST_GET_DETAILS - Return details about the listing file      */
/*    ================                          */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_GET_DETAILS (context_ptr, type, value)      */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Pointer to listing context block          */
/*  type     = unsigned long flag for type of info required     */
/*                                  */
/*    Outputs:                              */
/*  value    = Pointer to unsigned long returned value.     */
/*                                  */
/*    Returns:                              */
/*  PSL_LISTABORT = Invalid context pointer             */
/*                                  */
/*    This routine is responsible for returning the current page    */
/*    number.                               */
/*                                  */
/************************************************************************/

unsigned long   LIST_GET_DETAILS (         struct   list_ctx *context_ptr,
        unsigned long     type,
        unsigned long     *value )

{
    /*
     *    local data
     */
    unsigned long iret;         /* Status variable      */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *  This routine is currently only capable of returning the current
     *  page number. 
     *
     *  If Type is 1 => Just return the page number
     */
    if (type == LIST_DETAILS_PAGE)  /* If type 1,           */
    {
        *value  = (*context_ptr).page_nol;
        /* Return the page number   */
    }

    /*
     *  Exit here.
     */
    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    LIST_LINE - Output an unnumbered line to the listing file     */
/*    =========                             */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_LINE (context_ptr, line_ptr);           */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Pointer to listing context block          */
/*  line_ptr  = Pointer to null terminated line to output       */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success) or PSL_LISTABORT  */
/*                                  */
/*    This routine is responsible for outputing the input string    */
/*    to the (currently open) Listing file, WITHOUT a line number   */
/*    prefix, but indented to column 9.                 */
/*                                  */
/************************************************************************/

unsigned long   LIST_LINE (struct list_ctx *context_ptr,
                           const  char     *line_ptr)

{
    /*
     *  Local Data.
     */
    unsigned long   iret;           /* General status return    */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *  Check to see if Listing Initialized and, if not, output an error and
     *  exit.  Otherwise initialize Routine Return Status.
     */
    if (!(*context_ptr).list_initn) /* If not Initialized */
    {
        iret = PSL_LISTNOTINI;      /* Set up message status    */
        (*context_ptr).err_trappingl = ERT_DISABLED;
        /* Cancel Error Trapping */

        PSL_PUTDIAG ((*context_ptr).diag_ptrps , 1, iret);
        PSL_SIGNAL_ERROR ((*context_ptr).diag_ptrps , PSLERRCLS_LIST, iret);
        iret = PSL_LISTABORT;       /* Set up RETURN status     */
        goto list_line_end;     /* And Return at once       */
    }

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status     */

    /*
     *  If last output was an error message, output a blank line.
     */
    if ((*context_ptr).last_errl == ERR_YES)
        /* If we have just output an error */
    {
        iret = LIST_SKIPLINES (context_ptr, 1);
        /* ...then output "separator" line */
        (*context_ptr).last_errl = ERR_NO;
        /* ...and clear "last error" flag  */
    }

    /*
     *  Set up a page throw if the "current line" counter indicates so.
     */
    if ((*context_ptr).ln_on_pagl >= (*context_ptr).max_pag_sizl)
        /* If we are at the end of page */
    {
        iret = LIST_NEWPAGE (context_ptr);
        /* Throw a page         */
        if (!(psl_test_status (iret))) goto list_line_end;
        /* Return status, on error  */
    }

    (*context_ptr).ln_on_pagl = (*context_ptr).ln_on_pagl + 1;
    /* Count the line       */

    /*
     *    Write the line to the listing file.
     */
    iret = LIST__WRITE  (context_ptr,
                         1,     /* "Write" type         */
                         line_ptr,  /* String descriptor        */
                         0);        /* Omit line/page parameter     */
    if (!(psl_test_status (iret))) goto list_line_error;
    /* Report error, on failure */

    /*
     *  Exit, returning status to the caller.
     */
list_line_end:
    return iret;

    /*
     *  Error found outputing to the Listing file, report the error.
     */
list_line_error:
    LIST__FILE_ERROR ((*context_ptr).diag_ptrps,
                      PSL_WRITEERR,
                      (*context_ptr).prtfilc256);
    /* ...and report the error  */
    iret = PSL_LISTABORT;       /* Set failure return status    */
    goto list_line_end;         /* ...and return with status    */

}

/************************************************************************/
/*                                  */
/*    LIST_LINE_COLUMN_ONE  -  Output an unnumbered line to the     */
/*    ====================     listing file, at column one      */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_LINE_COLUMN_ONE (context_ptr, line_ptr);        */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Pointer to listing context block          */
/*  line_ptr  = Pointer to null terminated STRING to be output  */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success) or PSL_LISTABORT) */
/*                                  */
/*    This routine is responsible for outputing the input string    */
/*    to the (currently open) Listing file, WITHOUT a line number   */
/*    prefix, from column one.                      */
/*                                  */
/************************************************************************/

unsigned long   LIST_LINE_COLUMN_ONE (struct list_ctx *context_ptr,
                                      const  char     *line_ptr)
{

    /*
     *  Local Data.
     */
    unsigned long   iret;           /* General status return    */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *
     *  Check to see if Listing Initialized and, if not, output an error and
     *  exit.  Otherwise initialize Routine Return Status.
     *
     */
    if (!(*context_ptr).list_initn) /* If not Initialized */
    {
        iret = PSL_LISTNOTINI;      /* Set up message status    */
        (*context_ptr).err_trappingl = ERT_DISABLED;
        /* Cancel Error Trapping */

        PSL_PUTDIAG ((*context_ptr).diag_ptrps , 1, iret);
        PSL_SIGNAL_ERROR ((*context_ptr).diag_ptrps , PSLERRCLS_LIST, iret);
        iret = PSL_LISTABORT;       /* Set up RETURN status     */
        goto column_one_end;        /* And Return at once       */
    }

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status */

    /*
     *  If last output was an error message, output a blank line.
     */
    if ((*context_ptr).last_errl == ERR_YES)
    {
        iret = LIST_SKIPLINES (context_ptr, 1);
        /* Output a "separator" line    */
        (*context_ptr).last_errl = ERR_NO;
        /* Clear "last error" flag  */
    }

    /*
     *  Set up a page throw if the "current line" counter indicates so.
     */
    if ((*context_ptr).ln_on_pagl >= (*context_ptr).max_pag_sizl)
        /* If at end of page        */
    {
        iret = LIST_NEWPAGE (context_ptr);
        /* ...throw a new page      */
        if (!(psl_test_status (iret))) goto column_one_end;
        /* ...and return status, on error */
    }
    (*context_ptr).ln_on_pagl = (*context_ptr).ln_on_pagl + 1;
    /* Count the line       */

    /*
     *    Write the listing line to the listing file
     */
    iret = LIST__WRITE  (context_ptr,
                         4,      /* "Write" type         */
                         line_ptr,   /* String descriptor        */
                         0);     /* Omit line/column parameter   */
    if (!(psl_test_status (iret))) goto column_one_error;
    /* Report error, on failure     */

    /*
     *  Exit, returning status to the caller.
     */
column_one_end:
    return iret;

    /*
     *  Error found outputing to the Listing file, report the error.
     */
column_one_error:
    LIST__FILE_ERROR ((*context_ptr).diag_ptrps,
                      PSL_WRITEERR,
                      (*context_ptr).prtfilc256);
    /* ...and report the error  */
    iret = PSL_LISTABORT;       /* Set failure return status    */
    goto column_one_end;        /* and return with status   */
}

/************************************************************************/
/*                                  */
/*    LIST_LINE_REFERENCED - Output a numbered line to the listing  */
/*    ====================   file.                  */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_LINE_REFERENCED (context_ptr, line_ptr);        */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Pointer to listing context block          */
/*      line_ptr  = Pointer to null terminated STRING to be output  */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success) or PSL_LISTABORT) */
/*                                  */
/*    This routine is responsible for outputing the input string    */
/*    to the (currently open) Listing file, with a line number      */
/*    prefix.                               */
/*                                  */
/************************************************************************/

unsigned long   LIST_LINE_REFERENCED (struct list_ctx *context_ptr,
                                      const  char     *line_ptr)

{
    /*
     *  Local Data.
     */
    unsigned long iret;         /* General status return    */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *  Check to see if Listing Initialized and, if not, output an error and
     *  exit.  Otherwise initialize Routine Return Status.
     */
    if (!(*context_ptr).list_initn) /* If not Initialized */
    {
        iret = PSL_LISTNOTINI;      /* Set up message status    */
        (*context_ptr).err_trappingl = ERT_DISABLED;
        /* Cancel Error Trapping */

        PSL_PUTDIAG ((*context_ptr).diag_ptrps , 1, iret);
        PSL_SIGNAL_ERROR ((*context_ptr).diag_ptrps , PSLERRCLS_LIST, iret);
        iret = PSL_LISTABORT;       /* Set up RETURN status     */
        goto list_referenced_end;   /* And Return at once       */
    }

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status     */

    /*
     *  If last output was an error message, output a blank line.
     */
    if ((*context_ptr).last_errl == ERR_YES)
        /* If we just reported an error */
    {
        iret = LIST_SKIPLINES (context_ptr, 1);
        /* ...output a "separator" line */
        (*context_ptr).last_errl = ERR_NO;
        /* ...and clear "last error" flag */
    }

    /*
     *  Set up a page throw if the "current line" counter indicates so.
     */
    if ((*context_ptr).ln_on_pagl >= (*context_ptr).max_pag_sizl)
        /* If end of current page   */
    {
        iret = LIST_NEWPAGE (context_ptr);
        /* ...then throw a new page     */
        if (!(psl_test_status (iret))) goto list_referenced_end;
        /* Return status, on error  */
    }

    (*context_ptr).ln_refnol = (*context_ptr).ln_refnol + 1;
    /* Increment the ref. no.   */
    (*context_ptr).ln_on_pagl = (*context_ptr).ln_on_pagl +1;
    /* Count the new line       */

    /*
     *    ...and write the listing line to the listing file.
     */
    iret = LIST__WRITE  (context_ptr,
                         2,     /* "Write" type         */
                         line_ptr,  /* String descriptor        */
                         (*context_ptr).ln_refnol);
    /* Line number          */
    if (!(psl_test_status (iret))) goto list_referenced_error;
    /* Report error, on failure     */

    /*
     *  Exit, returning status to the caller.
     */
list_referenced_end:
    return iret;

    /*
     *  Error found outputing to the Listing file, report the error.
     */
list_referenced_error:
    LIST__FILE_ERROR ((*context_ptr).diag_ptrps,
                      PSL_WRITEERR,
                      (*context_ptr).prtfilc256);
    /* ...and report the error  */
    iret = PSL_LISTABORT;       /* ...set failure return status */
    goto list_referenced_end;       /* ...and return with status    */
}

/************************************************************************/
/*                                  */
/*    LIST_NEWPAGE - Output New page header and reset Line and      */
/*    ============   Page counts                    */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_NEWPAGE (context_ptr)               */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Pointer to listing context block          */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success) or PSL_LISTABORT) */
/*                                  */
/*    This routine is responsible for outputing a standard page     */
/*    header; "fixed" and any "floating" subheading lines, to the   */
/*    listing file. It also resets the count of lines on the        */
/*    current page and increments the page number count.        */
/*                                  */
/************************************************************************/

unsigned long   LIST_NEWPAGE (struct list_ctx *context_ptr)
{

    /*
     *  Local Data.
     */
    unsigned long   iret;           /* General status return    */
    int    page_no_len;        /* Length of page number text   */

    short  isub;           /* Do loop/Work counter     */
    char   *line_ptr;      /* Pointer to char string   */
    char   page [12];      /* Buffer to hold page value    */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *  Check to see if Listing Initialized and, if not, output an error and
     *  exit.  Otherwise initialize Routine Return Status.
     */
    if (!(*context_ptr).list_initn) /* If not Initialized */
    {
        iret = PSL_LISTNOTINI;      /* Set up message status    */
        (*context_ptr).err_trappingl = ERT_DISABLED;
        /* Cancel Error Trapping */

        PSL_PUTDIAG ((*context_ptr).diag_ptrps , 1, iret);
        PSL_SIGNAL_ERROR ((*context_ptr).diag_ptrps , PSLERRCLS_LIST, iret);
        iret = PSL_LISTABORT;       /* Set up RETURN status     */
        goto list_newpage_end;      /* And Return at once       */
    }

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status     */

    /*
     *  Show last output was not an error message.
     */
    (*context_ptr).last_errl = ERR_NO;  /* Clear "last error" flag  */

    /*
     *  Increment Page Number and initialize Line Number.
     */
    (*context_ptr).page_nol = (*context_ptr).page_nol + 1;
    /* Increment page number    */
    (*context_ptr).ln_on_pagl = 3;  /* Heading + 2 Blank Lines  */

    /*
     *    Add the page number to the main heading line and write it to the
     *    listing file.
     *    I (NF) tried to fix a lint warning regarding the next line by
     *    specifying %ln - this caused a different lint warning and the
     *    call no longer worked. So, be my guest.
     */
    (void) sprintf (page, "%ld%n",
                    (*context_ptr).page_nol, &page_no_len);
    /* Get textual page number  */
    line_ptr = (*context_ptr).headingc84;
    /* point to start of heading    */
    line_ptr += HEADING_PAGE_END - page_no_len;
    /* Point to start of page text   */
    (void) memcpy (line_ptr, page, page_no_len);
    /* Copy page number to buffer   */

    /*
     *    ...and write the heading line to the listing file.
     */
    iret = LIST__WRITE (context_ptr,
                        3,       /* "Write" type - Precede with FF */
                        (*context_ptr).headingc84,
                        /* Heading line string      */
                        0);      /* No last parameter        */
    if (!(psl_test_status (iret))) goto list_newpage_error;
    /* Report error, on failure     */

    /*
     *  Output Fixed Subheading Lines (those specified via LIST_INIT).
     */
    for (isub = 0; isub < 2; isub++)    /* for each subheading      */
    {
        if ((strlen ((*context_ptr).subhdgc256_6 [isub])) != 0)
            /* if its specified         */
        {
            iret = LIST__WRITE (context_ptr,
                                4,  /* "Write" type         */
                                (*context_ptr).subhdgc256_6 [isub],
                                /* Write subheading String  */
                                0); /* no line/page parameter   */
            if (!(psl_test_status (iret))) goto list_newpage_error;
            /* Report error, on failure     */

            (*context_ptr).ln_on_pagl = (*context_ptr).ln_on_pagl + 1;
            /* Count this line      */
        }
    }

    /*
     *  Output a Blank Line.
     */
    iret = LIST__WRITE (context_ptr,
                        5,      /* "Write" type         */
                        NULL,       /* No line to output        */
                        0);     /* No line/column parameter */
    if (!(psl_test_status (iret))) goto list_newpage_error;
    /* Report error, on failure     */

    /*
     *  Output Floating Subheading Lines (those specified via LIST_SUBHEADING).
     */
    for (isub = 2; isub < 6; isub++)    /* for each "floating" subhdg   */
    {
        if ((strlen ((*context_ptr).subhdgc256_6 [isub])) != 0)
        {
            iret = LIST__WRITE (context_ptr,
                                4,  /* "Write" type         */
                                (*context_ptr).subhdgc256_6 [isub],
                                /* String descriptor        */
                                0); /* No line/page parameter   */
            if (!(psl_test_status (iret))) goto list_newpage_error;
            /* Report error, on failure     */
            (*context_ptr).ln_on_pagl = (*context_ptr).ln_on_pagl + 1;
            /* Count this line      */
        }
    }

    /*
     *  Output a Second Blank Line.
     */
    iret = LIST__WRITE (context_ptr,
                        5,      /* "Write" type         */
                        NULL,       /* Mo string to output      */
                        0);     /* No line/page parameter   */
    if (!(psl_test_status (iret))) goto list_newpage_error;
    /* Report error, on failure     */

    /*
     *    Exit, returning status to the caller.
     */
list_newpage_end:
    return iret;

    /*
     *  Error found outputing to the Listing file, report the error.
     */
list_newpage_error:
    LIST__FILE_ERROR ((*context_ptr).diag_ptrps,
                      PSL_WRITEERR,
                      (*context_ptr).prtfilc256);
    /* ...and report the error  */
    iret = PSL_LISTABORT;       /* Set failure return status    */
    goto list_newpage_end;      /* and return with status   */
}

/************************************************************************/
/*                                  */
/*    LIST_SKIPLINES - Skip "n" lines in the listing file       */
/*    ==============                            */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_SKIPLINES (context_ptr, number)         */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Pointer to listing context block          */
/*  number  = unsigned LONG identifying number of lines     */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success), PSL_LISTABORT or */
/*            PSL_LISTNOROOM                    */
/*                                  */
/*    This routine is responsible for outputing the specified       */
/*    number of blank lines (i.e. lines of one space) to the        */
/*    listing file. It will return PSL_LISTNOROOM (and initiate     */
/*    a page throw) if there is insufficient room on the current    */
/*    page.                             */
/*                                  */
/************************************************************************/

unsigned long   LIST_SKIPLINES (struct list_ctx *context_ptr,
                                unsigned long no_of_lines)
{

    /*
     *  Local Data.
     */
    unsigned long   iret;           /* General status return */
    unsigned long   loop_count;     /* Loop counter */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *  Check to see if Listing Initialized and, if not, output an error and
     *  exit.  Otherwise initialize Routine Return Status.
     */
    if (!(*context_ptr).list_initn) /* If not Initialized */
    {
        iret = PSL_LISTNOTINI;      /* Set up message status    */
        (*context_ptr).err_trappingl = ERT_DISABLED;
        /* Cancel Error Trapping */

        PSL_PUTDIAG ((*context_ptr).diag_ptrps , 1, iret);
        PSL_SIGNAL_ERROR ((*context_ptr).diag_ptrps , PSLERRCLS_LIST, iret);
        iret = PSL_LISTABORT;       /* Set up RETURN status     */
        goto list_skiplines_end;    /* And Return at once       */
    }

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status */

    /*
     *  Show last output was not an error message.
     */
    (*context_ptr).last_errl = ERR_NO;  /* Negate "Last Error" flag */

    /*
     *  If there is insufficient room left on the page, return this status
     *  to the caller, otherwise, output the requisite number of blank lines.
     */
    if (((*context_ptr).ln_on_pagl + no_of_lines) > (unsigned long)
            (*context_ptr).max_pag_sizl)    /* If insufficient room on page */
    {
        (*context_ptr).ln_on_pagl = (*context_ptr).max_pag_sizl + 1;
        /* Initiate page throw      */
        iret = PSL_LISTNOROOM;      /* Tell the User "No room"! */
    }
    else
    {
        for (loop_count = 0;
                loop_count < no_of_lines;
                loop_count++, (*context_ptr).ln_on_pagl++)
        {
            iret = LIST__WRITE (context_ptr,
                                5,  /* "Write" type         */
                                NULL,   /* No string to list        */
                                0); /* No page/line number      */
            if (!(psl_test_status (iret))) goto list_skiplines_error;
            /* If error ===> Report it  */
        }
    }

    /*
     *  Exit, returning status to the caller.
     */
list_skiplines_end:
    return iret;

    /*
     *  Error found outputing to the Listing file, report the error.
     */
list_skiplines_error:
    LIST__FILE_ERROR ((*context_ptr).diag_ptrps,
                      PSL_WRITEERR,
                      (*context_ptr).prtfilc256);
    /* Report the error     */
    iret = PSL_LISTABORT;       /* Set failure return status    */
    goto list_skiplines_end;        /* and return with status   */
}

/************************************************************************/
/*                                  */
/*    LIST_SUBHEADING - Reset the "floating" subheadings        */
/*    ===============                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_SUBHEADING_C (context_ptr,              */
/*                subhdg3_ptr,              */
/*                subhdg4_ptr,              */
/*                subhdg5_ptr,              */
/*                subhdg6_ptr)              */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Pointer to listing context block          */
/*  subhdg%_ptr = Pointer to null terminated STRING representing    */
/*            subheading text                   */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            Will ALWAYS be PSL_NORMAL!            */
/*                                  */
/*    This routine is responsible for reseting the "floating"       */
/*    subheading lines for the Listing routine environment. It      */
/*    also ensures that the subheadings are NULL terminated.        */
/*                                  */
/************************************************************************/

unsigned long   LIST_SUBHEADING (struct list_ctx *context_ptr,
                                 const  char *subhdg3_ptr,
                                 const  char *subhdg4_ptr,
                                 const  char *subhdg5_ptr,
                                 const  char *subhdg6_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long  head_len;        /* Length of heading        */
    unsigned long  iret;            /* Status return        */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *    Copy Subheadings and lengths to common structure definitions, ensuring
     *    that they are null terminated and do not exceed the maximum length of
     *    256 characters (null terminated).
     */
    head_len = strlen (subhdg3_ptr);    /* Get the length of subheading */
    if (head_len > 255) head_len = 255;
    /* Ensure 255 chars or less */
    if ((head_len == 1)
            &&  (*subhdg3_ptr == ' '))      /* If heading is just 1 space   */
    {
        (*context_ptr).subhdgc256_6 [2] [0] = '\0';
        /* ...then clear subheading */
    }
    else
    {
        (void) memcpy ((*context_ptr).subhdgc256_6[2], subhdg3_ptr, head_len);
        /* ...else copy subheading  */
        (*context_ptr).subhdgc256_6 [2] [head_len] = '\0';
        /* ...and null terminate it.    */
    }

    /*
     *    Now handle subheading 4
     */
    head_len = strlen (subhdg4_ptr);    /* Get the length of subheading */
    if (head_len > 255) head_len = 255;
    /* Ensure 255 chars or less */
    if ((head_len == 1)
            &&  (*subhdg4_ptr == ' '))      /* If heading is just 1 space   */
    {
        (*context_ptr).subhdgc256_6 [3] [0] = '\0';
        /* ...then clear subheading */
    }
    else
    {
        (void) memcpy ((*context_ptr).subhdgc256_6 [3], subhdg4_ptr, head_len);
        /* ...else copy subheading  */
        (*context_ptr).subhdgc256_6 [3] [head_len] = '\0';
        /* ...and null terminate it.    */
    }

    /*
     *    ...and then subheading 5
     */
    head_len = strlen (subhdg5_ptr);    /* Get the length of subheading */
    if (head_len > 255) head_len = 255;
    /* Ensure 255 chars or less */
    if ((head_len == 1)
            &&  (*subhdg5_ptr == ' '))      /* If heading is just 1 space   */
    {
        (*context_ptr).subhdgc256_6 [4] [0] = '\0';
        /* ...then clear subheading */
    }
    else
    {
        (void) memcpy ((*context_ptr).subhdgc256_6[4], subhdg5_ptr, head_len);
        /* ...else copy subheading  */
        (*context_ptr).subhdgc256_6 [4] [head_len] = '\0';
        /* ...and null terminate it.    */
    }

    /*
     *    ...and finally subheading 6
     */
    head_len = strlen (subhdg6_ptr);    /* Get the length of subheading */
    if (head_len > 255) head_len = 255;
    /* Ensure 255 chars or less */
    if ((head_len == 1)
            &&  (*subhdg6_ptr == ' '))      /* If heading is just 1 space   */
    {
        (*context_ptr).subhdgc256_6 [5] [0] = '\0';
        /* ...then clear subheading */
    }
    else
    {
        (void) memcpy ((*context_ptr).subhdgc256_6[5], subhdg6_ptr, head_len);
        /* ...else copy subheading  */
        (*context_ptr).subhdgc256_6 [5] [head_len] = '\0';
        /* ...and null terminate it.    */
    }

    /*
     *  Exit, returning status to the caller.
     */
    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*    LIST_TESTPAGE - Tests for "n" free lines on current page      */
/*    =============                         */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_TESTPAGE (context_ptr, no_of_lines)         */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Pointer to listing context block          */
/*  no_of_lines = Signed LONG Number of lines to be tested      */
/*            If -1 then throw a new page immediately       */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success), PSL_LISTABORT or */
/*            PSL_LISTNOROOM                    */
/*                                  */
/*    This routine is responsible for testing to see if the     */
/*    requisite number of free lines are available on the current   */
/*    page, or, to force a page throw.                  */
/*                                  */
/************************************************************************/

unsigned long   LIST_TESTPAGE (struct   list_ctx *context_ptr,
                               long     no_of_lines)

{
    /*
     *  Local Data.
     */
    unsigned long    iret;          /* General status return    */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *  Initialize Routine Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status     */

    /*
     *  If there is insufficient room left on the page, return this status
     *  to the caller, and initiate a page throw.
     */
    if ((((*context_ptr).ln_on_pagl + no_of_lines) >
            (*context_ptr).max_pag_sizl) ||
            (no_of_lines == -1))
    {
        (*context_ptr).ln_on_pagl = (*context_ptr).max_pag_sizl + 1;
        /* Initiate forced page throw   */
        iret = PSL_LISTNOROOM;      /* Tell the User "No room"! */
    }

    /*
     *  Exit, returning status to the caller.
     */
    return iret;
}

/************************************************************************/
/*                                  */
/*    LIST__CHK_CONTEXT - Validate the context pointer.                 */
/*    ================                          */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST__CHK_CONTEXT (context_ptr)              */
/*                                  */
/*    Inputs:                               */
/*  context_ptr = Pointer to listing context block          */
/*                                  */
/*    Outputs:                              */
/*  None                                */
/*                                  */
/*    Returns:                              */
/*  PSL_NORMAL    = Valid context block             */
/*  PSL_LISTABORT = Invalid context block - Abort listing       */
/*                                  */
/*    This routine is responsible for validating the context block. */
/*    If it encounters an error then that error will be reported here   */
/*                                  */
/************************************************************************/

unsigned  long  LIST__CHK_CONTEXT (struct list_ctx *context_ptr)
{
    /*
     *    Local data
     */
    unsigned long iret;         /* Status variable      */

    if ((context_ptr == NULL)
            ||  (memcmp ((*context_ptr).ftypc4, "LIST", 4) != 0)
            ||  ((*context_ptr).fvrnl != LIST_FVRN))
    {
        iret = PSL_LISTABORT;       /* Primary status = list abort  */
        PSL_PUTDIAG ((*context_ptr).diag_ptrps , 2, iret, PSL_INVLISTCTX);
        PSL_SIGNAL_ERROR ((*context_ptr).diag_ptrps , PSLERRCLS_LIST, iret);
        return iret;            /* Return primary status    */
    }
    else
    {
        return PSL_NORMAL;      /* O.K. return success      */
    }
}

/************************************************************************/
/*
/*    LIST_EXPFIL - Get Expanded Listing File Specification         */
/*    ===========                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = LIST_EXPFIL (                        */
/*          context_ptr,                    */
/*          expfil_ptr,                 */
/*          expfil_size)                    */
/*                                  */
/*    Inputs:                               */
/*  context_ptr     = Pointer to listing context block      */
/*                                  */
/*    Outputs:                              */
/*  expfil_ptr  = Pointer to expanded file specification    */
/*  expfil_size)    = Output Expanded file specification size   */
/*                                  */
/*    Returns:                              */
/*  iret        = Unsigned Long PSL Routine Return Status       */
/*            May be PSL_NORMAL (Success) or PSL_LISTABORT) */
/*                                  */
/*  This routine is responsible returning the expanded listing      */
/*  file specification for the file identified by the supplied      */
/*  context block.                          */
/*                                  */
/************************************************************************/

unsigned long LIST_EXPFIL (struct list_ctx *context_ptr,
                           char *expfil_ptr,
                           unsigned long expfil_size)
{
    /*
     *  Local Data.
     */
    unsigned long     iret;         /* General status return    */
    unsigned long     length;       /* Length of filename */

    /*
     *  Initialize the Return Status.
     */
    iret = PSL_NORMAL;          /* Initialize Return Status     */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error.
     */
    iret = LIST__CHK_CONTEXT (context_ptr);
    if (!(psl_test_status (iret)))
    {
        return iret;
    }

    /*
     *    Get expanded file specification length.
     */
    length = strlen ((*context_ptr).prtfilc256);

    /*
     *    Return the expanded listing file specification - null-terminated.
     */
    if (length < expfil_size)
    {
        (void) strcpy (expfil_ptr, (*context_ptr).prtfilc256);
    }
    else
    {
        (void) memcpy (expfil_ptr, (*context_ptr).prtfilc256, expfil_size -1);
        *(expfil_ptr + expfil_size - 1) = '\0';
    }

    /*
     *  Exit, returning status to the caller.
     */
    return iret;
}






/************************************************************************/
/*                                  */
/*  LIST_OUTPUT_ERROR_BUFFER - Output Multi-Line Error Buffer   */
/*                                  */
/*  This routine output a multi-line error message. It is       */
/*  intended to be called from PSL_OUTPUT_DIAG.         */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*  iret = LIST_OUTPUT_ERROR_BUFFER (list_ctx_ptr, buf_ptr,     */
/*                   numlin, width)         */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret     = Unsigned Long PSL Routine Return Status       */
/*     list_ctx_ptr = Pointer to Listing Context Block      */
/*     buf_ptr  = Pointer to Error Text(s).             */
/*     numlin   = Number of Lines.                  */
/*     width    = Width of each Line.               */
/*                                  */
/************************************************************************/

unsigned long  LIST_OUTPUT_ERROR_BUFFER (struct list_ctx *list_ctx_ptr,
        const char *buf_ptr,
        unsigned long numlin,
        unsigned long width)
{

    /*   Local Data   */

    unsigned long       iret;       /* Return Status */
    unsigned long       jsub;       /* Loop Control */
    char       errmsg[256];    /* Error Message */
    char       *wrk_ptr;       /* Pointer to eack buffer */

    /*
     *    Check the validity of the context block. If this is invalid then return
     *    with an error. NOTE we cannot use the standard checking routine as this
     *    would cause us to loop if it encountered an error !!!
     */
    if ((list_ctx_ptr == NULL)
            ||  (memcmp ((*list_ctx_ptr).ftypc4, "LIST", 4) != 0)
            ||  ((*list_ctx_ptr).fvrnl != LIST_FVRN))
    {
        return PSL_NORMAL;
    }

    /*
     *  Set the error trap flag to disallow recursive error logging
     */
    if ((*list_ctx_ptr).err_trappingl == ERT_ENABLED)
        /* NOT in a recursive loop?...  */
    {
        (*list_ctx_ptr).err_trappingl = ERT_DISABLED;
        /* Prevent us being in one!     */

        /*
         *  If last output was not an error message, output a blank line first,
         *  and stop LIST_LINE_COLUMN_ONE from inserting one
         */
        if ((*list_ctx_ptr).last_errl == ERR_NO)
        {
            (void) LIST_SKIPLINES (list_ctx_ptr, 1); /* Skip a line here */
        }
        (*list_ctx_ptr).last_errl = ERR_NO; /* Avoid skipping a line */
        /* in LIST_LINE_COLUMN_ONE */

        /*
         *    We know that we have a multi-line error message and have a count of the 
         *    number of lines in that message. Ensure that there are sufficient lines
         *    on the page to fit in ALL of the message.
         */
        (void) LIST_TESTPAGE (list_ctx_ptr, numlin);
        /* Ensure enough room on page   */

        /*
         *  Write the buffers - NOTE: the messages may not always be Null-Terminated
         */
        for (jsub = 0, wrk_ptr = (char *) buf_ptr;
                jsub < numlin;
                jsub++, wrk_ptr += width)
        {
            (void) memcpy (errmsg, wrk_ptr, width);
            errmsg[width] = '\0';

            iret = LIST_LINE_COLUMN_ONE (list_ctx_ptr, errmsg);
            if ( !(psl_test_status (iret)) ) goto end;
        }

        /*
         *  Write a blank line
         */
        iret = LIST_SKIPLINES (list_ctx_ptr, 1);
        if ( !(psl_test_status (iret)) ) goto end;
    }

    /*
     *  Enable error trapping
     */
    (*list_ctx_ptr).err_trappingl = ERT_ENABLED;

    /*
     *  Indicate last error.
     */
    (*list_ctx_ptr).last_errl = ERR_YES;

    /*
     *   Return with Status
     */
end:
    return iret;
}
