/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_DATE_RTNS               */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains the platform-independent routines           */
/*  that manipulate "portable" dates and times.              */
/*                                       */
/*===========================================================================*/

/*   Define useful Include structures   */

#include  <psl_macros.h>        /* PSL Standard Macros */
#include  <string.h>            /* C String Functions */
#include  <ctype.h>         /* C Data Types & such-like */

/*   Define Function Prototypes for all external routines */
/*   used by the module (and for ourselves) */

#include  psl_common_file(psl_rtndef.h) /* Define ourselves */
#include  "psldate_intrtndef.h"     /* and our internal routines */

/*   Define Function Prototypes for all local routines */



static unsigned long FORMAT_CTMDAT (unsigned long date,
                                    unsigned long time,
                                    char *buf_ptr,
                                    unsigned long buflen,
                                    unsigned long flags);



static unsigned long FORMAT_PSLDAT (unsigned long date,
                                    unsigned long time,
                                    char *buf_ptr,
                                    unsigned long buflen,
                                    unsigned long flags);

static unsigned long PARSE_PSLDAT (const char *buf_ptr,
                                   unsigned long *date_ptr,
                                   unsigned long *time_ptr,
                                   unsigned long inflags,
                                   unsigned long *outflags_ptr);

static unsigned long PARSE_SQLDAT (char *buf_ptr,
                                   unsigned long *date_ptr,
                                   unsigned long *time_ptr,
                                   unsigned long inflags,
                                   unsigned long *outflags_ptr);

static Boolean VALIDATE_DATE (unsigned long year,
                              unsigned long month,
                              unsigned long day,
                              Boolean *leapyear_ptr);

static Boolean VALIDATE_TIME (unsigned long hours,
                              unsigned long mins,
                              unsigned long secs,
                              unsigned long msecs);

/*   Define Data Structures */

#include  psl_common_file(shrmsg.h) /* Message Statuses */
#include  psl_common_file(binasc_prmdef.h)
/* PSL_BINASC Parameters */
#include  psl_common_file(psldate_prmdef.h)
/* Routine Parameters */
#include  <time.h>          /* C Time Functions */


/************************************************************************/
/*                                  */
/*  Define Symbolic Names and Global Data for the Calculations  */
/*                                  */
/************************************************************************/

/*   Number of days in a Quadyear (4 Years + 1 Leap Day) */

#define QYD ((365*4)+1)

/*   Number of days in a Century (25 Quadyears - 1 Leap Day) */

#define CYD ((25*QYD)-1)

/*   Number of days in a Quadricentury (4 Centuries + 1 Leap Day) */

#define QCD ((4*CYD)+1)

/*   Number of days from 1-JAN-1601 to 17-NOV-1858 */

#define QC1 ((1858-1601)/400)       /* Quadricenturies */
#define CY1 (((1858-1601)/100)-(4*QC1)) /* Centuries */
#define QY1 (((1858-1601)/4)-(25*CY1)-(100*QC1))
/* Quadyears */
#define YR1 ((1858-1601)-(4*QY1)-(100*CY1)-(400*QC1))
/* Years */
#define BASOFF ((QC1*QCD) + (CY1*CYD) + (QY1*QYD) + (YR1*365) \
+31+28+31+30+31+30+31+31+30+31+17)

/*   Lowest year that we support.  This must obviously not be less  */
/*   than the base date, but does not have to be the same.      */

#define BASYEAR 1860

/*   Day of the week that the base date corresponds to */

#define BAS_WKDY PSWKDY_WEDNESDAY   /* 17-NOV-1858 was Wednesday */

/*   Number of Days in Leap Year to 1st of each month */

static unsigned long days_tab[12] = {  0,  31,  60,  91, 121, 152,
                                       182, 213, 244, 274, 305, 335};

/*   Number of Days in each month */

static unsigned long month_days[12] = {31, 28, 31, 30, 31, 30,
                                       31, 31, 30, 31, 30, 31};

/*   Table of Valid Month Abbreviations */

static char month_asc[] = "-JAN-FEB-MAR-APR-MAY-JUN-JUL-AUG-SEP-OCT-NOV-DEC";
static char month_asc_ctm[] = "janfebmaraprmayjunjulaugsepoctnovdec";


/************************************************************************/
/*                                  */
/*  PSL_COMPRESS_DATE - Create PSL Date from Constituent Parts  */
/*                                  */
/*  This routine creates a PSL Date from a Year, Month and Day.     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_COMPRESS_DATE (year, month, day, date_ptr);   */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret     = Unsigned Long PSL Return Status           */
/*          = PSL_BADPARAM if supplied date is invalid      */
/*     year     = Unsigned Long Year                */
/*     month    = Unsigned Long Month               */
/*     day      = Unsigned Long Day                 */
/*     date_ptr = Pointer to Unsigned Long to hold PSL Date     */
/*                                  */
/************************************************************************/

unsigned long PSL_COMPRESS_DATE (unsigned long year,
                                 unsigned long month,
                                 unsigned long day,
                                 unsigned long *date_ptr)
{

    /*   Local Data   */

    unsigned long       calc_date;      /* Date being calculated */
    unsigned long       calc_years;     /* Calculated Years */
    unsigned long       calc_qyrs;      /* Calculated Quadyears */
    unsigned long       calc_cents;     /* Calculated Centuries */
    unsigned long       calc_qcnts;     /* Calculated Quadricenturies */
    Boolean    leap_year;      /* Leap Year Flag */
    Boolean    rtnsts;     /* Routine Return Status */

    /*   The first thing to do is to check that we have been given a    */
    /*   valid date, and see if we have a leap year at the same time.   */

    rtnsts = VALIDATE_DATE (year, month, day, &leap_year);
    /* Validate the date */
    if  (!rtnsts) return PSL_BADPARAM;
    /* Exit with error if invalid */

    /*   Although we are really looking for days since 17/11/58, the    */
    /*   leap year calculations are easier if we actually calculate a   */
    /*   day count from a more 'logical' date (1-JAN-1601) and then     */
    /*   subtract the number of days from there to 17-NOV-1958.  1601   */
    /*   is chosen so that we correctly count a century for dates in    */
    /*   1701 on (i.e. the century in 1700) but not for dates before    */
    /*   that - using 1600 would not work.  We must first check,        */
    /*   obviously, that the year is at least 1601.             */

    /*   Let's sort out the number of days from the beginning of this   */
    /*   year first.                            */

    calc_date = days_tab[month-1];  /* Get days to start of month */
    calc_date += day;       /* Hence days to today */
    if  ((!leap_year)       /* If this is not a Leap Year */
            &&  (month > 2))       /* and it's March or later */
    {
        calc_date -= 1;     /* Discount 29th February */
    }

    /*   Sort out the year into Quadricenturies, Centuries, Quadyears   */
    /*   and Years.                             */

    calc_years = year - 1601;   /* Get Year since 1601 (must be +ve) */
    calc_qcnts = calc_years / 400;  /* Get number of Quadricenturies */
    calc_years -= (calc_qcnts * 400); /* And discount them */
    calc_cents = calc_years / 100;  /* Get number of Centuries */
    calc_years -= (calc_cents * 100); /* And discount them */
    calc_qyrs = calc_years / 4; /* Get number of Quadyears */
    calc_years -= (calc_qyrs * 4);  /* And discount them */

    /*   We now add up the number of days accounted for by the      */
    /*   Quadricenturies, the Centuries, the Quadyears and the Years to */
    /*   get a final total.                         */

    calc_date += (calc_years * 365); /* Add in Number of days for Years */
    calc_date += (calc_qyrs * QYD);  /* Plus number of days for Quadyears */
    calc_date += (calc_cents * CYD); /* Plus number of days for Centuries */
    calc_date += (calc_qcnts * QCD); /* Plus number of days for Quadricenturies */

    /*   Lastly we adjust for our base offset (17-NOV-1858) and return  */
    /*   the value to the caller.                       */

    *date_ptr = calc_date - BASOFF; /* Allow for base date */
    return PSL_NORMAL;      /* And return success */
}


/************************************************************************/
/*                                  */
/*  PSL_COMPRESS_TIME - Create PSL Time from Constituent Parts  */
/*                                  */
/*  This routine creates a PSL Time from a Year, Month and Day.     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_COMPRESS_TIME (hours, mins, secs, msecs, time_ptr);*/
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret     = Unsigned Long PSL Return Status           */
/*          = PSL_BADPARAM if supplied time is invalid      */
/*     hours    = Unsigned Long Hours               */
/*     mins     = Unsigned Long Minutes             */
/*     secs     = Unsigned Long Seconds             */
/*     msecs    = Unsigned Long Milli-seconds           */
/*     time_ptr = Pointer to Unsigned Long to hold PSL Time     */
/*                                  */
/************************************************************************/

unsigned long PSL_COMPRESS_TIME (unsigned long hours,
                                 unsigned long mins,
                                 unsigned long secs,
                                 unsigned long msecs,
                                 unsigned long *time_ptr)
{

    /*   Local Data   */

    Boolean    rtnsts;     /* Routine Return Status */

    /*   The first thing to do is to check that we have been given a    */
    /*   valid time.                            */

    rtnsts = VALIDATE_TIME (hours, mins, secs, msecs);
    /* Validate the time */
    if  (!rtnsts) return PSL_BADPARAM;
    /* Exit with error if invalid */

    /*   Just do the numbers and return */

    *time_ptr = (hours * 60 * 60 * 1000)
                /* First the hours */
                + (mins * 60 * 1000)  /* then the minutes */
                + (secs * 1000)   /* then the seconds */
                + msecs;      /* and lastly the milli-seconds */
    return PSL_NORMAL;      /* Return Success */
}


/************************************************************************/
/*                                  */
/*  PSL_EXPAND_DATE - Expand PSL Date into Constituent Parts    */
/*                                  */
/*  This routine expands a PSL Date into Year, Month and Day, and   */
/*  also calculates the weekday.                    */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_EXPAND_DATE (psl_date, year_ptr, month_ptr,   */
/*                 day_ptr, wkdy_ptr);          */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret      = Unsigned Long PSL Return Status (always success) */
/*     psl_date  = Unsigned Long PSL Date to expand         */
/*     year_ptr  = Pointer to Unsigned Long to hold Year        */
/*     month_ptr = Pointer to Unsigned Long to hold Month       */
/*     day_ptr   = Pointer to Unsigned Long to hold Day     */
/*     wkdy_ptr  = Pointer to Unsigned Long to hold Weekday     */
/*                                  */
/************************************************************************/

unsigned long   PSL_EXPAND_DATE (unsigned long psl_date,
                                 unsigned long *year_ptr,
                                 unsigned long *month_ptr,
                                 unsigned long *day_ptr,
                                 unsigned long *weekday_ptr)
{

    /*   Local Data   */

    unsigned long       calc_date;      /* Date being calculated */
    unsigned long       calc_qcnts;     /* Calculated Quadricenturies */
    unsigned long       calc_cents;     /* Calculated Centuries */
    unsigned long       calc_qyrs;      /* Calculated Quadyears */
    unsigned long       calc_years;     /* Calculated Years */
    unsigned long       calc_month;     /* Calculated Month */
    Boolean    leap_year;      /* Leap Year Flag */

    /*   This code is basically the inverse of the code in          */
    /*   PSL_COMPRESS_DATE, and probably makes most sense that way      */
    /*   round.  Basically we reset our date to be number of days since */
    /*   1st January 1601, and then calculate how many quadricenturies, */
    /*   centuries, quadyears, years, months and days that represents.  */
    /*   The calculation is done this way to ensure that Leap Years are */
    /*   correctly catered for.                     */

    /*   Firstly we adjust our date to an unsigned day count since      */
    /*   17/11/1858 and hence into an unsigned day count since      */
    /*   1/1/1601.                              */

    calc_date = psl_date + BASOFF;  /* Convert to days since 1/1/1601 */

    /*   We now sort out the number of days accounted for by the        */
    /*   Quadricenturies, the Centuries, the Quadyears and the Years.   */
    /*   Note that 31/12/1601 is counted as 0 years 365 days, rather    */
    /*   than 1 year 0 days, so to get the divides to work OK we must   */
    /*   decrement the total day count by 1 and then add it back at the */
    /*   end.  We must also cope with 31st December in a Leap Year as a */
    /*   special case.                          */

    calc_date -= 1;         /* Decrement Day Count */
    calc_qcnts = (calc_date / QCD); /* Calculate number of Quadricenturies */
    calc_date -= (calc_qcnts * QCD);/* And discount them */
    calc_cents = (calc_date / CYD); /* Calculate number of Centuries */
    calc_date -= (calc_cents * CYD);/* And discount them */
    calc_qyrs = (calc_date / QYD);  /* Calculate number of Quadyears */
    calc_date -= (calc_qyrs * QYD); /* And discount them */
    calc_years = (calc_date / 365); /* Calculate number of Years */
    calc_date -= (calc_years * 365);/* And discount them */
    calc_date += 1;         /* Reset day to 1 - 365 */

    /*   Because of the vagaries of leap-years, QCD is slightly more    */
    /*   than 4*CYD and QYD is slightly more than 4*365.  This means    */
    /*   that we have awkward boundary conditions whereby we could get  */
    /*   a result of 4 for either CALC_CENTS or CALC_YEARS, which isn't */
    /*   really valid.  We ripple this through so that we end up with   */
    /*   valid values for all the rest, but with 366 days for the       */
    /*   number of days.                            */

    if  (calc_cents == 4)       /* If we had 4 centuries */
    {
        calc_cents = 3;     /* Reset it to 3 centuries */
        calc_qyrs = 24;     /* + 24 quadyears */
        calc_years = 4;     /* + 4 years */
    }

    if  (calc_years == 4)       /* If we had 4 years */
    {
        calc_years = 3;     /* Reset to 3 years */
        calc_date = 366;        /* And 366 days */
    }

    /*   Now see if this Year is a Leap Year */

    if  ((calc_years == 3)      /* If Year divisible by 4 */
            &&  ((calc_qyrs != 24)     /* and not divisible by 100 */
                 ||   (calc_cents == 3)))   /* or also divisible by 400 */
    {
        leap_year = PTRUE;      /* Then it's a Leap Year */
    }
    else                /* Otherwise */
    {
        leap_year = PFALSE;     /* It isn't */
    }

    /*   Now calculate (and return) the real year number */

    *year_ptr = calc_years      /* Real Year is years */
                + (calc_qyrs * 4) /* plus quadyears */
                + (calc_cents * 100)  /* plus centuries */
                + (calc_qcnts * 400)  /* plus quadricenturies */
                + 1601;       /* plus baseline date */

    /*   We now have the number of days since the start of the year and */
    /*   a Leap Year flag.  We will dispose of the latter first by      */
    /*   adding a day if a non-leap year and after 28th February (Day   */
    /*   59) - we can then treat all years as if February had 29 days.  */

    if  ((!leap_year)       /* If not a leap year */
            &&  (calc_date > 59))      /* and after day 59 */
    {
        calc_date += 1;     /* Add on another day */
    }

    /*   Now search the days-per-month table to find the month number */

    for (calc_month = 11;       /* Start with December */
            calc_date <= days_tab[calc_month];
            /* While before this month */
            calc_month--)      /* Stepping Backwards */
    {
    }

    /*   We now have the month number and hence can calculate the day   */
    /*   number and return them both to the caller.             */

    *month_ptr = calc_month + 1;    /* Return Month */
    *day_ptr = calc_date - days_tab[calc_month];
    /* And day in month */

    /*   Lastly we need to calculate the weekday corresponding to this  */
    /*   date. We know what weekday the base date was so we just need   */
    /*   to add that on to the number of days since then, divide by     */
    /*   seven, and take the remainder.                 */

    calc_date = (psl_date + BAS_WKDY) / 7;
    /* Number of weeks since base date */
    *weekday_ptr = (psl_date + BAS_WKDY) - (calc_date * 7);
    /* Hence weekday number */

    /*   All done - exit with success */

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*  PSL_EXPAND_TIME - Expand PSL Time into Constituent Parts    */
/*                                  */
/*  This routine expands a PSL Time into Hours, Minutes, Seconds    */
/*  also Milli-seconds.                     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_EXPAND_TIME (psl_time, hours_ptr, mins_ptr,   */
/*                 secs_ptr, msecs_ptr);        */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret      = Unsigned Long PSL Return Status          */
/*           = PSL_BADPARAM if supplied time is invalid     */
/*     psl_time  = Unsigned Long PSL Time to expand         */
/*     hours_ptr = Pointer to Unsigned Long to hold Hours       */
/*     mins_ptr  = Pointer to Unsigned Long to hold Minutes     */
/*     secs_ptr  = Pointer to Unsigned Long to hold Seconds     */
/*     msecs_ptr = Pointer to Unsigned Long to hold Milli-seconds   */
/*                                  */
/************************************************************************/

unsigned long   PSL_EXPAND_TIME (unsigned long psl_time,
                                 unsigned long *hours_ptr,
                                 unsigned long *mins_ptr,
                                 unsigned long *secs_ptr,
                                 unsigned long *msecs_ptr)
{

    /*   Local Data   */

    unsigned long       calc_time;      /* Time being calculated */

    /*   The first thing to do is to check that we have been given a    */
    /*   valid time.                            */

    if  (psl_time >= (24 * 60 * 60 * 1000)) return PSL_BADPARAM;
    /* Exit with error if invalid */

    /*   We just work out each of the values in turn by dividing by the */
    /*   appropriate factor.                        */

    calc_time = psl_time;       /* Copy time to work variable */
    *hours_ptr = calc_time/(60 * 60 * 1000);
    /* Get number of hours */
    calc_time -= (*hours_ptr * (60 * 60 * 1000));
    /* And discount them */
    *mins_ptr = calc_time/(60 * 1000);
    /* Get number of minutes */
    calc_time -= (*mins_ptr * (60 * 1000));
    /* And discount them */
    *secs_ptr = calc_time/1000; /* Get number of seconds */
    *msecs_ptr = calc_time - (*secs_ptr * 1000);
    /* And milli-seconds */
    return PSL_NORMAL;      /* And return success */
}


/************************************************************************/
/*                                  */
/*  PSL_FMT_DATTIM - Format a PSL Date and/or Time          */
/*                                  */
/*  This routine formats a PSL Date and/or a PSL Time into a    */
/*  Character Buffer in the format requested by the caller.     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_FMT_DATTIM (psl_date, psl_time, buf_ptr, buflen,  */
/*                   dattyp, flags);            */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret      = Unsigned Long PSL Return Status          */
/*           = PSL_BADPARAM if supplied time is invalid     */
/*     psl_date  = Unsigned Long PSL Date to format         */
/*     psl_time  = Unsigned Long PSL Time to format         */
/*     buf_ptr   = Pointer to Buffer to hold formatted date     */
/*     buflen    = Length of Buffer pointed to by BUF_PTR       */
/*     dattyp    = Unsigned Long Date Type              */
/*     flags     = Unsigned Long Flags controlling conversion   */
/*                                  */
/************************************************************************/

unsigned long   PSL_FMT_DATTIM (unsigned long psl_date,
                                unsigned long psl_time,
                                char *buf_ptr,
                                unsigned long buflen,
                                unsigned long dattyp,
                                unsigned long flags)
{

    /*   Local Data   */

    unsigned long       iret;       /* Routine return status */

    /*   The first thing to do is to check that we have been given a    */
    /*   valid time.  (All PSL dates are valid.)                */

    if  (psl_time >= (24 * 60 * 60 * 1000)) return PSL_BADPARAM;
    /* Exit with error if invalid */

    /*   Now case on the date type and call the appropriate routine to  */
    /*   handle the formatting.                     */

    switch (dattyp)         /* Case on date type */
    {
    case PSDTYP_SYSTEM:     /* If Default System Type */
        {
            iret = PSL_FMT_SYSDAT (psl_date, psl_time,
                                   buf_ptr, buflen, flags);
            /* Call platform-specific routine */
            break;          /* And carry on */
        }
    case PSDTYP_PSL:        /* If Standard PSL Type */
        {
            iret = FORMAT_PSLDAT (psl_date, psl_time,
                                  buf_ptr, buflen, flags);
            /* Call local routine */
            break;          /* And carry on */
        }

    case PSDTYP_CTM:        /* If CTM Type */
        {
            iret = FORMAT_CTMDAT (psl_date, psl_time,
                                  buf_ptr, buflen, flags);
            /* Call local routine */
            break;          /* And carry on */
        }

    default:            /* Owt else */
        {
            iret = PSL_BADPARAM;    /* Reject with Error */
        }
    }               /* End of (Date Type switch) */

    /*   Now just return the final status to the caller */

    return iret;            /* Return status to caller */
}


/************************************************************************/
/*                                  */
/*  PSL_GET_DATTIM - Get current Date and Time          */
/*                                  */
/*  This routine retrieves the current Date and Time in PSL Date    */
/*  and Time format.  Note that it only provides a granularity  */
/*  of one second.                          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     PSL_GET_DATTIM (date_ptr, time_ptr);             */
/*                                  */
/*  Where:                              */
/*                                  */
/*     date_ptr = Pointer to Unsigned Long to hold PSL Date     */
/*     time_ptr = Pointer to Unsigned Long to hold PSL Time     */
/*                                  */
/************************************************************************/

void   PSL_GET_DATTIM (unsigned long *date_ptr,
                       unsigned long *time_ptr)
{

    /*   Local Data   */

    struct   tm     *tm_ptr;        /* Pointer to tm buffer */
    time_t     cur_time;       /* Current Time in internal format */

    /*   Get the current date and time via the standard routines */

    cur_time = time(NULL);      /* Get Current Time */
    tm_ptr = localtime(&cur_time);  /* Convert to tm structure */

    /*   Convert the time into milli-seconds since midnight */

    (void) PSL_COMPRESS_TIME ((*tm_ptr).tm_hour,
                              (*tm_ptr).tm_min,
                              (*tm_ptr).tm_sec,
                              0,
                              time_ptr);

    /*   Convert the date into days since Base Date via local routine */

    (void) PSL_COMPRESS_DATE ((*tm_ptr).tm_year + 1900, 1, 1, date_ptr);

    /* Calculate for 1st January */
    *date_ptr += (*tm_ptr).tm_yday; /* And add in number of days since */

    /*   All done - just return */

    return;
}


/************************************************************************/
/*                                  */
/*  PSL_PARSE_DATTIM - Parse an ASCII Date and/or Time      */
/*                                  */
/*  This routine parses a buffer containing an ASCII Date and Time  */
/*  into a PSL Date and Time value.                 */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_PARSE_DATTIM (buf_ptr, date_ptr, time_ptr,    */
/*                  dattyp, inflags, outflags_ptr); */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Return Status       */
/*              = PSL_BADPARAM if supplied date/time is invalid */
/*     buf_ptr  = Pointer to null-terminated string to Parse    */
/*     date_ptr = Pointer to Unsigned Long to hold PSL Date */
/*     time_ptr = Pointer to Unsigned Long to hold PSL Time */
/*     dattyp       = Unsigned Long Date Type           */
/*     inflags  = Unsigned Long flags related to the input  */
/*     outflags_ptr = Pointer to Unsigned Long to hold flags    */
/*            defining what was parsed.         */
/*                                  */
/************************************************************************/

unsigned long   PSL_PARSE_DATTIM (const char *buf_ptr,
                                  unsigned long *date_ptr,
                                  unsigned long *time_ptr,
                                  unsigned long dattyp,
                                  unsigned long inflags,
                                  unsigned long *outflags_ptr)
{

    /*   Local Data   */

    unsigned long       iret;       /* Routine return status */

    /*   Just case on the date type and call the appropriate routine to */
    /*   handle the formatting.                     */

    switch (dattyp)         /* Case on date type */
    {
    case PSDTYP_SYSTEM:     /* If Default System Type */
        {
            iret = PSL_PARSE_SYSDAT (
                       /* Call platform-specific routine */
                       buf_ptr,
                       date_ptr,
                       time_ptr,
                       inflags,
                       outflags_ptr);
            break;          /* And carry on */
        }
    case PSDTYP_PSL:        /* If Standard PSL Type */
        {
            iret = PARSE_PSLDAT (   /* Call local routine */
                       buf_ptr,
                       date_ptr,
                       time_ptr,
                       inflags,
                       outflags_ptr);
            break;          /* And carry on */
        }

    case PSDTYP_SQL:        /* If SYBASE SQL Type */
        {
            iret = PARSE_SQLDAT (   /* Call local routine */
                       (char *) buf_ptr,
                       date_ptr,
                       time_ptr,
                       inflags,
                       outflags_ptr);
            break;          /* And carry on */
        }

    default:            /* Owt else */
        {
            iret = PSL_BADPARAM;    /* Reject with Error */
        }
    }               /* End of (Date Type switch) */

    /*   Now just return the final status to the caller */

    return iret;            /* Return status to caller */
}





/************************************************************************/
/*                                  */
/*  FORMAT_CTMDAT - Format a PSL Date and/or Time in CTM Format */
/*                                  */
/*  This routine formats a PSL Date and/or a PSL Time into a    */
/*  Character Buffer in the CTM format.             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = FORMAT_CTMDAT (psl_date, psl_time,            */
/*                  buf_ptr, buflen, flags);    */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret      = Unsigned Long PSL Return Status          */
/*           = PSL_BADPARAM if supplied time is invalid     */
/*     psl_date  = Unsigned Long PSL Date to format         */
/*     psl_time  = Unsigned Long PSL Time to format         */
/*     buf_ptr   = Pointer to Buffer to hold formatted date     */
/*     buflen    = Length of Buffer pointed to by BUF_PTR       */
/*     flags     = Unsigned Long Flags controlling conversion   */
/*                                  */
/*  The approach this routine uses is to format the information */
/*  into a local buffer and then copy that back to the user.    */
/*  This makes it much easier to sort out truncating the output,    */
/*  if required.                            */
/*                                  */
/************************************************************************/

static unsigned long FORMAT_CTMDAT (unsigned long psl_date,
                                    unsigned long psl_time,
                                    char *buf_ptr,
                                    unsigned long buflen,
                                    unsigned long flags)
{

    /*   Fields required by PSL_BINASC */

    unsigned char       radix=10;       /* Radix */
    unsigned short      binasc_flags;   /* Conversion Flags */
    unsigned char       numchr;     /* Number of characters */

    /*   Local Data   */

    unsigned long       year;       /* Year Field */
    unsigned long       month;      /* Month Field */
    unsigned long       day;        /* Day Field */
    unsigned long       weekday;        /* Weekday Field */
    unsigned long       hours;      /* Hours Field */
    unsigned long       mins;       /* Minutes Field */
    unsigned long       secs;       /* Seconds Field */
    unsigned long       msecs;      /* Milli-seconds Field */
    unsigned long       iret;       /* PSL Routine Return Status */
    unsigned long       isub;       /* General subscript */
    char       wrk_buf [40];   /* Temporary Work Buffer */
    char       *wrk_ptr;       /* Pointer to output buffer */
    char       *month_ptr;     /* Pointer to month string */
    unsigned long       formatted_len;  /* Length of formatted data */

    /*   Use the standard PSL routines to expand the PSL date and time */

    (void) PSL_EXPAND_DATE (    /* Expand PSL Date */
        psl_date,       /* from this date */
        &year,      /* returning year */
        &month,     /* ... and month */
        &day,       /* ... and day */
        &weekday);      /* ... and weekday */

    (void) PSL_EXPAND_TIME (    /* Expand PSL Time */
        psl_time,       /* from this time */
        &hours,     /* returning hours */
        &mins,      /* ... and minutes */
        &secs,      /* ... and seconds */
        &msecs);        /* ... and milli-seconds */

    /*   Initialize work buffer and pointer to it */

    (void) memset (wrk_buf, '\0', sizeof (wrk_buf));
    wrk_ptr = wrk_buf;

    /*   Insert the date if required */

    if ((flags & PSDFMT_DATE) == 0) goto format_time;
    /* Skip Date if not required */

    /*   Insert day - this requires 2 characters */

    numchr = 2;         /* Number of characters required */
    binasc_flags = 2;       /* Leading zeros required */
    iret = PSL_BINASC (&day,    /* Insert day */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   Insert month - this requires 3 characters */

    numchr = 3;         /* Number of characters required */
    month_ptr = month_asc_ctm + (numchr * (month-1));
    /* Point to month string */
    for (isub = 0;
            isub < numchr;
            isub++, wrk_ptr++, month_ptr++)
    {
        *wrk_ptr = *month_ptr;
    }

    /*   Insert year - this requires 2 or 4 characters, depending on    */
    /*   whether or not we want the Century.                */

    if ((flags & PSDFMT_CENTURY) != 0)
        /* If century required */
    {
        numchr = 4;         /* 4 characters needed */
    }
    else                /* Otherwise (century NOT required) */
    {
        numchr = 2;         /* 2 characters needed */
        year %= 100;        /* Convert to modulus 100 */
    }

    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&year,   /* Insert year */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   We have done the date - insert a separating underscore before the time */

    *wrk_ptr = '_';         /* Insert underscore */
    wrk_ptr += 1;           /* Point to next free character */

    /*   FORMAT_TIME */

    /*   We will branch here if the date was omitted, or fall through   */
    /*   here after formatting the date.  WRK_PTR points to the next    */
    /*   free character of the output buffer.               */

format_time:

    /*   Insert the time if required */

    if ((flags & PSDFMT_TIME) == 0) goto copy_output;
    /* Skip Time if not required */

    /*   Insert hours - this requires 2 characters */

    numchr = 2;         /* Number of characters required */
    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&hours,  /* Insert hours */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   Insert minutes - this requires 2 characters */

    numchr = 2;         /* Number of characters required */
    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&mins,   /* Insert minutes */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   Insert the second part if required */

    if ((flags & PSDFMT_NOSEC) != 0) goto copy_output;
    /* Skip Seconds if not required */

    /*   Insert seconds - this requires 2 characters */

    numchr = 2;         /* Number of characters required */
    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&secs,   /* Insert seconds */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   Insert the milli-second part if required */

    if ((flags & PSDFMT_MILLISEC) == 0) goto copy_output;
    /* Skip Milliseconds if not required */

    /*   Insert milliseconds - this requires up to 3 characters */

    numchr = 3;         /* Number of characters required */
    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&msecs,  /* Insert milli-seconds */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   COPY_OUTPUT */

    /*   We will branch or drop through to here after formatting the    */
    /*   information. We now have to copy the string to the user        */
    /*   buffer, truncating if required, and ensuring that it is null   */
    /*   terminated.                            */

copy_output:

    formatted_len = wrk_ptr - wrk_buf;
    /* Get length of formatted string */

    if (formatted_len < buflen) /* If user buffer longer than data */
    {
        (void) memcpy (buf_ptr, wrk_buf, formatted_len+1);
        /* Copy all data + terminating null */
    }
    else                /* Otherwise */
    {
        (void) memcpy (buf_ptr, wrk_buf, buflen-1);
        /* Copy as much as will fill buffer */
        /* ... allowing for terminating null */
        wrk_ptr = buf_ptr + buflen - 1;
        /* Point to last character in buffer */
        *wrk_ptr = '\0';        /* Null terminate */
    }

    /*   All done */

    iret = PSL_NORMAL;      /* Return Success */

end:
    return iret;
}




/************************************************************************/
/*                                  */
/*  FORMAT_PSLDAT - Format a PSL Date and/or Time in PSL Format */
/*                                  */
/*  This routine formats a PSL Date and/or a PSL Time into a    */
/*  Character Buffer in the standard PSL format.            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = FORMAT_PSLDAT (psl_date, psl_time,            */
/*                  buf_ptr, buflen, flags);    */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret      = Unsigned Long PSL Return Status          */
/*           = PSL_BADPARAM if supplied time is invalid     */
/*     psl_date  = Unsigned Long PSL Date to format         */
/*     psl_time  = Unsigned Long PSL Time to format         */
/*     buf_ptr   = Pointer to Buffer to hold formatted date     */
/*     buflen    = Length of Buffer pointed to by BUF_PTR       */
/*     flags     = Unsigned Long Flags controlling conversion   */
/*                                  */
/*  The approach this routine uses is to format the information */
/*  into a local buffer and then copy that back to the user.    */
/*  This makes it much easier to sort out truncating the output,    */
/*  if required.                            */
/*                                  */
/************************************************************************/

static unsigned long FORMAT_PSLDAT (unsigned long psl_date,
                                    unsigned long psl_time,
                                    char *buf_ptr,
                                    unsigned long buflen,
                                    unsigned long flags)
{

    /*   Fields required by PSL_BINASC */

    unsigned char       radix=10;       /* Radix */
    unsigned short      binasc_flags;   /* Conversion Flags */
    unsigned char       numchr;     /* Number of characters */

    /*   Local Data   */

    unsigned long       year;       /* Year Field */
    unsigned long       month;      /* Month Field */
    unsigned long       day;        /* Day Field */
    unsigned long       weekday;        /* Weekday Field */
    unsigned long       hours;      /* Hours Field */
    unsigned long       mins;       /* Minutes Field */
    unsigned long       secs;       /* Seconds Field */
    unsigned long       msecs;      /* Milli-seconds Field */
    unsigned long       iret;       /* PSL Routine Return Status */
    unsigned long       isub;       /* General subscript */
    char       wrk_buf [40];   /* Temporary Work Buffer */
    char       *wrk_ptr;       /* Pointer to output buffer */
    char       *month_ptr;     /* Pointer to month string */
    unsigned long       formatted_len;  /* Length of formatted data */

    /*   Use the standard PSL routines to expand the PSL date and time */

    (void) PSL_EXPAND_DATE (    /* Expand PSL Date */
        psl_date,       /* from this date */
        &year,      /* returning year */
        &month,     /* ... and month */
        &day,       /* ... and day */
        &weekday);      /* ... and weekday */

    (void) PSL_EXPAND_TIME (    /* Expand PSL Time */
        psl_time,       /* from this time */
        &hours,     /* returning hours */
        &mins,      /* ... and minutes */
        &secs,      /* ... and seconds */
        &msecs);        /* ... and milli-seconds */

    /*   Initialize work buffer and pointer to it */

    (void) memset (wrk_buf, '\0', sizeof (wrk_buf));
    wrk_ptr = wrk_buf;

    /*   Insert the date if required */

    if ((flags & PSDFMT_DATE) == 0) goto format_time;
    /* Skip Date if not required */

    /*   Insert day - this requires 2 characters */

    numchr = 2;         /* Number of characters required */
    binasc_flags = 0;       /* No special conversion flags */
    iret = PSL_BINASC (&day,    /* Insert day */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   Insert month (including leading '-') - this requires 4 characters */

    numchr = 4;         /* Number of characters required */
    month_ptr = month_asc + (numchr * (month-1));
    /* Point to month string */
    for (isub = 0;
            isub < numchr;
            isub++, wrk_ptr++, month_ptr++)
    {
        *wrk_ptr = *month_ptr;
    }

    /*   Insert a separating '-' */

    *wrk_ptr = '-';         /* Insert '-'*/
    wrk_ptr += 1;           /* Point to next free character */

    /*   Insert year - this requires 2 or 4 characters, depending on    */
    /*   whether or not we want the Century.                */

    if ((flags & PSDFMT_CENTURY) != 0)
        /* If century required */
    {
        numchr = 4;         /* 4 characters needed */
    }
    else                /* Otherwise (century NOT required) */
    {
        numchr = 2;         /* 2 characters needed */
        year %= 100;        /* Convert to modulus 100 */
    }

    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&year,   /* Insert year */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   We have done the date - insert a separating space before the time */

    *wrk_ptr = ' ';         /* Insert space */
    wrk_ptr += 1;           /* Point to next free character */

    /*   FORMAT_TIME */

    /*   We will branch here if the date was omitted, or fall through   */
    /*   here after formatting the date.  WRK_PTR points to the next    */
    /*   free character of the output buffer.               */

format_time:

    /*   Insert the time if required */

    if ((flags & PSDFMT_TIME) == 0) goto copy_output;
    /* Skip Time if not required */

    /*   Insert hours - this requires 2 characters */

    numchr = 2;         /* Number of characters required */
    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&hours,  /* Insert hours */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   Insert a separating colon */

    *wrk_ptr = ':';         /* Insert colon */
    wrk_ptr += 1;           /* Point to next free character */

    /*   Insert minutes - this requires 2 characters */

    numchr = 2;         /* Number of characters required */
    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&mins,   /* Insert minutes */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */

    /*   Insert a separating colon */

    *wrk_ptr = ':';         /* Insert colon */
    wrk_ptr += 1;           /* Point to next free character */

    /*   Insert seconds - this requires 2 characters */

    numchr = 2;         /* Number of characters required */
    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&secs,   /* Insert seconds */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */




    /*   Insert the milli-second part if required */

    if ((flags & PSDFMT_MILLISEC) == 0) goto copy_output;
    /* Skip Milliseconds if not required */

    /*   Insert a separating period */

    *wrk_ptr = '.';         /* Insert period */
    wrk_ptr += 1;           /* Point to next free character */

    /*   Insert milliseconds - this requires up to 3 characters */

    numchr = 3;         /* Number of characters required */
    binasc_flags = BINASC_FILL; /* Zero-fill number */
    iret = PSL_BINASC (&msecs,  /* Insert milli-seconds */
                       &numchr,        /* Number of characters */
                       &radix,     /* Radix */
                       wrk_ptr,        /* Output buffer */
                       &binasc_flags); /* Conversion Flags */
    if (!(psl_test_status (iret))) goto end;
    /* Exit if it failed */
    wrk_ptr += numchr;      /* Point to next free character */



    /*   COPY_OUTPUT */

    /*   We will branch or drop through to here after formatting the    */
    /*   information. We now have to copy the string to the user        */
    /*   buffer, truncating if required, and ensuring that it is null   */
    /*   terminated.                            */

copy_output:

    formatted_len = wrk_ptr - wrk_buf;
    /* Get length of formatted string */

    if (formatted_len < buflen) /* If user buffer longer than data */
    {
        (void) memcpy (buf_ptr, wrk_buf, formatted_len+1);
        /* Copy all data + terminating null */
    }
    else                /* Otherwise */
    {
        (void) memcpy (buf_ptr, wrk_buf, buflen-1);
        /* Copy as much as will fill buffer */
        /* ... allowing for terminating null */
        wrk_ptr = buf_ptr + buflen - 1;
        /* Point to last character in buffer */
        *wrk_ptr = '\0';        /* Null terminate */
    }

    /*   All done */

    iret = PSL_NORMAL;      /* Return Success */

end:
    return iret;
}


/************************************************************************/
/*                                  */
/*  PARSE_PSLDAT - Parse an ASCII Date and/or Time          */
/*                                  */
/*  This routine parses a buffer containing an ASCII Date and Time  */
/*  in the standard PSL format into a PSL Date and Time value.  */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PARSE_PSLDAT (buf_ptr, date_ptr, time_ptr,        */
/*              inflags, outflags_ptr);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Return Status       */
/*              = PSL_BADPARAM if supplied date/time is invalid */
/*     buf_ptr  = Pointer to null-terminated string to Parse    */
/*     date_ptr = Pointer to Unsigned Long to hold PSL Date */
/*     time_ptr = Pointer to Unsigned Long to hold PSL Time */
/*     inflags  = Unsigned Long flags related to the input  */
/*     outflags_ptr = Pointer to Unsigned Long to hold flags    */
/*            defining what was parsed.         */
/*                                  */
/************************************************************************/

static unsigned long PARSE_PSLDAT (const char *buf_ptr,
                                   unsigned long *psl_date_ptr,
                                   unsigned long *psl_time_ptr,
                                   unsigned long inflags,
                                   unsigned long *outflags_ptr)
{

    /*   Fields required by PSL_ASCBIN */

    unsigned char       buf_length;     /* Buffer Length */
    unsigned char       radix=10;       /* Radix */
    unsigned short      chrcnt;     /* Unconverted characters */

    /*   Local Data   */

    long       year;       /* Year Field */
    long       month;      /* Month Field */
    long       day;        /* Day Field */
    long       hours;      /* Hours Field */
    long       mins;       /* Minutes Field */
    long       secs;       /* Seconds Field */
    long       msecs;      /* Milli-seconds Field */

    char       *date_ptr;      /* Pointer to date string */
    char       month_name[5];  /* Local month name */
    char       *month_ptr;     /* Pointer to month in array */
    unsigned long       isub;       /* General subscript */
    unsigned long       iret;       /* PSL Routine Return Status */

    /*   First initialize the output flags to say that nothing was      */
    /*   parsed so that we can cop out whenever we want to.         */

    *outflags_ptr = 0;      /* No Fields Parsed */

    /*   In our generosity we will allow any ASCII string of the        */
    /*   general form:                          */
    /*                                  */
    /*  <gap> <date> <divider> <time> <gap>             */
    /*                                  */
    /*   Where each gap may contain zero or more spaces, the date and   */
    /*   time fields may be omitted, and the divider may be " ", ","    */
    /*   or ":", optionally surrounded by spaces, and is only required  */
    /*   if both date and time are specified.               */

    /*   First step over leading spaces to find the start of the date   */
    /*   field, and just exit if that's all there is.           */

    for (date_ptr = (char *) buf_ptr; *date_ptr == ' '; date_ptr++)
        /* Step over leading spaces */
    {
    }
    if  (*date_ptr == '\0') return PSL_NORMAL;
    /* Exit if that's all there is */

    /*   Check to see if we are allowed to parse a date.            */

    if ((inflags & PSDPRS_DATE) == 0) goto parse_time;
    /* Date not allowed, just parse time */

    /*   We know we have something, but we don't yet know if it's a     */
    /*   date or time field, so do a quick check for a '-' somewhere in */
    /*   the string before going any further.               */

    if  (strchr(date_ptr, '-') == NULL) goto parse_time;
    /* If no date, just parse time */

    /*   We know there is a date, and the rules say it must be here, so */
    /*   let's try to parse it.  The first bit should be a number and   */
    /*   must not be the only thing around.                 */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &day,    /* into day field */
                       &chrcnt);    /* returning character count */
    if  ((iret != PSL_UNCONVCHARS)  /* If we had an error */
            /* .. OR exhausted the string */
            ||  (day < 1))         /* or value was 0 or negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_DAY;    /* Show we parsed a day */

    /*   The next four characters should be '-' followed by a month     */
    /*   name.  We will check the first in hand and then look for all   */
    /*   four in our master array about (the leading '-' preventing any */
    /*   chance of overlap.  We need it in upper case for this, so we   */
    /*   copy it to a local buffer first.                   */

    if  (*date_ptr != '-') return PSL_BADPARAM;
    /* Error if next bit not '-' */
    for (isub = 0; isub < 4; isub ++)
        /* For next four characters */
    {
        month_name[isub] = toupper(*(date_ptr + isub));
        /* Copy over upper-case equivalent */
    }
    month_name[4] = '\0';       /* Null-terminate the string */
    month_ptr = strstr (month_asc, month_name);
    /* Scan array for it */
    if  (month_ptr == NULL) return PSL_BADPARAM;
    /* Return error if not found */
    month = ((month_ptr - month_asc) / 4) + 1;
    /* Else get month number (1-12) */
    if  (*(date_ptr + 4) != '-') return PSL_BADPARAM;
    /* Error if not following '-' as well */
    date_ptr += 5;          /* Else step over "-MMM-" */
    *outflags_ptr |= PSDFLD_MONTH;  /* Show we parsed a month */

    /*   The next bit should be a year, which might or might not have a */
    /*   leading century.  If the resulting figure is in the range 0-99 */
    /*   we assume that it is 20th century and missing the leading 19.  */
    /*   Note however, that the user may not want to allow this.        */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &year,   /* into year field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (year < 0))        /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_YEAR;   /* Show we parsed a year */
    if  (year < 100)        /* If year was 0 - 99 */
    {
        if ((inflags & PSDPRS_CENTURY) != 0)
            /* If century must be present */
        {
            return PSL_BADPARAM;    /* then exit with error */
        }
        else            /* Else (century not required) */
        {

            /*
             *    If the user has requested a Y2K compliant date then we must be a bit more
             *    intelligent about our defaulting of the century. Anydate between 00 and
             *    38 will be treated as a 21st century date. Above 38 will be twentieth
             */
            if ((inflags & PSDPRS_Y2KCOMPLIANT) != 0)
            {
                if (year <= 38)
                {
                    year += 2000;
                }
                else
                {
                    year += 1900;
                }
            }

            else
            {
                year += 1900;       /* Assume 20th century */
            }
        }
    }
    else                /* Else (year > 99) */
    {
        *outflags_ptr |= PSDFLD_CENTURY;
        /* Show we parsed a century */
    }

    /*   We should now have all the date fields we need, so let's       */
    /*   convert them to a PSL date.  This will also validate the       */
    /*   various fields for us on the side.                 */

    iret = PSL_COMPRESS_DATE (year, month, day, psl_date_ptr);
    /* Convert to PSL date */
    if  (!psl_test_status(iret)) return iret;
    /* If an error, exit with it */

    /*   Check to see if we are allowed to parse a time. If not, there  */
    /*   is nothing else to do so just get out.             */

    if ((inflags & PSDPRS_TIME) == 0) goto end;
    /* Time not allowed, just exit */

    /*   Next we have a divider.  We don't need to get too paranoid     */
    /*   about this as most ambiguities will result in errors       */
    /*   elsewhere, so we'll just check that the next character is a    */
    /*   space or a divider and then ignore spaces, divider and spaces. */

    if  ((*date_ptr != ' ')     /* If next character not a space */
            &&  (*date_ptr != ':')     /* and not a colon */
            &&  (*date_ptr != ','))    /* and not a comma */
    {
        return PSL_BADPARAM;    /* Return an Error */
    }

    for ( ; *date_ptr == ' '; date_ptr++)
        /* Step over leading spaces */
    {
    }

    if  ((*date_ptr == ':')     /* If a colon */
            ||  (*date_ptr == ','))    /* or a comma */
    {
        date_ptr += 1;      /* Step over it */
        for ( ; *date_ptr == ' '; date_ptr++)
            /* Step over trailing spaces */
        {
        }
    }
    if  (*date_ptr == '\0') return PSL_NORMAL;
    /* Exit if that's all there is */

    /*   PARSE_TIME */

    /*   We will branch here if the date was omitted, or fall through   */
    /*   here after parsing the date.  DATE_PTR points to the time part */
    /*   of the string.                         */

parse_time:

    /*   Check to see if we are allowed to parse a time.            */

    if ((inflags & PSDPRS_TIME) == 0) goto end;
    /* Time not allowed, just exit */

    /*   The next bit should be a number representing the hours */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &hours,  /* into hours field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (hours < 0))       /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_HOUR;   /* Show we parsed an hour */
    if  (*date_ptr != ':')      /* If next character isn't a colon */
    {
        mins = 0;           /* Set minutes to zero */
        secs = 0;           /* And seconds */
        msecs = 0;          /* And milli-seconds */
        goto convert_time;      /* And convert what we have */
    }
    date_ptr += 1;          /* Else step over ':' */

    /*   The next bit should be a number representing the minutes */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &mins,   /* into minutes field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (mins < 0))        /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_MINUTE; /* Show we parsed some minutes */
    if  (*date_ptr != ':')      /* If next character isn't a colon */
    {
        secs = 0;           /* Set seconds to zero */
        msecs = 0;          /* And milli-seconds */
        goto convert_time;      /* And convert what we have */
    }
    date_ptr += 1;          /* Else step over ':' */

    /*   The next bit should be a number representing the seconds */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &secs,   /* into seconds field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (secs < 0))        /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_SECOND; /* Show we parsed some seconds */
    if  (*date_ptr != '.')      /* If next character isn't a dot */
    {
        msecs = 0;          /* Set milli-seconds to zero */
        goto convert_time;      /* And convert what we have */
    }
    date_ptr += 1;          /* Else step over '.' */

    /*   The next bit should be a number representing the           */
    /*   milli-seconds.  Note that these are actually specified as a    */
    /*   fraction so we need to allow for 1, 2 or 3 digits.         */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    if  (buf_length > 3) buf_length = 3;
    /* Only allow 3 at most */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &msecs,  /* into milli-seconds field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (msecs < 0))       /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    if  ((buf_length - chrcnt) == 1) /* If only one digit */
    {
        msecs *= 100;       /* We had tenths */
    }
    else if  ((buf_length - chrcnt) == 2)
        /* Else if only two digits */
    {
        msecs *= 10;        /* We had hundredths */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_MSECS;  /* Show we parsed some milli-seconds */

    /*   CONVERT_TIME */

    /*   We branch here when we have finished parsing the time string,  */
    /*   and just convert it to the PSL time.               */

convert_time:
    iret = PSL_COMPRESS_TIME (hours, mins, secs, msecs, psl_time_ptr);
    /* Convert to PSL time */
    if  (!psl_test_status(iret)) return iret;
    /* If an error, exit with it */

    /*   All done OK - return success */

end:
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*  VALIDATE_DATE - Validate a Date supplied as Constituent Parts   */
/*                                  */
/*  This routine validates a trio of Year, Month and Day, and   */
/*  indicates whether or not the year is a leap year.       */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     rtnsts = VALIDATE_DATE (year, month, day, leapyear_ptr); */
/*                                  */
/*  Where:                              */
/*                                  */
/*     rtnsts       = Boolean return status: PTRUE if date is valid */


/*     year         = Unsigned Long Year                */
/*     month        = Unsigned Long Month               */
/*     day          = Unsigned Long Day             */
/*     leapyear_ptr = Pointer to Boolean for leap year flag     */
/*                                  */
/************************************************************************/

static Boolean VALIDATE_DATE (unsigned long year,
                              unsigned long month,
                              unsigned long day,
                              Boolean *leapyear_ptr)
{

    /*   First do the simple checks, viz that the month is between 1    */
    /*   and 12 inclusive, that the day is not zero and that the year   */
    /*   is at least BASYEAR.                       */

    if  ((day == 0)         /* If day is illegal */
            ||  (month == 0)       /* or month is illegal */
            ||  (month > 12)
            ||  (year < BASYEAR))      /* or year is illegal */
    {
        return PFALSE;      /* return an error */
    }

    /*   Next do the Leap Year calculation as we need this to validate  */
    /*   the day field precisely.  The rules for this are slightly      */
    /*   devious.  A year is a leap year if it is divisible by 4,       */
    /*   unless it is divisible by 100 (when it isn't) except for those */
    /*   which are divisble by 400 (which are).             */

    if  ((year & 3) != 0)       /* If Year not divisible by 4 */
    {
        *leapyear_ptr = PFALSE; /* can't be leapyear */
    }
    else if (year != ((year / 100) * 100))
        /* If divisble by 4, but not by 100 */
    {
        *leapyear_ptr = PTRUE;  /* then must be leapyear */
    }
    else if (year != ((year / 400) * 400))
        /* If divisble by 100, but not by 400 */
    {
        *leapyear_ptr = PFALSE; /* isn't leapyear */
    }
    else                /* Else (divisible by 400) */
    {
        *leapyear_ptr = PTRUE;  /* then is leapyear */
    }

    /*   Now we need to sort out if the day is valid for the month in   */
    /*   question We check the basic number of days per month, then     */
    /*   check for a leap year and allow for 29th February as well.     */

    if  ((day <= month_days[month-1]) /* If day is in normal range */
            ||  ((*leapyear_ptr == PTRUE)  /* or we have a leap year */
                 &&   (month == 2)      /* and it's February */
                 &&   (day == 29)))     /* and the 29th */
    {
        return PTRUE;       /* then the date is valid */
    }
    else                /* Otherwise */
    {
        return PFALSE;      /* else the date is invalid */
    }
}


/************************************************************************/
/*                                  */
/*  VALIDATE_TIME - Validate a Time supplied as Constituent Parts   */
/*                                  */
/*  This routine validates a quartet of Hour, Minutes, Seconds and  */
/*  Milli-seconds.                          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     rtnsts = VALIDATE_TIME (hours, mins, secs, msecs);       */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret     = Unsigned Long PSL Return Status           */
/*          = PSL_BADPARAM if supplied time is invalid      */
/*     hours    = Unsigned Long Hours               */
/*     mins     = Unsigned Long Minutes             */
/*     secs     = Unsigned Long Seconds             */
/*                                  */
/************************************************************************/

static Boolean VALIDATE_TIME (unsigned long hours,
                              unsigned long mins,
                              unsigned long secs,
                              unsigned long msecs)
{

    /*   Just check the valid ranges */

    if  ((hours < 24)       /* If hours are 0 - 23 */
            &&  (mins < 60)        /* and minutes are 0-59 */
            &&  (secs < 60))       /* and seconds are 0-59 */
    {
        return PTRUE;       /* return success */
    }
    else                /* Otherwise */
    {
        return PFALSE;      /* return an error */
    }
}


/************************************************************************/
/*                                  */
/*  PARSE_SQLDAT - Parse an SQL Server format Date and Time     */
/*                                  */
/*  This routine parses a buffer containing an ASCII Date and Time  */
/*  in the standard SQL server format - YYMMDD HH:MM:SS             */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PARSE_SQLDAT (buf_ptr, date_ptr, time_ptr,        */
/*              inflags, outflags_ptr);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Return Status       */
/*              = PSL_BADPARAM if supplied date/time is invalid */
/*     buf_ptr  = Pointer to null-terminated string to Parse    */
/*     date_ptr = Pointer to Unsigned Long to hold PSL Date */
/*     time_ptr = Pointer to Unsigned Long to hold PSL Time */
/*     inflags  = Unsigned Long flags related to the input  */
/*     outflags_ptr = Pointer to Unsigned Long to hold flags    */
/*            defining what was parsed.         */
/*                                  */
/************************************************************************/

static unsigned long PARSE_SQLDAT (char *buf_ptr,
                                   unsigned long *psl_date_ptr,
                                   unsigned long *psl_time_ptr,
                                   unsigned long inflags,
                                   unsigned long *outflags_ptr)
{

    /*   Fields required by PSL_ASCBIN */

    unsigned char       buf_length;     /* Buffer Length */
    unsigned char       radix=10;       /* Radix */
    unsigned short      chrcnt;     /* Unconverted characters */

    /*   Local Data   */

    long       year;       /* Year Field */
    long       month;      /* Month Field */
    long       day;        /* Day Field */

    char        *date_ptr;      /* Pointer to date string */
    unsigned long       iret;       /* PSL Routine Return Status */
    unsigned long       time_flags;

    /*   First initialize the output flags to say that nothing was      */
    /*   parsed so that we can cop out whenever we want to.         */

    *outflags_ptr = 0;      /* No Fields Parsed */

    date_ptr = buf_ptr;

    /*   We know there is a date, and it should be of the format YYYYMMDD, so
     *   let's try to parse it.
     */
    buf_length = 4;         /* Allow 2 characters       */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &year,   /* into year field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (year < 0))        /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_YEAR;   /* Show we parsed a year */

    /*
     *  Now Parse the month.
     */
    buf_length = 2;     /* Allow 2 characters */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &month,  /* into month field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (month < 1))       /* or value was 0 or negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_MONTH;  /* Show we parsed a month */

    /*
     *  Now Parse the days.
     */
    buf_length = 2;     /* Allow 2 characters */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &day,    /* into day field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (day < 1))         /* or value was 0 or negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_DAY;    /* Show we parsed a day */


    /*   We should now have all the date fields we need, so let's       */
    /*   convert them to a PSL date.  This will also validate the       */
    /*   various fields for us on the side.                 */

    iret = PSL_COMPRESS_DATE (year, month, day, psl_date_ptr);
    /* Convert to PSL date */
    if  (!psl_test_status(iret)) return iret;
    /* If an error, exit with it */

    /*   Next we have a divider.  We don't need to get too paranoid     */
    /*   about this as most ambiguities will result in errors       */
    /*   elsewhere, so we'll just check that the next character is a    */
    /*   space or a divider and then ignore spaces, divider and spaces. */

    for ( ; *date_ptr == ' '; date_ptr++)
        /* Step over leading spaces */
    {
    }

    /*
     *  Now parse the time in the starndard HH:MM:SS format.
     */
    iret = PARSE_PSLDAT (date_ptr, NULL, psl_time_ptr,
                         PSDPRS_TIME, &time_flags);
    if (!psl_test_status(iret)) return iret;
    *outflags_ptr |= time_flags;


    /*   All done OK - return success */

    return PSL_NORMAL;
}

