/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_DATE_RTNS_WIN32             */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This module contains the date and time routines that are specific    */
/*  to WNT.                                  */
/*                                       */
/*===========================================================================*/

/*   Define useful Include structures   */

#include  <psl_macros.h>        /* PSL Standard Macros */
#ifdef WIN32

#include <string.h>

/*   Define Function Prototypes for all external routines */
/*   used by the module (and for ourselves) */

#include psl_common_file (psl_rtndef.h) /* PSL Routines */
#include "psldate_intrtndef.h"      /* Internal Routines */

/*   Define Data Structures */

#include psl_common_file (shrmsg.h) /* Message Statuses */
#include psl_common_file (psldate_prmdef.h)
/* Routine Parameters */
#include <time.h>           /* C Time Functions */


/************************************************************************/
/*                                  */
/*  PSL_FMT_SYSDAT - Format Date and/or Time in System Format   */
/*                                  */
/*  This routine formats a PSL Date and/or a PSL Time into a    */
/*  Character Buffer in the standard WNT system format.     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_FMT_SYSDAT (date, time, buf_ptr, buflen, flags);  */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret      = Unsigned Long PSL Return Status          */
/*           = PSL_BADPARAM if supplied time is invalid     */
/*     date      = Unsigned Long PSL Date to format         */
/*     time      = Unsigned Long PSL Time to format         */
/*     buf_ptr   = Pointer to Buffer to hold formatted date     */
/*     buflen    = Length of Buffer pointed to by BUF_PTR       */
/*     flags     = Unsigned Long Flags controlling conversion   */
/*                                  */
/************************************************************************/

unsigned long   PSL_FMT_SYSDAT (unsigned long psl_date,
                                unsigned long psl_time,
                                char *buf_ptr,
                                unsigned long buflen,
                                unsigned long flags)
{
    /*   Local Data   */

    unsigned long       year;       /* Year Field */
    unsigned long       month;      /* Month Field */
    unsigned long       day;        /* Day Field */
    unsigned long       weekday;        /* Weekday Field */
    unsigned long       hours;      /* Hours Field */
    unsigned long       mins;       /* Minutes Field */
    unsigned long       secs;       /* Seconds Field */
    unsigned long       msecs;      /* Milli-seconds Field */
    struct   tm     tm;         /* tm buffer */
    unsigned long       date_len;       /* Length thereof */
    char       *format_ptr;    /* Format string pointer */

    /*   Use the standard PSL routines to expand the PSL date and time */

    (void) PSL_EXPAND_DATE (    /* Expand PSL Date */
        psl_date,       /* from this date */
        &year,      /* returning year */
        &month,     /* ... and month */
        &day,       /* ... and day */
        &weekday);      /* ... and weekday */

    (void) PSL_EXPAND_TIME (    /* Expand PSL Time */
        psl_time,       /* from this time */
        &hours,     /* returning hours */
        &mins,      /* ... and minutes */
        &secs,      /* ... and seconds */
        &msecs);        /* ... and milli-seconds */

    /*   Set up tm buffer */

    tm.tm_isdst = -1;       /* Information Not available */

    tm.tm_hour = hours;     /* Copy Hours */
    tm.tm_min = mins;       /* Copy Minutes */
    tm.tm_sec = secs;       /* Copy Seconds */

    tm.tm_year = year - 1900;   /* Copy Year */
    tm.tm_mon = month - 1;      /* Copy Month */
    tm.tm_mday = day;       /* Copy Day of Month */
    tm.tm_yday = 0;         /* Err, not sure if this is required */

    if (((flags & PSDFMT_DATE) != 0)
            && ((flags & PSDFMT_TIME) != 0))
        /* If both date and time wanted */
    {
        format_ptr = "%c";
    }
    else if ((flags & PSDFMT_DATE) != 0)
        /* If just date wanted */
    {
        format_ptr = "%x";
    }
    else                /* Else just time */
    {
        format_ptr = "%X";
    }
    date_len = strftime (buf_ptr, buflen, format_ptr, &tm);
    /* Format the date/time */
    return PSL_NORMAL;
}

/************************************************************************/
/*                                  */
/*                                  */
/*  PSL_PARSE_SYSDAT - Parse an ASCII Date and/or Time      */
/*                                  */
/*  This routine parses a buffer containing an ASCII Date and Time  */
/*  in the standard WNT system format into a PSL Date and Time  */
/*  value.                              */
/*  Note that this routine expects the date/time in a similar   */
/*  format to the PSL format, except that the month must be     */
/*  specified as defined for the abbreviated month name in the  */
/*  current locale (case insensitive).              */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = PSL_PARSE_SYSDAT (buf_ptr, psl_date_ptr, psl_time_ptr,*/
/*                  inflags, outflags_ptr);     */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret         = Unsigned Long PSL Return Status       */
/*              = PSL_BADPARAM if supplied date/time is invalid */
/*     buf_ptr  = Pointer to null-terminated string to Parse    */
/*     psl_date_ptr = Pointer to Unsigned Long to hold PSL Date */
/*     psl_time_ptr = Pointer to Unsigned Long to hold PSL Time */
/*     inflags  = Unsigned Long flags related to the input  */
/*     outflags_ptr = Pointer to Unsigned Long to hold flags    */
/*            defining what was parsed.         */
/*                                  */
/************************************************************************/

unsigned long   PSL_PARSE_SYSDAT (const char *buf_ptr,
                                  unsigned long *psl_date_ptr,
                                  unsigned long *psl_time_ptr,
                                  unsigned long inflags,
                                  unsigned long *outflags_ptr)
{
    /*   Fields required by PSL_ASCBIN */

    unsigned char       buf_length;     /* Buffer Length */
    unsigned char       radix=10;       /* Radix */
    unsigned short      chrcnt;     /* Unconverted characters */

    /*   Local Data   */

    long       year;       /* Year Field */
    long       month;      /* Month Field */
    long       day;        /* Day Field */
    long       hours;      /* Hours Field */
    long       mins;       /* Minutes Field */
    long       secs;       /* Seconds Field */
    long       msecs;      /* Milli-seconds Field */

    char       *date_ptr;      /* Pointer to date string */
    unsigned long       iret;       /* PSL Routine Return Status */

    /*   First initialize the output flags to say that nothing was      */
    /*   parsed so that we can cop out whenever we want to.         */

    *outflags_ptr = 0;      /* No Fields Parsed */

    /*   In our generosity we will allow any ASCII string of the        */
    /*   general form:                          */
    /*                                  */
    /*  <gap> <date> <divider> <time> <gap>             */
    /*                                  */
    /*   Where each gap may contain zero or more spaces, the date and   */
    /*   time fields may be omitted, and the divider may be " ", ","    */
    /*   or ":", optionally surrounded by spaces, and is only required  */
    /*   if both date and time are specified.               */

    /*   First step over leading spaces to find the start of the date   */
    /*   field, and just exit if that's all there is.           */

    for (date_ptr = (char *) buf_ptr; *date_ptr == ' '; date_ptr++)
        /* Step over leading spaces */
    {
    }
    if  (*date_ptr == '\0') goto end;
    /* Exit if that's all there is */

    /*   Check to see if we are allowed to parse a date.            */

    if ((inflags & PSDPRS_DATE) == 0) goto parse_time;
    /* Date not allowed, just parse time */

    /*   We know we have something, but we don't yet know if it's a     */
    /*   date or time field, so do a quick check for a '/' somewhere in */
    /*   the string before going any further.               */

    if  (strchr(date_ptr, '/') == NULL) goto parse_time;
    /* If no date, just parse time */

    /*   We know there is a date, and the rules say it must be here, so */
    /*   let's try to parse it.  The first bit should be a number and   */
    /*   must not be the only thing around.                 */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &day,    /* into day field */
                       &chrcnt);    /* returning character count */
    if  ((iret != PSL_UNCONVCHARS)  /* If we had an error */
            /* .. OR exhausted the string */
            ||  (day < 1))         /* or value was 0 or negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_DAY;    /* Show we parsed a day */

    /*   The next string should be '/' followed by a month,                     */

    if  (*date_ptr != '/') return PSL_BADPARAM;
    /* Error if next bit not '/' */
    date_ptr += 1;          /* Step over the '/' */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &month,  /* into month field */
                       &chrcnt);    /* returning character count */
    if  ((iret != PSL_UNCONVCHARS)  /* If we had an error */
            /* .. OR exhausted the string */
            ||  (month < 1))           /* or value was 0 or negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_MONTH;  /* Show we parsed a month */

    /*   The next bit should be a year, which might or might not have a */
    /*   leading century.  If the resulting figure is in the range 0-99 */
    /*   we assume that it is 20th century and missing the leading 19.  */
    /*   Note however, that the user may not want to allow this.        */

    date_ptr += 1;          /* Step over the '-' */
    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &year,   /* into year field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (year < 0))        /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_YEAR;   /* Show we parsed a year */
    if  (year < 100)        /* If year was 0 - 99 */
    {
        if ((inflags & PSDPRS_CENTURY) != 0)
            /* If century must be present */
        {
            return PSL_BADPARAM;    /* then exit with error */
        }
        else            /* Else (century not required) */
        {
            /*
             *    If the user has requested a Y2K compliant date then we must be a bit more
             *    intelligent about our defaulting of the century. Anydate between 00 and
             *    38 will be treated as a 21st century date. Above 38 will be twentieth
             */
            if (((inflags & PSDPRS_Y2KCOMPLIANT) != 0) && (year <= 38))
            {
                year += 2000;
            }
            else
            {
                year += 1900;
            }

        }
    }
    else                /* Else (year > 99) */
    {
        *outflags_ptr |= PSDFLD_CENTURY;
        /* Show we parsed a century */
    }

    /*   We should now have all the date fields we need, so let's       */
    /*   convert them to a PSL date.  This will also validate the       */
    /*   various fields for us on the side.                 */

    iret = PSL_COMPRESS_DATE (year, month, day, psl_date_ptr);
    /* Convert to PSL date */
    if  (!psl_test_status(iret)) return iret;
    /* If an error, exit with it */

    /*   Check to see if we are allowed to parse a time. If not, there  */
    /*   is nothing else to do so just get out.             */

    if ((inflags & PSDPRS_TIME) == 0) goto end;
    /* Time not allowed, just exit */
    if  (*date_ptr == '\0') goto end;
    /* Exit if that's all there is */

    /*   Next we have a divider.  We don't need to get too paranoid     */
    /*   about this as most ambiguities will result in errors       */
    /*   elsewhere, so we'll just check that the next character is a    */
    /*   space or a divider and then ignore spaces, divider and spaces. */

    if  ((*date_ptr != ' ')     /* If next character not a space */
            &&  (*date_ptr != ':')     /* and not a colon */
            &&  (*date_ptr != ','))    /* and not a comma */
    {
        return PSL_BADPARAM;    /* Return an Error */
    }

    for ( ; *date_ptr == ' '; date_ptr++)
        /* Step over leading spaces */
    {
    }

    if  ((*date_ptr == ':')     /* If a colon */
            ||  (*date_ptr == ','))    /* or a comma */
    {
        date_ptr += 1;      /* Step over it */
        for ( ; *date_ptr == ' '; date_ptr++)
            /* Step over trailing spaces */
        {
        }
    }
    if  (*date_ptr == '\0') goto end;
    /* Exit if that's all there is */

    /*   PARSE_TIME */

    /*   We will branch here if the date was omitted, or fall through   */
    /*   here after parsing the date.  DATE_PTR points to the time part */
    /*   of the string.                         */

parse_time:

    /*   Check to see if we are allowed to parse a time.            */

    if ((inflags & PSDPRS_TIME) == 0) goto end;
    /* Time not allowed, just exit */

    /*   The next bit should be a number representing the hours */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &hours,  /* into hours field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (hours < 0))       /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_HOUR;   /* Show we parsed an hour */
    if  (*date_ptr != ':')      /* If next character isn't a colon */
    {
        mins = 0;           /* Set minutes to zero */
        secs = 0;           /* And seconds */
        msecs = 0;          /* And milli-seconds */
        goto convert_time;      /* And convert what we have */
    }
    date_ptr += 1;          /* Else step over ':' */

    /*   The next bit should be a number representing the minutes */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &mins,   /* into minutes field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (mins < 0))        /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_MINUTE; /* Show we parsed some minutes */
    if  (*date_ptr != ':')      /* If next character isn't a colon */
    {
        secs = 0;           /* Set seconds to zero */
        msecs = 0;          /* And milli-seconds */
        goto convert_time;      /* And convert what we have */
    }
    date_ptr += 1;          /* Else step over ':' */

    /*   The next bit should be a number representing the seconds */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &secs,   /* into seconds field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (secs < 0))        /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_SECOND; /* Show we parsed some seconds */
    if  (*date_ptr != '.')      /* If next character isn't a dot */
    {
        msecs = 0;          /* Set milli-seconds to zero */
        goto convert_time;      /* And convert what we have */
    }
    date_ptr += 1;          /* Else step over '.' */

    /*   The next bit should be a number representing the           */
    /*   milli-seconds.  Note that these are actually specified as a    */
    /*   fraction so we need to allow for 1, 2 or 3 digits.         */

    buf_length = strlen(date_ptr);  /* Get length of string remaining */
    if  (buf_length > 3) buf_length = 3;
    /* Only allow 3 at most */
    iret = PSL_ASCBIN (&buf_length, /* Try to convert a number */
                       date_ptr,    /* from here */
                       &radix,  /* in decimal */
                       &msecs,  /* into milli-seconds field */
                       &chrcnt);    /* returning character count */
    if  (((!psl_test_status(iret))  /* If we had an error */
            &&   (iret != PSL_UNCONVCHARS)) /* other than the normal */
            ||  (msecs < 0))       /* or number was negative */
    {
        return PSL_BADPARAM;    /* then exit with error */
    }
    if  ((buf_length - chrcnt) == 1) /* If only one digit */
    {
        msecs *= 100;       /* We had tenths */
    }
    else if  ((buf_length - chrcnt) == 2)
        /* Else if only two digits */
    {
        msecs *= 10;        /* We had hundredths */
    }
    date_ptr += (buf_length - chrcnt);
    /* Step over number */
    *outflags_ptr |= PSDFLD_MSECS;  /* Show we parsed some milli-seconds */

    /*   CONVERT_TIME */

    /*   We branch here when we have finished parsing the time string,  */
    /*   and just convert it to the PSL time.               */

convert_time:
    iret = PSL_COMPRESS_TIME (hours, mins, secs, msecs, psl_time_ptr);
    /* Convert to PSL time */
    if  (!psl_test_status(iret)) return iret;
    /* If an error, exit with it */

    /*   All done OK - return success */

end:
    return PSL_NORMAL;
}
#endif  /* WIN32 */
