/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*/
/*                                       */
/*               **** C O M M O N ****                   */
/*                                       */
/*===========================================================================*/
/*                                       */
/*          Module Name: PSL_FORMAT_DUMP                 */
/*          ============                         */
/*                                       */
/* Description:                                  */
/* ============                                  */
/*                                       */
/*  This routine formats the contents of a buffer into a 'side-by-side'  */
/*  display identical to that produced by the DUMP utility.          */
/*                                       */
/*===========================================================================*/


#include  <psl_macros.h>        /* PSL Standard Macros */

/*    Define useful Include structures   */

#include  <stdio.h>         /* Standard I/O */
#include  <string.h>            /* String functions */
#include  <ctype.h>         /* Character functions */

/*   Define Function Prototypes for all routines used by the module */

#include  psl_common_file(psl_rtndef.h) /* Define ourselves */

void FORMAT_ASCII (char *inp_ptr,
                   unsigned long bytcnt,
                   char *out_ptr);

void FORMAT_DATA (char *inp_ptr,
                  unsigned long bytcnt,
                  unsigned long outseg,
                  unsigned long segsiz,
                  unsigned long flags,
                  char *out_ptr);

void FORMAT_OFFSET (unsigned long bufoff,
                    unsigned long bufadr,
                    unsigned long flags,
                    char *out_ptr);

/*   Symbolic Values */

#define FMT_OCTAL   1
#define FMT_EXPADD  2


/************************************************************************/
/*                                  */
/*  PSL_FORMAT_DUMP - Format a string into 'dump' format        */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     Call  PSL_FORMAT_DUMP (buf_ptr, buflen, bufoff, bufadr   */
/*                out_ptr, outlen, flags)       */
/*                                  */
/*  Where:                              */
/*                                  */
/*     buf_ptr = Pointer to Input Buffer                */
/*     buflen  = Length of Input Buffer             */
/*     bufoff  = Input Buffer Offset                */
/*     bufadr  = Input Buffer Address               */
/*     out_ptr = Pointer to Output Buffer               */
/*     outlen  = Length of Output Buffer                */
/*     flags   = Control Flags:                 */
/*           1 => Output in Octal (default = Hexadecimal)   */
/*           2 => Include Address & Expand Offset       */
/*                                  */
/*                                  */
/*  This routine formats data in a 'Side-by-Side' format:       */
/*                                  */
/*  n Longwords in Hexadecimal (or Octal)               */
/*  n*4 ASCII Characters with non-printing chars suppressed     */
/*  Offsets (Absolute and/or Relative) (in Hexadecimal or Octal)    */
/*                                  */
/*  or the exact opposite for machines with byte-ordering opposite  */
/*  to that of the VAX/Alpha.                   */
/*                                  */
/************************************************************************/

void PSL_FORMAT_DUMP (char *inp_ptr,
                      unsigned short buflen,
                      unsigned long bufoff,
                      unsigned long bufadr,
                      char *out_ptr,
                      unsigned short outlen,
                      unsigned long flags)
{
    /*   Local Data */

    unsigned long       outseg;     /* Number of output segments */
    unsigned long       segsiz;     /* Size of each segment */
    unsigned long       segovr;     /* Overhead in each line */
    unsigned long       bytcnt;     /* Count of bytes to process */
    char      *dat_ptr;        /* Pointer to data display */
    char      *asc_ptr;        /* Pointer to ASCII display */
    char      *off_ptr;        /* Pointer to address display */

    /*    First see how many segments we have room for.  This depends   */
    /*    on the various flag settings.                 */

    if  ((flags & FMT_OCTAL) != 0)  /* If formatting in octal */
    {
        segsiz = 12;        /* Each segment is 12 bytes */
    }
    else                /* Else (hexadecimal) */
    {
        segsiz = 9;         /* Each segment is 9 bytes */
    }
    if  ((flags & FMT_EXPADD) != 0) /* If expanded format */
    {
        segovr = 21;        /* Overhead is 21 bytes */
    }
    else                /* Else (normal) */
    {
        segovr = 8;         /* Overhead is 8 bytes */
    }
    outseg = (outlen - segovr) / (segsiz + 4);
    /* Calculate number of chunks */

    /*    Hence calculate the offset of each section of output. */

#ifdef  PSL_LITTLE_ENDIAN       /* If right to left */
    dat_ptr = out_ptr;      /* Start with data */
    asc_ptr = dat_ptr + (outseg * segsiz) + 1;
    /* Then the ASCII bits */
    off_ptr = asc_ptr + (outseg * 4) + 1;
    /* Lastly the offset */
#else   /* PSL_LITTLE_ENDIAN */     /* Else (left to right) */
off_ptr = out_ptr;      /* Start with offset */
    asc_ptr = off_ptr + (segovr - 1);
    /* Then the ASCII bits */
    dat_ptr = asc_ptr + (outseg * 4) + 1;
    /* Lastly the data */
#endif  /* PSL_LITTLE_ENDIAN */

    /*   Next initialize the whole buffer to spaces */

    (void) memset (out_ptr, ' ', outlen);

    /*    Next see if we have a full set of values.  If we have fewer   */
    /*    than a full set of bytes, then we need to adjust for the      */
    /*    first part.  This involves skipping a designated number of    */
    /*    spaces and possibly doing some special processing for the     */
    /*    first segment (if it is incomplete).              */

    if  (buflen >= (outseg * 4))    /* If at least enough */
    {
        bytcnt = outseg * 4;    /* Only output as many as will fit */
    }
    else                /* Else (fewer than maximum) */
    {
        bytcnt = buflen;        /* Set number to output */
    }

    /*   First output the data in octal/hexadecimal */

    FORMAT_DATA (inp_ptr, bytcnt, outseg, segsiz, flags, dat_ptr);

    /*   Next we output the data in ASCII */

    FORMAT_ASCII (inp_ptr, bytcnt, asc_ptr);

    /*   Lastly we output the address and/or the offset of the current  */
    /*   buffer.                                */

    FORMAT_OFFSET (bufoff, bufadr, flags, off_ptr);

    /*   All done - just return to the caller */

    return;
}

/************************************************************************/
/*                                  */
/*  FORMAT_ASCII - Format the data as ASCII characters      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     FORMAT_ASCII (inp_ptr, bytcnt, out_ptr);         */
/*                                  */
/*  Where:                              */
/*                                  */
/*     inp_ptr = Pointer to Input Buffer                */
/*     bytcnt  = Count of bytes to display              */
/*     out_ptr = Pointer to Output Buffer               */
/*                                  */
/*  This routine formats the ASCII data part of the line.       */
/*                                  */
/************************************************************************/

void FORMAT_ASCII (char *inp_ptr,
                   unsigned long bytcnt,
                   char *out_ptr)
{
    /*   Local Data */

    unsigned long       isub;       /* General loop variable */

    /*   This routine outputs the text in ASCII, replacing any      */
    /*   non-printing characters with periods.              */

    for (isub = 0; isub < bytcnt; isub++)
        /* For each byte in the buffer */
    {
        if  (isprint(*inp_ptr)) /* If a printing character */
        {
            *out_ptr = *inp_ptr;    /* Just copy it over */
            out_ptr += 1;       /* And increment the pointers */
            inp_ptr += 1;
        }
        else            /* Else (non-printing) */
        {
            *out_ptr = '.';     /* Else insert a period */
            out_ptr += 1;       /* And increment the pointer */
            inp_ptr += 1;       /* And step past input character */
        }
    }

    /*    All done - just return to the caller */

    return;
}

#ifdef  PSL_LITTLE_ENDIAN
/************************************************************************/
/*                                  */
/*  FORMAT_DATA - Format the data as octal/hexadecimal      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     FORMAT_DATA (inp_ptr, bytcnt, outseg, segsiz,        */
/*          flags, out_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     inp_ptr = Pointer to Input Buffer                */
/*     bytcnt  = Count of bytes to display              */
/*     outseg  = Number of Segments to Output           */
/*     segsiz  = Width of data in each segment          */
/*     flags   = Control Flags:                 */
/*           1 => Output in Octal (default = Hexadecimal)   */
/*           2 => Include Address & Expand Offset       */
/*     out_ptr = Pointer to Output Buffer               */
/*                                  */
/*  This routine formats the octal/hexadecimal part of the line.    */
/*  This is the VAX-like version that displays from right to left.  */
/*                                  */
/************************************************************************/

void FORMAT_DATA (char *inp_ptr,
                  unsigned long bytcnt,
                  unsigned long outseg,
                  unsigned long segsiz,
                  unsigned long flags,
                  char *out_ptr)
{
    /*   Local Data */

    unsigned long       misseg;     /* Number of missing segments */
    unsigned long       misbyt;     /* Number of missing bytes */
    char       fmtbuf[20];     /* Interim buffer for formatting */
    long      *long_ptr;       /* Ditto as longwords */
    unsigned long       tempval;        /* Temporary octal value */
    static   char       octsiz[3] = {3, 5, 8};
    /* Widths of octal fragments */

    /*   See if we have fewer than a lineful of data and, if so, skip   */
    /*   over the first part.  Note that we must set odd bytes to 0 in  */
    /*   octal as a single byte spans multiple octal digits.        */

    inp_ptr += bytcnt;      /* Point after last character in input */
    if  (bytcnt < (outseg * 4)) /* If less than a full set */
    {
        misbyt = (outseg * 4) - bytcnt;
        /* Count number of missing bytes */
        misseg = misbyt / 4;    /* Hence number of missing segments */
        out_ptr += (misseg * segsiz);
        /* Step over them */
        misbyt -= (misseg * 4); /* And remove from count */
        if  (misbyt != 0) misbyt = 4 - misbyt;
        /* Hence number in first segment */
        bytcnt -= misbyt;       /* And adjust amount left to do */
        if  (misbyt != 0)       /* If some odd missing bytes */
        {
            if  ((flags & FMT_OCTAL) != 0)
                /* If formatting in octal */
            {
                tempval = 0;    /* Initialize temporary value */
                inp_ptr -= misbyt;  /* Decrement buffer pointer */
                (void) memcpy ((char *) &tempval, inp_ptr, misbyt);
                /* Copy in the missing bytes */
                out_ptr += 12 - octsiz[misbyt-1];
                /* Indent to appropriate byte */
                (void) sprintf (fmtbuf, "%0*lo", octsiz[misbyt-1], tempval);
                /* Format next byte */
                (void) memcpy (out_ptr, fmtbuf, octsiz[misbyt-1]);
                /* Copy to output buffer */
                out_ptr += octsiz[misbyt-1];
                /* And increment buffer pointer */
            }
            else            /* Else (hexadecimal) */
            {
                out_ptr += ((4 - misbyt) * 2) + 1;
                /* Indent to appropriate byte */
                for (; misbyt != 0; misbyt--)
                {
                    inp_ptr -= 1;   /* Decrement buffer pointer */
#ifdef  VMS
                    (void) sprintf (fmtbuf, "%02X", *inp_ptr);
                    /* Format next byte */
#else   /* VMS */
(void) sprintf (fmtbuf, "%02x", *inp_ptr);
                    /* Format next byte */
#endif  /* VMS */
                    (void) memcpy (out_ptr, fmtbuf, 2);
                    /* Copy to output buffer */
                    out_ptr += 2;   /* And increment buffer pointer */
                }           /* End of (for loop) */
            }           /* End of (if hex/octal) */
        }               /* End of (if odd missing bytes) */
    }               /* End of (if full set) */

    /*    Now format the main segments in hex/octal */

    for (; bytcnt != 0; bytcnt -= 4)
        /* For each four bytes */
    {
        inp_ptr -= 4;       /* Step to front of next longword */
        *out_ptr = ' ';     /* Insert a space */
        out_ptr += 1;       /* And increment the pointer */
        if  ((flags & FMT_OCTAL) != 0)
            /* If formatting in octal */
        {
            long_ptr = (void *) inp_ptr;
            /* Force Pointer to long * */
            (void) sprintf (fmtbuf, "%011lo", *long_ptr);
            /* Format next longword */
            (void) memcpy (out_ptr, fmtbuf, 11);
            /* Copy to output buffer */
            out_ptr += 11;      /* And increment buffer pointer */
        }
        else            /* Else (hexadecimal) */
        {
            long_ptr = (void *) inp_ptr;
            /* Force Pointer to long * */
#ifdef  VMS
            (void) sprintf (fmtbuf, "%08lX", *long_ptr);
            /* Format next longword */
#else   /* VMS */
            (void) sprintf (fmtbuf, "%08lx", *long_ptr);
            /* Format next longword */
#endif  /* VMS */
            (void) memcpy (out_ptr, fmtbuf, 8);
            /* Copy to output buffer */
            out_ptr += 8;       /* And increment buffer pointer */
        }
    }

    /*    All done - just return to the caller */

    return;
}

#else   /* PSL_LITTLE_ENDIAN */
/************************************************************************/
/*                                  */
/*  FORMAT_DATA - Format the data as octal/hexadecimal      */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     FORMAT_DATA (inp_ptr, bytcnt, outseg, segsiz,        */
/*          flags, out_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     inp_ptr = Pointer to Input Buffer                */
/*     bytcnt  = Count of bytes to display              */
/*     outseg  = Number of Segments to Output           */
/*     segsiz  = Width of data in each segment          */
/*     flags   = Control Flags:                 */
/*           1 => Output in Octal (default = Hexadecimal)   */
/*           2 => Include Address & Expand Offset       */
/*     out_ptr = Pointer to Output Buffer               */
/*                                  */
/*  This routine formats the octal/hexadecimal part of the line.    */
/*  This is the SUN-like version that displays from left to right.  */
/*                                  */
/************************************************************************/

void FORMAT_DATA (char *inp_ptr,
                  unsigned long bytcnt,
                  unsigned long outseg,
                  unsigned long segsiz,
                  unsigned long flags,
                  char *out_ptr)
{
    /*   Local Data */

    char       fmtbuf[20];     /* Interim buffer for formatting */
    long      *long_ptr;       /* Ditto as longwords */
    unsigned long       tempval;        /* Temporary octal value */
    static   char       octsiz[3] = {3, 5, 8};
    /* Widths of octal fragments */

    /*   First format the main segments in hex/octal */

    for (; bytcnt >= 4; bytcnt -= 4) /* For each four bytes */
    {
        if  ((flags & FMT_OCTAL) != 0)
            /* If formatting in octal */
        {
            long_ptr = (void *) inp_ptr;
            /* Force Pointer to long * */
            (void) sprintf (fmtbuf, "%011lo", *long_ptr);
            /* Format next longword */
            (void) memcpy (out_ptr, fmtbuf, 11);
            /* Copy to output buffer */
            out_ptr += 11;      /* And increment buffer pointer */
        }
        else            /* Else (hexadecimal) */
        {
            long_ptr = (void *) inp_ptr;
            /* Force Pointer to long * */
            (void) sprintf (fmtbuf, "%08lx", *long_ptr);
            /* Format next longword */
            (void) memcpy (out_ptr, fmtbuf, 8);
            /* Copy to output buffer */
            out_ptr += 8;       /* And increment buffer pointer */
        }
        inp_ptr += 4;       /* Step to front of next longword */
        *out_ptr = ' ';     /* Insert a space */
        out_ptr += 1;       /* And increment the pointer */
    }

    /*   See if we have any odd bytes left over to convert.  Note that  */
    /*   we must set odd bytes to 0 in octal as a single byte spans     */
    /*   multiple octal digits.                     */

    if  (bytcnt != 0)       /* If some left over */
    {
        if  ((flags & FMT_OCTAL) != 0)
            /* If formatting in octal */
        {
            tempval = 0;        /* Initialize temporary value */
            (void) memcpy ((char *) &tempval, inp_ptr, bytcnt);
            /* Copy in the missing bytes */
            (void) sprintf (fmtbuf, "%0*lo", octsiz[bytcnt-1], tempval);
            /* Format next byte */
            (void) memcpy (out_ptr, fmtbuf, octsiz[bytcnt-1]);
            /* Copy to output buffer */
            out_ptr += octsiz[bytcnt-1];
            /* And increment buffer pointer */
        }
        else            /* Else (hexadecimal) */
        {
            for (; bytcnt != 0; bytcnt--)
            {
                (void) sprintf (fmtbuf, "%02x", *inp_ptr);
                /* Format next byte */
                (void) memcpy (out_ptr, fmtbuf, 2);
                /* Copy to output buffer */
                inp_ptr += 1;   /* Decrement input buffer pointer */
                out_ptr += 2;   /* And increment output pointer */
            }           /* End of (for loop) */
        }               /* End of (if hex/octal) */
    }               /* End of (if odd missing bytes) */

    /*    All done - just return to the caller */

    return;
}
#endif  /* PSL_LITTLE_ENDIAN */

/************************************************************************/
/*                                  */
/*  FORMAT_OFFSET - Format the address &/or offset          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     FORMAT_OFFSET (bufoff, bufadr, flags, out_ptr);      */
/*                                  */
/*  Where:                              */
/*                                  */
/*     bufoff  = Input Buffer Offset                */
/*     bufadr  = Input Buffer Address               */
/*     flags   = Control Flags:                 */
/*           1 => Output in Octal (default = Hexadecimal)   */
/*           2 => Include Address & Expand Offset       */
/*     out_ptr = Pointer to Output Buffer               */
/*                                  */
/*                                  */
/*  This routine formats the address and/or offset part of the line */
/*                                  */
/************************************************************************/

void FORMAT_OFFSET (unsigned long bufoff,
                    unsigned long bufadr,
                    unsigned long flags,
                    char *out_ptr)
{
    /*   Local Data */

    char       fmtbuf[20];     /* Interim buffer for formatting */

    /*    This routine outputs either the address and offset of the     */
    /*    current buffer (both as eight digits), or the offset only (as */
    /*    six digits).                          */

    if  ((flags & FMT_EXPADD) != 0) /* If expanded format */
    {
        if  ((flags & FMT_OCTAL) != 0)
            /* If formatting in octal */
        {
            (void) sprintf (fmtbuf, "%08lo (%08lo)", bufadr, bufoff);
            /* Format next longword */
            (void) memcpy (out_ptr, fmtbuf, 19);
            /* Copy to output buffer */
        }
        else            /* Else (hexadecimal) */
        {
#ifdef  VMS
            (void) sprintf (fmtbuf, "%08lX (%08lX)", bufadr, bufoff);
            /* Format address and offset */
#else   /* VMS */
            (void) sprintf (fmtbuf, "%08lx (%08lx)", bufadr, bufoff);
            /* Format address and offset */
#endif  /* VMS */
            (void) memcpy (out_ptr, fmtbuf, 19);
            /* Copy to output buffer */
        }
    }
    else                /* Else (normal format) */
    {
        if  ((flags & FMT_OCTAL) != 0)
            /* If formatting in octal */
        {
            (void) sprintf (fmtbuf, "%06lo", bufoff);
            /* Format next longword */
            (void) memcpy (out_ptr, fmtbuf, 6);
            /* Copy to output buffer */
        }
        else            /* Else (hexadecimal) */
        {
#ifdef  VMS
            (void) sprintf (fmtbuf, "%06lX", bufoff);
            /* Format offset only */
#else   /* VMS */
            (void) sprintf (fmtbuf, "%06lx", bufoff);
            /* Format offset only */
#endif  /* VMS */
            (void) memcpy (out_ptr, fmtbuf, 6);
            /* Copy to output buffer */
        }
    }

    /*    All done - just return to the caller */

    return;
}
