/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
/*===========================================================================*/
/*									     */
/*			  **** C O M M O N ****   			     */
/*									     */
/*===========================================================================*/
/*									     */
/*			Module Name: PSL_ROOT_UNIX                   	     */
/*			============					     */
/*									     */
/*  Description:							     */
/*  ============							     */
/*									     */
/*	These routines are used, under UNIX, to return or concatenate        */
/*	the translation of the "psl_root" locations to the user.	     */
/*									     */
/*									     */
/*===========================================================================*/

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  This module is UNIX-Specific...
 */
#ifdef UNIX					/* On UNIX */

/*
 *  Standard ANSI header files.
 */
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <pwd.h>
#include <unistd.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>                         /* PSL Routines */
static void ADD_STR_TO_ROOT(const char *rootp,
                            const char *insp,
                            char *outsp,
                            size_t outs_size);
/*
 *  Structure Definitions.
 */
#include <psl_root.h>                           /* Root macros */
#include <shrmsg.h>                             /* PSL Messages */


/************************************************************************/
/*									*/
/*	PSL_GET_NODEDIR - Return node-specific directory            	*/
/*									*/
/*	Calling Convention:						*/
/*									*/
/*	   PSL_GET_NODEDIR (outbuf_ptr, outbuf_size)        		*/
/*									*/
/*	Where:								*/
/*									*/
/*	   outbuf_ptr  = Pointer to output string for result		*/
/*	   outbuf_size = Size of output string                       	*/
/*									*/
/*	This routine will return the current psl_root/nodename location	*/
/*									*/
/************************************************************************/
void PSL_GET_NODEDIR (char *outbuf_ptr,
                      unsigned long outbuf_size)
{
#ifdef HPV9							      
    char	 nodename[256];                 /* Nodename */

/*
 *  First get the current nodename - prefixed with a '/'.
 */
    nodename[0] = '/';
    (void) PSL_GET_NODENAME (NULL, &nodename[1], sizeof (nodename) -1);

/*
 *  Prefix this name with the psl_root location.
 */
    PSL_ADDVARROOT (nodename, outbuf_ptr, outbuf_size);

#else /* not HPV9 */
    PSL_ADDVARROOT("/tmp", outbuf_ptr, outbuf_size);
#endif /* not HP */
    return;
}


/************************************************************************/
/*									*/
/*	PSL_ADDROOT    - Prefix Specified String with PSL product root  */
/*      PSL_ADDVARROOT - Prefix specified string with PSL var root      */
/*      PSL_ADDETCROOT - Prefix specified string with PSL etc root      */
/*									*/
/*	Calling Convention:						*/
/*									*/
/*	   PSL_xxxROOT (insp, outsp, outs_size)          	        */
/*									*/
/*	Where:								*/
/*									*/
/*	   insp         = Pointer to string to be appended to PSL root	*/
/*	   outsp        = Pointer to output string for result		*/
/*			  May be NULL if just psl_root to be returned	*/
/*	   outs_size    = Size of output string                       	*/
/*									*/
/*	These routines will prefix the (location of) PSL product root   */
/*      directory to the specified string.                              */
/*									*/
/************************************************************************/
void PSL_ADDROOT (const char *insp,
                  char *outsp,
                  size_t outs_size)
{
    ADD_STR_TO_ROOT(PSL_ROOT_DIR, insp, outsp, outs_size);
    return;
}

void PSL_ADDVARROOT (const char *insp,
                     char *outsp,
                     size_t outs_size)
{
    ADD_STR_TO_ROOT(PSL_VAR_ROOT_DIR, insp, outsp, outs_size);
    return;
}

void PSL_ADDETCROOT (const char *insp,
                     char *outsp,
                     size_t outs_size)
{
    ADD_STR_TO_ROOT(PSL_ETC_ROOT_DIR, insp, outsp, outs_size);
    return;
}

/*
 * This function is basically the old PSL_ADDROOT with one extra parameter
 * and the code to obtain the root directory removed.
 */
/************************************************************************/
/*									*/
/*	ADD_STR_TO_ROOT - Add a string to the specified root directory  */
/*									*/
/*	Calling Convention:						*/
/*									*/
/*	   ADD_STR_TO_ROOT (rootp, insp, outsp, outs_size)          	*/
/*									*/
/*	Where:								*/
/*									*/
/*	   string_ptr  = Pointer to string to be appended to /psl_root	*/
/*	   outbuf_ptr  = Pointer to output string for result		*/
/*			 May be NULL if just psl_root to be returned	*/
/*	   outbuf_size = Size of output string                       	*/
/*									*/
/*	These routines will prefix the (location of) PSL product root   */
/*      directory to the specified string. It will use a hard-wired     */
/*      value depending on the platform.                                */
/*									*/
/************************************************************************/
static void ADD_STR_TO_ROOT(const char *rootp,
                            const char *insp,
                            char *outsp,
                            size_t outs_size)
{
    char                pslroot[512];		/* Work String */
    char                *wrk_ptr;		/* Work Pointer */
    unsigned long       wrk_len;		/* Length of Work String */


/*
 *  Save the root directory string and obtain its length.
 */
    (void) strcpy (pslroot, rootp);
    wrk_len = strlen (pslroot);

/*
 *  Copy as many characters as we can.
 */
    if (wrk_len < outs_size)
    {
	(void) memcpy (outsp, pslroot, wrk_len + 1);
    }
    else
    {
	(void) memcpy (outsp, pslroot, outs_size - 1);
	*(outsp + outs_size - 1) = '\0';
    }

/*
 *  How much have we left to copy.
 */
    if ((insp != NULL) &&
	(strlen (outsp) < (outs_size - 2)))
    {

/*
 *	Append the string.
 */
	wrk_ptr = strchr (outsp, '\0');
	wrk_len = wrk_ptr - outsp;			/* Space used */
	wrk_len = outs_size - wrk_len;		/* Space available */
	if (wrk_len > strlen (insp))
	{
	    wrk_len = strlen (insp) + 1;		/* Size of string */
	}

	(void) memcpy (wrk_ptr, insp, wrk_len - 1);
	*(wrk_ptr + wrk_len - 1) = '\0';
    }

    return;
}
/************************************************************************/
/*									*/
/*	psl_addroot_1 - Prefix Specified String with /psl_root (location)*/
/*									*/
/*	Calling Convention:						*/
/*									*/
/*	   psl_addroot_1 (string_ptr, outbuf_ptr, outbuf_size)          */
/*									*/
/*	Where:								*/
/*									*/
/*	   string_ptr  = Pointer to string to be appended to /psl_root	*/
/*	   outbuf_ptr  = Pointer to output string for result		*/
/*			 May be NULL if just psl_root to be returned	*/
/*	   outbuf_size = Size of output string                       	*/
/*									*/
/*	This routine will prefix the (location of) psl_root to the    	*/
/*	specified string.  It will first attempt to translate the	*/
/*	"PSL_ROOT" environment variable.  if that fails it will		*/
/*	use a hard-wired value of "/usr/psl_root".                     	*/
/*									*/
/************************************************************************/
void psl_addroot_1 (const char *string_ptr,                           
                    char *outbuf_ptr,
                    unsigned long outbuf_size)
{
/*
 *  Local Data
 */
	 char	 pslroot[256];		/* Work String */
	 char	 *wrk_ptr;		/* Work Pointer */
unsigned long	 wrk_len;		/* Length of Work String */

/*
 *  First attempt to translate the "PSL_ROOT" environment variable.
 */
    wrk_ptr = getenv ("PSL_ROOT");

/*
 *  Did we succeed?
 */
    if (wrk_ptr == NULL)
    {
	(void) strcpy (pslroot, "/usr/psl_root");	/* No - Hardwire */
    }
    else
    {
	(void) strcpy (pslroot, wrk_ptr); 		/* Yes - Get Result */
    }

/*
 *  Get the length.
 */
    wrk_len = strlen (pslroot);

/*
 *  Copy as many characters as we can.
 */
    if (wrk_len < outbuf_size)
    {
	(void) memcpy (outbuf_ptr, pslroot, wrk_len + 1);
    }
    else
    {
	(void) memcpy (outbuf_ptr, pslroot, outbuf_size - 1);
	*(outbuf_ptr + outbuf_size - 1) = '\0';
    }

/*
 *  How much have we left to copy.
 */
    if ((string_ptr != NULL) &&
	(strlen (outbuf_ptr) < (outbuf_size - 2)))
    {

/*
 *	Append the string.
 */
	wrk_ptr = strchr (outbuf_ptr, '\0');
	wrk_len = wrk_ptr - outbuf_ptr;			/* Space used */
	wrk_len = outbuf_size - wrk_len;		/* Space available */
	if (wrk_len > strlen (string_ptr))
	{
	    wrk_len = strlen (string_ptr) + 1;		/* Size of string */
	}

	(void) memcpy (wrk_ptr, string_ptr, wrk_len - 1);
	*(wrk_ptr + wrk_len - 1) = '\0';
    }

    return;
}


/************************************************************************/
/*									*/
/*	psl_get_nodedir_1 - Return node-specific directory            	*/
/*									*/
/*	Calling Convention:						*/
/*									*/
/*	   psl_get_nodedir_1 (outbuf_ptr, outbuf_size)        		*/
/*									*/
/*	Where:								*/
/*									*/
/*	   outbuf_ptr  = Pointer to output string for result		*/
/*	   outbuf_size = Size of output string                       	*/
/*									*/
/*	This routine will return the current psl_root/nodename location	*/
/*									*/
/************************************************************************/
void psl_get_nodedir_1 (char *outbuf_ptr,                             
                        unsigned long outbuf_size)
{
/*
 *  Local Data
 */
	 char	 nodename[256];		/* Nodename */

/*
 *  First get the current nodename - prefixed with a '/'.
 */
    nodename[0] = '/';
    (void) PSL_GET_NODENAME (NULL, &nodename[1], sizeof (nodename) -1);

/*
 *  Prefix this name with the psl_root location.
 */
    PSL_ADDROOT (nodename, outbuf_ptr, outbuf_size);

    return;
}

/************************************************************************/
/*									*/
/*	psl_get_psuser_1 - Return PSL username and UID        		*/
/*									*/
/*	Calling Convention:						*/
/*									*/
/*	   iret = psl_get_psuser_1 (diag_ptr, username_ptr, 		*/
/*				  username_size, uid_ptr, gid_ptr)	*/
/*									*/
/*	Where:								*/
/*									*/
/*	   iret	            = Unsigned Long PSL Routine Return Status	*/
/*	   diag_ptr         = Pointer to Diagnostics Stack		*/
/*	   username_ptr     = Pointer for username return		*/
/*	                      May be NULL if not required		*/
/*	   username_size    = Size of username string			*/
/*	   uid_ptr          = Pointer for UID return                  	*/
/*	   gid_ptr          = Pointer for GID return                  	*/
/*									*/
/*	This routine will return the User information selected during 	*/
/*	the installation. It will get this information from the owner	*/
/*	of the following file:						*/
/*									*/
/*	    <psl_root>/common_root					*/
/*									*/
/*      THIS FUNCTION IS NOW OBSOLETE - DO NOT USE IT                   */
/*									*/
/************************************************************************/
unsigned long	psl_get_psuser_1 (struct diagstk *diag_ptr,           
                                  char *username_ptr,
                                  unsigned long username_size,
                                  uid_t *uid_ptr,
                                  gid_t *gid_ptr)                     
{
    unsigned long       iret;			/* Routine Return Status */
    char                filename[256];		/* File name */
    long                status;                 /* Status from UNIX functions */
    unsigned long       wrk_len;		/* Length of String */
    struct stat         stat_buf;		/* stat() returned buffer */
    struct passwd       *passwd_ptr;		/* Pointer to Password Structure */

/*
 *  Build up the full data file name.
 */
    PSL_ADDROOT ("/common_root", filename, sizeof (filename));

/*
 *  Attempt to retrieve the statistics on the file (symlink).
 */
    status = stat (filename, &stat_buf);
    if (status == 0)			/* File exists - get the details */
    {
	*uid_ptr = stat_buf.st_uid;
	*gid_ptr = stat_buf.st_gid;				      

/*
 *	Do we require the username string?
 */
	if (username_ptr != NULL)
	{

/*
 *	    Get a password information structure for this UID.
 */
	    setpwent ();
	    passwd_ptr = getpwuid (stat_buf.st_uid);
	    if (passwd_ptr == NULL)
	    {
	        *username_ptr = '\0';		/* NULL Username */
	    }
	    else
	    {
	     
/*
 *	        Get the length of the username to copy.
 */
	        wrk_len = strlen ((*passwd_ptr).pw_name);

/*
 *	        Copy as many characters as we can.
 */
	        if (wrk_len < username_size)
	        {
		    (void) memcpy (username_ptr,
				    (*passwd_ptr).pw_name,
				    wrk_len + 1);
	        }
	        else
	        {
		    (void) memcpy (username_ptr,
			           (*passwd_ptr).pw_name,
				    username_size - 1);
		    *(username_ptr + username_size - 1) = '\0';
	        }
	    }

/*
 *	    Ensure that the password file is closed.
 */
	    endpwent ();
	}
    }
    else
    {
	iret = PSL_NOPSUSER;

	PSL_LOG_CIOERR (diag_ptr, iret, filename);
	goto end;
    }

/*
 *  Return success.
 */
    iret = PSL_NORMAL;

/*
 *  END:
 *
 *  At this point IRET will contain the final return status.
 */
end:
    return iret;
}

#endif		/* UNIX-Specific */
