/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*===========================================================================*
 *                                                                           *
 *    PSL_TARGET_RTNS.C                                  *
 *    ==================                             *
 *                                       *
 *    This module contains the routines that are responsible for maintaining *
 *    and distributing knowledge of the various test executers and target    *
 *    environments supported by the CYRANO Test product.             *
 *                                       *
 *===========================================================================*/

#include <stdio.h>
#include <string.h>

/*
 * PSL Macros.
 */
#include <psl_macros.h>
#include <psl_rtndef.h>

/*
 *    ...and our headers
 */
#include "target_rtndef.h"
#include "tofprmdef.h"

/*
 *   Define the structure that will hold the target details.
 */
struct Target
{
    char        Text[12];
    int     ExecuterID;
    char        Suffix[4];
    char        DisplayedText[24];
    int     Authorization;
    int     OtaAvailable;
    int     TofID;
};

/*
 *    Define the table of Targets. This is of primary use when determining the OTA modules and
 *    provides a mapping between the OTA targets and the actual executers.
 *
 *    Note:
 *      Parameters that define the indecis into this array are defined in the header file and 
 *      should be kept up to date with this list.
 */
static struct Target    TargetList[] =
    {"INVALID"      ,EXECUTER_UNKNOWN   ,"???"  ,"INVALID"  ,0  ,0  ,0
        ,"SYBASE"      ,EXECUTER_SYBASE    ,"syb"  ,"Sybase"   ,0  ,1  ,TOF_ID_SYBASE
        ,"ORACLE"      ,EXECUTER_ORACLE    ,"ora"  ,"Oracle"   ,0  ,1  ,TOF_ID_ORACLE
        ,"SQL-SERVER"      ,EXECUTER_SQLSERVER ,"mss"  ,"MS SQL Server",0  ,1  ,TOF_ID_SQLSERVER
        ,"ORACLE8"     ,EXECUTER_ORACLE8   ,"or8"  ,"Oracle 8" ,0  ,1  ,TOF_ID_ORACLE8
        ,"ORACLE-APP"      ,EXECUTER_ORACLEAPP ,"ora"  ,"Oracle Applications"  ,0 ,0   ,TOF_ID_ORACLEAPP
        ,"TUXEDO"      ,EXECUTER_TUXEDO    ,"tux"  ,"Tuxedo"   ,0  ,0  ,TOF_ID_TUXEDO
        ,"SAP"         ,EXECUTER_SAP       ,"sap"  ,"SAP"      ,0  ,0  ,TOF_ID_SAP
        ,"TCPIP"       ,EXECUTER_TCPIP     ,"tcp"  ,"TCP/IP"   ,0  ,0  ,TOF_ID_TCPIP
        ,"COMMAND"     ,EXECUTER_CMD       ,"cmp"  ,"Command"  ,0  ,0  ,TOF_ID_COMMAND
        ,"HTP"         ,EXECUTER_HTTP      ,"htp"  ,"Htp"      ,0  ,0  ,TOF_ID_HTP
        ,"WDT"      ,EXECUTER_HTTP_DT   ,"wdt"  ,"Wdt"      ,0  ,0  ,TOF_ID_WEB_DT
        ,"NTP"      ,EXECUTER_NTP   ,"ntp"  ,"ntp"      ,0  ,0  ,TOF_ID_NTP
        ,"SMP"      ,EXECUTER_SMP   ,"smp"  ,"smp"      ,0  ,0  ,TOF_ID_SMP
        ,"SYM"      ,EXECUTER_SYM   ,"sym"  ,"Sym"      ,0  ,0  ,TOF_ID_SYM
        ,"MMO"      ,EXECUTER_MMO   ,"mmo"  ,"Mmo"      ,0  ,0  ,TOF_ID_MMO
        ,"ORM"      ,EXECUTER_ORM   ,"orm"  ,"Orm"      ,0  ,0  ,TOF_ID_ORM
        ,"UXM"      ,EXECUTER_UXM   ,"uxm"  ,"Uxm"      ,0  ,0  ,TOF_ID_UXM
    };


/*
 *    Define the list of Executers that can be specified in the node definitions.
 */
static char*    Executers[] =
    {"CYRANO-TEST" ,"CYRANO-ROBOT" ,"WINDOWS", "CYRANO-OTA-SYBASE"
     ,"CYRANO-WEBTESTER" ,"CYRANO-OTA-ORACLE" ,"CYRANO-OTA-SQL-SERVER"};


/*
 *    Define counts of targets and executers.
 *    THIS MUST BE KEPT IN STEP WITH THE ABOVE LISTS
 */
static int      TargetCount = 17;
static int      ExecuterCount = 8;

/*
 *    Tgt_GetTargetID
 *    ===============
 *
 *    Return the numeric Target ID that corresponds to the supplied TARGET string
 *
 *    Calling format:
 *
 *      ID = Tgt_GetTargetID (TargetString);
 *
 *    Parameters:
 *      TargetString    Pointer to char string containing the target string as returned from a
 *                      previous call to GetTargetList(), GetDisplayTargetList(), GetTarget() or 
 *                      GetDisplayTarget()
 *
 *    Returns:
 *      0 based ID of the Target string supplied. 
 *      TARGET_UNKNOWN is returned if no such target exists
 */
enum TARGETID       Tgt_GetTargetID     (char* pTargetString)
{
    int isub;
    enum TARGETID TargetID = TARGET_UNKNOWN;

    /*
     *   Loop through the list of targets looking for the one supplied and return its index if found.
     */
    for (isub = (int)TARGET_SYBASE; isub < TargetCount; isub++)
    {
        if ((strcmp (pTargetString ,TargetList[isub].Text) == 0)
                ||  (strcmp (pTargetString ,TargetList[isub].DisplayedText) == 0))
        {
            TargetID = (enum TARGETID) isub;
            break;
        }
    }

    return TargetID;
}

/*
 *    Tgt_GetTarget
 *    =============
 *
 *    Get the target string from its ID.
 *
 *    Calling Format:
 *      status = Tgt_GetTarget (TargetID ,Buff ,BuffLen ,ReturnDisplayed);
 *
 *    Parameters:
 *      TargetID    - int 0 based TargetID in the range 0 - (TargetCount-1)
 *      Buff        - Pointer to char buffer to hold the returned target string
 *      BuffLen     - int length in bytes of the above buffer
 *      ReturnDisplayed - return the displayed target name not the internal target name
 *
 *    Returns:
 *      0           - Success
 *      -1          - No such target 
 */
int Tgt_GetTarget   (enum   TARGETID TargetID
                     ,       char*    pBuff
                     ,   int      BuffLen
                     ,   Boolean  ReturnDisplayed)
{
    char*   pTarget;
    int     CopyLen;

    if ((TargetID < TARGET_SYBASE) || ((int)TargetID > TargetCount)) return -1;

    /*
     *    Get a pointer to the required target string and copy as much as possible into the target 
     *    remembering to ensure that it is null terminated
     */
    if (ReturnDisplayed)
    {
        pTarget = TargetList[TargetID].DisplayedText;
    }
    else
    {
        pTarget = TargetList[TargetID].Text;
    }

    CopyLen = strlen (pTarget);

    if (CopyLen > (BuffLen - 1)) CopyLen = BuffLen - 1; /* restrict length if necessary     */

    /*
     *    ...and copy the target text into the buffer (only if required though)
     */
    if (pBuff != NULL)
    {
        (void) memcpy (pBuff ,pTarget ,CopyLen);
        /* ...and copy in data as required  */
        *(pBuff + CopyLen) = '\0';
        /* ...and null terminator           */
    }

    return 0;
}

/*
 *    Tgt_GetSuffix
 *    =============
 *
 *    Get the Suffix string from its ID.
 *
 *    Calling Format:
 *      status = Tgt_GetSuffix (TargetID ,Buff ,BuffLen);
 *
 *    Parameters:
 *      TargetID    - int 0 based TargetID in the range 0 - (TargetCount-1)
 *      Buff        - Pointer to char buffer to hold the returned suffix string
 *      BuffLen     - int length in bytes of the above buffer
 *
 *    Returns:
 *      0           - Success
 *      -1          - No such target 
 */
int     Tgt_GetSuffix       (enum TARGETID TargetID ,char* pBuff ,int BuffLen)
{
    char*   pTarget;
    int     CopyLen;

    if ((TargetID < TARGET_SYBASE) || ((int)TargetID > TargetCount)) return -1;

    /*
     *    Get a pointer to the required suffix string and copy as much as possible into the target 
     *    remembering to ensure that it is null terminated
     */
    pTarget = TargetList[TargetID].Suffix;
    CopyLen = strlen (pTarget);

    if (CopyLen > (BuffLen - 1)) CopyLen = BuffLen - 1; /* restrict length if necessary     */
    if (pBuff != NULL)
    {
        (void) memcpy (pBuff ,pTarget ,CopyLen);            /* ...and copy in data as required  */
        *(pBuff + CopyLen) = '\0';                          /* ...and null terminator           */
    }

    return 0;
}


/*
 *    Tgt_GetExecuterList
 *    ===================
 *
 *    return a list of the available executers, separated by nulls with a double null terminator
 *
 *    Calling Format:
 *      status = Tgt_GetExecuterList (Buff ,BuffLen);
 *
 *    Parameters:
 *      Buff        - Pointer to char buffer to hold the returned list of executers
 *      BuffLen     - int length in bytes of the above buffer
 *
 *    Returns:
 *      0           - Success
 *      -1          - Insufficient space in buffer
 */
int     Tgt_GetExecuterList (char* pBuff ,int BuffLen)
{
    int SpaceRequired;
    int isub;
    int Status = 0;                         /* Assume all will go well          */
    int SpaceLeft = BuffLen - 1;            /* Set up space available           */

    /*
     *   Loop through the list of executers and add each tgo the list if space is available.
     */
    for (isub = EXECUTER_CTEST; ((isub < ExecuterCount) && (Status != -1)); isub++)
    {

        /*
         *    Check to see if we have sufficient space to add all of the target to it and if not do not
         *    add any to it at all, but simply return an error.
         */
        SpaceRequired = strlen (Executers[isub]);
        if (SpaceRequired > SpaceLeft)
        {
            Status = -1;
        }
        else
        {
            (void)strcpy (pBuff ,Executers[isub]);
            pBuff += SpaceRequired + 1;
        }
    }

    /*
     *    Add an additional terminating NULL to the list
     */
    *pBuff = '\0';

    return Status;
}

/*
 *    Tgt_GetTargetList
 *    =================
 *
 *    return a list of the available targets, separated by nulls with a double null terminator
 *
 *    Calling Format:
 *      status = Tgt_GetTargetList (Buff ,BuffLen);
 *
 *    Parameters:
 *      Buff        - Pointer to char buffer to hold the returned list of targets
 *      BuffLen     - int length in bytes of the above buffer
 *
 *    Returns:
 *      0           - Success
 *      -1          - Insufficient space in buffer
 */
int     Tgt_GetTargetList   (char* pBuff ,int BuffLen ,Boolean ReturnDisplayed)
{
    int SpaceRequired;
    int isub;
    int Status = 0;                         /* Assume all will go well          */
    int SpaceLeft = BuffLen - 1;            /* Set up space available           */
    char* pTarget;

    /*
     *   Loop through the list of targets and add each tgo the list if space is available.
     */
    for (isub = (int)TARGET_SYBASE; ((isub < TargetCount) && (Status != -1)); isub++)
    {

        /*
         *    Get a pointer to the required target string and copy as much as possible into the target 
         *    remembering to ensure that it is null terminated
         */
        if (ReturnDisplayed)
        {
            pTarget = TargetList[isub].DisplayedText;
        }
        else
        {
            pTarget = TargetList[isub].Text;
        }

        /*
         *    Check to see if we have sufficient space to add all of the target to it and if not do not
         *    add any to it at all, but simply return an error.
         */
        SpaceRequired = strlen (pTarget);
        if (SpaceRequired > SpaceLeft)
        {
            Status = -1;
        }
        else
        {
            (void)strcpy (pBuff ,pTarget);
            pBuff += SpaceRequired + 1;
        }
    }

    /*
     *    Add an additional terminating NULL to the list
     */
    *pBuff = '\0';

    return Status;
}


/*
 *    Tgt_OtaForTarget
 *    ============
 *
 *    Determine if OTA is available for the specified target.
 *
 *    Calling format:
 *
 *      status = Tgt_OtaForTarget (TargetID);
 *
 *    Parameters:
 *      TargetID    - int Target ID to be checked
 *
 *    Returns:
 *      0 if OTA is applicable for this target
 *      -1 is returned if not
 */
int     Tgt_OtaForTarget    (enum TARGETID TargetID)
{
    if ((TargetID < TARGET_SYBASE) || ((int)TargetID > TargetCount)) return -1;
    return 0;           /* ALL VALID TARGETS HAVE AN OTA    */
}


/*
 *  Tgt_GetTargetExecutable
 *  ===================
 *
 *    This function will find the executable name of the target image we wish to invoke. 
 *
 *    Calling Format:
 *      Tgt_GetTargetExecutable (TargetID ,ImageType ,Buff ,BuffLen);
 *
 *    Parameters:
 *      TargetID    TARGETID target for which to get the image
 *      ImageType   TARGETIMAGE type of image for the target
 *      Buff        Pointer to char to hold the executable name
 *      BuffLen     int length of the above buffer
 */
void    Tgt_GetTargetExecutable (enum TARGETID TargetID ,enum TARGETIMAGE image ,char* Buff ,int BuffLen)
{
    char ThisSuffix[4];
    char WorkBuffer[32];
    int  length;

    /*
     *    Clear the buffer supplied to assume that we did not find the image.
     */
    (void) memset (Buff ,'\0' ,BuffLen);

    /*
     *    Get the suffix which we will append to the image name
     */
    Tgt_GetSuffix (TargetID ,ThisSuffix ,sizeof (ThisSuffix));

    /*
     *   ...and construct the image type
     */
    if      (image == SELECT)
    {
        (void) sprintf (WorkBuffer ,"TRecorder_%s" ,ThisSuffix);
    }

    else if (image == WHERE)
    {
        (void) sprintf (WorkBuffer ,"TModeller_%s" ,ThisSuffix);
    }

    else if (image == WHILE)
    {
        (void) sprintf (WorkBuffer ,"while_%s" ,ThisSuffix);
    }

    else if (image == TEXECUTER)
    {
        (void) sprintf (WorkBuffer ,"TExecuter_%s" ,ThisSuffix);
    }

    else if (image == TMANAGER)
    {
        (void) sprintf (WorkBuffer ,"TestManager");
    }

    /*remove .exe
     *    On all platforms except UNIX append '.exe' onto the file name
     */
    /*#ifndef   UNIX
        (void) strcat (WorkBuffer ,".exe");
    #endif */
    /* Not UNIX */

    /*
     *    ...and copy as much as possible into the users buffer
     */
    length = strlen (WorkBuffer);
    if (length >= BuffLen) length = BuffLen - 1;
    (void) memcpy (Buff ,WorkBuffer ,length);
    *(Buff + length) = '\0';

    return;
}
