/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*
 *  scl_getcmd.c:
 *  =============
 *
 *  This module contains the routine SCL_GETCMD. This retrieves and parses
 *  the command line.
 */

/*
 *  PSL Macros.
 */
#include <psl_macros.h>
#include <vtest_macros.h>

/*
 *  Standard ANSI headers.
 */
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>
#include "scl_rtndef.h"         /* SCL Routines */

/*
 *  Local Function prototypes.
 */
static void INI_SCL_CMD (struct scl_cmd *scl_cmd_ptr);

/*
 *  Structure definitions and parameters.
 */

#include <shrmsg.h>         /* PSL messages          */
#include <getopt.h>         /* getopt symbols        */
#include <psldiag_prmdef.h>     /* PSL Diagnostic parameters */
#include "scl_comdef.h"         /* Global Data structure     */
#include "scl_cmd.h"            /* Command Line Data Structure */

/*
 *    For WINDOWS/UNIX define the NULL file specification.
 */
#ifdef UNIX
#define NULL_FILE   "/dev/null"
#endif

#ifdef  WIN32
#define NULL_FILE   "NUL"
#endif

/************************************************************************/
/*                                                                      */
/*  SCL_GETCMD - Get and Parse Command Line                             */
/*  ==========                                                          */
/*                                                                      */
/*  Calling Convention:                                                 */
/*                                                                      */
/*     iret = SCL_GETCMD (scl_com_ptr, argc, argv_ptr);                 */
/*                                                                      */
/*  Where:                                                              */
/*                                                                      */
/*     iret        = Unsigned Long PSL Routine Return Status            */
/*     scl_com_ptr = Pointer to Common Data Structure                   */
/*     argc        = long int argument count                            */
/*     argv_ptr    = pointer to array of char arguments                 */
/*                                                                      */
/*   This routine parses the command line. The information              */
/*   obtained is stored in the global data structure.                   */
/*                                                                      */
/************************************************************************/

unsigned long SCL_GETCMD (struct scl_com *scl_com_ptr,
                          long argc,
                          char *argv_ptr [])
{
    /*
     *  Local Data,
     */
    unsigned long   iret = PSL_NORMAL;  /* Return status */
    unsigned long   isub;               /* Loop Control */
    unsigned long   jsub;               /* Loop Control */
    long            chk_data_len;       /* Length of CHK_DATANAME */
    long            status;             /* Signed status */
    char            **string_ptr_ptr;   /* Pointer to pointer to string */
    struct scl_cmd  *scl_cmd_ptr;       /* Pointer to CMD structure */
#define PDEFVAR_CNT 11
    static char *pdefvar [PDEFVAR_CNT] = {  /* Predefined variants */
        "%UNIX", "%VMS", "%WIN32",
        "%ALPHA", "%HP", "%IBM", "%NCR", "%OSF", "%SUN", "%NTI", "%VAX"};
    int        optchr;     /* Option character or -1 (end of options) */
    int        option_index;       /* `getopt_long' stores the option index here */
    static   char       *scl_file = NULL;   /* Pointer to SCL filename */
    char       msgbuf [256];   /* Message buffer */

    static const struct option long_options[] =
        {
            /* name,    has_arg,           flag, value */
            {"help", no_argument, 0, 'h'},
            {"confirm", no_argument, 0, 'i'},
            {"include", required_argument, 0, 'I'},
            {"list", required_argument, 0, 'l'},
            {"object", required_argument, 0, 'o'},
            {"log", no_argument, 0, 'v'},
            {"type",    required_argument, 0, 't'},
            {"variant", required_argument, 0, 'V'},
            {0, 0, 0, 0}
        };

    const char *const usage = "Usage: scl [OPTION] [SCL_FILE]\n"
                              "-h                      --help\n"
                              "-i                      --confirm\n"
                              "-I inc_file             --include=inc_file\n"
                              "-l [on|off]             --list=[on|off]\n"
                              "-o obj_file             --object=obj_file\n"
                              "-t [script|test|thread] --type\n"
                              "-v                      --log\n"
                              "-V v1+v2{+...}          --variant=v1+v2{+...}";

    /*
     *  Get pointer to CMD structure.
     */
    scl_cmd_ptr = (*scl_com_ptr).cmd_ptrps;

    /* Initialize Command Line Data */
    INI_SCL_CMD (scl_cmd_ptr);

#if defined(UNIX) || defined(WIN32)
    /*
     *  If there are no arguments, just output the usage summary.
     */
    if (argc == 1)          /* If no arguments */
    {
        iret = PSL_BADPARAM;        /* Force an error */
        goto output_usage;      /* And output usage summary */
    }
#endif /* UNIX */

    /*
     *  Parse the command line: Suppress error reporting in getopt_long so that
     *  we can report errors in our own style, when "?" is returned.
     */
    opterr = 0;             /* Suppress getopt_long error reporting */

    while (1)   /* Loop indefinitely - rely on 'break' and 'goto' to exit loop */
    {
        option_index = 0;       /* Reset option index for each option */

        /* Parse the next command line option */
        optchr = getopt_long (argc, argv_ptr, "hiI:l:o:t:vV:", long_options, &option_index);

        /* Detect the end of the options */
        if (optchr == -1) break;

        /* Process this command line option */
        switch (optchr)
        {
        case 0:
            fprintf (stderr, "getopt_long returns 0 - should never happen\n");
            break;

        case 'h':           /* --help */
            /*puts("option -h\n"); */
            goto output_usage;

        case 'i':           /* --confirm */
            /*puts("option -i\n"); */
            (*scl_cmd_ptr).confirmn = PTRUE;
            (*scl_cmd_ptr).confirm_negatedn = PFALSE;
            (*scl_cmd_ptr).confirm_valuen = PFALSE;
            break;

        case 'I':           /* --include=inc_file */
            /* printf ("option -I with value `%s'\n", optarg); */
            (*scl_cmd_ptr).includen = PTRUE;
            (*scl_cmd_ptr).include_negatedn = PFALSE;
            (*scl_cmd_ptr).include_valuen = PTRUE;
            (*scl_cmd_ptr).include_value_ptrpc = optarg;
            break;

        case 'l':           /* --list=[on|off] */
            /*printf("option -l with value `%s'\n", optarg); */
            if (strcmp (optarg, "on") == 0) /* --list=on */
            {
                (*scl_cmd_ptr).listn = PTRUE;
            }
            else if (strcmp (optarg, "off") == 0) /* --list=off */
            {
                (*scl_cmd_ptr).listn = PFALSE;
            }
            else                /* Unrecognised --list keyword */
            {
                iret = PSL_IVKEYW;      /* Set up error: Unrecognised keyword */
                PSL_PUTDIAG (           /* Put Diagnostic */
                    (*scl_com_ptr).diag_ptrps,  /* Use this stack */
                    1,              /* Number of message segments */
                    iret,           /* Message */
                    optarg);            /* Unrecognised keyword */
                PSL_OUTPUT_DIAG (       /* Output the diagnostic */
                    (*scl_com_ptr).diag_ptrps,  /* From this stack */
                    0,              /* No Context */
                    DIOUFL_LSTERR);     /* Last Error */
                goto output_usage;      /* And output usage summary */
            }
            break;

        case 'o':           /* --object=obj_file */
            /*printf ("option -o with value `%s'\n", optarg); */
            (*scl_cmd_ptr).objectn = PTRUE;
            (*scl_cmd_ptr).object_negatedn = PFALSE;
            (*scl_cmd_ptr).object_valuen = PTRUE;
            (*scl_cmd_ptr).object_value_ptrpc = optarg;
            break;

        case 't':           /* --type=[script|test|thread] */
            /*printf("option -t with value `%s'\n", optarg); */
            if (strcmp (optarg, "script") == 0) /* --type=script */
            {
                (*scl_cmd_ptr).type.scriptn = PTRUE;
                (*scl_cmd_ptr).type.testn = PFALSE;
                (*scl_cmd_ptr).type.threadn = PFALSE;
            }
            else if (strcmp (optarg, "test") == 0) /* --type=test */
            {
                (*scl_cmd_ptr).type.scriptn = PFALSE;
                (*scl_cmd_ptr).type.testn = PTRUE;
                (*scl_cmd_ptr).type.threadn = PFALSE;
            }
            else if (strcmp (optarg, "thread") == 0) /* --type=thread */
            {
                (*scl_cmd_ptr).type.scriptn = PFALSE;
                (*scl_cmd_ptr).type.testn = PFALSE;
                (*scl_cmd_ptr).type.threadn = PTRUE;
            }
            else                /* Unrecognised --type keyword */
            {
                iret = PSL_IVKEYW;      /* Set up error: Unrecognised keyword */
                PSL_PUTDIAG (           /* Put Diagnostic */
                    (*scl_com_ptr).diag_ptrps,  /* Use this stack */
                    1,              /* Number of message segments */
                    iret,           /* Message */
                    optarg);            /* Unrecognised keyword */
                PSL_OUTPUT_DIAG (       /* Output the diagnostic */
                    (*scl_com_ptr).diag_ptrps,  /* From this stack */
                    0,              /* No Context */
                    DIOUFL_LSTERR);     /* Last Error */
                goto output_usage;      /* And output usage summary */
            }
            break;

        case 'v':           /* --log */
            /*puts("option -v\n"); */
            (*scl_cmd_ptr).logn = PTRUE;
            (*scl_cmd_ptr).log_negatedn = PFALSE;
            (*scl_cmd_ptr).log_valuen = PFALSE;
            break;

        case '?':               /* Unrecognised option switch */
        default:                /* ... */
            msgbuf[0] = optopt;         /* Unrecognised option switch */
            msgbuf[1] = '\0';           /* ... */

            iret = PSL_IVSWIT;          /* Set up error: Unrecognised switch */
            PSL_PUTDIAG (           /* Put Diagnostic */
                (*scl_com_ptr).diag_ptrps,  /* Use this stack */
                1,              /* Number of message segments */
                iret,           /* Message */
                msgbuf);            /* Unrecognised switch */
            PSL_OUTPUT_DIAG (           /* Output the diagnostic */
                (*scl_com_ptr).diag_ptrps,  /* From this stack */
                0,              /* No Context */
                DIOUFL_LSTERR);     /* Last Error */
            goto output_usage;          /* And output usage summary */
        }
    }

    if (optind == argc - 1)     /* If correct number of parameters */
    {
        /*printf("filename %s \n", argv_ptr[optind]); */
        (*scl_cmd_ptr).scln = PTRUE;
        (*scl_cmd_ptr).scl_negatedn = PFALSE;
        (*scl_cmd_ptr).scl_valuen = PTRUE;
        (*scl_cmd_ptr).scl_value_ptrppc = &scl_file;
        scl_file = argv_ptr[optind++];
        (*scl_cmd_ptr).scl_value_countl += 1;
    }
    else if (optind >= argc)        /* Else if too few parameters */
    {
        iret = PSL_INSFPRM;     /* Set up error: Insufficient parameters */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*scl_com_ptr).diag_ptrps,  /* Use this stack */
            1,             /* Number of message segments */
            iret);         /* Message */
        PSL_OUTPUT_DIAG (       /* Output the diagnostic */
            (*scl_com_ptr).diag_ptrps,  /* From this stack */
            0,              /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        goto end;           /* and exit! */
    }
    else                /* Else too many parameters */
    {
        iret = PSL_MAXPARM;     /* Set up error: Too many parameters */
        optind++;           /* Move to next option */
        PSL_PUTDIAG (           /* Put Diagnostic */
            (*scl_com_ptr).diag_ptrps,  /* Use this stack */
            1,             /* Number of message segments */
            iret,          /* Message */
            argv_ptr[optind]);     /* First 'extra' Parameter */
        PSL_OUTPUT_DIAG (       /* Output the diagnostic */
            (*scl_com_ptr).diag_ptrps,  /* From this stack */
            0,              /* No Context */
            DIOUFL_LSTERR);     /* Last Error */
        goto end;           /* and exit! */
    }

    /*
     *  INCLUDE
     *
     */
    if ((*scl_cmd_ptr).includen)
    {
        (void) strcpy ((*scl_com_ptr).include_valuec256,
                       (*scl_cmd_ptr).include_value_ptrpc);
    }
    else
    {
        (*scl_com_ptr).include_valuec256[0] = '\0';
    }

    /*
     *  LIST
     */
    if ((*scl_cmd_ptr).listn)
    {
        (void) LIST_ALLOCATE ((*scl_com_ptr).diag_ptrps,
                              &(*scl_com_ptr).lst_contextps);
    }

    /*
     *  OBJECT
     *
     *  Unless explicitly negated, OBJECT is the default.
     */
    if ((*scl_cmd_ptr).objectn)
    {
        (void) strcpy ((*scl_com_ptr).object_valuec256,
                       (*scl_cmd_ptr).object_value_ptrpc);
    }
    else
    {
        (*scl_com_ptr).object_valuec256[0] = '\0';
    }

    if (!(*scl_cmd_ptr).object_negatedn)
    {

        /*
         *    On WINDOWS and UNIX there is no way of turning off the creation of an
         *    object file and so the typical approach is to redirect it to NULL. We 
         *    detect this here and turn off the object file creation to save us the 
         *    effort of producing a file that is to be thrown away later!!!
         */
        if (strcmp ((*scl_com_ptr).object_valuec256 ,NULL_FILE) != 0)
        {
            (*scl_cmd_ptr).objectn = PTRUE;
        }
        else
        {
            (*scl_cmd_ptr).objectn = PFALSE;
        }
    }

    /*
     *  TYPE.
     */
    (*scl_com_ptr).type_scriptn = (*scl_cmd_ptr).type.scriptn;
    (*scl_com_ptr).type_testn   = (*scl_cmd_ptr).type.testn;
    (*scl_com_ptr).type_threadn = (*scl_cmd_ptr).type.threadn;

    /*
     *  If /TYPE has not been specified, default type to script.
     */
    if (!(*scl_com_ptr).type_scriptn
            && !(*scl_com_ptr).type_testn
            && !(*scl_com_ptr).type_threadn)
    {
        (*scl_com_ptr).type_scriptn = PTRUE;
    }

    /*
     *  VARIANT.
     *
     *  Check that the values are valid.
     */
    for (isub = 0, string_ptr_ptr = (*scl_cmd_ptr).variant_value_ptrppc;
            isub < (*scl_cmd_ptr).variant_value_countl;
            isub++, string_ptr_ptr++)
    {
        chk_data_len = strlen (*string_ptr_ptr);

        for (jsub = 0; jsub < PDEFVAR_CNT; jsub++)
        {
            if (strcmp (*string_ptr_ptr, pdefvar [jsub]) == 0) goto nxtvar;
        }

        (void) CHK_DATANAME (*string_ptr_ptr, &chk_data_len, &status);

        if (status != 0)
        {
            iret = VTEST_INVVARIANT;    /* Set up error */
            PSL_PUTDIAG (       /* Put Diagnostic */
                (*scl_com_ptr).diag_ptrps,
                /* Use this stack */
                1,          /* Number of message segments */
                iret);          /* Message */
            PSL_OUTPUT_DIAG (       /* Output the diagnostic */
                (*scl_com_ptr).diag_ptrps,
                /* From this stack */
                0,          /* No Context */
                DIOUFL_LSTERR);     /* Last Error */
            goto end;           /* and exit! */
        }
nxtvar:
        ;
    }

    goto end;               /* And exit */

    /*
     *  OUTPUT_USAGE:
     */
output_usage:
    puts (usage);

    /*
     *  END:
     *
     *  Routine exit point.
     */
end:
    return iret;
}

/************************************************************************/
/*                                  */
/*  INI_SCL_CMD - Initialize Command Line Data          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     INI_SCL_GETCMD (scl_cmd_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     scl_cmd_ptr = Pointer to Command Line Data Structure     */
/*                                  */
/*  This routine initializes the Command Line Data Structure.   */
/*                                  */
/************************************************************************/
static void INI_SCL_CMD (struct scl_cmd *scl_cmd_ptr)
{

    (*scl_cmd_ptr).confirmn = PFALSE;
    (*scl_cmd_ptr).confirm_negatedn = PFALSE;
    (*scl_cmd_ptr).confirm_valuen = PFALSE;

    (*scl_cmd_ptr).includen = PFALSE;
    (*scl_cmd_ptr).include_negatedn = PFALSE;
    (*scl_cmd_ptr).include_valuen = PFALSE;
    (*scl_cmd_ptr).include_value_ptrpc = NULL;

    (*scl_cmd_ptr).listn = PFALSE;
    (*scl_cmd_ptr).list_negatedn = PFALSE;
    (*scl_cmd_ptr).list_valuen = PFALSE;
    (*scl_cmd_ptr).list_value_ptrpc = NULL;

    (*scl_cmd_ptr).objectn = PFALSE;
    (*scl_cmd_ptr).object_negatedn = PFALSE;
    (*scl_cmd_ptr).object_valuen = PFALSE;
    (*scl_cmd_ptr).object_value_ptrpc = NULL;

    (*scl_cmd_ptr).logn = PFALSE;
    (*scl_cmd_ptr).log_negatedn = PFALSE;
    (*scl_cmd_ptr).log_valuen = PFALSE;

    (*scl_cmd_ptr).scln = PFALSE;
    (*scl_cmd_ptr).scl_negatedn = PFALSE;
    (*scl_cmd_ptr).scl_valuen = PFALSE;
    (*scl_cmd_ptr).scl_value_ptrppc = NULL;
    (*scl_cmd_ptr).scl_value_countl = 0;

    (*scl_cmd_ptr).typen = PFALSE;
    (*scl_cmd_ptr).type_negatedn = PFALSE;
    (*scl_cmd_ptr).type_valuen = PFALSE;
    (*scl_cmd_ptr).type.scriptn = PFALSE;
    (*scl_cmd_ptr).type.testn = PFALSE;
    (*scl_cmd_ptr).type.threadn = PFALSE;

    (*scl_cmd_ptr).variantn = PFALSE;
    (*scl_cmd_ptr).variant_negatedn = PFALSE;
    (*scl_cmd_ptr).variant_valuen = PFALSE;
    /*** char **variant_value_ptrppc; ***/
    /*** unsigned long variant_value_countl; ***/

    return;
}
