/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

// TCommonRcrd.cpp

#include <psl_macros.h>

// PSL function prototypes.
extern "C"
{
#include <psl_rtndef.h>
}

#include "tofprmdef.h"

#include "stdafx.h"

#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif

#include "LogTrace.hpp"

#ifndef CYRANO_TP_STANDALONE
#   include "CECSemaphore.h"
#endif

#include "TCommonRcrd.h"
#include "TRecMake.h"
#include "TContext.h"

#include "ociphases.h"
#include "psl_rtndef.h"         // PSL routines
#include "ascbin_prmdef.h"      // PSL_ASCBIN routine symbols
#include "binasc_prmdef.h"      // PSL_BINASC routine symbols
#include "shrmsg.h"             // PSL messages

#include "AsyncVirtualUser.h"

static const char trc_prefix[] = "T* ";
static const char wrn_prefix[] = "W* ";
static const char err_prefix[] = "E* ";


// Helper Functions
void MSGERR(const string& msg)
{
    g_pLogTrace->AddTrace(TRCCAT_RUNNING, TRCLEV_NONE, err_prefix + msg);
}

void MSGTRC(const string& msg)
{
    g_pLogTrace->AddTrace(TRCCAT_RUNNING, TRCLEV_HIGH, msg);
}

int goto_destination(int destination,TContext& ctxt);
int PrepareGoto(const TRecord* rec, const string& errlab, const TScript* script, bool required);
int ExecuteGoto(const TRecord* rec, TContext& ctxt, const string& errlab, int* _cur_label_index, CFastLock *pFL = NULL);
int ExecuteGotoLabel(const TRecord* rec, TContext& ctxt, const string& label_name, CFastLock *pFL = NULL);

extern RecCreator_map_t* recCreatorMap = NULL;

template<class TRcrd> const bool RegisterCreator()
{
    if (!recCreatorMap) recCreatorMap = new RecCreator_map_t;
    string thisKey = TRcrd::KEY.toString();
    (*recCreatorMap)[TRcrd::KEY] = SharedPtr<TRecCreator>(new TRecMake<TRcrd>());
    size_t numRegd = recCreatorMap->size();
    return true;
}

#include "TCommonRcrd.inl"

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Abort record 'execute' function. The Abort record has the following
//  	    format:
//
//	        <PRE>
//	        Abort
//          target <target_code>        // TOF_ABTTGT_*
//	        </PRE>
//
// @rdesc   <PRE>
//	        Return status:
//	        0 (zero) = Success (always).
//	        </PRE>
//
// @parm    Reference to thread context object.
//

int TRcrdAbort::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ABORT);
    int	target = intPrmVal(TParam::stTarget, ctxt);
    MSGTRC ("Entered: TRcrdAbort::execute; target = " + int_to_string(target));  //**RAC

    switch(target)
    {
    case TOF_ABTTGT_THREAD:     // Detatch thread
        return TOF_RES_KILL_THREAD;

    default:                    // Do nothing
        {
            MSGERR("TRcrdAbort: target value not handled: " + int_to_string(target));
            break;
        }
    }

    ctxt.set_phase_text(PHASE_NONE);
    return TOF_RES_OK;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Conditional (If ... Goto) record 'prepare' function. The
//	    Conditional record has the following format:
//
//	    <PRE>
//	    Conditional
//	    expr <condition>
//	    name <label>
//	    </PRE>
//
//	    where <condition> is a boolean expression string (empty for
//	    unconditional goto) and <label> is the target label.
//
//	    This function encodes the conditional expression string into a more
//	    suitable form for run-time evaluation, validates the target label
//	    and saves label information. The encoded expression and label
//	    information are used by the ::execute() function. The function logs
//	    an error message and returns a failure status if the conditional
//	    expression cannot be encoded or the target label is invalid.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdCond::prepare(TScript* script, int objIndex, bool verbose)
{

    // Get conditional expression string (from 'expr' parameter).
    const TParam* prmP = param(TParam::stExpression);
    string l_condition = prmP->strVal();

    // Get 'goto' label (from 'name' parameter); report error if label not found.
    _labelIndex = PrepareGoto(this, TParam::stName, script, true);
    // If conditional expression string not empty, encode it. Report error if encode fails,
    if (l_condition.length())
    {
        if (!_exprP) _exprP = new TtBoolExpr(script);

        if (_exprP->Parse(l_condition, 0) == false)
        {
            prmP = param(TParam::stName);
            string s_name = prmP->strVal();
            MSGERR ( "CONDITIONAL: IF " + l_condition + " GOTO " + s_name +
                     " Failed during PREPARE with: " + _exprP->Reason());
            return -1;	// Return 'failure' status
        }

    }

    // If verbose set, output tracing information.
    if(verbose)
    {
        prmP = param(TParam::stName);
        string s_name = prmP->strVal();
        MSGTRC ("IF " + l_condition + " GOTO: " + s_name +
                "(" + int_to_string(_labelIndex) + ")");
    }

    return 0;	// Return success
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Conditional (If ... Goto) record 'execute' function. The
//	    Conditional record has the following format:
//
//	    <PRE>
//	    Conditional
//	    expr <condition>
//	    name <label>
//	    </PRE>
//
//	    This function executes a Conditional record, i.e. it evaluates the
//	    encoded conditional expression (if any) set up by ::prepare() and
//	    determines whether the condition is true or false and so whether
//	    the jump is to be made. The return parameter indicates the result.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Execution failed; error message logged by this routine.
//	    0 (zero) = Condition false, jump NOT to be taken.
//	    n (n>0) = Condition true, index of 'goto' target record (1 based)
//		to jump to.
//	    </PRE>
// @parm    Reference to thread context object.
//

int TRcrdCond::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_CONDITIONAL);
    int destination=-1;
    // Get 'goto' label index from 'name' parameter.
    if (_exprP)	// Null expression is unconditional Jump
    {
        // Evaluate Conditional Expression, 0=false, non ZERO=true, NULL=Error
        string res;
        res = _exprP->Evaluate(ctxt);

        if (res.empty())
        {
            string subt_expr = strPrmVal(TParam::stExpression, ctxt);
            ctxt.log("CONDITIONAL FAILED: '" + subt_expr + "' \n " + ctxt.ErrorMessage(), ERRMSG);
            throw mdbParseException("CONDITIONAL FAILED: '" + subt_expr + "' \n " + ctxt.ErrorMessage());
        }
        else if (res != CECTofTxt::stFALSE)
            destination = ExecuteGoto(this, ctxt, TParam::stName, &_labelIndex, m_pFastLock);
        else
        {
            destination=0;
        }
    }
    else
        destination = ExecuteGoto(this, ctxt, TParam::stName, &_labelIndex, m_pFastLock);

    if(ctxt.GetVerbose(TRACEMSG))
    {
        string subt_expr = strPrmVal(TParam::stExpression, ctxt);
        const TParam* prmP = param(TParam::stExpression);
        string raw_expr  = prmP->strVal();
        string s_name = strPrmVal(TParam::stName,ctxt);
        ctxt.log("CONDITIONAL: IF "+ raw_expr + " GOTO " + s_name +
                 " { " + subt_expr + " } " + string(destination?"TRUE":"FALSE"), TRACEMSG);
    }
    ctxt.set_phase_text(TP_PHASE_NONE);

    return destination;
} // End TRcrdCond::execute()

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Convert record PREPARE function. The Convert record has the
//  	    following format:
//
//	        <PRE>
//	        Convert
//          from <from_variable>
//	        to <to_variable>
//	        radix <radix>
//	        options <opt_mask>
//	        {errlab <label>}
//	        </PRE>
//
// @rdesc   <PRE>
//	        Return status:
//	        -1 = Prepare failed; error message logged by this routine.
//	        0 (zero) = Success.
//	        </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdConvert::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prmP = param(TParam::stFrom);
    if (prmP == NULL) throw mdbParseException("No Source Variable in Convert");
    string srcvar = prmP->strVal();

    prmP = param(TParam::stTo);
    if (prmP == NULL) throw mdbParseException("No Target Variable in Convert");
    string tgtvar = prmP->strVal();

    prmP = param(TParam::stRadix);
    if (prmP==NULL) throw mdbParseException("No Radix Expression in Convert");
    string radix = prmP->strVal();

    // Check that the options mask is present.
    prmP = param(TParam::stOptions);
    if (prmP == NULL)
    {
        MSGERR("CONVERT: " + srcvar + " to " + tgtvar +
               " Failed during PREPARE with: Convert options missing");
        return -1;
    }

    // If verbose set, output tracing information.
    if (verbose)
    {
        MSGTRC("CONVERT: " + srcvar + " TO " + tgtvar + " RADIX " + radix);
    }

    // Encode the source variable; report an error if it is missing or encode fails.
    if (srcvar.length())
    {
        _srcVarP = new TtVarRef(script);
        if (!(_srcVarP->Parse(srcvar, 0)))
        {
            MSGERR("CONVERT: Failed during PREPARE of Source Variable: " + srcvar +
                   " with: " + _srcVarP->Reason());
            return -1;
        }
    }
    else
    {
        MSGERR("CONVERT: Failed during PREPARE, Source Variable Missing");
        return -1;
    }

    // Get pointer to TRecord for the source variable - as we need to know the variable
    // type in order to determine what type of conversion to do.
    TRecord* trecP = (*script).GetVardef(_srcVarP->Varnam());
    if (trecP == NULL) // If not a defined variable
    {
        MSGERR("CONVERT: " + srcvar + " to " + tgtvar +
               " Failed during PREPARE with: Variable '" + _srcVarP->Varnam() + "' not defined");
        return -1;
    }

    // Get source variable options mask so that we can check the variable type.
    prmP = trecP->param(TParam::stOptions);
    if (prmP == NULL) // If no 'options'  parameter (should always be present).
    {
        MSGERR("CONVERT: " + srcvar + " to " + tgtvar +
               " Failed during 'srcvar' PREPARE with: Variable options missing");
        return -1;
    }

    // Set the conversion type based on the source variable type.
    int srcopt = prmP->intVal();    // Source variable options mask
    if (srcopt & VTINT)
    {
        _toInteger = false;
    }
    else
    {
        _toInteger = true;
    }

    // Encode the target variable; report an error if it is missing or encode fails.
    if (tgtvar.length())
    {
        _tgtVarP = new TtVarRef(script);
        if (!(_tgtVarP->Parse(tgtvar, 0)))
        {
            MSGERR("CONVERT: Failed during PREPARE of Target Variable: " + tgtvar +
                   " with: " + _tgtVarP->Reason());
            return -1;
        }
    }
    else
    {
        MSGERR("CONVERT: Failed during PREPARE, Target Variable Missing");
        return -1;
    }

    // Encode the Radix expression - report error if missing.
    if (radix.length())
    {
        _radixExprP = new TtIntExpr(script);
        if (!(_radixExprP->Parse(radix, 0)))
        {
            MSGERR("CONVERT: Failed in PREPARE of Radix: " +
                   radix + " with: " + _radixExprP->Reason());
            return -1;
        }
    }
    else
    {
        MSGERR("CONVERT: Failed during PREPARE, Missing Radix Expression");
        return -1;
    }

    return 0;
} // end of TRcrdConvert::prepare

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Convert record 'execute' function. The Convert record has the
//  	    following format:
//
//	        <PRE>
//	        Convert
//          from <from_variable>
//	        to <to_variable>
//	        options <opt_mask>
//	        {radix <radix>}
//	        {errlab <label>}
//	        </PRE>
//
// @rdesc   <PRE>
//	        Return status:
//	        -1 = Execution failed; error message logged by this routine.
//	        0 (zero) = Success.
//	        n (n>0) = Execution failed and errlab specified, record to
//		    jump to (1 based).
//	        </PRE>
//
// @parm    Reference to thread context object.
//
int TRcrdConvert::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_CONVERT);

    // Evaluate source variable value.
    string srcval;  // Value of source variable
    string varidx = _srcVarP->Evaluate(ctxt); // Evaluate source variable subscript, if any
    bool bFound;
    srcval = ctxt.get_variable(_srcVarP->Varnam(), bFound, false, varidx);
    if(!bFound)
    {
        string srcnam = strPrmVal(TParam::stFrom, ctxt);    // Source variable name
        ctxt.log("CONVERT FAILED: No value available for source variable (" + srcnam + ")", ERRMSG);
        throw mdbParseException("CONVERT FAILED: No value available for source variable (" + srcnam + ")");
    }

    // Get the conversion radix, if any, as an unsigned char - this is the format
    // required by PSL_BINASC and PSL_ASCBIN.
    string radval("10");    // Conversion radix - default 10.
    if (_radixExprP)        // If radix expression specified, evaluate it
    {
        radval = _radixExprP->Evaluate(ctxt);
        if (radval.empty())
        {
            string radprm = strPrmVal(TParam::stRadix, ctxt);
            ctxt.log("CONVERT FAILED: '" + radprm + "' \n " + ctxt.ErrorMessage(), ERRMSG);
            throw mdbParseException("CONVERT FAILED: '" + radprm + "' \n " + ctxt.ErrorMessage());
        }
    }

    unsigned char radix_c = (unsigned char)atol(radval.c_str());

    // Get the conversion options.
    int options = intPrmVal(TParam::stOptions, ctxt);

    // Perform the conversion.
    int status = 0;     // Assume conversion successful
    string result;      // Put result as string here

    // If conversion from CHARACTER to INTEGER.
    if (_toInteger)
    {
        // Strip off any leading or trailing spaces as PSL_ASCBIN does not
        // like them and will report an error if any are present.
        long isub;      // Loop control
        char tmpchrbuf [256];
        char *chr_ptr = &tmpchrbuf[0];
        strcpy (tmpchrbuf, srcval.c_str());
        long srclen = srcval.length();   // Source value length
        for (isub = 0;
                ((isub < srclen) && (*chr_ptr == ' '));
                isub++, chr_ptr++);

        // Reduce source value length by the number of leading spaces stripped.
        srclen -= isub;

        // Copy source string length into an unsigned char - as required by PSL_ASCBIN.
        unsigned char chrlen_c = (unsigned char) srclen;

        // Perform CHARACTER to INTEGER conversion.
        long numval;                        // Numeric value represented by ASCII string
        unsigned short unconverted;         // Number of unconverted characters
        unsigned long iret;                 // PSL_ASCBIN return status
        iret = PSL_ASCBIN (&chrlen_c, chr_ptr, &radix_c, &numval, &unconverted);
        if (!(psl_test_status (iret)) || (unconverted != 0))
        {
            ctxt.log("CONVERT FAILED: Cannot convert \"" + srcval +
                     "\" to integer; Radix=" + radval + ".", ERRMSG);
            status = -1; // Conversion failed
        }

        else
        {
            char resStr[32];
            sprintf(resStr, "%li", numval);
            result = string(resStr);
        }
    }

    // Else must be conversion from INTEGER to CHARACTER.
    else
    {
        // Set up the PSL_BINASC options from the Convert options mask. Note that
        // in the TOF options mask CLEAR indicates SIGNED, LEFT JUSTIFY and
        // LEADING ZEROS - it is the reverse in PSL_BINASC(!) so we must SET the
        // corresponding bits.
        unsigned short binasc_options = 0; // Clear PSL_BINASC options
        if ((options & TOF_CNVOPT_SIGN) == 0) binasc_options |= BINASC_SIGN;
        if ((options & TOF_CNVOPT_LEAD) == 0) binasc_options |= BINASC_FILL;
        if ((options & TOF_CNVOPT_JUST) == 0) binasc_options |= BINASC_JUSTIFY;

        // PSL_BINASC will only accept a char length whereas SCL variables have a
        // LONG length. Therefore we must check to see if the variable in question
        // has a length > 255 characters and handle it as a special case.
        unsigned char outlen_c = 255;     // Output string length
        char tmpchrbuf [256];
        char *chr_ptr = &tmpchrbuf[0];

        // Perform INTEGER to CHARACTER conversion.
        unsigned long iret;                     // PSL_ASCBIN return status
        long srcval_l = atol(srcval.c_str());   // Source value as a long
        iret = PSL_BINASC (&srcval_l, &outlen_c, &radix_c,
                           chr_ptr, &binasc_options);
        if (!(psl_test_status (iret)))
        {
            if (iret == PSL_BADRADIX)
            {
                ctxt.log("CONVERT FAILED: Cannot convert \"" + srcval +
                         "\" to character; Radix=" + radval + ".", ERRMSG);
                status = -1; // Conversion failed
            }

            else
            {
                result = "*";
            }
        }
        else
        {
            char resStr[256];
            sprintf(resStr, "%s", chr_ptr);
            chr_ptr = (char *) memchr (resStr, ' ', 256);
            *chr_ptr = '\0';
            result = string(resStr);
        }
    }

    // If conversion successful, set the target variable with the result of the conversion.
    int dest = 0;
    if (status == 0)
    {
        varidx = _tgtVarP->Evaluate(ctxt); // Evaluate target variable subscripts, if any
        if((ctxt.set_variable(_tgtVarP->Varnam(), result, varidx)) == -1)
        {
            string tgtnam = strPrmVal(TParam::stTo, ctxt);  // Target variable name
            ctxt.log("CONVERT FAILED: Cannot set target variable (" + tgtnam + ")", ERRMSG);
            throw mdbParseException("CONVERT FAILED: Cannot set target variable (" + tgtnam + ")");
        }
    }

    // Else conversion failed:
    // 1. If no ON ERROR GOTO <label> parameter, throw an exception. The error will be
    //    trapped by the script 'ON ERROR' handler, if any.
    // 2. If ON ERROR GOTO <label> specified, set up target destination return.
    else
    {
        int labelIndex;
        try
        {
            labelIndex = PrepareGoto(this, TParam::stErrlab, ctxt.Script(), false);
        }
        catch(...) // No valid ON ERROR GOTO <label>
        {
            throw mdbParseException("Unable to perform Convert");
        }
        dest = ExecuteGoto(this, ctxt, TParam::stErrlab, &labelIndex);
    }

    ctxt.set_phase_text(TP_PHASE_NONE);

    return dest;
} // end of TRcrdConvert::execute

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Else (Else/Else If ...) record 'prepare' function. The Else record
//	    has the following format:
//
//	    <PRE>
//	    Else
//	    {expr <condition>}
//	    </PRE>
//
//	    where <condition> is a boolean expression string.
//
//	    This function encodes the conditional expression string (if any)
//	    into a more suitable form for run-time evaluation. The encoded
//	    expression is used by the ::execute() function. The function logs
//	    an error message and returns a failure status if the conditional
//	    expression cannot be encoded.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdElse::prepare(TScript* script, int objIndex, bool verbose)
{
    if (script->OpenObjects().empty())
    {
        MSGERR("ELSE Failed during PREPARE: Missing parent IF-THEN");
        return -1;
    }

    // Get conditional expression string (from 'expr' parameter).
    const TParam* prmP = param(TParam::stExpression);
    string l_condition = prmP->strVal();

    // If this is an Else If, encode the conditional expression.
    if (l_condition.length())	// Optional conditional, makes this an ELSE-IF
    {
        if (!_exprP) _exprP = new TtBoolExpr(script);

        if (_exprP->Parse(l_condition, 0) == false)
        {
            MSGERR ("ELSE: ELSE IF " + l_condition + " THEN <statements>"
                    " Failed during PREPARE with: " + _exprP->Reason());
            return -1;
        }
    }

    _ifThenElse = static_cast<TRcrdFlowCtrl*>(script->OpenObjects().top());
    if (_ifThenElse)
    {
        if (_ifThenElse->isType(TRcrdIF_Then::KEY) || _ifThenElse->isType(TRcrdElse::KEY))
        {
            _ifThenElse->SetEndIndex(objIndex);	// Update parent statement with of end point
        }

        else
        {
            MSGERR ("ELSE Failed during PREPARE: Missing parent IF-THEN");
            return -1;
        }
    }
    else
    {
        MSGERR("ELSE Failed during PREPARE: Missing parent IF-THEN");
        return -1;
    }

    _startIndex = objIndex;
    script->OpenObjects().pop();	// Remove the IfThen record
    script->OpenObjects().push(this);	// Make IfElse available to Eif (end if) processing

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Else (Else/Else If ...) record 'execute' function. The Else record
//	    has the following format:
//
//	    <PRE>
//	    Else
//	    {expr <condition>}
//	    </PRE>
//
//	    This function executes an Else record, i.e. it evaluates the
//	    encoded conditional expression (if any) set up by ::prepare() and
//	    determines whether the condition is true or false and so whether
//	    the jump is to be made. The return parameter indicates the result.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Execution failed; error message logged by this routine.
//	    0 (zero) = Condition true (or not present), execute following
//		else/else if block.
//	    n (n>0) = Condition false, index of next 'if' block (i.e. Else If,
//		Else or Endif) record to jump to (1 based).
//	    </PRE>
// @parm    Reference to thread context object.
//
int TRcrdElse::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ELSE);
    // Check for illegal block structure
    if (ctxt.OpenObjects().empty() || ctxt.OpenObjects().top() != _ifThenElse)
    {
        ctxt.log("ELSE Failed: Missing IF-THEN, Jumped into IF-THEN <statements> ELSE block", ERRMSG);
        throw mdbParseException("ELSE Failed: Missing IF-THEN, Jumped into IF-THEN <statements> ELSE block");
        //return -1;
    }


    ctxt.OpenObjects().pop();		// Remove the IfThen record
    ctxt.OpenObjects().push(this);	// Make IfElse available to EndIF processing

    if (ctxt.MaskIsSet(CTXT_SKIP_ELSE))	// Reached here from preceding THEN/ELSE-IF block
        return _endIndex + 1;			// So skip ELSE block and goto END-IF

    int destination = -1;
    if (_exprP) // this is an ELSE IF, not simple ELSE, so evaluate condition
    {
        string res;
        res = _exprP->Evaluate(ctxt);
        if (res.empty())
        {
            string subt_expr = strPrmVal(TParam::stExpression, ctxt);
            ctxt.log("ELSE IF Failed: '" + subt_expr + "' \n " + ctxt.ErrorMessage(), ERRMSG);
            throw mdbParseException("ELSE IF Failed: '" + subt_expr + "' \n " + ctxt.ErrorMessage());
            //return -1;
        }

        else if (res == CECTofTxt::stFALSE) // Condition failed goto ELSE/END-IF
        {
            destination = _endIndex;
            ctxt.ClearMask(CTXT_SKIP_ELSE);		// Ensure ELSE part is executed
        }

        else
            ctxt.SetMask(CTXT_SKIP_ELSE);		// Ensure ELSE part is bypassed
    }


    ++destination; // Adjust index to 1 based as 0=Success in Script

    if(ctxt.GetVerbose(TRACEMSG))
    {
        if (_exprP)
        {
            string subt_expr = strPrmVal(TParam::stExpression, ctxt);
            const TParam* prmP = param(TParam::stExpression);
            string raw_expr  = prmP->strVal();
            ctxt.log("ELSE IF "+ raw_expr + " THEN <statements> END-IF"
                     " { " + subt_expr + " } " + string(destination?"FALSE":"TRUE"), TRACEMSG);
        }
        else
        {
            ctxt.log("ELSE ...", TRACEMSG);
        }
    }

    ctxt.set_phase_text(PHASE_NONE);

    return destination;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Eloop (End loop) record 'prepare' function. The Eloop record
//	    has the following format:
//
//	    <PRE>
//	    Eloop
//	    </PRE>
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdEndLoop::prepare(TScript* script, int objIndex, bool verbose)
{
    // Report error if no corresponding Loop for Eloop.
    if (script->OpenObjects().empty())
    {
        MSGERR("END-LOOP Failed during PREPARE: Missing main LOOP");
        return -1;
    }

    // Update loop data
    _mainLoop = static_cast<TRcrdLoop*>(script->OpenObjects().top()); // Get parent loop
    _mainLoop->SetEndIndex(objIndex);	// Update parent loop with location of end point
    script->OpenObjects().pop();	// Remove current loop from active Loops

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Eloop (End loop) record 'execute' function. The Eloop record
//	    has the following format:
//
//	    <PRE>
//	    Eloop
//	    </PRE>
//
// @rdesc   <PRE>
//	    Return status:
//	    0 (zero) = ???Should never be zero???
//	    n (n>0) = Jump to start of loop, index of start of loop (1 based)
//	    </PRE>
// @parm    Reference to thread context object.
//
int TRcrdEndLoop::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ENDLOOP);
    int destination = -1;

    // Check if Loop has completed.
    if (!ctxt.OpenObjects().empty() && ctxt.OpenObjects().top() == _mainLoop)
    {
        destination = _mainLoop->StartIndex(); // Jump back to start of Loop
    }


    ++destination; // Adjust index to 1 based as 0=Success in Script

    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("END-LOOP " + string(destination?"Next Iteration":"Finished"), TRACEMSG);

    ctxt.set_phase_text(PHASE_NONE);

    return destination;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Eif (Endif) record 'prepare' function. The Eif record has the
//	    following format:
//
//	    <PRE>
//	    Eif
//	    </PRE>
//
//	    This function updates the parent 'If' record with the index of this
//	    Eif record and removes the current 'If' from the active stack.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdEndIF::prepare(TScript* script, int objIndex, bool verbose)
{
    if (script->OpenObjects().empty())
    {
        MSGERR ("END-IF Failed during PREPARE: Missing parent IF-THEN");
        return -1;
    }

    _ifThenElse = static_cast<TRcrdFlowCtrl*>(script->OpenObjects().top());
    if (_ifThenElse)
    {
        if (_ifThenElse->isType(TRcrdIF_Then::KEY) || _ifThenElse->isType(TRcrdElse::KEY))
        {
            _ifThenElse->SetEndIndex(objIndex);	// Update parent IF statement with location of end point
            script->OpenObjects().pop();	// Remove current statement from active stack
        }

        else
        {
            MSGERR ("END-IF Failed during PREPARE: Missing parent IF-THEN");
            return -1;
        }
    }
    else
    {
        MSGERR ("END-IF Failed during PREPARE: Missing parent IF-THEN");
        return -1;
    }

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Eif (Endif) record 'execute' function. The Eif record has the
//	    following format:
//
//	    <PRE>
//	    Eif
//	    </PRE>
//
//	    This function executes an Eif record, i.e. it removes the 'If .. Then'
//	    object from the open objects stack.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Execution failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
// @parm    Reference to thread context object.
//
int TRcrdEndIF::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_ENDIF);
    // Check for illegal block structure
    if (ctxt.OpenObjects().empty() || ctxt.OpenObjects().top() != _ifThenElse)
    {
        ctxt.log("END-IF Failed: Missing IF-THEN/ELSE, Jumped into IF-THEN <statements> END-IF block", ERRMSG);
        throw mdbParseException("END-IF Failed: Missing IF-THEN/ELSE, Jumped into IF-THEN <statements> END-IF block");
    }

    ctxt.OpenObjects().pop();		// Remove IF-THEN/ELSE from open objects
    ctxt.SetMask(CTXT_SKIP_ELSE);	// Ensure next ELSE part is bypassed

    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("END-IF", TRACEMSG);

    ctxt.set_phase_text(PHASE_NONE);

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Format record PREPARE function. The Format record has the
//	    following format:
//
//	    <PRE>
//	    Format
//	    to <target_variable>
//	    string <format_string>
//	    sValues <values>
//	    {errlab <label>}
//	    </PRE>
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdFormat::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prmP = param(TParam::stTo);
    if (prmP == NULL) throw mdbParseException("No Target Variable in Format");
    string tgtvar = prmP->strVal();

    prmP = param(TParam::stString);
    if (prmP==NULL) throw mdbParseException("No Format String in Format");
    string s_format = prmP->strVal();

    prmP = param(TParam::stSvalues);
    if (prmP == NULL) throw mdbParseException("No Values in Format");
    string s_values = prmP->strVal();

    // Create variable reference object, for format target variable and
    // encode the variable reference; report an error if encode fails.
    _varP = new TtVarRef(script);
    if (!(_varP->Parse(tgtvar, 0)))
    {
        MSGERR("FORMAT: Failed during 'name' (" + tgtvar + ") PREPARE with: " +
               _varP->Reason());
        return -1;
    }

    // Encode the format string; report an error if it is missing or encode fails.
    if (s_format.length())
    {
        _fmtExprP = new TtChrExpr(script);

        if (!(_fmtExprP->Parse(s_format, 0)))
        {
            MSGERR("FORMAT: Failed in PREPARE of Format String: " +
                   s_format + " with: " + _fmtExprP->Reason());
            return -1;
        }
    }
    else
    {
        MSGERR("FORMAT: Failed during PREPARE, Missing Format String");
        return -1;
    }

    if (s_values.empty())
    {
        MSGERR("FORMAT: Failed during PREPARE, Values Missing");
        return -1;
    }

    // Create value list object, for value list.
    _valListP = new TtValList(script);

    if (!_valListP->Parse(s_values, 0))
    {
        MSGERR ("FORMAT: into: " + tgtvar + " format: " + s_format +
                " values: " + s_values + " Failed during PREPARE with: " + _valListP->Reason());
        return -1;	// Return 'failure' status
    }

    // If verbose set, output tracing information.

    if (verbose)
    {
        MSGTRC ("FORMAT: into: " + tgtvar + " format: " + s_format + " values: " + s_values);
    }

    return 0;
} // end of TRcrdFormat::prepare

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Format record 'execute' function. The Format record has the
//	    following format:
//
//	    <PRE>
//	    Format
//	    to <target_variable>
//	    string <format_string>
//	    sValues <values>
//	    {errlab <label>}
//	    </PRE>
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Execution failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    n (n>0) = Execution failed and errlab specified, record to
//		jump to (1 based).
//	    </PRE>
//
// @parm    Reference to thread context object.
//
int TRcrdFormat::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_FORMAT);

    // Evaluate format string.
    string fmtstr = _fmtExprP->Evaluate(ctxt);
    if (fmtstr.empty())
    {
        string prmVal = strPrmVal(TParam::stString, ctxt);
        ctxt.log("FORMAT FAILED: '" + prmVal + "' \n " + ctxt.ErrorMessage(), ERRMSG);
        throw mdbParseException("FORMAT FAILED: '" + prmVal + "' \n " + ctxt.ErrorMessage());
        //return -1;
    }

    // Evaluate the target variable subscripts within the current context.

    string varidx = _varP->Evaluate(ctxt);

    // Perform the format.
    int status;
    try
    {
        status = ctxt.Format(_varP->Varnam(), varidx, fmtstr, _valListP);
    }
    catch (...)
    {
        status = -1; // Flag format failure.
    }


    int dest = 0;
    if (status != 0)
    {
        int labelIndex;
        try //Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex = PrepareGoto(this,TParam::stErrlab,ctxt.Script(),false);
        }
        catch(...) // No valid goto statement
        {
            throw mdbParseException("Unable to perform Format");
        }
        dest = ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(TP_PHASE_NONE);
    return dest;
} // end of TRcrdFormat::execute

///////////////////////////////////////////////////////////////////////////////
// @mfunc   IfThen (If ... Then) record 'prepare' function. The IfThen record
//	    has the following format:
//
//	    <PRE>
//	    IfThen
//	    expr <condition>
//	    </PRE>
//
//	    where <condition> is a boolean expression string.
//
//	    This function encodes the conditional expression string (if any)
//	    into a more suitable form for run-time evaluation. The encoded
//	    expression is used by the ::execute() function. The function logs
//	    an error message and returns a failure status if the conditional
//	    expression cannot be encoded.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdIF_Then::prepare(TScript* script, int objIndex, bool verbose)
{
    // We DONT want variable substitutions during the PREPARE phase
    const TParam* prmP = param(TParam::stExpression);
    string l_condition = prmP->strVal();

    if (l_condition.empty())
    {
        MSGERR ("IF-THEN: IF ?Unknown? THEN <statements>"
                " Failed during PREPARE, Missing Condition");
        return -1;
    }

    if (!_exprP) _exprP = new TtBoolExpr(script);

    if (_exprP->Parse(l_condition, 0) == false)
    {
        MSGERR ("IF-THEN: IF " + l_condition + " THEN <statements>"
                " Failed during PREPARE with: " + _exprP->Reason());
        return -1;
    }

    if(verbose)
    {
        MSGTRC("IF-THEN " + l_condition + " THEN <statements> END-IF");
    }

    _startIndex = objIndex;
    script->OpenObjects().push(this);	// Make IfThen available to EndIF processing
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   IfThen (If ... Then) record 'execute' function. The IfThen record
//	    has the following format:
//
//	    <PRE>
//	    IfThen
//	    expr <condition>
//	    </PRE>
//
//	    This function executes an IfThen record, i.e. it evaluates the
//	    encoded conditional expression set up by ::prepare() and
//	    determines whether the condition is true or false and so whether
//	    the jump is to be made. The return parameter indicates the result.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Execution failed; error message logged by this routine.
//	    0 (zero) = Condition true (or not present), execute following
//		'then' block.
//	    n (n>0) = Condition false, index of next 'if' block (i.e. Else If,
//		Else or Endif) record to jump to (1 based).
//	    </PRE>
// @parm    Reference to thread context object.
//
int TRcrdIF_Then::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_IFTHEN);
    if (_endIndex == -1) // Error, missing END-IF
    {
        ctxt.log("IF-THEN Failed: Missing END-IF", ERRMSG);
        throw mdbParseException("IF-THEN Failed: Missing END-IF");
        //return -1;
    }


    int destination = -1;
    // Evaluate Conditional Expression, 0=false, non ZERO=true, NULL=Error
    string res;
    res = _exprP->Evaluate(ctxt);

    if (res.empty())
    {
        string subt_expr = strPrmVal(TParam::stExpression, ctxt);
        ctxt.log("IF-THEN Failed: '" + subt_expr + "' \n " + ctxt.ErrorMessage(), ERRMSG);
        throw mdbParseException("IF-THEN Failed: '" + subt_expr + "' \n " + ctxt.ErrorMessage());
        //return -1;
    }

    else if (res == CECTofTxt::stFALSE)		// Condition failed goto ELSE/END-IF
    {
        destination = _endIndex;
        ctxt.ClearMask(CTXT_SKIP_ELSE);		// Ensure ELSE part is executed
    }

    else
    {
        ctxt.SetMask(CTXT_SKIP_ELSE);		// Ensure ELSE part is bypassed
    }


    ctxt.OpenObjects().push(this);

    ++destination; // Adjust index to 1 based as 0=Success in Script

    if(ctxt.GetVerbose(TRACEMSG))
    {
        string subt_expr = strPrmVal(TParam::stExpression, ctxt);
        const TParam* prmP = param(TParam::stExpression);
        string raw_expr = "IF-THEN: IF "+ prmP->strVal() + " THEN <statements> END-IF"
                  " { " + subt_expr + " } " + string(destination?"FALSE":"TRUE");
        ctxt.log (raw_expr, TRACEMSG);
    }

    ctxt.set_phase_text(PHASE_NONE);

    return destination;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Label (Label) record 'prepare' function. The Label record has the
//	    following format:
//
//	    <PRE>
//	    Label
//	    name <label>
//	    </PRE>
//
//	    This function adds the specified label name to the labels map.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdLabel::prepare(TScript* script, int objIndex, bool verbose)
{
    // Get label name from 'name' parameter.
    const TParam* prmP = param(TParam::stName);
    string s_name = prmP->strVal();

    if (s_name.empty())
    {
        MSGERR("LABEL: ?Unknown? Failed during PREPARE, Missing Label");
        return -1;
    }

    // Add label to label map.
    script->AddLabel(s_name, objIndex);

    // If verbose set, output tracing information.
    if(verbose)
    {
        MSGTRC("LABEL: " + s_name + " at position: " + int_to_string(objIndex));
    }

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Label (Label) record 'execute' function. The Label record has the
//	    following format:
//
//	    <PRE>
//	    Label
//	    name <label>
//	    </PRE>
//
//	    This function simply outputs tracing information if 'verbose' is
//	    set.
//
// @rdesc   <PRE>
//	    Return status:
//	    0 (zero) = Success (always).
//	    </PRE>
// @parm    Reference to thread context object.
//
int TRcrdLabel::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_LABEL);

    if(ctxt.GetVerbose(TRACEMSG))
    {
        string s_name = strPrmVal(TParam::stName, ctxt);
        ctxt.log("LABEL: " + s_name, TRACEMSG);
    }
    ctxt.set_phase_text(PHASE_NONE);

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Load record PREPARE function. The Load record has the
//  	    following format:
//
//	        <PRE>
//	        Load
//	        data <data_code>
//	        {from <timer_name>}
//	        to <variable_name>
//	        </PRE>
//
// @rdesc   <PRE>
//	        Return status:
//	        -1 = Prepare failed; error message logged by this routine.
//	        0 (zero) = Success.
//	        </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdLoad::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prmP = param(TParam::stData);
    if (prmP == NULL) throw mdbParseException("No Load Type in Load");
    _ldData = prmP->intVal();

    prmP = param(TParam::stFrom);
    if (prmP != NULL) _fromTimnam = prmP->strVal();

    prmP = param(TParam::stInto);
	if (prmP == NULL)
		throw mdbParseException("No Target variable in Load");
    string tgtvar = prmP->strVal();
    if (tgtvar.empty()) 
	{
	    prmP = param(TParam::stIntoTimer);
		if (prmP == NULL)
			throw mdbParseException("No Target variable or timer in Load");
		_toTimnam = prmP->strVal();
	}
	else
	{
	    string tgtvar = prmP->strVal();

		// If verbose set, output tracing information.
		if (verbose)
		{
			string msgstr;
			msgstr = "LOAD: ";
			if (_fromTimnam != "") msgstr += " FOR TIMER " + _fromTimnam;
			msgstr += " INTO " + tgtvar;
			MSGTRC(msgstr);
		}

		// Encode the target variable; report an error if it is missing or encode fails.
		if (!tgtvar.empty())
		{
			_tgtVarP = new TtVarRef(script);
			if (!(_tgtVarP->Parse(tgtvar, 0)))
			{
				MSGERR("LOAD: Failed during PREPARE of Target Variable: " + tgtvar +
					   " with: " + _tgtVarP->Reason());
				return -1;
			}
		}
		else
		{
			MSGERR("LOAD: Failed during PREPARE, Target Variable Missing");
			return -1;
		}

		// Get pointer to TRecord for variable - as we need the variable type.
		TRecord* trecP = (*script).GetVardef(_tgtVarP->Varnam());
		if (trecP == NULL) // If not a defined variable
		{
			MSGERR("LOAD: " + tgtvar +
				   " Failed during PREPARE with: Variable '" + _tgtVarP->Varnam() + "' not defined");
			return -1;
		}
		// Set the load type based on the target variable type.
		prmP = trecP->param(TParam::stOptions);
		if (prmP == NULL)
		{
			MSGERR("LOAD: " + tgtvar +
				   " Failed during PREPARE with: Load options missing");
			return -1;
		}
		int tgtopt = prmP-> intVal();       // Target variable options mask
		_toInteger = (tgtopt & VTINT) != 0;
	}

    return 0;
} // end of TRcrdLoad::prepare

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Load record 'execute' function. The Load record has the following
//	    format:
//
//	    <PRE>
//	    Load
//	    data <data_code>
//	    {from <timer_name>}
//	    to <variable_name>
//	    </PRE>
//
//	    This method calls a method on the context to load the data
//	    identified by the data code (and for Load Timer the timer name in
//	    the 'from' parameter) into the specified variable.
//
// @rdesc   <PRE>
//	    Return status:
//	    0 (zero) = Success (always).
//	    </PRE>
// @parm    Reference to thread context object.
//
int TRcrdLoad::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_LOAD);

	if (_toTimnam.size() > 0)
	{
		if(ctxt.GetVerbose(TRACEMSG))
			ctxt.log("LOAD TIMER " + _fromTimnam + " INTO TIMER " + _toTimnam, TRACEMSG);
		if (!ctxt.LoadTimerInTimer(_fromTimnam, _toTimnam))
			ctxt.log("LOAD TIMER FAILED. The FROM timer must be stopped before executing LOAD TIMER.", WRNMSG);
 
		ctxt.StopTimer(_toTimnam) ;
	}
	else
	{
	    if(ctxt.GetVerbose(TRACEMSG))
		    ctxt.log("LOAD TIMER " + _fromTimnam + " INTO " + _tgtVarP->Varnam() , TRACEMSG);
	    ctxt.Load((LOADTYPE)_ldData, _fromTimnam, _tgtVarP, _toInteger);
	}

    ctxt.set_phase_text(PHASE_NONE);

    return 0;
} // end of TRcrdLoad::execute

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Logmsg (Log message) record 'execute' function. The Logmsg record
//	    has the following format:
//
//	    <PRE>
//	    Logmsg
//	    destination <dest_code>
//	    options <opt_code>
//	    lmsg <ltxt_delimiter>
//	    <lmsg_operands>
//	    <ltxt_delimiter>
//	    </PRE>
//
//	    where <lmsg_operands> is a segmented string (as understood by the
//	    TSegStr class) optionally containing text and variable references.
//
// @rdesc   <PRE>
//	    Return status:
//	    0 (zero) = Success (always).
//	    </PRE>
// @parm    Reference to thread context object.
//
int TRcrdLogmsg::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_LOG);

    string s_lmsg = strPrmVal(TParam::stLogMsg, ctxt);//This should convert an expression if required

    // destination defaults to 0
    int s_dest = 0;
    s_dest = intPrmVal(TParam::stDestination, ctxt);

#ifdef __DLLSA__
    // Trace all messages on unique replay
    s_dest = 4;
#endif

    if(ctxt.GetVerbose((LogTypes)s_dest))
        ctxt.log("LOG: " + s_lmsg, (LogTypes)s_dest);

    ctxt.set_phase_text(PHASE_NONE);

    if(s_dest == NOTEMSG)
    {
        CAsyncVirtualUser& VU = (CAsyncVirtualUser&)ctxt;
        VU.OnNoteText(s_lmsg);
    }

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Loop (Loop) record 'prepare' function. The Loop record has the
//	    following format:
//
//	    <PRE>
//	    Loop
//	    name <variable>
//	    initial <expression>
//	    eValue <expression>
//	    {step <expression>}
//	    </PRE>
//
//	    This function encodes the loop control variable reference and the
//	    initial, end and step (if present) value expression strings, into a
//	    more suitable form for run-time evaluation. The encoded expressions
//	    are used by the ::execute() function. The function logs an error
//	    message and returns a failure status if any of the items cannot be
//	    encoded.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
//
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdLoop::prepare(TScript* script, int objIndex, bool verbose)
{
    g_pLogTrace->AddTrace(TRCCAT_RUNNING, TRCLEV_HIGH, "PREPARE FOR LOOP");

    // Get the control variable and expression strings.
    const TParam* prmP = param(TParam::stName);
    if (prmP==NULL) throw mdbParseException("No Control Variable in Loop");
    string l_name = prmP->strVal();

    prmP = param(TParam::stInitial);
    if (prmP==NULL) throw mdbParseException("No Initialisation Expression in Loop");
    string  s_initialExpr = prmP->strVal();

    prmP = param(TParam::stEndValue);
    if (prmP==NULL) throw mdbParseException("No Termination Expression in Loop");
    string s_endExpr  = prmP->strVal();

    prmP = param(TParam::stStep);
    string s_stepExpr = prmP->strVal();

    // If verbose set, output tracing information.
    if (verbose)
    {
        MSGTRC("LOOP: " + l_name + " FROM " + s_initialExpr + " TO " + s_endExpr +
               " STEP BY " + s_stepExpr);
    }

    // Encode the variable reference; report an error if it is missing or encode fails.
    if (l_name.length())
    {
        _varP = new TtVarRef(script);
        if (!(_varP->Parse(l_name, 0)))
        {
            MSGERR("LOOP: Failed during PREPARE of Control Variable: " + l_name +
                   " with: " + _varP->Reason());
            return -1;
        }
    }
    else
    {
        MSGERR("LOOP: Failed during PREPARE, Control Variable Missing");
        return -1;
    }

    // Encode the Initialisation expression - report error if missing.
    if (s_initialExpr.length())
    {
        if (!_initialExprP) _initialExprP = new TtIntExpr(script);

        if (_initialExprP->Parse(s_initialExpr, 0) == false)
        {
            MSGERR("LOOP: Failed in PREPARE of Initialisation Expression: " +
                   s_initialExpr + " with: " + _initialExprP->Reason());
            return -1;
        }
    }
    else
    {
        MSGERR("LOOP: Failed during PREPARE, Missing Initialisation Expression");
        return -1;
    }

    // Encode the Loop End expression - report error if missing.
    if (s_endExpr.length())
    {
        if (!_endExprP) _endExprP = new TtIntExpr(script);

        if (_endExprP->Parse(s_endExpr, 0) == false)
        {
            MSGERR("LOOP: Failed in PREPARE of End Expression: " +
                   s_endExpr + " with: " + _endExprP->Reason());
            return -1;
        }
    }
    else
    {
        MSGERR("LOOP: Failed during PREPARE, Missing End Expression");
        return -1;
    }

    // Encode the Step Value expression - defaults to 1 if not specified.
    if (s_stepExpr.length())
    {
        if (!_stepExprP) _stepExprP = new TtIntExpr(script);

        if (_stepExprP->Parse(s_stepExpr, 0) == false)
        {
            MSGERR("LOOP: Failed in PREPARE of Step Expression: " +
                   s_stepExpr + " with: " + _stepExprP->Reason());
            return -1;
        }
    }

    _startIndex = objIndex;		// Save start location (end loop goto destination)
    script->OpenObjects().push(this);	// Make loop available to end loop processing

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Loop (Loop) record 'execute' function. The Loop record has the
//	    following format:
//
//	    <PRE>
//	    Loop
//	    name <variable>
//	    initial <expression>
//	    eValue <expression>
//	    {step <expression>}
//	    </PRE>
//
//	    This function performs the following tasks:
//
//	    <OL>
//	    <LI>Checks that the Loop is terminated by an Eloop record</LI>
//	    <LI>Initialises the control variable, if this is the first time in.
//	    </LI>
//	    <LI>Adds the step value to the control variable if this is not the
//	    first time in and a step expression has been specified.</LI>
//	    <LI>Evaluates the loop end expression (if any) and sets the
//	    function return value accordingly.</LI>
//	    </OL>
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Execution failed; error message logged by this routine.
//	    0 (zero) = Loop end NOT reached, execute following 'loop' block.
//	    n (n>0) = Loop end reached, index of first object record beyond
//		Eloop to jump to (1 based).
//	    </PRE>
// @parm    Reference to thread context object.
//
int TRcrdLoop::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_LOOP);
    // If no END-LOOP, report and return an error.
    if (_endIndex == -1) // END-LOOP not specified
    {
        ctxt.log("LOOP FAILED: Missing END-LOOP.", ERRMSG);
        throw mdbParseException("LOOP FAILED: Missing END-LOOP.");
    }

    string l_name = strPrmVal(TParam::stName, ctxt);	// Control variable name
    string s_varval;					// Control variable value as string
    long varval;					// Control variable value as numeric
    string result;					// Expression result as string
    long step;						// Step value as numeric
    bool first_iteration = false;			// First loop iteration flag
    string varidx;					// Control variable index

    // If first iteration of loop, initialise loop control variable
    if (ctxt.OpenObjects().empty() || ctxt.OpenObjects().top() != this)
    {
        first_iteration = true;
        if (_initialExprP) // If Intialisation expression specified
        {
            result = _initialExprP->Evaluate(ctxt);

            if (result.empty())
            {
                string s_initialExpr = strPrmVal(TParam::stValue, ctxt);
                ctxt.log("LOOP FAILED: '" + s_initialExpr + "' \n " + ctxt.ErrorMessage(), ERRMSG);
                throw mdbParseException("LOOP FAILED: '" + s_initialExpr + "' \n " + ctxt.ErrorMessage());
            }

            // Set control variable to value of Initialisation expression.
            varval = atol(result.c_str());  // Control variable value
            varidx = _varP->Evaluate(ctxt); // Evaluate control variable index, if any

            if((ctxt.set_variable(_varP->Varnam(), result, varidx)) == -1)
            {
                ctxt.log("LOOP: set control variable failed(" + l_name + ")", ERRMSG);
                throw mdbParseException("LOOP: set control variable failed(" + l_name + ")");
            }
        }
        ctxt.OpenObjects().push(this); // Mark loop as active
    }

    // Always get 'step' value - we use it even on first iteration.

    if (_stepExprP)	// Get user specified step expression, if any
    {
        result = _stepExprP->Evaluate(ctxt);

        if (result.empty())
        {
            string s_stepExpr = strPrmVal(TParam::stStep, ctxt);
            ctxt.log("LOOP FAILED: '" + s_stepExpr + "' \n " + ctxt.ErrorMessage(), ERRMSG);
            throw mdbParseException("LOOP FAILED: '" + s_stepExpr + "' \n " + ctxt.ErrorMessage());
            //return -1;
        }

    }
    else
    {
        result = "1";   // Default step value
    }


    step = atol(result.c_str());    // Step value

    // If this is NOT the first loop iteration, add step value to control variable value.
    if (!first_iteration)
    {
        varidx = _varP->Evaluate(ctxt); // Evaluate control variable index, if any

        // Get current control variable value.
        bool bFound;
        s_varval = ctxt.get_variable(_varP->Varnam(), bFound, false, varidx);
        if(!bFound)
        {
            ctxt.log("LOOP FAILED: No value available for variable (" + l_name + ")", ERRMSG);
            throw mdbParseException("LOOP FAILED: No value available for variable (" + l_name + ")");
        }

        // Get new control variable value and reset control variable.
        varval = atol(s_varval.c_str()) + step;
        char resStr[32];
        sprintf(resStr, "%li", varval);
        result = string(resStr);
        varidx = _varP->Evaluate(ctxt); // Evaluate control variable index, if any

        if((ctxt.set_variable(_varP->Varnam(), result, varidx)) == -1)
        {
            ctxt.log("LOOP FAILED: Cannot 'step' control variable (" + l_name + ")", ERRMSG);
            throw mdbParseException("LOOP FAILED: Cannot 'step' control variable (" + l_name + ")");
        }
    }
    else
    {
        // If verbose set, output tracing information about the loop parameters.
        if(ctxt.GetVerbose(TRACEMSG))
        {
            string trctxt;
            const TParam* prmP = param(TParam::stEndValue);
            string endval = prmP->strVal();

            if (endval.empty())
            {
                trctxt = "Forever";
            }
            else if (step >= 0)
            {
                trctxt = "Until: (" + l_name + " > " + endval + ')';
            }
            else
            {
                trctxt = "Until: (" + l_name + " < " + endval + ')';
            }

            ctxt.log("LOOP " + trctxt, TRACEMSG);
        }
    }

    // Evaluate Loop End Expression and set 'destination' record index.
    int destination = -1;
    if (_endExprP)
    {
        result = _endExprP->Evaluate(ctxt);

        if (result.empty())
        {
            string s_endExpr = strPrmVal(TParam::stEndValue, ctxt);
            ctxt.log("LOOP FAILED: '" + s_endExpr + "' \n " + ctxt.ErrorMessage(), ERRMSG);
            throw mdbParseException("LOOP FAILED: '" + s_endExpr + "' \n " + ctxt.ErrorMessage());
        }

        // If end value exceeded set destination to exit loop. If step value is positive
        // terminate loop if (varval > endval), otherwise terminate it if (varval < endval).
        long endval = atol(result.c_str());
        if (((step >= 0) && (varval > endval)) ||
                ((step < 0) && (varval < endval)))
        {
            destination = _endIndex;	// Terminate loop destination
            ctxt.OpenObjects().pop();	// Terminate Loop
        }

    }

    // If verbose set, output tracing information.
    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("END-LOOP " + string(destination ? "Next Iteration (" + int_to_string(varval) + ')' : "Finished"), TRACEMSG);

    ctxt.set_phase_text(PHASE_NONE);

    return destination+1; // Adjust index to 1 based as 0=Success in Script
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Setvar (set variable) record 'prepare' function. The Setvar record
//	    has the following format:
//
//	    <PRE>
//	    Setvar
//	    name <variable>
//	    value <expression>
//	    {errlab <label>}
//	    </PRE>
//
//	    This function encodes the variable string and expression string into
//	    more suitable forms for run-time evaluation. The encoded expressions
//	    are used by the ::execute() function. This function logs an error
//	    message and returns a failure status if the expression
//	    cannot be encoded.
//
//	    The variable string needs to be encoded as it may be subscripted, in
//	    which case the subscripts need to be evaluated within the execute
//	    function.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
//
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//

int TRcrdSetvar::prepare (TScript* script, int objIndex, bool verbose)
{
    // Get variable string and expression string from 'name' and 'value' parameters.
    const TParam* prmP = param(TParam::stName);
    string l_name = prmP->strVal();
    prmP = param(TParam::stValue);
    string l_value = prmP->strVal();

    // Create variable reference object, for variable to be set.
    _varP = new TtVarRef(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_varP->Parse(l_name, 0)))
    {
        MSGERR("SET VARIABLE: " + l_name + " to " + l_value +
               " Failed during 'name' PREPARE with: " + _varP->Reason());
        return -1;
    }

    // Get pointer to TRecord for variable - as we need information about the variable
    // to process the expression string.
    TRecord* trecP = (*script).GetVardef(_varP->Varnam());
    if (trecP == NULL) // If not a defined variable
    {
        MSGERR("SET VARIABLE: " + l_name + " to " + l_value +
               " Failed during PREPARE with: Variable '" + _varP->Varnam() + "' not defined");
        return -1;
    }

    // Get variable options mask so that we can check the variable type and so
    // determine what type of expression we need (integer or character).
    prmP = trecP->param(TParam::stOptions);
    if (prmP == NULL) // If no 'options'  parameter (should always be present).
    {
        MSGERR("SET VARIABLE: " + l_name + " to " + l_value +
               " Failed during 'value' PREPARE with: Variable options missing");
        return -1;
    }
    int options = prmP->intVal();   // Variable options mask

    // Create an expression object (of the appropriate type)
    if (_exprP) delete _exprP;	// Delete any existing object first

    if (options & VTCHAR)
    {
        _exprP = new TtChrExpr(script);
    }
    else if (options & VTINT)
    {
        _exprP = new TtIntExpr(script);
    }
    else // Inrecognised variable type.
    {
        MSGERR("SET VARIABLE: " + l_name + " to " + l_value +
               " Failed during 'value' PREPARE with: Unrecognised variable type");
        return -1;
    }

    // Encode the expression; report an error if encode fails.
    if (!(_exprP->Parse(l_value, 0)))
    {
        MSGERR("SET VARIABLE: " + l_name + " to " + l_value +
               " Failed during 'value' PREPARE with: " + _exprP->Reason());
        return -1;
    }
    _labelIndex=PrepareGoto(this,TParam::stErrlab,script,false);

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Setvar (set variable) record 'execute' function. The Setvar record
//	    has the following format:
//
//	    <PRE>
//	    Setvar
//	    name <variable>
//	    value <expression>
//	    {errlab <label>}
//	    </PRE>
//
//	    This function evaluates the expression within the current context
//	    and sets the specified variable to this value.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Execution failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
//
// @parm    Reference to thread context object.
//
int TRcrdSetvar::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_SETVAR);
    bool had_error=false;
    // Get the name of the variable to set from the 'name' parameter and
    // the expression string (from which to set it) from the 'value' parameter.
    // These are used for logging purposes only.
    string l_name  = strPrmVal(TParam::stName, ctxt);	// Name of variable to set
    string l_value = strPrmVal(TParam::stValue, ctxt);	// Expression used to set variable

    // Evaluate the variable subscripts within the current context.
    string varidx;
    {
        varidx = _varP->Evaluate(ctxt);

        // Evaluate the expression whose value is to be assigned to the target variable
        // within the current context.
        if (_exprP)
        {
            l_value = _exprP->Evaluate(ctxt);
        }

        if(ctxt.GetVerbose(TRACEMSG))
            ctxt.log("SET VARIABLE: " + l_name + " to value " + l_value + ",Index " + varidx, TRACEMSG);

        if((ctxt.set_variable(_varP->Varnam(), l_value, varidx)) == -1)
        {
            had_error=true;
        }
        if(had_error)
        {
            if(_labelIndex==0)//no jump set use On Error Goto
                throw mdbParseException("set variable failed(" + l_name + ")");
            //if jump set use this
            return ExecuteGoto(this,ctxt,TParam::stErrlab,&_labelIndex);
        }
    }//end of mutex guard
    ctxt.set_phase_text(PHASE_NONE);
    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for GenVar Record
// ======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdGenvar::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_GENVAR);
    string  l_name = strPrmVal(TParam::stName, ctxt);

    CLock guard (ctxt.GetVarFl());

    CECVar* var=ctxt.get_varlist().get_variable_ptr(l_name);
    if(var==NULL)
    {
        throw mdbParseException("Variable " + l_name + " does not exist in this scope");
    }
    string strCmd;
    int l_mode=intPrmVal(TParam::stMode,ctxt);
    // Mode option enum (Sequential=0, Random=1, Reset=2)
    switch (l_mode)
    {
    case 0:
        strCmd = "NEXT: ";
        var->next_value();
        break;
    case 1:
        strCmd = "GENERATE: ";
        var->generate_value();
        break;
    case 2:
        strCmd = "RESET: ";
        var->reset_seed();
        break;
    default:
        mdbParseException("Unknown Mode for Generate Command");
    }
    if(ctxt.GetVerbose(TRACEMSG))
    {
        ctxt.log(strCmd + l_name, TRACEMSG);
    }
    ctxt.set_phase_text(PHASE_NONE);
    return 0;
} // end of TRcrdGenvar::execute
///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Seed Record
// ======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdSeed::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_SEED);
    string  l_name = strPrmVal(TParam::stName, ctxt);

    CLock guard (ctxt.GetVarFl());

    CECVar* var=ctxt.get_varlist().get_variable_ptr(l_name);
    if(var==NULL)
    {
        string msg="Variable "+l_name+" does not exist in this scope";
        throw mdbParseException(msg);
    }
    int l_value=intPrmVal(TParam::stValue,ctxt);
    var->reset_seed(l_value);
    ctxt.set_phase_text(PHASE_NONE);
    return 0;
} // end of TRcrdSeed::execute

// *****************************************************************************
// * SYNCHRONISATION
// *****************************************************************************

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Wait Record
// ==============================
///////////////////////////////////////////////////////////////////////////////
// @mfunc

int TRcrdWait::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = dynamic_cast<CAsyncVirtualUser&> (ctxt);

    switch(VU.GetTRecordCallStep())
    {
    case 0:
        {
            ctxt.set_phase_text(PHASE_WAIT);

            unsigned nDelay = intPrmVal(TParam::stTDelay, ctxt);

            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("WAIT : Delay: " + int_to_string(nDelay), TRACEMSG);

            if(VU.GetTrace() && VU.GetTrace()->IsTrace("sync"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdWait: Initiating a wait operation for user " << VU.GetManagerNumber() << ". Duration is " << nDelay << ".\r\n";
                ReleaseTraceStream();
            }

            VU.SetTRecordCallStep(1);   // Will enter the "case 1" section next time execute is called.
            if(VU.StartEventTimer(nDelay) == FALSE)
            {
                assert(0);
                return TOF_RES_FAILED;
            }

            return TOF_RES_ASYNC_STARTED;  // Should let the framework call us again when the timer op. is completed.
        }

    case 1:
        {
            if(VU.GetTrace() && VU.GetTrace()->IsTrace("sync"))
            {
                GetTraceStream(VU.GetTrace()) << "TRcrdWait: Waking up after a wait operation for user " << VU.GetManagerNumber() << ".\r\n";
                ReleaseTraceStream();
            }

            ctxt.set_phase_text(PHASE_NONE);

            VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.

            return TOF_RES_OK;   // Move to next TOF object.
        }

    default:
        VU.SetTRecordCallStep(0);   // Reset call phase for the next call to an async TRecord::execute.
        ctxt.set_phase_text(PHASE_NONE);
        assert(0);

        return TOF_RES_FAILED;
    };
}

///////////////////////////////////////////////////////////////////////////////
// PREPARE method for BREAK AND CONTINUE Record
// ============================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdBreakContinue::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prmP = param(TParam::stDepth);
    unsigned int depth = prmP->intVal();
    if (depth < 1)
    {
        MSGERR("Break/Conctinue Failed during PREPARE: Illegal 'depth' value(" +
               int_to_string(depth) + ")");
        return -1;
    }

    // Check for attempt to break out of too many Loops
    if (script->OpenObjects().size() < depth)
    {
        MSGERR("Break/Conctinue Failed during PREPARE: " +
               string(depth==1?"Missing main LOOP":"Depth exceeds number of open LOOPs"));
        return -1;
    }

    prmP = param(TParam::stAction);
    unsigned action = prmP->intVal();
    if(action != 0 && action != 1)
    {
        MSGERR("Break/Conctinue Failed during PREPARE: Unknown Action Type");
        return -1;
    }

    // We DONT want variable substitutions during the PREPARE phase
    prmP = param(TParam::stExpression);
    string l_condition = prmP->strVal();

    if (l_condition.length())	// Optional condition
    {
        if (!_expr)
            _expr = new TExpression();

        if (_expr->SetExpression(l_condition, 0) == false)
        {
            MSGERR("Break/Contnue: IF " + l_condition +
                   " Failed during PREPARE with: " + _expr->Reason());
            return -1;
        }
    }

    return 0;
}


///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for BREAK CONTINUE Record  NOT SUPPORTED (YET!!!!)
// ========================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdBreakContinue::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_BREAKCONTINUE);
    unsigned int depth = intPrmVal(TParam::stDepth, ctxt);

    if (ctxt.OpenObjects().size() < depth)
    {
        ctxt.log("Break/Conctinue Failed: " +
                 string(depth==1?"Missing main LOOP":"Depth exceeds number of open LOOPs") +
                 ", Jumped into middle of LOOP?", ERRMSG);
        throw mdbParseException("Break/Conctinue Failed: " +
                                string(depth==1?"Missing main LOOP":"Depth exceeds number of open LOOPs") +
                                ", Jumped into middle of LOOP?");
    }

    unsigned action = intPrmVal(TParam::stAction, ctxt);
    int destination = -1;
    string result;

    if (_expr)
    {
        result = _expr->Evaluate(ctxt);
    }
    else
        result = TExpression::stTRUE;

    if (result.empty())
    {
        string	s_expr	= strPrmVal(TParam::stExpression, ctxt);
        ctxt.log("Break/Continue Failed: '" + s_expr + "' \n " + ctxt.ErrorMessage(), ERRMSG);
        throw mdbParseException("Break/Continue Failed: '" + s_expr + "' \n " + ctxt.ErrorMessage());
    }
    else if (result != TExpression::stFALSE)
    {
        // If depth > 1 Terminate inner Loops (regardless of action)
        TRecord* flowObject = NULL;
        while (depth)
        {
            flowObject = ctxt.OpenObjects().top();
            // Terminate any IF-THENs and/or inner Loop(s)
            if ( flowObject->isType(TRcrdLoop::KEY) == false || --depth)
                ctxt.OpenObjects().pop();	// Remove from active Objects
        }

        // Perform selected action (break or continue) on outer most Loop

        TRcrdLoop* loop = static_cast<TRcrdLoop*>(flowObject);
        if (action == 0) // Break Loop
        {
            ctxt.OpenObjects().pop();		// Terminate Loop
            destination = loop->EndIndex();	// Jump to end of Loop
        }

        else if (action == 1) // Continue Loop (Next Iteration)
        {
            destination = loop->EndIndex();	// Jump to end of Loop
        }

    }

    ++destination; // Adjust index to 1 based as 0=Success in Script

    if(ctxt.GetVerbose(TRACEMSG))
    {
        string	subt_expr	= strPrmVal(TParam::stExpression, ctxt);
        const TParam* prmP = param(TParam::stExpression);
        string raw_expr  = prmP->strVal();
        ctxt.log("Break/Continue: IF " + raw_expr + string(action?" Continue":" Break") +
                 " LOOP { " + subt_expr + " } " + string(destination?"TRUE":"FALSE"), TRACEMSG);
    }

    ctxt.set_phase_text(PHASE_NONE);

    return destination;
}

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Subroutine record PREPARE function. The Subroutine record has the
//	    following format:
//
//	    <PRE>
//	    Subroutine
//	    name <name>
//	    {parameters <parameter_list>}
//	    </PRE>
//
//	    This function:
//
//	    <PRE>
//	    - Checks that a subroutine name has been specified.
//	    - Saves the Trecord index of the start of the subroutine.
//	    - Encodes any subroutine parameters (as they may be subscripted
//	      in which case the subscripts will need evaluating when the
//	       subroutine is called).
//	    - Saves the subroutine name and a pointer to the parameters
//	      object (TtValList) in a Tscript-wide map for use when the
//	      subroutine is called.
//	    </PRE>
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
//
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdSubroutine::prepare(TScript* script, int objIndex, bool verbose)
{
    // Get subroutine name.
    const TParam* prmP = param(TParam::stName);
    string s_name = prmP->strVal();

    // If there is no subroutine name, report an error.
    if (s_name.empty())
    {
        MSGERR("Subroutine: ?Unknown? Failed during prepare, Missing Name");
        return -1;
    }

    script->AddSubroutineLabel(s_name, objIndex);

    // Get subroutine parameter list, if any.
    prmP = param(TParam::stParameters);
    string s_parameters;
    if (prmP != NULL) s_parameters = prmP->strVal();

    // Create value list object, for parameters, if any.
    TtValList* prmListP = new TtValList(script);

    if (!prmListP->Parse(s_parameters, 0))
    {
        MSGERR ("SUBROUTINE: name: " + s_name + " parameters: " + s_parameters +
                " Failed during PREPARE of parameters with: " + prmListP->Reason());
        return -1;	// Return 'failure' status
    }

    // Add the subroutine details to the subroutine name/parameters map.

    script->AddSubroutineName(s_name, prmListP);

    // Output tracing information.
    MSGTRC ("SUBROUTINE: name: " + s_name + " at position: " +
            int_to_string(objIndex) + " parameters: " + s_parameters);

    return 0;
} // end of TRcrdSubroutine::prepare

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Subroutine record EXECUTE function. The Subroutine record has the
//	    following format:
//
//	    <PRE>
//	    Subroutine
//	    name <name>
//	    {parameters <parameter_list>}
//	    </PRE>
//
//	    This function outputs tracing information.
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Execute failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
//
// @parm    Reference to thread context object.
//
int TRcrdSubroutine::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_SUBROUTINE);
    // Get subroutine name.
    string s_name = strPrmVal(TParam::stName, ctxt);

    // Output tracing information.
    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("SUBROUTINE: " + s_name, TRACEMSG);

    ctxt.set_phase_text(PHASE_NONE);
    return 0;

}// end of TRcrdSubroutine::execute

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Csub (call subroutine) record PREPARE function. The Csub record has
//	    the following format:
//
//	    <PRE>
//	    Csub
//	    name <subroutine_name>
//	    {parameters <parameter_list>}
//	    </PRE>
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Prepare failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
// @parm    Pointer to TScript object for this TOF file.
// @parm    Index of current TRecord object in record list.
// @parm    Verbose flag. true = Verbose set; false = Verbose NOT set.
//
int TRcrdCsub::prepare(TScript* script, int objIndex, bool verbose)
{
    // Get subroutine name and parameters parameter values.
    const TParam* prmP = param(TParam::stName);
    prmP = param(TParam::stName);
    if (prmP == NULL) throw mdbParseException("No Name in Call Subroutine");
    string s_name = prmP->strVal();

    prmP = param(TParam::stParameters);
    string s_parameters;
    if (prmP != NULL) s_parameters = prmP->strVal();

    // Log an error if there is no subroutine name.
    if (s_name.empty())
    {
        MSGERR("Call Subroutine ?Unknown?, Failed during PREPARE, Missing Name");
        return -1;
    }

    // Create and populate a value list object for the parameter list (if any).
    if (!s_parameters.empty())
    {
        _prmListP = new TtValList(script);

        if (!_prmListP->Parse(s_parameters, 0))
        {
            MSGERR ("CALL SUBROUTINE: name: " + s_name + " parameters: " + s_parameters +
                    " Failed during PREPARE with: " + _prmListP->Reason());
            return -1;	// Return 'failure' status
        }

    }

    // Save return location (ReturnProc destination).
    _returnIndex = objIndex + 1;
    return 0;
} // End of TRcrdCsub::prepare

///////////////////////////////////////////////////////////////////////////////
// @mfunc   Csub (call subroutine) record EXECUTE function. The Csub record has
//	    the following format:
//
//	    <PRE>
//	    Csub
//	    name <subroutine_name>
//	    {parameters <parameter_list>}
//	    </PRE>
//
// @rdesc   <PRE>
//	    Return status:
//	    -1 = Execute failed; error message logged by this routine.
//	    0 (zero) = Success.
//	    </PRE>
//.
// @parm    Reference to thread context object.
//
int TRcrdCsub::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_CALLSUB);
    // Get subroutine name.
    string s_name = strPrmVal(TParam::stName, ctxt);

    // Set up Trecord index of the first command in the subroutine (if not already done).
    int labelIndex(-1);
    {
        labelIndex = _labelIndex;
        if (labelIndex == -1)
        {
            // Resolve Jump
            labelIndex = ctxt.Script()->GetSubroutineLabelIndex(s_name);

            if(ctxt.GetVerbose())
                ctxt.log("Call Subroutine: " + s_name + '(' + int_to_string(labelIndex) + ')');

            if (labelIndex == -1)
            {
                // Unknown label, Error
                ctxt.log("Unknown LABEL: " + s_name, ERRMSG);
                throw mdbParseException("Unknown Subroutine: " + s_name);
            }
            _labelIndex = labelIndex;
        }
    }// End mutex guard

    // Save Trecord index of the command at which to resume execution on return from
    // the subroutine.
    ctxt.OpenObjects().push(this);
    ctxt.PushSubroutine(s_name, _returnIndex);

    // Evaluate the actual argument values and set the corresponding parameter variables.
    ctxt.SubParameters(ctxt.Script(), s_name, _prmListP);

    // Return with index of first Trecord in subroutine.
    ctxt.set_phase_text(PHASE_NONE);
    return ++labelIndex;// Adjust index to 1 based as 0=Success in Script

} // End of TRcrdCsub::execute

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Return Record
// ================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc

int TRcrdReturn::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_RETURN);
    TRecord* rec = NULL;

    bool enclosingObjects = (ctxt.OpenObjects().empty() == false);
    // Check if jump exits any active Loops or IF-THEN-ELSEs
    while (enclosingObjects)
    {
        TRcrdFlowCtrl* flowObject = dynamic_cast<TRcrdFlowCtrl*>(ctxt.OpenObjects().top());
        if(flowObject)
        {
            // Tidy up Flow Control structures.
            ctxt.OpenObjects().pop();
            enclosingObjects = (ctxt.OpenObjects().empty() == false);
        }
        else
            enclosingObjects = false;
    }

    if (!ctxt.OpenObjects().empty())
        rec = ctxt.OpenObjects().top();
    if (rec == NULL || rec->isType(TRcrdCsub::KEY) == false)
        throw mdbParseException("Return Proc Failed: CallStack underflow, or Invalid return target");

    ctxt.OpenObjects().pop();
    int destination=ctxt.GetCurrentSubroutineData().ReturnIndex;

    if(ctxt.GetVerbose(TRACEMSG))
        ctxt.log("RETURN PROC to "+ int_to_string(++destination), TRACEMSG);

    ctxt.PopSubroutine();
    ctxt.set_phase_text(PHASE_NONE);
    return destination;
}// end of TRcrdReturn::execute

///////////////////////////////////////////////////////////////////////////////
// PREPARE method for GOTO Record
// ======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc prepare method for Goto
int TRcrdGoto::prepare(TScript* script, int objIndex, bool verbose)
{
    _labelIndex=PrepareGoto(this,TParam::stGolab,script,true);
    return 0;

}// end of TRcrdGoto::prepare

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for GOTO Record
// ==============================
///////////////////////////////////////////////////////////////////////////////
int TRcrdGoto::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_GOTO);

    int iret = ExecuteGoto(this, ctxt, TParam::stGolab, &_labelIndex, m_pFastLock);

    ctxt.set_phase_text(PHASE_NONE);
    return iret;
}// end of TRcrdGoto::execute


// *****************************************************************************
// * ERROR HANDLING/TERMINATION COMMANDS
// *****************************************************************************
int TRcrdGoError::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_GOERROR);
    string  s_errlab = strPrmVal(TParam::stErrlab, ctxt);
    int labelIndex = ctxt.GetLabelIndex(s_errlab);
    ctxt.SetErrorLabel(labelIndex);
    ctxt.set_phase_text(PHASE_NONE);
    return 0;

}

int TRcrdDisable::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_DISABLE);
    int	l_target = intPrmVal(TParam::stTarget, ctxt);

    switch(l_target)
    {
    case TOF_ENDTGT_ONERROR:    //Cancel On Error
        ctxt.CancelOnError();
        break;
    default://do nowt
        MSGERR("TRcrdDisable: target value not handled: " + int_to_string(l_target));
        break;
    }

    ctxt.set_phase_text(PHASE_NONE);
    return 0;
}

int TRcrdExit::execute(TContext& ctxt)
{
    CAsyncVirtualUser& VU = (CAsyncVirtualUser&)ctxt;

    DWORD dwCallStep = VU.GetTRecordCallStep();
    switch(dwCallStep)
    {
    case 0:
        {
            ctxt.set_phase_text(PHASE_EXIT);
            // Post a cancel message to all active connections.
            VU.CancelAllQueries();

            // Wait for all the requests to finish.
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("TRcrdExit, step 0: Cancelled all requests. Now waiting for the terminating script requests to complete.", TRACEMSG);

            VU.SetTRecordCallStep(1);       // Will enter the "case 1" section next time execute is called.
            if(VU.SynchroniseAllHTTPClients() == FALSE)
            {
                return TOF_RES_ASYNC_STARTED;   // Should let the framework call us again when the timer op. is completed.
            }

        }
        // No break.

    case 1:
        {
            // Close all the connections.
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("TRcrdExit, step 1: Script requests completed. Disconnecting all clients.", TRACEMSG);

            VU.SetTRecordCallStep(2);       // Will enter the "case 1" section next time execute is called.
            if(VU.DisconnectAllClients() == TOF_RES_ASYNC_STARTED)
                return TOF_RES_ASYNC_STARTED;
        }
        // No break.

    case 2:
        {
            if(ctxt.GetVerbose(TRACEMSG))
                ctxt.log("TRcrdExit, step 2: Script clients disconnected.", TRACEMSG);

            int	    l_options = intPrmVal(TParam::stOptions, ctxt);
            string  s_status = strPrmVal(TParam::stStatus, ctxt);

            bool keepalive = (l_options==1) ? true : false;

            ctxt.ExitCommand(keepalive, s_status);

            break;
        }

    default:
        {
            // Unknown error.
            ctxt.set_phase_text(PHASE_NONE);
            assert(0);
            return TOF_RES_FAILED;
        }
    };


    // Do not set this since TRcrdCscr will have to use the calling script value. VU.SetTRecordCallStep(0);
    ctxt.set_phase_text(PHASE_NONE);
    return TOF_RES_OVER;  // End Script here
}

// *****************************************************************************
// * FILE HANDLING
// *****************************************************************************

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Open Record
// ==============================
///////////////////////////////////////////////////////////////////////////////
// @mfunc

int TRcrdOpen::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_OPEN);
    string	s_filename	= strPrmVal(TParam::stName, ctxt);
    string	s_fileid	= strPrmVal(TParam::stFileId, ctxt);
    int		l_access	= intPrmVal(TParam::stAccess, ctxt);
    int		l_options	= intPrmVal(TParam::stOptions, ctxt);
    bool error=false;
    try
    {
        ctxt.FileOpen(s_filename, s_fileid, l_access, l_options);
    }
    catch(...)
    {
        error=true;//have caught an error that is not a timeout
    }

    int dest=0;
    if(error)// only do error processing if an error(keep efficiency for 'correct' runs)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException(string("Unable to open file ")+s_filename);
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(PHASE_NONE);
    return dest;

}// end of TRcrdOpen::execute

///////////////////////////////////////////////////////////////////////////////
// Prepare method for Read Record
// ==============================
///////////////////////////////////////////////////////////////////////////////
int TRcrdRead::prepare(TScript* script, int objIndex, bool verbose)
{
    const TParam* prm = param("variable");
    if(prm == NULL) return -1;
    string l_varref=prm->strVal();
    _varP = new TtVarRef(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_varP->Parse(l_varref, 0)))
    {
        MSGERR("Read File: " + l_varref +
               " Failed during 'name' PREPARE with: " + _varP->Reason());
        return -1;
    }

    // Get pointer to TRecord for variable - as we need information about the variable
    // to process the expression string.
    TRecord* trecP = (*script).GetVardef(_varP->Varnam());
    if (trecP == NULL) // If not a defined variable
    {
        MSGERR("Read File: " + l_varref +
               " Failed during PREPARE with: Variable '" + _varP->Varnam() + "' not defined");
        return -1;
    }

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Read Record
// ==============================
///////////////////////////////////////////////////////////////////////////////

int TRcrdRead::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_READ);
    string	s_fileid	= strPrmVal(TParam::stFileId, ctxt);
    int filestat=0;
    bool error=false;
    try
    {
        filestat=ctxt.FileRead(s_fileid, _varP);
    }
    catch(...)
    {
        error=true;//have caught an error that is not a timeout
    }

    int dest=0;
    if(filestat==-1)//End of File
    {
        try
        {
            int labelIndex;
            // will set up dest to be 0 if no goto label
            labelIndex=PrepareGoto(this,"endlab",ctxt.Script(),false);
            dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
            if(dest==0)//if no goto rewind file
                ctxt.FileControl(TOF_FCNOP_REWIND,s_fileid);
        }
        catch(...)
        {
            error=true;
        }
    }
    if(error)// only do error processing if an error(keep efficiency for 'correct' runs)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException(string("Unable to Read from fileid ")+s_fileid);
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(PHASE_NONE);
    return dest;
}// end of TRcrdRead::execute


///////////////////////////////////////////////////////////////////////////////
// Prepare method for Write Record
// ==============================
///////////////////////////////////////////////////////////////////////////////
int TRcrdWrite::prepare(TScript* script, int objIndex, bool bVerbose)
{
    const TParam* prm = param("string");
    if(prm == NULL)
        return -1;

    string l_data = prm->strVal();
    _exprP = new TtChrExpr(script);

    // Encode the variable reference; report an error if encode fails.
    if (!(_exprP->Parse(l_data, 0)))
    {
        MSGERR("Write File: " + l_data + " Failed during 'name' PREPARE with: " + _exprP->Reason());
        return -1;
    }

    return 0;
}

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Write Record
// ======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc

int TRcrdWrite::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_WRITE);

    string	s_fileid	= strPrmVal(TParam::stFileId, ctxt);
    string	s_data;
    s_data = _exprP->Evaluate(ctxt);
    bool error=false;
    try
    {
        ctxt.FileWrite(s_fileid, s_data);
    }
    catch(...)
    {
        error=true;//have caught an error that is not a timeout
    }

    int dest=0;
    if(error)// only do error processing if an error(keep efficiency for 'correct' runs)
    {
        int labelIndex;
        try//Prepare the goto if there is no goto throw to ON error goto
        {
            labelIndex=PrepareGoto(this,TParam::stErrlab,ctxt.Script(),true);
        }
        catch(...)//no valid goto statement
        {
            throw mdbParseException(string("Unable to Write to fileid ")+s_fileid);
        }
        dest=ExecuteGoto(this,ctxt,TParam::stErrlab,&labelIndex);
    }
    ctxt.set_phase_text(PHASE_NONE);
    return dest;

}// end of TRcrdWrite::execute

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Fcntl Record
// ======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc

int TRcrdFcntl::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_FCNTL);
    int		l_operation	= intPrmVal(TParam::stFileOperation, ctxt);
    string	s_fileid	= strPrmVal(TParam::stFileId, ctxt);
    string	s_errlab	= strPrmVal(TParam::stErrlab, ctxt);

    ctxt.FileControl(l_operation, s_fileid);
    // On Error GOTO label
    ctxt.set_phase_text(PHASE_NONE);
    return 0;
}// end of TRcrdFcntl::execute

// *****************************************************************************
// * FORMAL TEST CONTROL
// *****************************************************************************

///////////////////////////////////////////////////////////////////////////////
// PREPARE method for Tcstart Record
// ======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc Empty all relevent information is done at execute time
int TRcrdTcstart::prepare(TScript* script, int objIndex, bool verbose)
{
    script->TestCaseStarted();
    return 0;

}// end of TRcrdTcstart::prepare

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Tcstart Record
// ======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc execute method for Starting Testcases
int TRcrdTcstart::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_TCSTART);
    string desc = strPrmVal(TParam::stDescription, ctxt);
    ctxt.TestCaseDesc(desc);
    ctxt.set_phase_text(PHASE_NONE);
    return 0;
}// end of TRcrdTcstart::execute


///////////////////////////////////////////////////////////////////////////////
// PREPARE method for Tcterm Record
// ======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc prepare method for Terminating Testcases with pass or fail
int TRcrdTcterm::prepare(TScript* script, int objIndex, bool verbose)
{
    string l_label="";
    // Get and save label index.if one
    _labelIndex=PrepareGoto(this,TParam::stGolab,script,false);
    return 0;
}// end of TRcrdTcterm::prepare

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Tcterm Record
// ================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc
int TRcrdTcterm::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_TCTERM);
    // DO NOT use standard Goto commands(ExecuteGoto) as if there is no label
    // the EndTestCase statement is used
    int	    l_result	= intPrmVal(TParam::stResult, ctxt);
    //get current test case description from script for use in output
    string _description=ctxt.TestCaseDesc();
    ctxt.TCterminate(l_result,_description);
    if(_labelIndex==0)//if no explicit goto use next end test case command object index
        _labelIndex=ctxt.GetNextEndTestCaseIndex();
    ctxt.set_phase_text(PHASE_NONE);
    return goto_destination(_labelIndex+1,ctxt);//jump to record after the EndTestcase
}// end of TRcrdTcterm::execute

///////////////////////////////////////////////////////////////////////////////
// PREPARE method for Tcend Record
// ======================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc prepare method for Ending Testcases, add index of this end test case to script

int TRcrdTcend::prepare(TScript* script, int objIndex, bool verbose)
{
    if(script->IsTestCaseStarted())
        script->AddEndTestCase(objIndex);
    else
        throw mdbParseException("End Test-Case statement without corresponding Start Test-Case");
    return 0;
}// end of TRcrdTcend::prepare

///////////////////////////////////////////////////////////////////////////////
// EXECUTE method for Tcend Record
// =====================================
///////////////////////////////////////////////////////////////////////////////
// @mfunc Treat as a test case passed, jumps from pass or fail test
//	  case should start at the line after this
int TRcrdTcend::execute(TContext& ctxt)
{
    ctxt.set_phase_text(PHASE_TCEND);
    //get current test case description from script for use in output
    ctxt.TCterminate(TOF_TCTRES_PASS, ctxt.TestCaseDesc());
    ctxt.set_phase_text(PHASE_NONE);
    return 0;
}// end of TRcrdTcend::execute


// *****************************************************************************
// NEW TEST CONTROL INSTRUCTION RECORDS
// *****************************************************************************

#include "TInterRecords.cpp"

#ifdef CYRANO_TEST_MANAGER
#   include "TManagerRcrd.cpp"
#elif defined (CYRANO_TEST_EXECUTER)
#   include "TEGenericRcrd.cpp"
#   include "TExecuterRcrd.cpp"
#elif defined (CYRANO_HTTP_EXECUTER)
#   include "TExecuterRcrd.cpp"
#else
#   error Code not yet complete TRecord type not defined
#endif

///////////////////////////////////////////////////////////////////////////////
// HELPER FUNCTIONS
// ================
///////////////////////////////////////////////////////////////////////////////
int goto_destination(int destination,TContext& ctxt)
{
    // Jumping to new position
    if (destination >= 0)
    {
        bool enclosingObjects = (ctxt.OpenObjects().empty() == false);
        // Check if Jump exits any active Loops or IF-THEN-ELSEs
        while (enclosingObjects)
        {
            TRcrdFlowCtrl* flowObject = dynamic_cast<TRcrdFlowCtrl*>(ctxt.OpenObjects().top());
            if (flowObject &&
                (destination < flowObject->StartIndex() || destination > flowObject->EndIndex())
               )
            {
                // Tidy up Flow Control structures.
                ctxt.OpenObjects().pop();
                enclosingObjects = (ctxt.OpenObjects().empty() == false);
            }
            else
                enclosingObjects = false;
        }
        ++destination;
    }

    // Adjust index to 1 based as 0=Success in Script
    return destination;
}

// @func For all TRecords that require a Goto statement
//	 Set up the data in the TRecord prepare method for jumping
// @parm Record undergoing prepare
// @parm name of parameter to convert to index label
// @parm current script
// @parm is a label mandatory
// @rdesc index in script to jump to
// -1 for forward jump (unable to resolve yet)
// 0  for no jump (blank error label)
// >0 for label to jump to

int PrepareGoto(const TRecord* rec, const string& errlab, const TScript* script, bool required)
{
    int labelIndex=0;//no goto is 0 (i.e. jump nowhere)
    const TParam* prmP = rec->param(errlab);
    string _errlab("");

    if(prmP!= NULL)
        _errlab= prmP->strVal();

    if((_errlab.empty()||prmP==NULL)&&required)
        throw mdbParseException(" GOTO ?Unknown? Failed during PREPARE, Missing Label");

    // Get and save label index.if one set to -1 if forward jump
    if(!_errlab.empty())
        labelIndex = script->GetLabelIndex(_errlab,script->GetCurrentSubroutineName());
    return labelIndex;
}

// @func For all TRecords that require a Goto statement
//	 calculate the required jump
// @parm Record undergoing execute
// @parm Current Context(thread)
// @parm name of parameter to convert to index label
// @parm    static member variable of TRecord that contains a label index
//	    this should be mutex locked on access to set up
//	    if this index is >0 then the calculation has already been done
//	    for another thread and the calculation does not need to be repeated
// @parm static mutex member variable for class to lock previous parameter
// @rdesc return value including jump
// 0  for no jump (successful record completion)
// >0 for label to jump to
int ExecuteGoto(const TRecord* rec,TContext& ctxt,const string& errlab,int* _cur_label_index, CFastLock *pFL)
{
    int labelIndex = -1;

    {
        labelIndex = *_cur_label_index;
        if(labelIndex==0)
            return labelIndex;//if set to no jump (0) then return no jump (i.e. success)

        // Get 'goto' label index from 'name' parameter.
        string s_name = rec->strPrmVal(errlab, ctxt);
        if (labelIndex == -1) // Computed Goto or Forward Jump
        {
            labelIndex = ctxt.GetLabelIndex(s_name); // Resolve Jump
            if (labelIndex == -1) // Unknown label, Error
            {
                ctxt.log("Unknown LABEL: " + s_name, ERRMSG);
                throw mdbParseException("Unknown LABEL: " + s_name);
            }
            // Transaction records are common across ALL threads,
            // therefore need to Mutex this run-phase assignment
            *_cur_label_index = labelIndex;
        }
    }

    if (labelIndex>= 0) // Jumping to new position
    {
        bool enclosingObjects = (ctxt.OpenObjects().empty() == false);
        // Check if Jump exits any active Loops or IF-THEN-ELSEs
        while (enclosingObjects)
        {
            TRcrdFlowCtrl* flowObject = dynamic_cast<TRcrdFlowCtrl*>(ctxt.OpenObjects().top());
            if (flowObject &&
                    (labelIndex < flowObject->StartIndex() ||
                     labelIndex > flowObject->EndIndex()
                    )
               )
            {
                // Tidy up Flow Control structures.
                ctxt.OpenObjects().pop();
                enclosingObjects = (ctxt.OpenObjects().empty() == false);
            }
            else
                enclosingObjects = false;
        }
    }
    ++labelIndex; // Adjust index to 1 based as 0=Success in Script
    return labelIndex;
}

int ExecuteGotoLabel(const TRecord* rec, TContext& ctxt, const string& label_name, CFastLock *pFL)
{
    int labelIndex;
    labelIndex = ctxt.GetLabelIndex(label_name); // Resolve Jump

    if (labelIndex == -1) // Unknown label, Error
    {
        ctxt.log("Unknown LABEL: " + label_name, ERRMSG);
        throw mdbParseException("Unknown LABEL: " + label_name);
    }

    if (labelIndex>= 0) // Jumping to new position
    {
        bool enclosingObjects = (ctxt.OpenObjects().empty() == false);
        // Check if Jump exits any active Loops or IF-THEN-ELSEs
        while (enclosingObjects)
        {
            TRcrdFlowCtrl* flowObject = dynamic_cast<TRcrdFlowCtrl*>(ctxt.OpenObjects().top());
            if (flowObject &&
                (labelIndex < flowObject->StartIndex() || labelIndex > flowObject->EndIndex())
               )
            {
                // Tidy up Flow Control structures.
                ctxt.OpenObjects().pop();
                enclosingObjects = (ctxt.OpenObjects().empty() == false);
            }
            else
                enclosingObjects = false;
        }
    }
    ++labelIndex; // Adjust index to 1 based as 0=Success in Script
    return labelIndex;
}

// *****************************************************************************
// TRecord & derivatives initialization / uninitialization.
// *****************************************************************************
bool TRecord::InitClass()
{
    extern void InitHTTP();
    bool bRes = true;

    TRcrdAbort          ::_RegisterCreator();
    TRcrdBreakContinue  ::_RegisterCreator();
    TRcrdCond           ::_RegisterCreator();
    TRcrdConvert        ::_RegisterCreator();
    TRcrdCsub           ::_RegisterCreator();
    TRcrdDisable        ::_RegisterCreator();
    TRcrdElse           ::_RegisterCreator();
    TRcrdEndIF          ::_RegisterCreator();
    TRcrdEndLoop        ::_RegisterCreator();
    TRcrdEtimer         ::_RegisterCreator();
    TRcrdExit           ::_RegisterCreator();
    TRcrdFcntl          ::_RegisterCreator();
    TRcrdFormat         ::_RegisterCreator();
    TRcrdGenvar         ::_RegisterCreator();
    TRcrdGoError        ::_RegisterCreator();
    TRcrdGoto           ::_RegisterCreator();
    TRcrdIF_Then        ::_RegisterCreator();
    TRcrdLabel          ::_RegisterCreator();
    TRcrdLoad           ::_RegisterCreator();
    TRcrdLogmsg         ::_RegisterCreator();
    TRcrdLoop           ::_RegisterCreator();
    TRcrdOpen           ::_RegisterCreator();
    TRcrdRead           ::_RegisterCreator();
    TRcrdReturn         ::_RegisterCreator();
    TRcrdSeed           ::_RegisterCreator();
    TRcrdSetvar         ::_RegisterCreator();
    TRcrdStimer         ::_RegisterCreator();
    TRcrdSubroutine     ::_RegisterCreator();
    TRcrdTcend          ::_RegisterCreator();
    TRcrdTcstart        ::_RegisterCreator();
    TRcrdTcterm         ::_RegisterCreator();
    TRcrdWait           ::_RegisterCreator();
    TRcrdWrite          ::_RegisterCreator();

    // Records not for the thread pool DLL.
#ifndef __DLLSA__
    TRcrdAcmutex        ::_RegisterCreator();
    TRcrdClsemaphore    ::_RegisterCreator();
    TRcrdRlmutex        ::_RegisterCreator();
    TRcrdStsemaphore    ::_RegisterCreator();
    TRcrdWtsemaphore    ::_RegisterCreator();
#endif
    // Specific records.
#ifdef CYRANO_TEST_MANAGER
    // Test Manager.
    TRcrdExecuteTest    ::_RegisterCreator();
    TRcrdExecuteThread  ::_RegisterCreator();
    TRcrdWtest          ::_RegisterCreator();
    ThrDefRcrd          ::_RegisterCreator();
#elif defined (CYRANO_TEST_EXECUTER)
    // Test Executer.
    TRcrdCscr           ::_RegisterCreator();
    TRcrdEntry          ::_RegisterCreator();
#elif defined (CYRANO_HTTP_EXECUTER)
    // Test Executer.
    TRcrdCscr           ::_RegisterCreator();
    TRcrdEntry          ::_RegisterCreator();

    // HTTP Test Executer.
    InitHTTP();
#endif

    assert(m_pFastLock == NULL);
    if(m_pFastLock == NULL)
        m_pFastLock = new CFastLock();

    if(m_pFastLock == NULL)
        bRes = false;

    return bRes;
}

void TRecord::UninitClass()
{
    if(recCreatorMap)
    {
        delete recCreatorMap;
        recCreatorMap = NULL;
    }

    if(m_pFastLock != NULL)
    {
        delete m_pFastLock;
        m_pFastLock = NULL;
    }
}
