/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifdef WIN32
#pragma warning(disable:4786)
#pragma warning(disable:4503)
#endif

#include "CORBARepository.h"
#include "threadcontrol.h"
#include "mdbdelay.h"
#include "mdbexcept.h"
#ifndef CYRANO_TP_STANDALONE
#include "TestProcess.h"
#include <Objects/SpawnProc.h>
#else
#include <ctime.hpp>
#endif

#include <assert.h>

int server_specific_init(void);
int server_specific_cleanup(void);

bool Cthreadcontrol::aborting = false;


/////////////////////////////////////////////////////////////////////////////
// Cthreadcontrol

// Constructor defining maximum threads
Cthreadcontrol::Cthreadcontrol(TScriptRule* rule, int max_threads)
        : scriptCache(rule), maxthrd(max_threads)
{
    //if 0 threads then have the default
    if(maxthrd<1)maxthrd=DEFAULT_THREAD_MAX;
    // Perform first time initialization
    threads.clear();
    server_specific_init();
}

// Destructor
Cthreadcontrol::~Cthreadcontrol()
{
    //
    //    Loop through the vector of threads and delete them.
    //
    // Set static abort flag
    aborting = true;

    // Attempt to abort running threads.
    if(!threads.empty())
    {
        int i;
        // Set end all threads normally
        logr->AddTrace(SHUTDOWN,DEBUGLEVEL,"Tell all threads to end");
        for(i = 0 ; i < (int)threads.size() ; i++)
        {
            if(threads[i] != NULL)
                threads[i]->end_thread();
        }
        //Wait for all threads to finish up to a maximum of 30 seconds
        CTTime tm;
        tm.now();
        unsigned long StartTime=tm.getSeconds();
        bool abort=false;
#ifndef CYR_NO_STD_NAMESPACE
        for (std::vector<Cmdbthread*>::iterator iter=threads.begin(); iter!=threads.end(); iter++)
#else
        for (vector<Cmdbthread*>::iterator iter=threads.begin(); iter!=threads.end(); iter++)
#endif
        {
            CTTime now;
            now.now();
            long t=now.getSeconds()-StartTime;
            if(t<0)t*=-1;
            if(t>30)abort=true;//if greater that 30 seconds wait 0
            if(*iter)
            {
                string msg="Wait for Thread:"+(*iter)->GetThreadID();
                logr->AddTrace(THREADING,DEBUGLEVEL,msg.c_str());
                (*iter)->Terminate();
                //report of status of thread before deleting
                (*iter)->ReportStatus();
                delete *iter;
            }
        }
    }// end of if

    server_specific_cleanup();//cleanup bit and pieces
}



void Cthreadcontrol::delay_percent(int d)
{
    Cmdbthread::delay_percent(d);
}
ThreadTask* Cthreadcontrol::RetrieveTask(const string& filename)
{
    CTLockGuard<CTMutex> guard(m_cachemutex);//this mutex may result in 2 mutex lock for same call
    return scriptCache.makeTask(filename);
}
bool    Cthreadcontrol::doesTaskExist(const string& filename)
{
    return scriptCache.doesTaskExist(filename);
}

/////////////////////////////////////////////////////////////////////
// @mfunc Waits for compleion of all threads for a set time
// @rdesc If threads complete in time return true else false
bool Cthreadcontrol::WaitForCompletion()
{
    CyrSleep(10000);//Wait a bit first
#ifndef CYR_NO_STD_NAMESPACE
    for (std::vector<Cmdbthread*>::iterator iter=threads.begin(); iter!=threads.end(); iter++)
#else
    for (vector<Cmdbthread*>::iterator iter=threads.begin(); iter!=threads.end(); iter++)
#endif
    {
        enumThreadState status;
        //wait up to 10 seconds for thread to finish before terminating
        status=(*iter)->get_state();
        if(status==THREAD_RUNNING||status==THREAD_CREATED)
            return false;//if state is not invalid or terminated then there are still threads running
    }
    return true;//all threads have completed or aborted
}
/////////////////////////////////////////////////////////////////////
// @mfunc Report on All threads status

bool Cthreadcontrol::ReportOnThreads()
{
    unsigned int passed=0;
    unsigned int failed=0;
#ifndef CYR_NO_STD_NAMESPACE
    for (std::vector<Cmdbthread*>::iterator iter=threads.begin(); iter!=threads.end(); iter++)
#else
    for (vector<Cmdbthread*>::iterator iter=threads.begin(); iter!=threads.end(); iter++)
#endif
    {
        bool bret=(*iter)->ReportStatus();
        if(bret)
        {
            passed++;
        }
        else
        {
            failed++;
        }
    }
    //if more than 50% passed return true
    string msg="Thread Counting, Passed "+int_to_string(passed)+" , Failed "+int_to_string(failed);
    logr->AddTrace(THREADING,DEBUGLEVEL,msg.c_str());
    return (failed>passed)?false:true;
}
