/* Copyright (C) 2000 Cyrano SA

   This file is part of OpenSTA

   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.

   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.

   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html

   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
#ifndef _CORBAREPOSITORY_H_
#define _CORBAREPOSITORY_H_
///////////////////////////////////////////////////////////////////////////////
// @doc     CORBARepository
// @module  CORBARepository | This is to be used a a single Global Class
//  to handle all interfaces with the repository. After Creation it will be shared 
//  accross all threads and must be MT-safe. It also looks after any differences between
//  standard and Stand Alone behaviour, usually doing nothing in standalone mode
//  ALL public methods (except those involved with creation) must be protected with a mutex
//  lock, this should ensure that only 1 thread acesses this object at any 1 time. This should 
//  prevent the need for mutexes around all members but will require that no public method
//  calls another public method as this will result in a deadlock situation. Hopefully this 
//  will not present a problem due to the limited functionality of this class
//
///////////////////////////////////////////////////////////////////////////////
#pragma warning (disable: 4786)
#include <string>
#ifdef SOLARIS
#include <algorithm>
#endif
#include <vector>
#include <map>

/*
#ifdef HPUX
#   include  <iostream.h>
#else
#   include  <iostream>
#endif
*/

#ifndef _WIN32_WINNT
#   define _WIN32_WINNT 0x0404
#endif
#ifndef WINVER
#   define WINVER       0x0400
#endif
#include "windows.h"
#include "FastLock.hpp"

#ifdef WIN32
#   pragma warning(disable:4786)
#   pragma warning(disable:4503)
#endif

#ifndef CYR_NO_STD_NAMESPACE
/*
    using std::cout;
    using std::endl;
*/
    using std::string;
    using std::vector;
    using std::map;

#endif

#ifndef CYRANO_TP_STANDALONE
#   include <CECNodeDefinition.h>
#   include "CyrWrapLib/CyrWrapLib.h"
#   include "CyrImpactLib/CECImpactFolders.h"
#   include "CyrImpactLib/CECImpactProject.h"
#else
#   include <CECSACORBA.h>
#endif

extern const string TESTRUNSTRING_SEP;

typedef map<string, string> LocalResultFiles_t;

typedef vector<string> TestRunNameList_t;

//@class CORBARepository
class CORBARepository
{
public:
    CORBARepository(CORBAServer* corba,const string& datanode,const string& user,const string& pass,int pid,const string& exename);
    virtual ~CORBARepository();

    static BOOL InitClass();
    static void UninitClass();

    //Create Projects, Folder and TestRuns
    virtual int	    Connect(const string& ProjectName ,const string& TestName,const string& TestRunString="");
    //Retrieve Tof file from repository
    virtual string  GetAsFile(const string& Filename);
    virtual string  GetTofFile(const string& Filename);
    virtual string  GetThreadFile(const string& Filename);
    virtual string  GetVariableFile(const string& Filename);
    virtual void	SetLocalResultsFile(const char* FullFilename, const char* Filename, bool del = true);
    virtual void	SetLocalResultsFile(const LocalResultFiles_t& resultFiles, bool del = true);

    //Retrieve Node file from repository
#ifndef CYRANO_TP_STANDALONE
    virtual CECNodeDefinition  GetNodeDefinition(const string& Filename);
#endif
    //Set new timer output
    virtual int	    TimerOutput(const string& TimerName,long ElapsedTime,const string& ScriptName,const string& ThreadID,const string& TestName ,int no_threads);
    //Set new log message
    virtual int	    LogMessage	(const string& Msg, int LogFile, const string& ScriptName="", const string& ThreadID="",int objindex=0);
    virtual int LogErrMessage(const string& Msg, int LogFile,const string ScriptName="",const string& ThreadID="",int objindex=0);
    virtual int LogWrnMessage(const string& Msg, int LogFile,const string& ScriptName=string(),const string& ThreadID="",int objindex=0);
#ifndef CYRANO_TP_STANDALONE
    virtual string  GetProjectName();
    virtual string  GetUserName();
    virtual string  GetPassword();
    virtual string  GetDataServerNodeName();
    virtual string  GetTestRunString();
    virtual string  GetPrimaryTestRunString();
    virtual CECOrbTestRun* GetTestRun()
    {
	return m_TestRun;
    }
    virtual CECDataFolder* GetDataFolder()
    {
	return m_DataFolder;
    }
#endif//#ifndef CYRANO_TP_STANDALONE

	long GetTimeOffset()
	{
		return m_timeoffset;
	}

protected:
#ifndef CYRANO_TP_STANDALONE
    virtual int	    MakeActualTestRun(const string& TestRunString,const string& TestName)=0;
    virtual CECOrbTestRun*    CreateParentTestRuns(const TestRunNameList_t& TestRunNameList,const string& TestRunString)=0;
    virtual CECOrbTestRun* GetParentTestRunFromString(const string& TestRunString);
#endif
    virtual string GetFile(const string& ShortFileName,const string& LongFilename,CECOrbFolder* Folder);
    //If there is more to add to TestRunString do it here
    virtual string  AppendOwnTestRunString(){return "";};
    
    string				m_DataServerNodeName;//Node on which Data server can be found
    int					m_pid;//PID of Process running
    string              m_szPid;    // String version of m_pid.
    string				m_UserName;//username required for repository
    string				m_Password;//password
    string				m_exename;//argv[0]

	LocalResultFiles_t	m_LocalResultFiles;

    bool			m_LocalResults;
    bool			m_DeleteLocalResults;
	long			m_timeoffset;//time offset between this process and Repository TimeService (assuming time to get time=0);

#ifndef CYRANO_TP_STANDALONE
    CECOrbServer*			m_Server;//Server on repository
    CECOrbProject*			m_Project;//Project for this entire Test
    CECOrbFolder*			m_RootFolder;//Root Folder in Project
    CECObjectCodeFolder*		m_ObjectCodeFolder;//Where files will be retrieved from
    CECNodeFolder*			m_NodeFolder;//Where Node files will be retrieved from
    CECDataFolder*			m_DataFolder;//Where Data files will be retrieved from
    CECTestFolder*			m_TestFolder;//Test Folder within Project
    CECTestStore*			m_TestStore;//Folder in TestFolder with name of test to be execute, this is where all info from this test is stored
    CECOrbTestRun*			m_TestRun;//Where information for 1 run of this test is stored
    CORBAServer*			m_corba;

    //If this is a child CORBARepository (bizzare isn't it)
    //Parent Server,Project and TestFolder should be the same
    typedef vector<CECOrbFolder*>	TestRunList_t;
    //list of any TestRuns that this one belongs to
    TestRunList_t			m_ParentTestRunList; 
#else
	CECOrbFolder*			m_ObjectCodeFolder;
	CECOrbFolder*			m_DataFolder;
#endif//#ifndef CYRANO_TP_STANDALONE    
    static CFastLock *s_pPublicMethodFl;    //Only 1 public method may be called at once, of course this means no public method can call another
};
extern CORBARepository* Repository;//Global Repository Object
#endif
