/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

/*  tpg_main.c:
 *  ===========
 *
 *  This is the main module of the command line utility TPG.
 *  TPG processes a '.tpa' file, a source file containing language state
 *  transition table definitions. It generates two 'C' language header
 *  files:
 *
 *  mylang.tpa  -->  mylang.h          State transition table header
 *                   mylang_rtndef.h   Action routine prototypes header
 */

/*
 *  PSL Macros.
 */
#include <psl_macros.h>

/*
 *  Standard ANSI header files
 */
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 *  Function Prototypes.
 */
#include <psl_rtndef.h>         /* PSL Routines */
#include "tpg_tpa_rtndef.h"     /* TPG Action Routines */
#include "tpg_rtndef.h"         /* TPG Routines */

static void STORE_ERROR (struct tpg_com *tpg_com_ptr,
                         unsigned long errnum);

static void UNDERLINE_ERROR (struct tpaarg *);

/*
 *  Structure definitions 
 */
#include <shrmsg.h>         /* PSL Messages */
#include <confirm_respdef.h>        /* Confirmation Symbols */
#include <psldate_prmdef.h>     /* PSL Date/Time Symbols */
#include <psldiag_prmdef.h>     /* Diagnostic Routine Symbols */
#include <pslerr_prmdef.h>      /* Error Routine Symbols */
#include <pslfil_prmdef.h>      /* File Routine Symbols */
#include <psl_tpaargdef.h>      /* TPARSE Argument Block */
#include <psl_tpatabdef.h>      /* TPARSE Table Entry */
#include "tpg_comdef.h"         /* Common Data Structure */
#include "tpg_cmd.h"            /* Command Line Data */

/*
 *  State structure - this holds information from a PSL$STATE definition
 */
struct state
{
    char  *name_ptr;        /* Pointer to state name */
    long tran_index;        /* Index in tran_table */
};
/*
 *  Tran structure - this holds information from a PSL$TRAN definition
 */
struct tran
{
    char *type_ptr;     /* Type pointer */
    char *label_ptr;        /* Label pointer */
    char *action_ptr;       /* Action pointer */
    char *mask_ptr;     /* Mask pointer */
    char *mask_adr_ptr;     /* Mask_adr pointer */
    char *argument_ptr;     /* Argument pointer */
    long state_index;       /* Index in state_table */
    long end_state;     /* End of state flag */
};
/*
 *  Global data
 */
static  char    source_line[132];   /* Source line */
static  char    parsable_line[132]; /* De-commented upper case line */
static  char    error_line[132];    /* Used for underlining errors */
static  long    line_size;      /* Length of Source_line */
static  long    parsable_size;      /* Length of Parsable_line */
static  Boolean eof_flag;       /* Set if EOF encountered */
static  char    table_name[80];     /* Table name */
static  char    *state_ptr;     /* State name pointer */
static  char    type[80];       /* Type */
static  char    *type_ptr;      /* Type pointer */
static  char    *label_ptr;     /* Label pointer */
static  char    *action_ptr;        /* Action pointer */
static  char    *mask_ptr;      /* Mask pointer */
static  char    *mask_adr_ptr;      /* Mask_adr pointer */
static  char    *argument_ptr;      /* Argument pointer */
static  struct  state   state_table[10000];
/* State table */
static  struct  tran    tran_table[10000];
/* Tran table */
static  long    nstate;         /* Number of entries in state table */
static  long    ntran;          /* Number of entries in tran table */
static  char    actnam[100000];     /* List of action routine names */
static  char    *next_actnam_ptr = actnam;
/* Pointer to next name in actnam */
/*
 *  When a source line contains several errors, all underlining is
 *  is merged together and all errors are output after the underline.
 */
#define MAX_ERRORS_IN_LINE 5

static  unsigned long   error_array [MAX_ERRORS_IN_LINE];
/* Array or Errors in Source Line */
static  unsigned long errors_in_line;   /* Number of Errors in Source Line */
static  Boolean underline_flag;     /* Underline Line Required */

/*
 *  TPARSE argument block
 */
static  struct tpaarg tparse_data;


int main (int argc, char *argv_ptr[])
{
    /*
     *  Local Data.
     */
    unsigned long       iret;       /* Return status */
    long       status;     /* File I/O status */
    char       subheading1 [256];  /* List file Subheading 1 */
    char       subheading2 [256];  /* List file Subheading 2 */
    char       listing_version [8];
    /* Listing Version Number */
    static struct tpg_com tpg_com;      /* Common Data */

    static struct tpg_cmd tpg_cmd;      /* Command Line Struct */
    struct   tpg_cmd    *tpg_cmd_ptr;   /* Pointer to Command Line Struct */

    /*
     *  TPARSE state table
     */
#include "tpg_tpa.h"

    /*
     *  Store the utility name.
     */
    PSL_SET_UTILNAME ("TPG");

    /*
     *  Set up pointer to Command Line Data in Common Data.
     */
    tpg_com.cmd_ptrps = &tpg_cmd;

    /*
     *  Copy pointers for ease of reference.
     */
    tpg_cmd_ptr = tpg_com.cmd_ptrps;

    /*
     *  Set up the Diagnostics Stack
     */
    tpg_com.diagstkps = PSL_ALLOCATE_DIAGNOSTICS ();
    /* Allocate the Stack */
    if (tpg_com.diagstkps == NULL)  /* If it failed */
    {
        PSL_SIGDIAG (           /* Add Diagnostic */
            NULL,           /* Use primary (fallback) stack */
            2,          /* Number of message segments */
            PSL_SYSERVERR,      /* First message */
            "PSL_ALLOCATE_DIAGNOSTICS",
            /* Parameter */
            PSL_INSVIRMEM);     /* Secondary message */
        goto end;           /* and exit at once */
    }

    /*
     *  Flag that, if any errors are signalled, we want to report them
     *  and exit.
     */
    (void) PSL_TRAP_ERRORS (        /* Say we want to trap errors */
        tpg_com.diagstkps,      /* Via this diagnostics pointer */
        0,          /* All classes */
        0,          /* All numbers */
        PSLERRACT_EXIT,     /* And report them */
        NULL,           /* With no setjmp environment */
        NULL);          /* No error handling routine */

    /*
     *  Parse the command line
     */
    iret = TPG_GETCMD (&tpg_com, argc, argv_ptr);
    if (!(psl_test_status (iret))) goto end;

    /*
     *  FIND_NEXT_FILE:
     *
     *  Find next Source File.
     */
find_next_file:
    ;

    iret = TPG_GETNXT_FILE (&tpg_com, tpg_com.tpg_namec256);
    if (!(psl_test_status (iret))) goto end;

    /*
     *  If we have a match then we need to ask the user for Confirmation
     *  if /CONFIRM was specified (and has not since been superceded by
     *  a response of 'All').
     */
    if ((*tpg_cmd_ptr).confirmn)    /* If /CONFIRM specified */
    {

        /*
         *  CONFIRM_PROMPT:
         *
         *  Prompt for Confirmation.
         */
confirm_prompt:
        ;

        (void) printf ("%s, compile? [N]: ", tpg_com.tpg_namec256);

        /*
         *  Get response.
         */
        status = CONFIRM_RESPONSE ();

        /*
         *  Analyze response.
         */
        if (status == CNF_NO)       /* If answer was No */
        {
            goto find_next_file;    /* ....Try next file */
        }
        else if (status == CNF_QUIT)    /* If answer was Quit */
        {
            goto end;           /* ....Exit from Loop */
        }
        else if (status == CNF_PROMPT)  /* If answer was Unknown */
        {
            goto confirm_prompt;    /* ....Repeat Prompt */
        }
        else if (status == CNF_ALL) /* If answer was All */
        {
            (*tpg_cmd_ptr).confirmn = PFALSE;
            /* ....Negate qualifier */
        }

    }

    /*
     *  Open the input file. We will not open the output files until we have
     *  sucessfully compiled the source file.
     *
     *  Get the name of the file.
     */
    iret = PSL_FILE_PARSE (     /* Parse filespec */
               tpg_com.tpg_namec256,   /* This File */
               NULL,           /* No Default File */
               NULL,           /* No Related File */
               NULL,           /* No Expected Related Extension */
               tpg_com.tpg_namec256,   /* Returned File */
               sizeof (tpg_com.tpg_namec256),
               /* Maximum length thereof */
               FILPRS_INPUT);      /* Flags */

    if (!(psl_test_status (iret)))  /* If an error */
    {
        PSL_SIGDIAG (           /* Signal Diagnostic */
            tpg_com.diagstkps,  /* Use this diagnostic stack */
            2,          /* Number of message segments */
            PSL_OPENIN,     /* First message */
            tpg_com.tpg_namec256,   /* Parameter */
            iret);          /* Secondary message */
    }

    PSL_CLEARERR ();            /* Clear error numbers */

    tpg_com.tpg_file_ptrpf = fopen (tpg_com.tpg_namec256, "r");
    if (tpg_com.tpg_file_ptrpf == NULL)
    {
        PSL_SIG_CIOERR (        /* Log I/O Error */
            tpg_com.diagstkps,
            /* Use this stack */
            PSL_OPENIN,     /* Message */
            tpg_com.tpg_namec256);  /* Parameter */
    }

    /*
     *  Get the full name.
     */
    PSL_GET_FILENAME (tpg_com.tpg_file_ptrpf,
                      tpg_com.tpg_namec256,
                      sizeof (tpg_com.tpg_namec256));

    /*
     *  ...and open the listing file immediately (If listing is enabled)
     */
    if ((*tpg_cmd_ptr).listn)
    {
        (void) strcpy (listing_version, "V1.0");

        /*
         *  Set up subheadings.
         */
        (void) sprintf (subheading1, "Command : %.*s",
                        (int) sizeof (subheading1) - 11,
                        (*tpg_cmd_ptr).command_line_value_ptrpc);

        (void) sprintf (subheading2, "TPA File: %.*s",
                        (int) sizeof (subheading2) - 11,
                        tpg_com.tpg_namec256);

        iret = LIST_INIT (      /* Initialize Listing */
                   tpg_com.list_ctx_ptrps, /* Pointer to Context */
                   "TPG",          /* Utility name */
                   ".prt",         /* Default file name */
                   tpg_com.tpg_namec256,   /* Related filename */
                   subheading1,        /* Command line as header */
                   subheading2,        /* File name as header */
                   listing_version);   /* Version Number */

        if (!(psl_test_status (iret))) goto end;

        /*
         *  Set up PSL_OUTPUT_DIAG to output errors to the Listing as well.
         */
        iret = PSL_SETUP_OUDIAG (   /* Setup PSL_OUTPUT_DIAG */
                   tpg_com.diagstkps,  /* Using this Diagnostics stack */
                   DIOUTY_LIST,        /* Output to Listing */
                   tpg_com.list_ctx_ptrps, /* Using this Listing context */
                   132);           /* Output Width */
        if (!(psl_test_status (iret)))
        {
            PSL_SIGDIAG (       /* Signal Diagnostic */
                tpg_com.diagstkps,      /* Use this stack */
                2,          /* Number of message segments */
                PSL_SYSERVERR,      /* First message */
                "PSL_SETUP_OUDIAG", /* Parameter */
                iret);          /* Secondary message */
        }
    }

    /*
     *  Set up data areas so that the Parsing Buffer is empty (force call to
     *  GET_NEXT_SOURCE_LINE) and then parse the input file through TPARSE
     *  tables.
     */
    tparse_data.string_ptrpc = parsable_line;
    /* Point to line to be Parsed */
    tparse_data.tracertn_ptrpf = NULL;
    tparse_data.user_ptrp = &tpg_com;

    tparse_data.tracelvll = 0;      /* No tracing */

    /*
     *  Initialize relevant data
     */
    errors_in_line = 0;         /* Show zero total errors */
    eof_flag = PFALSE;          /* Clear End of File flag */
    underline_flag = PFALSE;        /* Clear Underline flag */
    tpg_com.line_countl = 0;        /* Set to be on line 1 */
    tpg_com.err_countl = 0;     /* Set no errors */

    /*
     *   Perform the parsing
     */
    iret = PSL_TPARSE (&tparse_data, tpg_state_table);

    /*
     *  Close the files and exit
     */
    if (tpg_com.err_countl == 0)
    {
        status = fclose (tpg_com.tpg_file_ptrpf);
        if (status != 0)
        {
            PSL_SIG_CIOERR (        /* Signal file error */
                tpg_com.diagstkps,      /* Use this stack */
                PSL_CLOSEIN,        /* Error to report */
                tpg_com.tpg_namec256);  /* Filespec */
        }

        status = fclose (tpg_com.out_file_ptrpf);
        if (status != 0)
        {
            PSL_SIG_CIOERR (        /* Signal file error */
                tpg_com.diagstkps,  /* Use this stack */
                PSL_CLOSEOUT,       /* Error to report */
                tpg_com.out_namec256);  /* Filespec */
        }

        status = fclose (tpg_com.func_file_ptrpf);
        if (status != 0)
        {
            PSL_SIG_CIOERR (     /* Signal file error */
                tpg_com.diagstkps,  /* Use this stack */
                PSL_CLOSEOUT,       /* Error to report */
                tpg_com.func_namec256); /* Filespec */
        }
    }

    if ((*tpg_cmd_ptr).listn && (tpg_com.list_ctx_ptrps != NULL))
    {
        (void) LIST_CLOSE (tpg_com.list_ctx_ptrps);
        /* Close Listing */
        (void) PSL_SETUP_OUDIAG (   /* Cancel PSL_OUTPUT_DIAG */
            tpg_com.diagstkps,  /* Using this Diagnostics stack */
            DIOUTY_LIST,        /* Output to Listing */
            0,          /* Null Listing context */
            0);         /* Output Width */
    }

    /*
     *  Go back and try the next input file.
     */
    goto find_next_file;

end:

    /*
     *  Deallocate Diagnostics stack.
     */
    PSL_DEALLOCATE_DIAGNOSTICS (tpg_com.diagstkps);

    /*
     *  Exit with 'Worst Error' status
     */
    PSL_EXIT ();
    return 0;
}


/************************************************************************/
/*                                  */
/*    GET_NEXT_RECORD - Get the next line from the Source File          */
/*    ===============                           */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*  iret = GET_NEXT_RECORD (argblk_ptr);                */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret            = Unsigned Long PSL Routine Return Status   */
/*  argblk_ptr  = Pointer to TPARSE Argument Block Structure    */
/*                                  */
/*    This routine gets the next line from the source file.     */
/*                                  */
/*    It first checks to see if there are any errors in the current */
/*    source line and, if so, outputs the underline and reports the */
/*    errors.                               */
/*                                  */
/************************************************************************/

unsigned long GET_NEXT_RECORD (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    short      quote;
    long       isub;
    char       *recptr;
    struct tpg_com      *tpg_com_ptr;
    struct tpg_cmd      *tpg_cmd_ptr;

    tpg_com_ptr = (struct tpg_com *) (*argblk_ptr).user_ptrp;
    tpg_cmd_ptr = (*tpg_com_ptr).cmd_ptrps;

    /*
     *  If we have already found the End of File on a previous parsing 
     *  operation, return straight away. This will enable this routine to be 
     *  called at lots of places in the TPARSE state table, and will always 
     *  produce the same result.
     */
    if (eof_flag) return 0;

    /*
     *  Write out the current input line to the listing file if one was
     *  specified.
     */
get_record_start:
    if (((*tpg_cmd_ptr).listn) && ((*tpg_com_ptr).line_countl !=0))
    {
        (void) LIST_LINE_REFERENCED ((*tpg_com_ptr).list_ctx_ptrps, source_line);
    }

    /*
     *  Are there any errors to report?
     */
    if (errors_in_line != 0)
    {

        /*
         *  Output the underline buffer, if required.
         */
        if (underline_flag)
        {

            /*
             *  Print the source line and the error line to the screen
             */
            (void) printf ("\n%.*s", line_size, source_line);
            (void) printf ("%.*s\n", line_size, error_line);

            /*
             *  Write the error line to the listing file if one was specified.
             */
            if ((*tpg_cmd_ptr).listn)
            {
                (void) LIST_LINE ((*tpg_com_ptr).list_ctx_ptrps, error_line);
            }
        }

        /*
         *  Loop through the array of errors and report them.
         */
        for (isub = 0; isub < (long)errors_in_line; isub++)
        {
            PSL_PUTDIAG (       /* Add Diagnostic */
                (*tpg_com_ptr).diagstkps,
                /* Use this stack */
                1,          /* Number of message segments */
                error_array [isub]);
            /* This message */
            PSL_OUTPUT_DIAG (       /* Output Diagnostic */
                (*tpg_com_ptr).diagstkps,
                /* From this stack */
                0,          /* No Context */
                DIOUFL_LSTERR);     /* Output Last Error */
        }

        /*
         *  Reset counter and underline flag.
         */
        errors_in_line = 0;
        underline_flag = PFALSE;
    }

    /*
     *  Read a record
     */
    (*tpg_com_ptr).line_countl += 1;    /* Increment line counter */
    recptr = fgets (source_line, 132, (*tpg_com_ptr).tpg_file_ptrpf);
    /* Read next record */
    if (recptr == NULL)         /* Assume EOF */
    {
        eof_flag = PTRUE;
        return 0;
    }

    /*
     *  Set up the Parsable line to contain the same as the input line,
     *  without the Comments, and converted to Uppercase.
     */
    line_size = strlen (source_line);   /* Length of line */
    parsable_size = 0;          /* Initialize parsable size */

    /*
     *  No Quotes yet.
     */
    quote = 0;              /* Reset quote flag */

    for (isub = 0; isub < line_size; isub++)
    {

        /*
         *  If we have a Quote it either starts or ends a quoted string (we hope).
         */
        if ((source_line[isub] == '\"' ) ||
                (source_line[isub] == '\'' ))
        {
            if (!quote)
            {
                quote = source_line[isub];
            }
            else if (quote == source_line[isub])
            {
                quote = 0;
            }
        }

        /*
         *  If we have a comment that is NOT inside a quoted string, get out. 
         */
        if ((source_line[isub] == ';') && (!quote))
            goto end_line;

        /*
         *  Add uppercase character to parsable line
         */
        parsable_line[parsable_size] = toupper (source_line[isub]);
        /* Copy uppercase character */

        /*
         *  Set up the error line as a mask of the parsable line with TABS in the
         *  same places.
         */
        if (parsable_line[parsable_size] == '\t')
        {
            error_line[parsable_size] = '\t';
        }
        else
        {
            error_line[parsable_size] = ' ';
        }

        parsable_size += 1;     /* Increment parsable size */
    }

end_line:
    /*
     *  Strip off trailing non-printing characters
     */
    while ((parsable_line[parsable_size - 1] <= ' ') &&
            (parsable_size > 0))
    {
        parsable_size -= 1;     /* Decrement parsable size */
    }
    parsable_line[parsable_size] = '\0';
    /* Null-terminate line */
    /*
     *  Test if the whole parsable line was blank. If so, read the next line.
     */
    if (parsable_size <= 0)
        goto get_record_start;

    /*
     *  Here if there is a new line for parsing.
     */
    tparse_data.stringcntl = parsable_size;
    /* Store parse string length */
    tparse_data.string_ptrpc = parsable_line;
    /* Point to line to be Parsed */
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    GET_CONTUATION  - Read a continuation record, reporting an error  */
/*          if there were still characters in the input     */
/*          string.                     */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     iret = GET_CONTINUATION (argblk_ptr);            */
/*                                  */
/*  Where:                              */
/*                                  */
/*     iret       = Unsigned Long PSL Routine Return Status     */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/************************************************************************/

unsigned long GET_CONTINUATION (struct tpaarg *argblk_ptr)
{
    if (tparse_data.stringcntl > 0)
    {
        return PSL_CONTERR;     /* Return error */
    }
    else
    {
        return GET_NEXT_RECORD (argblk_ptr);
        /* Else get next record */
    }
}


/************************************************************************/
/*                                  */
/*    INIT_STATE      - This routine is called after we have parsed the */
/*          PSL$INIT_STATE clause in the source file.       */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = INIT_STATE (argblk_ptr);                 */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                  */
/*    This routine is called after parsing a PSL$INIT_STATE keyword.    */
/*    Its purpose is to initialize for new table            */
/*                                  */
/************************************************************************/

unsigned long INIT_STATE (struct tpaarg *argblk_ptr)
{

    /*
     *  Local data
     */
    char       *ch_ptr;        /* Character pointer */

    nstate = 0;             /* Initialize entries in state table */
    ntran = 0;              /* Initialize entries in tran table */
    (void) memset ((void *) type, '\0', 80);
    /* Initialize type */
    label_ptr = "";         /* Initialize label pointer */
    action_ptr = "";            /* Initialize action pointer */
    mask_ptr = "";          /* Initialize mask pointer */
    mask_adr_ptr = "";          /* Initialize mask_adr pointer */
    argument_ptr = "";          /* Initialize argument pointer */

    /*
     *  Copy table name and force into lower case 
     */
    (void) strncpy (table_name, tparse_data.token_ptrpc, tparse_data.tokencntl);
    /* Copy table name */
    ch_ptr = table_name;        /* Point to name */
    while (*ch_ptr != '\0')
    {
        *ch_ptr = tolower (*ch_ptr);/* Force character to lowercase */
        ch_ptr += 1;        /* Step pointer */
    }
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    STATE           - This routine is called after we have parsed the */
/*          PSL$STATE clause in the source file.        */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = STATE (argblk_ptr);                  */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                  */
/*    This routine is called after parsing a PSL$STATE keyword.         */
/*    Its purpose is to add another entry to the state table.       */
/*                                  */
/************************************************************************/

unsigned long STATE (struct tpaarg *argblk_ptr)
{
    /*
     *  Get state name 
     */
    state_ptr = (char *) calloc (1, tparse_data.tokencntl + 1);
    /* Allocate memory */
    (void) strncpy (state_ptr, tparse_data.token_ptrpc, tparse_data.tokencntl);
    /* Copy string */
    state_table[nstate].name_ptr = state_ptr;
    /* Store pointer to state name */
    state_table[nstate].tran_index = ntran;
    /* Store tran_table index */
    /*
     *  Set end of state flag in previous tran entry
     */
    if (ntran > 0)
    {
        tran_table[ntran-1].end_state = 1;
    }

    /*
     *  Store index in next tran entry
     */
    tran_table[ntran].state_index = nstate;
    nstate += 1;            /* Increment table entries */
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    STORE_TYPE      - This routine is called after we have parsed the */
/*          type field of a transition.                     */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = STORE_TYPE (argblk_ptr);                 */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                                                      */
/*  Processing the type field - param specifies action required:    */
/*      1 - Append current string to type                       */
/*      2 - Append current character to type            */
/*                                  */
/************************************************************************/

unsigned long STORE_TYPE (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    char       *ch_ptr;        /* Character pointer */

    if (tparse_data.paraml == 1)
    {

        /*
         *  Append string to type string - if string is quoted, then use the
         *  equivalent string in the original line in order to preserve the 
         *  original case
         */
        ch_ptr = tparse_data.token_ptrpc;
        /* Point to token */
        if ((*ch_ptr == '\'')   /* Quoted string */
                || (*ch_ptr == '\"'))
        {
            ch_ptr += &source_line [0] - &parsable_line [0];
            /* Point to original line */
        }
        (void) strncat (type, ch_ptr, tparse_data.tokencntl);
        /* Append string */
    }
    else
    {

        /*
         *  Append character to type string
         */
        type[strlen (type)] = tparse_data.charb;
        /* Append character */
    }
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    STORE_MASK - This routine is called to store the mask field   */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = STORE_MASK (argblk_ptr);                 */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                                                      */
/*  Having identified the argument token in the parsable line,  */
/*  we must actually use the equivalent string in the original  */
/*  line, as we need to preserve the original case          */
/*                                                                      */
/************************************************************************/

unsigned long STORE_MASK (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    unsigned long       offset;     /* Character String Offset */

    /*
     *  Allocate memory for mask
     */
    mask_ptr = (char *) calloc (1, tparse_data.tokencntl + 1);

    /*
     *  Calculate the offset of the character pointed to by token_ptr in
     *  the parsable_line buffer.
     */
    offset = tparse_data.token_ptrpc - parsable_line;

    /*
     *  Copy mask from the original buffer
     */
    (void) strncpy (mask_ptr, source_line + offset, tparse_data.tokencntl);

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    STORE_MASK_ADR - This routine is called to store the mask     */
/*          address                     */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = STORE_MASK_ADR (argblk_ptr);             */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                                                      */
/*  Having identified the argument token in the parsable line,  */
/*  we must actually use the equivalent string in the original  */
/*  line, as we need to preserve the original case          */
/*                                                                      */
/************************************************************************/

unsigned long STORE_MASK_ADR (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    unsigned long       offset;     /* Character String Offset */

    /*
     *  Allocate memory for mask_adr
     */
    mask_adr_ptr = (char *) calloc (1, tparse_data.tokencntl + 1);

    /*
     *  Calculate the offset of the character pointed to by token_ptr in
     *  the parsable_line buffer.
     */
    offset = tparse_data.token_ptrpc - parsable_line;

    /*
     *  Copy mask from the original buffer
     */
    (void) strncpy (mask_adr_ptr, source_line + offset, tparse_data.tokencntl);

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    STORE_ARGUMENT - This routine is called to store the argument */
/*          field                       */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = STORE_ARGUMENT (argblk_ptr);             */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                                                      */
/*  Having identified the argument token in the parsable line,  */
/*  we must actually use the equivalent string in the original  */
/*  line, as we need to preserve the original case          */
/*                                                                      */
/************************************************************************/

unsigned long STORE_ARGUMENT (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    unsigned long       offset;     /* Character String Offset */

    /*
     *  Allocate memory for argument
     */
    argument_ptr = (char *) calloc (1, tparse_data.tokencntl + 1);

    /*
     *  Calculate the offset of the character pointed to by token_ptr in
     *  the parsable_line buffer.
     */
    offset = tparse_data.token_ptrpc - parsable_line;

    /*
     *  Copy argument from the original buffer
     */
    (void) strncpy (argument_ptr, source_line + offset, tparse_data.tokencntl);

    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    TRAN            - This routine is called after we have parsed the */
/*          PSL$TRAN  clause in the source file.        */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = TRAN  (argblk_ptr);                  */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                  */
/*    This routine is called after parsing a PSL$TRAN  keyword.         */
/*    Its purpose is to add another entry to the transition table.  */
/*                                  */
/************************************************************************/

unsigned long TRAN (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    char       *actnam_ptr;    /* Pointer to action name in table */

    /*
     *  PSL$TRAN has been read - add another entry to tran_table
     */
    type_ptr = (char *) calloc (1, strlen (type) + 1);
    /* Allocate memory */
    (void) strncpy (type_ptr, type, strlen (type));
    /* Copy string */
    tran_table[ntran].type_ptr = type_ptr;
    /* Store type pointer */
    tran_table[ntran].label_ptr = label_ptr;
    /* Store label pointer */
    tran_table[ntran].action_ptr = action_ptr;
    /* Store action pointer */
    tran_table[ntran].mask_ptr = mask_ptr;
    /* Store mask pointer */
    tran_table[ntran].mask_adr_ptr = mask_adr_ptr;
    /* Store mask_adr pointer */
    tran_table[ntran].argument_ptr = argument_ptr;
    /* Store argument pointer */
    tran_table[ntran].end_state = 0;
    /* Assume not end of state */
    tran_table[ntran+1].state_index = -1;
    /* Set next tran not start of state */
    ntran += 1;             /* Increment table entries */

    /*
     *  Add the action name to the action name table if it is not already there
     */
    if (*action_ptr != '\0')        /* If there is an action routine */
    {
        actnam_ptr = actnam;        /* Point to start of table */
        while (actnam_ptr < next_actnam_ptr)
            /* While still entries in table */
        {
            if (strcmp (actnam_ptr, action_ptr) == 0)
                /* If name already in table */
            {
                goto end_actadd;    /* No need to add it */
            }
            actnam_ptr += strlen (actnam_ptr) + 1;
            /* Point to next name */
        }
        (void) strcpy (next_actnam_ptr, action_ptr);
        /* Add name to table */
        next_actnam_ptr += strlen (action_ptr) + 1;
        /* And update pointer */
    }

end_actadd:
    ;

    /*
     *  Initialize for next tran
     */
    (void) memset ((void *) type, '\0', 80);
    /* Initialize type */
    label_ptr = "";         /* Initialize label pointer */
    action_ptr = "";            /* Initialize action pointer */
    mask_ptr = "";          /* Initialize mask pointer */
    mask_adr_ptr = "";          /* Initialize mask_adr pointer */
    argument_ptr = "";          /* Initialize argument pointer */
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    END_STATE       - This routine is called after we have parsed the */
/*          PSL$END_STATE clause in the source file.    */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = END_STATE (argblk_ptr);                  */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                  */
/*    This routine is called after parsing a PSL$END_STATE keyword.     */
/*    Its purpose is to write out the data for the current table.   */
/*                                  */
/************************************************************************/

unsigned long END_STATE (struct tpaarg *argblk_ptr)
{
    /*
     *  Local data
     */
    long       isub;       /* Loop control */
    long       outcnt;     /* Output count */
    char       *ch_ptr;        /* Character pointer */
    short      quitb;      /* quitb value */
    long       state_index;    /* State table index */
    unsigned long       iret;       /* Return Status */
    unsigned long       date;       /* PSL Date */
    unsigned long       time;       /* PSL Time */
    char       dattim_buf [80];    /* PSL Date/Time buffer */
    char       wrkbuf [256];

    struct tpg_com      *tpg_com_ptr;
    struct tpg_cmd      *tpg_cmd_ptr;

    tpg_com_ptr = (struct tpg_com *) (*argblk_ptr).user_ptrp;
    tpg_cmd_ptr = (*tpg_com_ptr).cmd_ptrps;

    /*
     *  If we have had errors then report this and exit without writing any
     *  of the output files.
     */
    if ((*tpg_com_ptr).err_countl != 0)
    {
        PSL_PUTDIAG (           /* Add Diagnostic */
            (*tpg_com_ptr).diagstkps,
            /* Use this stack */
            1,          /* Number of message segments */
            PSL_ENDNOOBJ,       /* Message */
            (*tpg_com_ptr).tpg_namec256,
            /* First Parameter */
            (*tpg_com_ptr).err_countl);
        /* Second Parameter */
        PSL_OUTPUT_DIAG (       /* Output Diagnostics */
            (*tpg_com_ptr).diagstkps,
            /* Use this stack */
            0,          /* No Context */
            DIOUFL_LSTERR);     /* Output Last Error */
        goto end_state_end;     /* No need to write files then */
    }

    /*
     *  Set end of state flag in previous tran entry
     */
    tran_table[ntran-1].end_state = 1;

    /*
     *  Get date/time for file headers
     */
    PSL_GET_DATTIM (&date, &time);

    iret = PSL_FMT_DATTIM (date, time, dattim_buf, sizeof (dattim_buf),
                           PSDTYP_PSL, (PSDFMT_DATE | PSDFMT_TIME));
    if (!(psl_test_status (iret)))
    {
        PSL_SIGDIAG (           /* Signal Diagnostic */
            (*tpg_com_ptr).diagstkps,
            /* Use this stack */
            2,          /* Number of message segments */
            PSL_SYSERVERR,      /* First message */
            "PSL_FMT_DATTIM",   /* Parameter */
            iret);          /* Secondary message */
    }

    /*
     *  Open the output tparse structures file
     */
    PSL_CLEARERR ();        /* Clear error numbers */

    /*
     *  Get the name of the output file.
     */
    iret = PSL_FILE_PARSE (     /* Parse filespec */
               (*tpg_cmd_ptr).object_value_ptrpc,
               /* Specified File */
               ".h",           /* Default File */
               (*tpg_com_ptr).tpg_namec256,
               /* Related File */
               ".tpa",
               (*tpg_com_ptr).out_namec256,
               /* Returned File */
               255,            /* Maximum length thereof */
               0);         /* No Flags */

    if (!(psl_test_status (iret)))  /* If an error */
    {
        PSL_SIGDIAG (           /* Signal Diagnostic */
            (*tpg_com_ptr).diagstkps,
            /* Use this diagnostic stack */
            2,          /* Number of message segments */
            PSL_SYSERVERR,      /* First message */
            "PSL_FILE_PARSE for Header File",
            /* Parameter */
            iret);          /* Secondary message */
    }

    (void) strcpy ((*tpg_com_ptr).func_namec256, (*tpg_com_ptr).out_namec256);
    /* Copy output to function filename */

    (*tpg_com_ptr).out_file_ptrpf = fopen ((*tpg_com_ptr).out_namec256, "w");
    if ((*tpg_com_ptr).out_file_ptrpf == NULL)
    {
        PSL_SIG_CIOERR (        /* Log I/O Error */
            (*tpg_com_ptr).diagstkps,
            /* Use this stack */
            PSL_OPENOUT,        /* Message */
            (*tpg_com_ptr).out_namec256);
        /* Parameter */
    }

    /*
     *  ...then the output function prototypes file
     */
    ch_ptr = strrchr ((*tpg_com_ptr).func_namec256, '.');
    /* Locate extension */
    *ch_ptr = '\0';         /* Remove extension */
    (void) strcat ((*tpg_com_ptr).func_namec256, "_rtndef.h");
    /* Add new extension */

    (*tpg_com_ptr).func_file_ptrpf = fopen ((*tpg_com_ptr).func_namec256, "w");
    if ((*tpg_com_ptr).func_file_ptrpf == NULL)
    {
        PSL_SIG_CIOERR (        /* Log I/O Error */
            (*tpg_com_ptr).diagstkps,
            /* Use this stack */
            PSL_OPENOUT,    /* Message */
            (*tpg_com_ptr).func_namec256);
        /* Parameter */
    }

    /*
     *  Write out function prototypes for action routines
     */
    (void) fprintf ((*tpg_com_ptr).func_file_ptrpf,
                    "/*\n *\tInclude File Written by TPG on %s */\n",
                    dattim_buf);       /* Write header comment */
    (void) fprintf ((*tpg_com_ptr).func_file_ptrpf,
                    "/*\n *\tFunction Prototypes for Action Routines\n */\n\n");
    /* Write header comment */
    ch_ptr = actnam;            /* Point to action name table */
    while (ch_ptr < next_actnam_ptr)
        /* While still entries in table */
    {
        (void) fprintf ((*tpg_com_ptr).func_file_ptrpf,
                        "unsigned long %s (struct tpaarg *);\n", ch_ptr);
        /* Write prototype line */
        ch_ptr += strlen (ch_ptr) + 1;
        /* Point to next name */
    }

    /*
     *  Write out the table - first write a header
     */
    (void) fprintf (
        (*tpg_com_ptr).out_file_ptrpf,
        "/*\n *\tInclude File Written by TPG on %s */\n",
        dattim_buf);       /* Write header comment */
    (void) fprintf (
        (*tpg_com_ptr).out_file_ptrpf,
        "/*\n *\tTPARSE Table %s\n */\n\n", table_name);
    /* Write header comment */

    /*
     *  Then write a line to include the function prototypes file - note
     *  that we exclude any device/directory portion of the name, as this
     *  will not be portable
     */
    (void) strcpy (wrkbuf, (*tpg_com_ptr).func_namec256);

    for (isub = 0; isub < (long)strlen (wrkbuf); isub++)
    {
        wrkbuf [isub] = tolower (wrkbuf [isub]);
    }

    ch_ptr = strrchr (wrkbuf, '_');
    /* Locate '_rtndef.h' */
    while ((ch_ptr != wrkbuf) && (*ch_ptr != '\\'))
        /* Scan backwards looking for device/directory */
    {
        ch_ptr -= 1;            /* Move pointer back one character */
    }
    if (ch_ptr != wrkbuf) ch_ptr += 1;
    /* Step past dev/dir */
    (void) fprintf (
        (*tpg_com_ptr).out_file_ptrpf,
        "\n#include \"%s\"\n", ch_ptr);
    /* Write Include line */
    /*
     *  Write out #define statements to define the states
     */
    for (isub = 0; isub < nstate; isub++)
    {
        (void) fprintf (
            (*tpg_com_ptr).out_file_ptrpf,
            "#define\t%s\t%s[%ld]\n",
            state_table[isub].name_ptr,
            table_name,
            state_table[isub].tran_index);
    }

    /*
     *  Write out the table structure
     */
    (void) fprintf (
        (*tpg_com_ptr).out_file_ptrpf,
        "\nstatic struct tpatab\t%s[] = {\n", table_name);
    /* Write the struct definition */
    for (isub = 0; isub < ntran; isub++)
    {
        outcnt = 0;         /* Initialize output count */
        /*
         *  If this is the start of a state, write a comment
         */
        state_index = tran_table[isub].state_index;
        /* Get index to state table */
        if (state_index != -1)      /* -1 indicates not start of state */
        {
            (void) fprintf (
                (*tpg_com_ptr).out_file_ptrpf,
                "/* %ld: %s */\n",
                state_table[state_index].tran_index,
                state_table[state_index].name_ptr);
        }

        /*
         *  Write end_stateb
         */
        outcnt = fprintf (
                     (*tpg_com_ptr).out_file_ptrpf,
                     "   {%ld, ",
                     tran_table[isub].end_state);
        /*
         *  Write token_typeb, token_charb, quitb, keyword_ptrpc, subexpr_ptrps
         */
        if (strcmp (tran_table[isub].label_ptr, "PSLTPA_EXIT") == 0)
        {
            quitb = 1;          /* quitb = PSLTPA_EXIT */
        }
        else if (strcmp (tran_table[isub].label_ptr, "PSLTPA_FAIL") == 0)
        {
            quitb = 2;          /* quitb = PSLTPA_FAIL */
        }
        else
        {
            quitb = 0;          /* No quit action */
        }
        ch_ptr = tran_table[isub].type_ptr;
        /* Point to type string */
        if ((*ch_ptr == '\'')       /* Quoted string */
                || (*ch_ptr == '\"'))
        {
            if (*ch_ptr == *(ch_ptr + 2))
                /* If single character */
            {
                ch_ptr += 1;        /* Point to character */
                if (*ch_ptr == '\'')    /* If single quote */
                {
                    outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf,
                                       "PSLTPA_CHAR, '\\'', %d, 0, 0, ",
                                       quitb);
                }
                else
                {
                    outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf,
                                       "PSLTPA_CHAR, '%c', %d, 0, 0, ",
                                       *ch_ptr, quitb);
                }
            }
            else            /* Keyword */
            {
                ch_ptr += 1;        /* Skip over the initial quote */
                *(ch_ptr + strlen (ch_ptr) - 1) = '\0';
                /* Drop the final quote */
                if ((strlen (ch_ptr) == 2) &&
                        (*(ch_ptr + 1) == '*'))
                    /* If 'c*', then single char keyword */
                {
                    outcnt += fprintf (
                                  (*tpg_com_ptr).out_file_ptrpf,
                                  "PSLTPA_KEYWORD, 0, %d, \"%c\", 0, ",
                                  quitb, *ch_ptr);
                }
                else
                {
                    outcnt += fprintf (
                                  (*tpg_com_ptr).out_file_ptrpf,
                                  "PSLTPA_KEYWORD, 0, %d, \"%s\", 0, ",
                                  quitb, ch_ptr);
                }
            }
        }
        else if (*ch_ptr == '!')    /* Sub-expression */
        {
            ch_ptr += 1;        /* Point to sub-expression */
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf,
                               "PSLTPA_SUBXPR, 0, %d, 0, &%s, ",
                               quitb, ch_ptr);
        }

        /*
         *  If the token begins with a '%', we have entered the hexadecimal value
         *  of a non-printing character. This is stored as a single character with
         *  the hex value being the character.
         */
        else if (*ch_ptr == '%')    /* Hexadecimal value        */
        {
            ch_ptr += 1;        /* Skip to the value        */
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf
                               ,"PSLTPA_CHAR, 0X%s, %d, 0, 0, "
                               ,ch_ptr, quitb);
        }

        else                /* All other token types */
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf,
                               "%s, 0, %d, 0, 0, ",
                               ch_ptr, quitb);
        }

        /*
         *  Write target_ptrps
         */
        ch_ptr = tran_table[isub].label_ptr;
        /* Point to target string */
        if ((outcnt + strlen (ch_ptr) + 4) > 80)
        {
            (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "\n");
            /* Output newline */
            outcnt = fprintf ((*tpg_com_ptr).out_file_ptrpf, "       ");
            /* Output spaces */
        }
        if ((strlen (ch_ptr) == 0)  /* If no target */
                || (strcmp (ch_ptr, "PSLTPA_EXIT") == 0)
                || (strcmp (ch_ptr, "PSLTPA_FAIL") == 0))
            /* Or special cases */
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "0, ");
        }
        else
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "&%s, ", ch_ptr);
        }

        /*
         *  Write action_ptrpf
         */
        ch_ptr = tran_table[isub].action_ptr;
        /* Point to action string */
        if ((outcnt + strlen (ch_ptr) + 3) > 80)
        {
            (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "\n");
            /* Output newline */
            outcnt = fprintf ((*tpg_com_ptr).out_file_ptrpf, "       ");
            /* Output spaces */
        }
        if (strlen (ch_ptr) == 0)   /* If no action */
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "0, ");
        }
        else
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "%s, ", ch_ptr);
        }

        /*
         *  Write action_paraml
         */
        ch_ptr = tran_table[isub].argument_ptr;
        /* Point to argument string */
        if ((outcnt + strlen (ch_ptr) + 3) > 80)
        {
            (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "\n");
            /* Output newline */
            outcnt = fprintf ((*tpg_com_ptr).out_file_ptrpf, "       ");
            /* Output spaces */
        }
        if (strlen (ch_ptr) == 0)   /* If no argument */
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "0, ");
        }
        else
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "%s, ", ch_ptr);
        }

        /*
         *  Write data_ptrp
         */
        ch_ptr = tran_table[isub].mask_adr_ptr;
        /* Point to mask_adr string */
        if ((outcnt + strlen (ch_ptr) + 4) > 80)
        {
            (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "\n");
            /* Output newline */
            outcnt = fprintf ((*tpg_com_ptr).out_file_ptrpf, "       ");
            /* Output spaces */
        }
        if (strlen (ch_ptr) == 0)   /* If no argument */
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "0, ");
        }
        else
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "&%s, ", ch_ptr);
        }

        /*
         *  Write maskl
         */
        ch_ptr = tran_table[isub].mask_ptr;
        /* Point to mask string */
        if ((outcnt + strlen (ch_ptr) + 3) > 80)
        {
            (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "\n");
            /* Output newline */
            outcnt = fprintf ((*tpg_com_ptr).out_file_ptrpf, "       ");
            /* Output spaces */
        }
        if (strlen (ch_ptr) == 0)   /* If no mask */
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "0, ");
        }
        else
        {
            outcnt += fprintf ((*tpg_com_ptr).out_file_ptrpf, "%s, ", ch_ptr);
        }

        /*
         *  Write zero entries.
         */
        if ((outcnt + 11) > 80)
        {
            (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "\n");
            /* Output newline */
            outcnt = fprintf ((*tpg_com_ptr).out_file_ptrpf, "       ");
            /* Output spaces */
        }

        (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "0, 0, 0, 0");

        /*
         *  If this is not the last tran, write a comma 
         */
        if (isub < (ntran - 1))
        {
            (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "},\n");
        }
        else
        {
            (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "}\n");
        }
    }

    /*
     *  Terminate the structure
     */
    (void) fprintf ((*tpg_com_ptr).out_file_ptrpf, "};\n");

end_state_end:
    ;
    return PSL_NORMAL;
}


/************************************************************************/
/*                                  */
/*    CHECK_END_OF_FILE                                                 */
/*    =================                                             */
/*                                  */
/*    Calling Convention:                       */
/*                                  */
/*      iret = CHECK_END_OF_FILE (argblk_ptr);              */
/*                                  */
/*    Where:                                */
/*                                  */
/*  iret          = Unsigned Long PSL Routine Return Status     */
/*  argblk_ptr = Pointer to TPARSE Argument Block Structure     */
/*                                  */
/*    This routine is called to detect the end of the source file.      */
/*    Return success if EOF has been reached                */
/*                                  */
/************************************************************************/

unsigned long CHECK_END_OF_FILE (struct tpaarg *argblk_ptr)
{
    if (eof_flag)
    {
        return PSL_NORMAL;
    }
    else
    {
        return 0;
    }
}


/************************************************************************/
/*                                  */
/*  EXTEND_TOKEN_TO_EOL - Extend Token to End of Line       */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     EXTEND_TOKEN_TO_EOL (argblk_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine extends the current Token to the end of the    */
/*  current line.                           */
/*                                  */
/************************************************************************/

void    EXTEND_TOKEN_TO_EOL (struct tpaarg *argblk_ptr)
{
    (*argblk_ptr).tokencntl = (*argblk_ptr).stringcntl + 1;
    return;
}


/************************************************************************/
/*                                  */
/*  FIND_TOKEN_END - Find end of current Token          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     FIND_TOKEN_END (argblk_ptr);                 */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine finds the end of the current Token.        */
/*  Tokens are delimited by non-printing characters and commas. */
/*                                  */
/************************************************************************/

void    FIND_TOKEN_END (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       tokencnt;
    char       *wrk_ptr;

    /*
     *  Copy Token Count for manipulation.
     */
    tokencnt  = (*argblk_ptr).tokencntl;

    wrk_ptr = (*argblk_ptr).token_ptrpc + 1;

    /*
     *  Test if this is at the end of the input String.
     */
loop:
    if  ( (*argblk_ptr).stringcntl <= 0 ) goto end;

    if  ( *wrk_ptr <= ' ' ) goto end;
    if  ( *wrk_ptr == ',' ) goto end;

    tokencnt += 1;
    wrk_ptr += 1;

    (*argblk_ptr).string_ptrpc += 1;
    (*argblk_ptr).stringcntl -= 1;

    goto loop;

end:
    (*argblk_ptr).tokencntl = tokencnt;

    return;
}


/*----------------------------------------------------------------------*/
/*                                  */
/*    REP_ERROR_TOKEN                           */
/*    ===============                           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     REP_ERROR_TOKEN (argblk_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*    This routine is responsible for storing an Error in the array of  */
/*    errors for the current source line and then underlining the Token.*/
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   REP_ERROR_TOKEN (struct tpaarg *argblk_ptr)
{
    /*
     *  Store the error in the array, if possible.
     */
    STORE_ERROR ((struct tpg_com *)(*argblk_ptr).user_ptrp, (*argblk_ptr).paraml);

    FIND_TOKEN_END (argblk_ptr);    /* Extend to end of token */

    /*
     *  Underline the Token in Error.
     */
    UNDERLINE_ERROR (argblk_ptr);

    return PSL_NORMAL;
}


/*----------------------------------------------------------------------*/
/*                                  */
/*    REP_ERROR_LINE                            */
/*    ==============                            */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     REP_ERROR_LINE (argblk_ptr);                 */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*    This routine is responsible for storing an Error in the array of  */
/*    errors for the current source line and then underlining the rest  */
/*    of the current source line                    */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   REP_ERROR_LINE (struct tpaarg *argblk_ptr)
{
    /*
     *  Store the error in the array, if possible and expand the token
     *  to the end of the line.
     */
    STORE_ERROR ((struct tpg_com *)(*argblk_ptr).user_ptrp, (*argblk_ptr).paraml);
    EXTEND_TOKEN_TO_EOL (argblk_ptr);

    /*
     *  Underline the Token in Error.
     */
    UNDERLINE_ERROR (argblk_ptr);   /* Underline the error  */
    (*argblk_ptr).stringcntl = 0;   /* ...and terminate the line */
    return PSL_NORMAL;          /* Return success   */
}


/*----------------------------------------------------------------------*/
/*                                  */
/*    REP_ERROR_SYMBOL                          */
/*    ================                          */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     REP_ERROR_SYMBOL (argblk_ptr);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*    This routine is responsible for storing an Error in the array of  */
/*    errors for the current source line and then underlining the   */
/*    current token. It does not modify the token pointer       */
/*                                  */
/*----------------------------------------------------------------------*/

unsigned long   REP_ERROR_SYMBOL (struct tpaarg *argblk_ptr)
{
    /*
     *  Store the error in the array, if possible.
     */
    STORE_ERROR ((struct tpg_com *)(*argblk_ptr).user_ptrp, (*argblk_ptr).paraml);

    /*
     *  Underline the Token in Error.
     */
    UNDERLINE_ERROR (argblk_ptr);       /* Underline the error  */
    (*argblk_ptr).stringcntl = 0;       /* ...and terminate the line */
    return PSL_NORMAL;              /* Return success   */
}


/************************************************************************/
/*                                  */
/*  STORE_ERROR - Store an Error                    */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     STORE_ERROR (tpg_com_ptr, errnum);               */
/*                                  */
/*  Where:                              */
/*                                  */
/*     tpg_com_ptr = Pointer to Common Data Structure       */
/*                                  */
/*  This routine is responsible for storing an Error in the array   */
/*  of errors for the current source line and incrementing the  */
/*  appropriate counters.                       */
/*                                  */
/************************************************************************/

static void STORE_ERROR (struct tpg_com *tpg_com_ptr,
                         unsigned long errnum)
{
    /*
     *  Local Data.
     */
    unsigned long       jsub;       /* Loop Control */

    /*
     *  Check to see if we have already filled the error array.
     */
    if  ( errors_in_line == MAX_ERRORS_IN_LINE ) goto end;

    /*
     *  Loop through the error array to make sure that the new error
     *  is not already in there. If it is, we just underline it.
     */
    for (jsub = 0; jsub < errors_in_line; jsub++)
    {
        if ( error_array [jsub] == errnum ) goto end;
    }

    /*
     *  If we get here the error is not already in the array so we insert it
     *  in the first free slot and count it.
     */
    error_array [errors_in_line] = errnum;

    errors_in_line += 1;        /* Increment line error count   */

end:
    (*tpg_com_ptr).err_countl += 1; /* Increment error count    */
    return;
}


/************************************************************************/
/*                                  */
/*  UNDERLINE_ERROR - Underline the current Token           */
/*                                  */
/*  Calling Convention:                     */
/*                                  */
/*     UNDERLINE_ERROR (argblk_ptr);                */
/*                                  */
/*  Where:                              */
/*                                  */
/*     argblk_ptr = Pointer to TPARSE Argument Block Structure  */
/*                                  */
/*  This routine is responsible for underlining the current Token.  */
/*  The ERROR_LINE is set up with asterisks in the correct places.  */
/*                                  */
/************************************************************************/

static void UNDERLINE_ERROR (struct tpaarg *argblk_ptr)
{
    /*
     *  Local Data.
     */
    unsigned long       offset;     /* Character String Offset */
    unsigned long       jsub;       /* Loop Control */

    /*
     *  Calculate the offset of the character pointed to by token_ptr in
     *  the parsable_line buffer.
     */
    offset = (*argblk_ptr).token_ptrpc - parsable_line;

    /*
     *  Set the corresponding characters in the error_line buffer to
     *  asterisks, remembering not to overwrite tabs etc.
     */
    for (jsub = 0; jsub <= (unsigned long)(*argblk_ptr).tokencntl; jsub++)
    {
        if (error_line [offset+jsub] != '\t')
        {
            error_line [offset+jsub] = '*';
        }
    }

    /*
     *  Set the flag that indicates that an Underline Line is required.
     */
    underline_flag = PTRUE;

    return;
}
