/*
Module : SOCMFC.CPP
Purpose: implementation for an MFC wrapper class for sockets
Created: PJN / 05-08-1998
History: None

Copyright (c) 2001 by PJ Naughter.  (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


/////////////////// Includes //////////////////////////////////

#include "stdafx.h"
#include "SocMFC.h"



/////////////////// Macros / Defines //////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif



///////////////// Implementation //////////////////////////////

////////// Exception handling code

void AfxThrowWSocketException(int nError /* = 0 */)
{
	if (nError == 0)
		nError = ::WSAGetLastError();

	CWSocketException* pException = new CWSocketException(nError);

	TRACE(_T("Warning: throwing CWSocketException for error %d\n"), nError);
	THROW(pException);
}

BOOL CWSocketException::GetErrorMessage(LPTSTR pstrError, UINT nMaxError, PUINT pnHelpContext)
{
	ASSERT(pstrError != NULL && AfxIsValidString(pstrError, nMaxError));

	if (pnHelpContext != NULL)
		*pnHelpContext = 0;

	LPTSTR lpBuffer;
	BOOL bRet = FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
			                      NULL,  m_nError, MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT),
			                      (LPTSTR) &lpBuffer, 0, NULL);

	if (bRet == FALSE)
		*pstrError = '\0';
	else
	{
		lstrcpyn(pstrError, lpBuffer, nMaxError);
		bRet = TRUE;

		LocalFree(lpBuffer);
	}

	return bRet;
}

CString CWSocketException::GetErrorMessage()
{
  CString rVal;
  LPTSTR pstrError = rVal.GetBuffer(4096);
  GetErrorMessage(pstrError, 4096, NULL);
  rVal.ReleaseBuffer();
  return rVal;
}

CWSocketException::CWSocketException(int nError)
{
	m_nError = nError;
}

CWSocketException::~CWSocketException()
{
}

IMPLEMENT_DYNAMIC(CWSocketException, CException)

#ifdef _DEBUG
void CWSocketException::Dump(CDumpContext& dc) const
{
	CObject::Dump(dc);

	dc << "m_nError = " << m_nError;
}
#endif



////////// The main class /////////////////////////////////////

CWSocket::CWSocket()
{
  m_hSocket = INVALID_SOCKET; //default to an invalid scoket descriptor
}

CWSocket::~CWSocket()
{
  Close();
}

void CWSocket::Create()
{
  m_hSocket = socket(AF_INET, SOCK_STREAM, 0);
  if (m_hSocket == INVALID_SOCKET)
    AfxThrowWSocketException();
}

void CWSocket::Connect(LPCTSTR pszHostAddress, int nPort, LPCTSTR pszLocalBoundAddress)
{
	//For correct operation of the T2A macro, see MFC Tech Note 59
	USES_CONVERSION;

  ASSERT(IsCreated()); //must have been created first
  ASSERT(pszHostAddress); //Must have a valid host
  ASSERT(_tcslen(pszHostAddress)); //as above

  //Bind to the local address if need be
  if (pszLocalBoundAddress && _tcslen(pszLocalBoundAddress))
    Bind(0, pszLocalBoundAddress);
  
  //Work out the IP address of the machine we want to connect to
	LPSTR lpszAsciiDestination = T2A((LPTSTR)pszHostAddress);

	//Determine if the address is in dotted notation
	SOCKADDR_IN sockDestinationAddr;
	ZeroMemory(&sockDestinationAddr, sizeof(sockDestinationAddr));
	sockDestinationAddr.sin_family = AF_INET;
	sockDestinationAddr.sin_port = htons((u_short)nPort);
	sockDestinationAddr.sin_addr.s_addr = inet_addr(lpszAsciiDestination);

	//If the address is not dotted notation, then do a DNS 
	//lookup of it.
	if (sockDestinationAddr.sin_addr.s_addr == INADDR_NONE)
	{
		LPHOSTENT lphost;
		lphost = gethostbyname(lpszAsciiDestination);
		if (lphost != NULL)
			sockDestinationAddr.sin_addr.s_addr = ((LPIN_ADDR)lphost->h_addr)->s_addr;
		else
		{
      AfxThrowWSocketException(WSAEINVAL); 
			return;
		}
	}

	//Call the protected version which takes an address 
	//in the form of a standard C style struct.
	Connect((SOCKADDR*)&sockDestinationAddr, sizeof(sockDestinationAddr));
}

void CWSocket::Connect(const SOCKADDR* lpSockAddr, int nSockAddrLen)
{
  ASSERT(IsCreated()); //must have been created first

	if (connect(m_hSocket, lpSockAddr, nSockAddrLen) == SOCKET_ERROR)
    AfxThrowWSocketException();
}

void CWSocket::Send(void* pBuffer, int nBuf)
{
  ASSERT(IsCreated()); //must have been created first

  if (send(m_hSocket, (char*) pBuffer, nBuf, 0) == SOCKET_ERROR)
    AfxThrowWSocketException();
}

int CWSocket::Receive(void* pBuffer, int nBuf)
{
  ASSERT(IsCreated()); //must have been created first

  int nReceived = recv(m_hSocket, (char*) pBuffer, nBuf, 0); 
  if (nReceived == SOCKET_ERROR)
    AfxThrowWSocketException();

  return nReceived;
}

void CWSocket::Close()
{
	if (IsCreated())
	{
		closesocket(m_hSocket);
		m_hSocket = INVALID_SOCKET;
	}
}

BOOL CWSocket::IsReadible(DWORD dwTimeout)
{
  ASSERT(IsCreated()); //must have been created first

  timeval timeout;
  timeout.tv_sec = dwTimeout/1000;
  timeout.tv_usec = (dwTimeout%1000)*1000;
  fd_set fds;
  FD_ZERO(&fds);
  FD_SET(m_hSocket, &fds);
  int nStatus = select(0, &fds, NULL, NULL, &timeout);
  if (nStatus == SOCKET_ERROR)
    AfxThrowWSocketException();

  return !(nStatus == 0);
}

BOOL CWSocket::IsCreated() const
{
  return (m_hSocket != INVALID_SOCKET);
}

void CWSocket::Attach(SOCKET socket)
{
  if (IsCreated())
    Close();
  m_hSocket = socket;
}

SOCKET CWSocket::Detach()
{
  SOCKET socket = m_hSocket;
  m_hSocket = INVALID_SOCKET;
  return socket;
}

CWSocket::operator SOCKET()
{
  return m_hSocket;
}

void CWSocket::Listen(int nBacklog)
{
  ASSERT(IsCreated()); //must have been created first

  int nStatus = listen(m_hSocket, nBacklog);
  if (nStatus == SOCKET_ERROR)
    AfxThrowWSocketException();
}

void CWSocket::Bind(int nPort, LPCTSTR pszLocalBoundAddress)
{
	//For correct operation of the T2A macro, see MFC Tech Note 59
	USES_CONVERSION;

  ASSERT(IsCreated()); //must have been created first

  //Setup the structure used in sdk "bind" calls
  SOCKADDR_IN sockLocalAddress;
  ZeroMemory(&sockLocalAddress, sizeof(sockLocalAddress));
  sockLocalAddress.sin_family = AF_INET;
  sockLocalAddress.sin_port = htons((u_short)nPort);

  //Do we need to bind to a specific IP address?
  if (pszLocalBoundAddress && _tcslen(pszLocalBoundAddress))
  {
    //Convert to an ASCII string
    LPSTR lpszAsciiLocalAddress = T2A((LPTSTR)pszLocalBoundAddress);
    sockLocalAddress.sin_addr.s_addr = inet_addr(lpszAsciiLocalAddress);

	  //If the address is not dotted notation, then do a DNS 
	  //lookup of it.
	  if (sockLocalAddress.sin_addr.s_addr == INADDR_NONE)
	  {
		  LPHOSTENT lphost;
		  lphost = gethostbyname(lpszAsciiLocalAddress);
		  if (lphost != NULL)
			  sockLocalAddress.sin_addr.s_addr = ((LPIN_ADDR)lphost->h_addr)->s_addr;
		  else
		  {
        AfxThrowWSocketException(WSAEINVAL); 
			  return;
		  }
    }
  }
  else
    sockLocalAddress.sin_addr.s_addr = htonl(INADDR_ANY); //Bind to any IP address;

  //Finally call the SDK "bind" function
  if (bind(m_hSocket, (sockaddr*) &sockLocalAddress, sizeof(sockLocalAddress)) == SOCKET_ERROR)
    AfxThrowWSocketException(); 
}

void CWSocket::Accept(CWSocket& clientSocket, sockaddr_in& clientAddress)
{
  ASSERT(IsCreated()); //must have been created first
  ASSERT(!clientSocket.IsCreated()); //Must be an unitialized socket

  //Call the SDK accept function  
  int nSize = sizeof(sockaddr_in);
  SOCKET socket = ::accept(m_hSocket, (sockaddr*) &clientAddress, &nSize);
  if (socket == INVALID_SOCKET)
    AfxThrowWSocketException(); 

  //Wrap the return value up into a C++ instance
  clientSocket.Attach(socket);
}
