// TestDebug.cpp: implementation of the CTestDebug class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "TestDebug.h"
#include "CorbaInfo.h"
#include "RegFunctions.h"

extern void Trace( const char *format ,... );

#define SM_BREAK -4
#define SM_RUN	-3
#define SM_SINGLE_EVENT -2
#define SM_SINGLE_USER -1
//////////////////////////////////////////////////////////////////////
// Construction/Destruction
/////////////////////////////////////////////////////////////////////
CTestDebug::CTestDebug(const CString& strTestName,BOOL bUseBreakpoints,BOOL bRun,const CString& strHost):
	StepByStepCommanderImpl(strHost),
	m_strTestName(strTestName),
	m_nStepMode(SM_BREAK),
	m_nCurrentVU(-1),
	m_hRunEvent(INVALID_HANDLE_VALUE),
	m_bReadyToStep(FALSE),
	m_bTestRunning(FALSE),
	m_nRequestId(0),
	m_bUseBreakPoints(bUseBreakpoints),
	m_bAborted(FALSE)
{
	m_pComms = new CTestDebugCommsProxy(this);

	if(!m_CorbaInfo.HasServerBeenCreated())
		return;

	StepByStepCommanderImpl::Initialise(m_CorbaInfo.GetServer());
	SetCommander(TRUE);

	if(bRun)
		m_nStepMode = SM_RUN;

}

CTestDebug::~CTestDebug()
{
	SetCommander(FALSE);

	delete m_pComms;
}

void CTestDebug::RunTest()
{
	if(m_hRunEvent == INVALID_HANDLE_VALUE)
		m_hRunEvent = CreateEvent(NULL,TRUE,FALSE,NULL);
	
	//Just wait until the test has completed then shuts everything down
	if(!m_CorbaInfo.HasServerBeenCreated())
		return;

	CString strRegName;
	//Wait for test to start
	while(!m_CorbaInfo.IsTestActive(m_strTestName, strRegName))
		Sleep(1000);
	
	m_bTestRunning = true;
	
	//Check if test is complete
	while(m_CorbaInfo.IsTestActive(m_strTestName, strRegName))
		Sleep(5000);
	
	m_bTestRunning = false;

	//Shutdown comms
	m_pComms->NotifyTestComplete();

	CloseHandle(m_hRunEvent);
	m_hRunEvent = INVALID_HANDLE_VALUE;
}

CORBA::Boolean CTestDebug::IsStepByStep()
{
	Trace("IsStepByStep Called\n");
	return true;
}

void CTestDebug::NextRequest(const char* ScriptName, CORBA::Long lScriptLineNumber, CORBA::Long lTaskGroupID, 
									 CORBA::Long lVUId, CORBA::Long lConId, const char* strRequest)
{
	if(m_bAborted)
		return;

	CSingleLock lock(&m_CritSectCorba);
	lock.Lock();

	CNextRequestData NRD;
	NRD.SetScriptLine(lScriptLineNumber);
	NRD.SetTaskGroup(lTaskGroupID);
	NRD.SetVUId(lVUId);
	NRD.SetConId(lConId);
	NRD.SetRequest(strRequest);
	NRD.SetRequestId(m_nRequestId);
	m_nRequestId++;

	//Generate script name from tof name
	char *szFilename = strrchr(ScriptName,'\\');
	char *szExtension = strrchr(szFilename,'.');
	NRD.SetScript(GetRepositoryPath() + strScriptType + CString(szFilename,(szExtension-szFilename)) + ".HTP");
		
	lock.Unlock();

	//Process this notification
	DoNextRequest(&NRD);
}

void CTestDebug::RequestComplete(CORBA::Long lTaskGroupID, CORBA::Long lVUId, 
										 CORBA::Long lConId, CORBA::Boolean bIsBinaryData, 
										 const char* strHeaders, const char* strBody, CORBA::Long lBodyLength)
{
	if(m_bAborted)
		return;

	CSingleLock lock(&m_CritSectCorba);
	lock.Lock();
	
	CRequestCompleteData RCD;
	RCD.SetTaskGroup(lTaskGroupID);
	RCD.SetVUId(lVUId);
	RCD.SetConId(lConId);
	RCD.SetHeaders(strHeaders);
	RCD.SetBody(strBody,lBodyLength,bIsBinaryData);
	
	//Process this notification
	DoRequestComplete(&RCD);

	lock.Unlock();
}

void CTestDebug::DoRequestComplete(CRequestCompleteData* pRC)
{
	//Notify client of request complete
	m_pComms->NotifyRequestComplete(pRC);

	LogRequestComplete(pRC);
}

void CTestDebug::DoNextRequest(CNextRequestData* pNRD)
{
	CSingleLock lock(&m_CritSectCorba);
	lock.Lock();

	//Tell client about next request
	m_pComms->NotifyNextRequest(pNRD);

	//Write event to file
	LogNextRequest(pNRD);

	if(IsBreakpoint(pNRD->GetScript(),pNRD->GetScriptLine()))
	{
		m_bReadyToStep = true;
		m_pComms->NotifyReadyToStep();
		lock.Unlock();
		WaitForSingleObject(m_hRunEvent,INFINITE);
		lock.Lock();
		m_bReadyToStep = false;
	}
	else
	{
		switch(m_nStepMode)
		{
		case SM_BREAK:
		case SM_SINGLE_EVENT:
			m_bReadyToStep = true;
			m_pComms->NotifyReadyToStep();
			lock.Unlock();
			WaitForSingleObject(m_hRunEvent,INFINITE);
			lock.Lock();
			m_bReadyToStep = false;
		break;
		case SM_SINGLE_USER:
			if(pNRD->GetVUId() == m_nCurrentVU)
			{
				m_nStepMode = SM_BREAK;
				m_bReadyToStep = true;
				m_pComms->NotifyReadyToStep();
				lock.Unlock();
				WaitForSingleObject(m_hRunEvent,INFINITE);
				lock.Lock();
				m_bReadyToStep = false;
			}
		break;
		case SM_RUN:
		default:
		break;
		}
	}

	if(m_bAborted)
		return;

	ResetEvent(m_hRunEvent);
	
	Next(pNRD->GetTaskGroup(),pNRD->GetVUId());
	m_pComms->NotifyStepped(pNRD->GetVUId(),pNRD->GetConId());
	lock.Unlock();
}

BOOL CTestDebug::IsBreakpoint(const CString& strScript,int nLine)
{
	if(!m_bUseBreakPoints)
		return false;

	LoadBreakPoints(strScript);

	bool b;
	return m_mapBreakPoints.Lookup(nLine,b);
}

BOOL CTestDebug::LoadBreakPoints(const CString& strScript)
{
	if(m_strCurrentScript == strScript)
		return true;
	
	m_strCurrentScript = strScript;

	m_mapBreakPoints.RemoveAll();

	CFile f;
	int nPos = strScript.ReverseFind('.');
	CString strPath = strScript.Left(nPos) + ".dat";
	if(!f.Open(strPath,CFile::modeRead|CFile::shareDenyNone))
		return false;

	int nVersion;
	f.Read(&nVersion,sizeof(int));
	if(nVersion > 1)
		return false;

	DWORD dwState;
	int nUrl,nLine;
	while(1)
	{
		f.Read(&nUrl,sizeof(int));
		if(nUrl == -1)
			break;
		
		f.Read(&nLine,sizeof(int));
		f.Read(&dwState,sizeof(DWORD));

		if(dwState & 0x00000001)
			m_mapBreakPoints.SetAt(nLine,true);
	}

	f.Close();
	return true;
}

BOOL CTestDebug::StepSingleUser(int nVUId)
{
	CSingleLock lock(&m_CritSectCommand);
	lock.Lock();

	m_nStepMode = SM_SINGLE_USER;
	m_nCurrentVU = nVUId;

	SetEvent(m_hRunEvent);
	
	return true;
}

BOOL CTestDebug::StepSingleEvent()
{
	CSingleLock lock(&m_CritSectCommand);
	lock.Lock();

	m_nStepMode = SM_SINGLE_EVENT;
	
	SetEvent(m_hRunEvent);
	
	return true;
}

BOOL CTestDebug::Run()
{
	CSingleLock lock(&m_CritSectCommand);
	lock.Lock();

	m_nStepMode = SM_RUN;

	SetEvent(m_hRunEvent);

	return true;
}

BOOL  CTestDebug::Break()
{
	CSingleLock lock(&m_CritSectCommand);
	lock.Lock();

	m_nStepMode = SM_BREAK;

	return true;
}

BOOL CTestDebug::SetBreakPoint(const CString& strScript,int nLine,BOOL bSet)
{
	CSingleLock lock(&m_CritSectCommand);
	lock.Lock();

	if(strScript.CompareNoCase(m_strCurrentScript)==0)
	{
		if(bSet)
			m_mapBreakPoints.SetAt(nLine,true);
		else
			m_mapBreakPoints.RemoveKey(nLine);
	}
	return true;
}

BOOL CTestDebug::SendState()
{
	CSingleLock lock(&m_CritSectCommand);
	lock.Lock();

	if(!m_bTestRunning)
		m_pComms->NotifyReadyToStep();
	else	
	if(m_bReadyToStep)
		m_pComms->NotifyReadyToStep();
	else
		m_pComms->NotifyRunning();

	return true;
}

void CTestDebug::LogNextRequest(CNextRequestData* pNRD)
{
	if(!m_HttpSSFile.IsOpen())
	{
		CString strPath;
		GetTestRunPath(strPath);
		m_HttpSSFile.Open(strPath,false);
	}

	m_HttpSSFile.Write(pNRD);
}

void CTestDebug::LogRequestComplete(CRequestCompleteData* pRCD)
{
	if(!m_HttpSSFile.IsOpen())
	{
		CString strPath;
		GetTestRunPath(strPath);
		m_HttpSSFile.Open(strPath,false);
	}

	m_HttpSSFile.Write(pRCD);
}

void CTestDebug::GetTestRunPath(CString &strPath)
{
	CString strTestPath(GetRepositoryPath() + strTestType + "\\" + m_strTestName + "\\");

	CTime tCurrent,tNewest(0);
	CFileFind ff;
	//Use \? so ?? is not misinterpreted as a trigraph character
	BOOL bContinue = ff.FindFile(strTestPath + "\?\?-\?\?-\?\?\?\? \?\?-\?\?-\?\?.00\?");
	while(bContinue)
	{
		bContinue = ff.FindNextFile();

		if(ff.GetCreationTime(tCurrent) && (tCurrent>tNewest))
		{
			strPath = ff.GetFilePath();
			tNewest = tCurrent;
		}
	}
}

BOOL CTestDebug::Abort()
{
	m_bAborted = true;	
	SetEvent(m_hRunEvent);
	return true;
}