/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

////////////////////////////////////////////////////////////////////////////////
// Asynchronous TCP socket implementation file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AsyncTCPSocket.hpp"
#include <ctype.h>
#include <string>
#include <MSTcpIP.h>

#ifdef USE_PERF_DATA
LONG* CAsyncTCPSocket::m_splConnPerSec= 0;
#endif

CAsyncTCPSocket::CHostResolution CAsyncTCPSocket::s_hostResService;
BOOL CAsyncTCPSocket::m_bIsWindows2000 = FALSE;
CAsyncTCPSocket::SocketPool* CAsyncTCPSocket::s_pPool = NULL;


// ----------------------------------------------------------------------------
//
// CAsyncTCPSocket members definitions.
//
// ----------------------------------------------------------------------------

CAsyncTCPSocket::CAsyncTCPSocket(CIOCP& iocp, CTrace* pTrace)
    : CAsyncDevice(iocp, pTrace), m_dwTimestamp(0), m_pClientEventTarget(NULL),
      m_bTimedOut(FALSE)
{
    m_hDevice = (HANDLE)INVALID_SOCKET;
}

CAsyncTCPSocket::~CAsyncTCPSocket()
{
    if(m_hDevice != (HANDLE)INVALID_SOCKET)
        closesocket((SOCKET)m_hDevice);
}

void* CAsyncTCPSocket::operator new(size_t size)
{
    void* pRet = NULL;
    if(s_pPool)
    {
        // Allocate from the pool:
        CAsyncTCPSocket* p;
        if(s_pPool->AllocateObjects(&p))
            pRet = (void*)p;
    }
    else
    {
        // Delegate to global new:
        pRet = ::operator new(size);
    }

    return pRet;
}

void CAsyncTCPSocket::operator delete(void* p)
{
    if(s_pPool)
    {
        CAsyncTCPSocket* pS = (CAsyncTCPSocket*)p;
        s_pPool->FreeObjects(&pS);
    }
}

DWORD CAsyncTCPSocket::Open(LPCSTR zServer, DWORD dwPort, IocpEvent *pEventTarget, DWORD *pdwTime)
{
    CLock lock(m_sl);

    DWORD dwRes = 0;
    if(pdwTime)
        *pdwTime = 0;

    if(m_hDevice != (HANDLE)INVALID_SOCKET)
    {
        // Re-connection.
        dwRes = CAsyncTCPSocket::Close();
    }
    if(dwRes == 0)
    {
        // Initialize the structures.
        IN_ADDR RemoteIpAddress;
        ZeroMemory(&m_remoteAddr, sizeof (m_remoteAddr));
        m_remoteAddr.sin_family = AF_INET;
        m_remoteAddr.sin_port = htons((WORD)dwPort);

        RemoteIpAddress.s_addr = inet_addr(zServer);
        if(RemoteIpAddress.s_addr == INADDR_NONE)
        {
            if(s_hostResService.ResolveName(zServer, m_remoteAddr.sin_addr) == FALSE)
            {
                dwRes = (DWORD)WSAGetLastError();
                return SOCKET_ERROR;
            }
        }
        else
            m_remoteAddr.sin_addr = RemoteIpAddress;

        SOCKET s = INVALID_SOCKET;
        if(dwRes == 0)
        {
            // Create the socket as overlapped.
            GROUP g = 0;
            s = WSASocket(AF_INET, SOCK_STREAM, 0, NULL, g, WSA_FLAG_OVERLAPPED);
            if(s == INVALID_SOCKET)
                dwRes = (DWORD)WSAGetLastError();
        }

        if(dwRes == 0)
        {
/* Removed to avoid RESETS.
            // Set appropriate option for hard disconnect.
            static const LINGER s_linger = {1, 0};
            setsockopt((SOCKET)s, SOL_SOCKET, SO_LINGER, (const char *)&s_linger, sizeof(s_linger));
*/

            // Turn on TCP keepalive if asked to.
            extern BOOL g_bTCPKeepAlives;
            extern unsigned long g_ulKeepAliveInterval;
            if(g_bTCPKeepAlives)
            {
                if(m_bIsWindows2000)
                {
                    // For Windows 2000 and above:
                    struct tcp_keepalive ka =
                    {
                           1,                   // On / Off.
                        1000,                   // keepalivetime (ms): how often keepalives are sent
                                                //   (default is 1 second).
                        g_ulKeepAliveInterval   // keepaliveinterval (ms): idle time during which there
                                                //   are no sends or receives, and after which a
                                                //   keepalive is sent (default is 2 hours).
                    };
                    DWORD dwOutBytes = 0;
                    WSAIoctl(s, SIO_KEEPALIVE_VALS, (PVOID)&ka, sizeof(ka),
                        (PVOID)NULL, 0, &dwOutBytes, NULL, NULL);
                }
                else
                {
                    // For Windows NT 4.0:
                    BOOL bKA = TRUE;
                    setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (const char *)&bKA, sizeof(bKA));
                }
            }


#ifdef USE_PERF_DATA
#   ifdef LOCK_COUNTERS
            InterlockedIncrement(m_splConnPerSec);
#   else
            ++*m_splConnPerSec;
#   endif
#endif
            unsigned __int64 qwTime;
            FILETIME ftNow;
            if(pdwTime)
            {
                GetSystemTimeAsFileTime(&ftNow);
                qwTime = ftNow.dwHighDateTime;
                qwTime <<= 32;
                qwTime |= ftNow.dwLowDateTime;
            }
            if(connect((SOCKET)s, (PSOCKADDR)&m_remoteAddr, sizeof(SOCKADDR_IN)) == SOCKET_ERROR)
                dwRes = (DWORD)WSAGetLastError();
            else if(pdwTime)
            {
                unsigned __int64 qwTime2;
                GetSystemTimeAsFileTime(&ftNow);
                qwTime2 = ftNow.dwHighDateTime;
                qwTime2 <<= 32;
                qwTime2 |= ftNow.dwLowDateTime;
                qwTime = qwTime2 - qwTime;
                *pdwTime = (DWORD)(qwTime / 10000); // Convert to milliseconds.
            }

            if(dwRes == 0)
            {
                // Bind the socket to the completion port.
                dwRes = BindToIocp((HANDLE)s, this);

                if(dwRes == 0)
                {
                    m_hDevice = (HANDLE)s;
                    m_pClientEventTarget = pEventTarget;  // Cache the client's target.
                }
            }
            else
                CAsyncTCPSocket::Close();
        }
    }

    return dwRes;
}

DWORD CAsyncTCPSocket::Close()
{
    CLock lock(m_sl);

    DWORD dwRes = 0;
    if(m_hDevice != (HANDLE)INVALID_SOCKET)
    {
        if(closesocket((SOCKET)m_hDevice) == SOCKET_ERROR)
            dwRes = (DWORD)WSAGetLastError();
        m_hDevice = (HANDLE)INVALID_SOCKET;
    }
    m_dwTimestamp = 0;

    return dwRes;
}

DWORD CAsyncTCPSocket::Read(LPSTR zBuffer, DWORD dwBufferLength)
{
    CLock lock(m_sl);

    DWORD dwRes = 0;
    m_wsaRecvBuf.buf = zBuffer;
    m_wsaRecvBuf.len = dwBufferLength;
    ZeroMemory(&m_overlapped, sizeof(OVERLAPPED));
    DWORD dwBytesReceived = 0;
    DWORD dwFlags = 0;
    m_iocp.IncrementOutstandingIoCount();

    m_bTimedOut = FALSE;
    m_dwTimestamp = GetTickCount();

    if(WSARecv((SOCKET)m_hDevice, &m_wsaRecvBuf, 1, &dwBytesReceived, &dwFlags, &m_overlapped, NULL) == SOCKET_ERROR)
    {
        dwRes = (DWORD)WSAGetLastError();
        if(dwRes == WSA_IO_PENDING)
            dwRes = 0;  // IO pending is a normal case.
        else
        {
            m_iocp.DecrementOutstandingIoCount();
            m_dwTimestamp = 0;
        }
    }

    return dwRes;
}

DWORD CAsyncTCPSocket::Write(LPCSTR zBuffer, DWORD dwBufferLength)
{
    CLock lock(m_sl);

    if(dwBufferLength == 0)
    {
        dwBufferLength = lstrlen(zBuffer);
    }

    m_wsaSendBuf.buf = (LPSTR)zBuffer;
    m_wsaSendBuf.len = dwBufferLength;

    DWORD dwRes = 0;
    DWORD dwBytesSent = 0;
    ZeroMemory(&m_overlapped, sizeof(OVERLAPPED));
    m_iocp.IncrementOutstandingIoCount();

    m_bTimedOut = FALSE;
    m_dwTimestamp = GetTickCount();

    if(WSASend((SOCKET)m_hDevice, &m_wsaSendBuf, 1, &dwBytesSent, 0, &m_overlapped, NULL) == SOCKET_ERROR)
    {
        dwRes = (DWORD)WSAGetLastError();
        if(dwRes == WSA_IO_PENDING)
            dwRes = 0;  // IO pending is a normal case.
        else
        {
            // The socket send failed.
            m_iocp.DecrementOutstandingIoCount();
            m_dwTimestamp = 0;
        }
    }

    return dwRes;
}

DWORD CAsyncTCPSocket::Read(LPWSABUF lpBuffers, DWORD dwBufferCount)
{
    CLock lock(m_sl);

    DWORD dwRes = 0;
    ZeroMemory(&m_overlapped, sizeof(OVERLAPPED));
    DWORD dwBytesReceived = 0;
    DWORD dwFlags = 0;
    m_iocp.IncrementOutstandingIoCount();

    m_bTimedOut = FALSE;
    m_dwTimestamp = GetTickCount();

    dwRes = WSARecv((SOCKET)m_hDevice, lpBuffers, dwBufferCount, &dwBytesReceived, &dwFlags, &m_overlapped, NULL);
    if(dwRes == SOCKET_ERROR)
    {
        dwRes = (DWORD)WSAGetLastError();
        if(dwRes == WSA_IO_PENDING)
        {
            dwRes = 0;  // IO pending is a normal case.
        }
        else
        {
            m_iocp.DecrementOutstandingIoCount();
            m_dwTimestamp = 0;
        }
    }
    else
    {
        if(dwRes != 0)
            assert(0);
    }

    return dwRes;
}

DWORD CAsyncTCPSocket::Write(LPWSABUF lpBuffers, DWORD dwBufferCount)
{
    CLock lock(m_sl);

    DWORD dwRes = 0;
    DWORD dwBytesSent = 0;
    ZeroMemory(&m_overlapped, sizeof(OVERLAPPED));
    m_iocp.IncrementOutstandingIoCount();

    m_bTimedOut = FALSE;
    m_dwTimestamp = GetTickCount();

    if(WSASend((SOCKET)m_hDevice, lpBuffers, dwBufferCount, &dwBytesSent, 0, &m_overlapped, NULL) == SOCKET_ERROR)
    {
        dwRes = (DWORD)WSAGetLastError();
        if(dwRes == WSA_IO_PENDING)
            dwRes = 0;  // IO pending is a normal case.
        else
        {
            assert(0);
            m_iocp.DecrementOutstandingIoCount();
            m_dwTimestamp = 0;
        }
    }

    return dwRes;
}

BOOL CAsyncTCPSocket::IsOpen()
{
    CLock lock(m_sl);

    return m_hDevice != (HANDLE)INVALID_SOCKET;
}

BOOL CAsyncTCPSocket::GetRemoteHostName(LPSTR pHostName, DWORD dwSize) const
{
    CLock lock(m_sl);

    BOOL bRet = FALSE;
    std::string strHostName;
    if(s_hostResService.ResolveIPAddress(m_remoteAddr.sin_addr, strHostName)
        && strHostName.size() < dwSize)
    {
        lstrcpyA(pHostName,strHostName.c_str());
        bRet = TRUE;
    }
    return bRet;
}

BOOL CAsyncTCPSocket::Timeout(DWORD dwDuration, SOCKET& socket)
{
    BOOL bRes = FALSE;

    // If the timestamp is not zero, it means that an I/O is in progress.
    if(m_dwTimestamp)
    {
        DWORD dwElapsed;
        DWORD dwNow = GetTickCount();
        if(dwNow < m_dwTimestamp)
        {
            // m_dwTimestamp was taken before the system time reached its limit,
            // and dwNow after it loop around to zero.
            // Evaluate the duration with care:
            dwElapsed = dwNow + (MAXDWORD - m_dwTimestamp);
        }
        else
        {
            // Normal case.
            dwElapsed = dwNow - m_dwTimestamp;
        }

        if(dwElapsed >= dwDuration)
        {
            socket = (SOCKET)m_hDevice;
            m_bTimedOut = TRUE;         // Remember that a timeout occurred.
            CAsyncTCPSocket::Close();   // Forcibly close the socket!
            bRes= TRUE;
        }
    }

    return bRes;
}

// ----------------------------------------------------------------------------
//
// CAsyncTCPSocket::CHostResolution members definitions.
//
// ----------------------------------------------------------------------------

BOOL CAsyncTCPSocket::CHostResolution::ResolveName(LPCSTR zHostName, IN_ADDR& addr)
{
    BOOL bRet = TRUE;
    CLock lock(m_pFlName);

    std::string strHostName = zHostName;
    StringToAddr::iterator it = m_NameToAddr.find(strHostName);
    if(it != m_NameToAddr.end())
        addr.s_addr = (*it).second;     // Was already cached.
    else
    {
        // Resolve machine name to address.
        PHOSTENT pHostEnt = gethostbyname(zHostName);
        if(pHostEnt == NULL)
            bRet = FALSE;
        else
        {
            // Cache the address found.
            CopyMemory(&addr, pHostEnt->h_addr, pHostEnt->h_length);
            m_NameToAddr[strHostName] = addr.s_addr;
        }
    }

    return bRet;
}

BOOL CAsyncTCPSocket::CHostResolution::ResolveIPAddress(const IN_ADDR& addr, std::string &hostName)
{
    BOOL bRet = TRUE;
    CLock lock(m_pFlAddr);

    AddrToString::iterator it = m_AddrToName.find(addr.s_addr);
    if(it != m_AddrToName.end())
        hostName = (*it).second;    // Was already cached.
    else
    {
        // Resolve address to machine name.
        PHOSTENT pHostEnt = gethostbyaddr((const char*)&addr, 4, AF_INET);
        if(pHostEnt == NULL)
            bRet = FALSE;
        else
        {
            // Cache the name found.
            hostName = pHostEnt->h_name;
            m_AddrToName[addr.s_addr] = hostName;
        }
    }

    return bRet;
}

void CAsyncTCPSocket::OnEvent(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType)
{
    CLock lock(m_sl);

    m_dwTimestamp = 0;
    m_bTimedOut = FALSE;

    // Forward to the caller:
    if(m_pClientEventTarget)
        m_pClientEventTarget->OnEvent(dwNumberOfBytesTransferred, bIsIo, dwEventType);
}

void CAsyncTCPSocket::OnEventFailed(DWORD dwNumberOfBytesTransferred, BOOL bIsIo, DWORD dwEventType, DWORD dwError)
{
    CLock lock(m_sl);

    m_dwTimestamp = 0;

    // Forward to the caller:
    if(m_pClientEventTarget)
    {
        // Look specifically for a forced timeout.
        if(m_bTimedOut && dwError == ERROR_NETNAME_DELETED)
            dwError = WAIT_TIMEOUT;
        m_pClientEventTarget->OnEventFailed(dwNumberOfBytesTransferred, bIsIo, dwEventType, dwError);
    }
}
