/*
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _QUOTIUM_OBJECT_POOL_H_
#define _QUOTIUM_OBJECT_POOL_H_

#include <stddef.h>
#include "Fastlock.hpp"

////////////////////////////////////////////////////////////////////////////////
// General purpose object pool.
// THREAD SAFE.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

// The object pool class.
template<typename O>
class ObjectPool
{
public:
    // Construction / Destruction:
    ObjectPool(DWORD dwRegionSizeBytes = 5242880,   // 5 MB
               DWORD dwIncrementObjects= 256)
           : m_dwRegionSizeBytes(dwRegionSizeBytes),
             m_dwIncrementObjects(dwIncrementObjects),
             m_dwCommitedElements(0), m_pRegion(NULL),
             m_pBusyList(NULL), m_pFreeList(NULL), m_dwFreeListLength(0)
    {
        m_pRegion = VirtualAlloc(NULL, dwRegionSizeBytes,
            MEM_RESERVE | MEM_TOP_DOWN, PAGE_READWRITE);
    }
    ~ObjectPool()
    {
        VirtualFree(m_pRegion, 0, MEM_RELEASE);
    }

    // Validity check:
    BOOL IsValid() const {CLock lock(m_objectPoolLock); return m_pRegion != NULL;}

    // Allocating objects.
    // The ao array is filled on output with pointers
    // to uninitialized objects.
    // dwNumber is the number of objects to allocate.
    // Returns NULL if memory error.
    BOOL AllocateObjects(O* ao[], DWORD dwNumber = 1)
    {
        CLock lock(m_objectPoolLock);

        // Allocate new elements if necessary:
        while(dwNumber > m_dwFreeListLength)
        {
            if(!Allocate())
                return FALSE;
        }

        // Unhook the elements from the free list...
        Element* pEl = m_pFreeList;
        Element* pElLast;
        DWORD dw;
        for(dw = 0; dw < dwNumber; ++dw)
        {
            pElLast = m_pFreeList;
            ao[dw] = &pElLast->m_object;
            m_pFreeList = m_pFreeList->m_pNext;
        }
        if(m_pFreeList)
            m_pFreeList->m_pPrevious = NULL;

        // ... and hook them up at the front of the busy list:
        pElLast->m_pNext = m_pBusyList;
        if(m_pBusyList)
            m_pBusyList->m_pPrevious = pElLast;
        m_pBusyList = pEl;

        // Update free list length:
        m_dwFreeListLength -= dwNumber;

        return TRUE;
    }

    // Releasing objects.
    void FreeObjects(O* ao[], DWORD dwNumber = 1)
    {
        CLock lock(m_objectPoolLock);

        for(DWORD dw = 0; dw < dwNumber; ++dw)
        {
            // Get the element pointer from the m_object member:
            Element* pEl = (Element*)(((BYTE*)(ao[dw])) - offsetof(Element, m_object));

            // Unhook the element from somewhere in the busy list:
            if(pEl->m_pPrevious)
                pEl->m_pPrevious->m_pNext = pEl->m_pNext;
            if(pEl->m_pNext)
                pEl->m_pNext->m_pPrevious = pEl->m_pPrevious;
            if(m_pBusyList == pEl)
                m_pBusyList = m_pBusyList->m_pNext;

            // And hook it up at the front of the free list:
            pEl->m_pNext = m_pFreeList;
            if(m_pFreeList)
                m_pFreeList->m_pPrevious = pEl;
            m_pFreeList = pEl;
        }
        if(m_pFreeList)
            m_pFreeList->m_pPrevious = NULL;

        // Update free list length:
        m_dwFreeListLength += dwNumber;
    }

    // Enumerating objects in the pool.
    // To support object enumeration, the caller of Enumerate must be a class
    // implementing the following interface:
    struct PoolObjectEnumerator
    {
        // This function should return TRUE to continue the
        // enumeration, FALSE to stop it.
        virtual BOOL NextObject(O& object) = 0;
    };
    void Enumerate(PoolObjectEnumerator& poe)
    {
        CLock lock(m_objectPoolLock);

        BOOL bContinue = TRUE;
        for(Element* pEl = m_pBusyList; bContinue && pEl != NULL; pEl = pEl->m_pNext)
            bContinue = poe.NextObject(pEl->m_object);
    }

private:
    //// Data Types.
    struct Element
    {
        Element    *m_pNext, *m_pPrevious;
        O           m_object;
    };
    typedef Element* PElement;

    //// Helpers.

    // Allocate an increment worth of elements.
    BOOL Allocate()
    {
        BOOL bRet = TRUE;

        // Allocate an increment:
        bRet = VirtualAlloc((PBYTE)m_pRegion + m_dwCommitedElements * sizeof(Element),
            m_dwIncrementObjects * sizeof(Element), MEM_COMMIT, PAGE_READWRITE) != NULL;
        if(bRet)
        {
            // Initialize the elements: prepend them to the free list:
            Element* pUpperLimit = (PElement)m_pRegion + m_dwCommitedElements + m_dwIncrementObjects;
            for(Element* pEl = (PElement)m_pRegion + m_dwCommitedElements;
                         pEl < pUpperLimit;
                       ++pEl)
            {
                // Hook up the element at the front of the free list:
                pEl->m_pNext = m_pFreeList;
                if(m_pFreeList)
                    m_pFreeList->m_pPrevious = pEl;
                m_pFreeList = pEl;
            }
            m_pFreeList->m_pPrevious = NULL;
            m_dwCommitedElements += m_dwIncrementObjects;
            m_dwFreeListLength += m_dwIncrementObjects;
        }

        return bRet;
    }

    //// Data.
    void*       m_pRegion;
    Element    *m_pBusyList, *m_pFreeList;
    DWORD       m_dwFreeListLength;
    DWORD       m_dwRegionSizeBytes;
    DWORD       m_dwCommitedElements;
    DWORD       m_dwIncrementObjects;
    mutable CFastLock m_objectPoolLock;
};

#endif // _QUOTIUM_OBJECT_POOL_H_
