/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

////////////////////////////////////////////////////////////////////////////////
// Thread pool implementation file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "ThreadPool.hpp"
#include "TimerQueue.hpp"
#include "Common.hpp"
#include <process.h>
#include <malloc.h>


CThreadPool::CThreadPool(CIOCP& iocp, DWORD dwSize, CTrace* pTrace)
    : m_iocp(iocp), m_ppThreads(NULL), m_dwSize(dwSize),
      m_dwPeakRunnableThreads(0), m_hTerminationEvent(NULL),
      m_dwState(STATE_STOPPED), m_dwReadyThreads(0),
      m_pTrace(pTrace)
#ifdef USE_PERF_DATA
	  , m_plActiveThreadsPerfCtr(NULL)
#endif
{
}

CThreadPool::~CThreadPool()
{
    Cleanup();
}

void CThreadPool::Cleanup()
{
    // Delete the thread objects.
    if(m_ppThreads)
    {
        for(DWORD dw = 0; dw < m_dwSize; dw++)
        {
            if(m_ppThreads[dw])
            {
                m_ppThreads[dw]->CloseHandle();
                delete m_ppThreads[dw];
            }
            m_ppThreads[dw] = NULL;
        }
        delete [] m_ppThreads;
        m_ppThreads = NULL;
    }

    // Close the termination event handle.
    if(m_hTerminationEvent)
    {
        CloseHandle(m_hTerminationEvent);
        m_hTerminationEvent = NULL;
    }

    m_dwState = STATE_STOPPED;
}

DWORD CThreadPool::Run(IocpEventDispatcher *pEventDispatcher, DWORD m_dwTimeoutMilliseconds)
{
    DWORD dwRes = 0;

#ifdef USE_PERF_DATA
    // Note: we assume there's only ONE instance of the CThreadPool class.
    m_plActiveThreadsPerfCtr    = g_pPerformance->GetCounterValue(PRFCNT_OPENSTA_THREADPOOL_ACTIVETHREADS);
    *m_plActiveThreadsPerfCtr   = 0;
#endif

    m_dwPeakRunnableThreads = 0;
    ClearLiveThreadCount();
    m_dwState = STATE_STARTING;

    // Create the termination event.
    m_hTerminationEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

    // Create the thread pool.
    m_dwReadyThreads = 0;
    DWORD dw, dwError;
    m_ppThreads = new PTHREAD[m_dwSize];
    ZeroMemory(m_ppThreads, m_dwSize * sizeof(PTHREAD));
    for(dw = 0; dw < m_dwSize; dw++)
    {
        // Create the thread object.
        m_ppThreads[dw] = new CThread(*this, m_dwTimeoutMilliseconds, m_pTrace);
        if(m_ppThreads[dw] == NULL)
        {
            dwRes = ERROR_NOT_ENOUGH_MEMORY;
            break;
        }

        // Create and start the real OS thread.
        dwError = m_ppThreads[dw]->Create();
        if(dwError != 0)
        {
            dwRes = dwError;
            if(m_pTrace)
            {
                GetTraceStream(m_pTrace) << "ERROR: could not create thread " << dw << ". Error is: " << dwRes << "\r\n";
                ReleaseTraceStream();
            }
            break;
        }
    }

    if(dwRes == 0)
    {
        // Wait for the termination event to be set. In this particular case,
        // we wait for all threads to be started and blocked on the IOCP.
        WaitForSingleObject(m_hTerminationEvent, INFINITE);
        ResetEvent(m_hTerminationEvent);

        m_dwState = STATE_RUNNING;

        dwRes = OnStart();
        if(dwRes == 0)
        {
            // Posting a start event to the I/O completion queue.
            if(pEventDispatcher)
            {
                dwRes = m_iocp.PostCompletedIO(0, pEventDispatcher, (LPOVERLAPPED)IocpEvent::EVENT_TYPE_THREADPOOL_START);
            }

            if(dwRes == 0)
            {
                // Wait for the termination event to be set.
                WaitForSingleObject(m_hTerminationEvent, INFINITE);
            }
            else
            {
                if(m_pTrace)
                {
                    GetTraceStream(m_pTrace) << "ERROR: could not start the thread engine. Error is " << dwRes << "\r\n";
                    ReleaseTraceStream();
                }
            }
        }
    }

    StopThreads();
    Cleanup();
    dwRes = OnExit();

    return dwRes;
}

void CThreadPool::StopThreads()
{
    m_dwState = STATE_STOPPING;

    // Build an array of all thread handles.
    HANDLE *phThread = (HANDLE*)_alloca((size_t)(m_dwSize * sizeof(HANDLE)));
    DWORD dw = 0;
    for(dw = 0; dw < m_dwSize; dw++)
    {
        assert(m_ppThreads[dw] != NULL);
        phThread[dw] = *m_ppThreads[dw];
    }

    // Stop the threads.
    for(dw = 0; dw < m_dwSize; dw++)
        m_iocp.PostCompletedIO(0, NULL, (LPOVERLAPPED)IocpEvent::EVENT_TYPE_THREADPOOL_STOP_THREAD);

    // Wait for the end of the threads (by chunks of MAXIMUM_WAIT_OBJECTS threads).
    DWORD dwFirst = 0, dwLast;  // The dwLast bound is excluded from the interval.
    do
    {
        dwLast = __min(dwFirst + MAXIMUM_WAIT_OBJECTS, m_dwSize);
        WaitForMultipleObjects(dwLast - dwFirst, phThread + dwFirst, TRUE, INFINITE);
        dwFirst = dwLast;
    } while(dwFirst < m_dwSize);

    // Close all thread handles.
    for(dw = 0; dw < m_dwSize; dw++)
        m_ppThreads[dw]->CloseHandle();
}

DWORD CThreadPool::IncrementRunnableThreadCount()
{
#ifdef USE_PERF_DATA
    DWORD dwRunnableThreadCount = (DWORD)InterlockedIncrement(m_plActiveThreadsPerfCtr);

    if(m_dwState == STATE_RUNNING)
    {
        // Adjust the peak runnable thread count if need.
        if(dwRunnableThreadCount > m_dwPeakRunnableThreads)
            InterlockedExchange((LPLONG)&m_dwPeakRunnableThreads, (LONG)dwRunnableThreadCount);
    }

    return dwRunnableThreadCount;
#else
    return 0;
#endif
}

DWORD CThreadPool::DecrementRunnableThreadCount()
{
#ifdef USE_PERF_DATA
    return (DWORD)InterlockedDecrement(m_plActiveThreadsPerfCtr);
#else
    return 0;
#endif
}

void CThreadPool::ThreadWillBlockOnIocp()
{
    if(m_dwState == STATE_STARTING)
    {
        if(InterlockedIncrement((LPLONG)&m_dwReadyThreads) == (LONG)m_dwSize)
        {
            // The thread pool is ready.
            SetEvent(m_hTerminationEvent);
        }
    }
}

////////////////////////// CThread class //////////////////////////////


CThreadPool::CThread::CThread(CThreadPool& threadPool, DWORD dwTimeoutMilliseconds, CTrace* pTrace)
    : m_threadPool(threadPool), m_hThread(NULL), m_dwTimeoutMilliseconds(dwTimeoutMilliseconds),
      m_pTrace(pTrace)
{
}

CThreadPool::CThread::~CThread()
{
    CloseHandle();
}

DWORD CThreadPool::CThread::Create()
{
    DWORD dwRes = 0;

    // Actually create the thread.
    m_hThread = BeginThreadEx(
        NULL,
        0,
        ThreadFunction,
        (void *)this,
        0,
        NULL
    );
    if(m_hThread == 0)
        dwRes = errno;
    else
        OnInit();

    return dwRes;
}

DWORD CThreadPool::CThread::Resume()
{
    DWORD dwRes = 0;

    // Resume the thread.
    if(ResumeThread(m_hThread) == 0xFFFFFFFF)
        dwRes = GetLastError();

    return dwRes;
}

DWORD CThreadPool::CThread::Suspend()
{
    DWORD dwRes = 0;

    // Suspend the thread.
    if(SuspendThread(m_hThread) == 0xFFFFFFFF)
        dwRes = GetLastError();

    return dwRes;
}

DWORD WINAPI CThreadPool::CThread::ThreadFunction(LPVOID pContext)
{
    CThread *pThread = (CThread *)pContext;
    pThread->m_threadPool.IncrementLiveThreadCount();
    pThread->m_threadPool.IncrementRunnableThreadCount();
    srand(GetCurrentThreadId());  // Randomize in thread with original seed.
    DWORD dwError = pThread->OnRun();
    pThread->OnExit();
    pThread->m_threadPool.OnExitThread(*pThread);
    pThread->m_threadPool.DecrementRunnableThreadCount();
    pThread->m_threadPool.DecrementLiveThreadCount();

    return dwError;
}

DWORD CThreadPool::CThread::OnRun()
{
    DWORD dwRes = 0;
    CIOCP& iocp = m_threadPool.m_iocp;
    DWORD dwBytesTransferred;
    LPOVERLAPPED pOverlapped;
    BOOL bContinue = TRUE;

    if(m_pTrace)
    {
        GetTraceStream(m_pTrace)
            << "Pooled Thread " << GetCurrentThreadId() << " starting..." << "\r\n";
        ReleaseTraceStream();
    }

    while(bContinue)
    {
        dwBytesTransferred = 0;
        pOverlapped = NULL;

        // Wait for the next completed I/O.
        IocpEventDispatcher *pEventDispatcher = NULL;
        m_threadPool.ThreadWillBlockOnIocp();
        m_threadPool.DecrementRunnableThreadCount();
        DWORD dwError = iocp.GetNextCompletionStatus(&dwBytesTransferred, pEventDispatcher, &pOverlapped, m_dwTimeoutMilliseconds);
        m_threadPool.IncrementRunnableThreadCount();

        if(dwError == 0)
        {
            // Stop event?
            if((DWORD)pOverlapped == IocpEvent::EVENT_TYPE_THREADPOOL_STOP_THREAD)
                bContinue = FALSE;  // Stop the thread.
            else
                pEventDispatcher->GetIocpEventTarget()->OnEvent(dwBytesTransferred, ((DWORD)pOverlapped > 0xFFFF), (DWORD)pOverlapped);
        }
        else
        {
            if(pOverlapped != NULL) // Failed I/O.
                pEventDispatcher->GetIocpEventTarget()->OnEventFailed(dwBytesTransferred, ((DWORD)pOverlapped > 0xFFFF), (DWORD)pOverlapped, dwError);
            else
            {
                if(dwError == WAIT_TIMEOUT)
                {
                    // Timed out while waiting for the next completed I/O.
                    OnTimeout();

                    if(m_pTrace)
                    {
                        GetTraceStream(m_pTrace)
                            << "Thread " << GetCurrentThreadId() << " timed out: exiting..." << "\r\n";
                        ReleaseTraceStream();
                    }

                    bContinue = FALSE;  // Stop the thread.
                }
                else
                {
                    OnError(dwError);
                }
            }
        }
    }

    if(m_pTrace)
    {
        GetTraceStream(m_pTrace)
            << "Pooled Thread " << GetCurrentThreadId() << " normally exiting..." << "\r\n";
        ReleaseTraceStream();
    }

    return dwRes;
}
