/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

#ifndef _OPENSTA_TIMERQUEUE_H_
#define _OPENSTA_TIMERQUEUE_H_

////////////////////////////////////////////////////////////////////////////////
// Timer queue definition file.
//
// Return values: unless otherwise stated, a DWORD return value is a
// GetLastError() error code.
////////////////////////////////////////////////////////////////////////////////

#include "Iocp.hpp"
#include "Trace.hpp"
#pragma warning (disable: 4786)
#include <queue>
#pragma warning (disable: 4786)

// This class implements a queue of timer durations, to simulate overlapped waits.
// THREAD SAFE as soon as Init() has been succesfully called.
class CTimerQueue
{
public:
    //// --------------------------------------------------------------------------------
    //// Construction / Destruction.
    //// --------------------------------------------------------------------------------
    CTimerQueue(CIOCP& iocp, DWORD dwTimeOutMilliseconds, CTrace* pTrace = NULL)
        : m_iocp(iocp), m_dwTimeOutMilliseconds(dwTimeOutMilliseconds),
          m_hTimerThread(NULL), m_bQueueOpen(FALSE),
          m_dwPeakSize(0),
          m_pTrace(pTrace),
          m_sl(4000)
#ifdef USE_PERF_DATA
          ,m_pTimerQueueLengthPerfCount(NULL)
#endif
    {}
    ~CTimerQueue();

    //// --------------------------------------------------------------------------------
    //// User methods.
    //// --------------------------------------------------------------------------------

    // Mandatory initialization function.
    // Creates and starts the timer thread. Creates support objects.
    DWORD Init();

    // Pushes a duration into the timer queue. dwDuration is expressed
    // in milliseconds. To stop the timer thread, push an entry having
    // pEventDispatcher = NULL.
    // If bForce is TRUE, then a new entry is inserted into the queue
    // even if it's closed.
    // dwExtraInfo is any useful info which will be passed as the
    // dwNumberOfBytes argument to the IOCP's completion queue.
    DWORD Push(IocpEventDispatcher *pEventDispatcher, DWORD dwDuration, BOOL bForce = FALSE, DWORD dwExtraInfo = 0);

    // Retrieving the peak size of the queue.
    DWORD GetPeakQueueSize() const {return m_dwPeakSize;}

    // HANDLE cast operator yielding the timer thread handle.
    operator HANDLE() const {return m_hTimerThread;}

private:
    //// --------------------------------------------------------------------------------
    //// Helpers.
    //// --------------------------------------------------------------------------------

    // Pops durations from the timer queue when the time is up, and sends them to
    // the IOCP's completion queue.
    // pEventDispatcher is an output argument. Called internally by the timer thread.
    // This function returns only when the loop breaks, that is, when an entry
    // with a NULL IocpEventDispatcher* is popped.
    void PopLoop();

    // Timer queue entry.
    struct CTimerQueueEntry
    {
        __int64              m_qwTimestamp;
        IocpEventDispatcher* m_pEventDispatcher;
        DWORD                m_dwExtraInfo;
        bool operator< (const CTimerQueueEntry& ct) const {return m_qwTimestamp < ct.m_qwTimestamp;}
        bool operator> (const CTimerQueueEntry& ct) const {return m_qwTimestamp > ct.m_qwTimestamp;}
        bool operator==(const CTimerQueueEntry& ct) const {return m_qwTimestamp == ct.m_qwTimestamp;}
        CTimerQueueEntry() : m_qwTimestamp(0), m_pEventDispatcher(NULL), m_dwExtraInfo(0) {}
        CTimerQueueEntry(const CTimerQueueEntry& ct) : m_qwTimestamp(ct.m_qwTimestamp), m_pEventDispatcher(ct.m_pEventDispatcher), m_dwExtraInfo(ct.m_dwExtraInfo) {}
        CTimerQueueEntry(__int64 qwTimestamp, IocpEventDispatcher* pEventDispatcher, DWORD dwExtraInfo = 0) : m_qwTimestamp(qwTimestamp), m_pEventDispatcher(pEventDispatcher), m_dwExtraInfo(dwExtraInfo) {}
        void operator=(const CTimerQueueEntry& ct) {m_qwTimestamp = ct.m_qwTimestamp; m_pEventDispatcher = ct.m_pEventDispatcher; m_dwExtraInfo = ct.m_dwExtraInfo;}
    };
    typedef std::priority_queue<CTimerQueueEntry, std::vector<CTimerQueueEntry>, std::greater<CTimerQueueEntry> > CTimerPriorityQueue;

    // Timer thread function. The this pointer is transmitted via pContext.
    // Starts pumping durations from the queue with the Pop() function.
    // As a new duration is popped, a notification is posted
    // to the IOCP via m_iocp.PostCompletedIO().
    // If a popped duration contains a NULL pointer, the thread exits.
    static DWORD WINAPI TimerThreadFunction(LPVOID pContext);

    //// --------------------------------------------------------------------------------
    //// Data.
    //// --------------------------------------------------------------------------------
    CIOCP&              m_iocp;                 // The associated IOCP.
    HANDLE              m_hTimer;               // Waitable timer used to implement the waiting.
    CFastLock           m_sl;                   // Protect the object against concurrent accesses.
    CTimerPriorityQueue m_queue;                // The priority queue implementing asynchronous waits.
    DWORD               m_dwTimeOutMilliseconds;// Timeout for the timer queue.
    HANDLE              m_hTimerThread;         // The timer thread's handle.
    DWORD               m_dwPeakSize;           // The peak queue size.
    BOOL                m_bQueueOpen;           // If TRUE, the Push() method is allowed to insert an entry
                                                // into the queue. This allows to close the queue when
                                                // the timer thread is exiting, so that other threads
                                                // do not fill it up.
    CTrace             *m_pTrace;               // The ostream trace support object.

#ifdef USE_PERF_DATA
    LONG *              m_pTimerQueueLengthPerfCount; // Pointer to performance counter.
#endif

private:
    //// --------------------------------------------------------------------------------
    //// Forbidden.
    //// --------------------------------------------------------------------------------
    void operator=(const CTimerQueue&);
};

#endif  // _OPENSTA_TIMERQUEUE_H_
