/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
                        /**********************************
                        *              PHOENIX            *
                        * Base class for parser.          *
                        * Class definition.               *
                        **********************************/


#include "poptions.h"                   // Compilation options. Generated by parser generator or lexical generator.
                                        // Includes choice of target platform and array packing
#include "wsparrt.h"                    // User code and support routines, Part 1 and Part 3 of the parser source program
#include "wsparseref.h"                 // Static production length and left side arrays


// Lookup in the parsing table.
// See "Compilers. Principles, Techniques, and Tools" by
// Aho, Sethi and Ullman [1986], page 219.
// Arguments :	ulState (IN) : the current state.
//              ulSymbol (IN): the current input symbol (terminal or nonterminal).
// Return    :  parsing table entry value: an action or a goto state.
template <typename U>
inline long PhParser<U>::ActionGoto(unsigned long ulState, unsigned long ulSymbol)
{
    unsigned int uiAGPoolIndex, uiAGs;

    if(ulSymbol < NONTERMINAL_START)
    {
        /* Action */
        if(ActionState[ulState] < 0)
            return ActionState[ulState];   // Simple reduction!
        uiAGPoolIndex = ActionState[ulState] * 2;
        while((uiAGs = AGPool[uiAGPoolIndex]) != ulSymbol && uiAGs != END_OF_LIST)
            uiAGPoolIndex += 2;
    }
    else
    {
        /* Goto */
        uiAGPoolIndex = GotoSymbol[ulSymbol - NONTERMINAL_START] * 2;
        while((uiAGs = AGPool[uiAGPoolIndex]) != ulState && uiAGs != END_OF_LIST)
            uiAGPoolIndex += 2;
    }

    return AGPool[uiAGPoolIndex + 1];
}


// Parsing algorithm.
// Return:  TRUE if success, FALSE if fatal parser error.
template <typename U>
inline BOOL PhParser<U>::Parse(const AddressQueryList& aql, AddressValueList& avl, unsigned long* pulErrorLine, unsigned long* pulErrorCol, BOOL bProduceTree)
{
    BOOL bError = FALSE;
    bProduceTree_ = bProduceTree;
    BOOL bContinueParsing = TRUE;
    ParsingPhase ulPhase = PHASE_NORMAL;
    long lToken = 0, lAction = 0, lProduction = 0, lState = 0, lPlen = 0;
    unsigned long ulTokenLine = 0, ulTokenCol = 0;
    unsigned long ulShiftError = 0;
    BOOL bNeedNextToken = TRUE;
    CSymbolAttributes<U> tokenAttributes = CSymbolAttributes<U>();
    CSymbolAttributes<U> leftSymbolAttributes = CSymbolAttributes<U>();
    bHasHTMLTag_ = bHasHeadTag_ = bHasBodyTag_ = FALSE;

    // Fill in the avl.
    avl.clear();
    for(AddressQueryList::const_iterator it = aql.begin(); it != aql.end(); it++)
        avl.push_back(AddressValue(*it));

    // Push state 0 on parser stack.
    stack_.push_back(StackEntry());

    do
    {
        if(bNeedNextToken)
        {
            // Read next token.
            lToken = lexer_.FindToken(ulTokenLine, ulTokenCol, &tokenAttributes);
            if(lToken == -1 || (bError && lToken == PhLexer<U>::ENDMARKER))
            {
                OnFatalError(ulPhase == PHASE_ERROR);
                return FALSE;
            }

            if(pulErrorLine)
                *pulErrorLine = ulTokenLine;
            if(pulErrorCol)
                *pulErrorCol  = ulTokenCol;
        }

        AfterErrorAction afterErrorAction;
        lAction = ActionGoto(stack_.back().ulSymbol_, lToken);
        switch(lAction)
        {
            case BLANK:           // Syntax error
                bError = TRUE;

                // Signal the error.
                afterErrorAction = OnError(ulTokenLine, ulTokenCol, ulPhase == PHASE_ERROR);
                switch(afterErrorAction)
                {
                    case RAISE_FATAL_ERROR:
                        OnFatalError(TRUE);
                        bContinueParsing = FALSE;
                        break;

                    case STOP_PARSER:
                        OnAccept();
                        bContinueParsing = FALSE;
                        bError = FALSE;
                        break;

                    case TRY_TO_RECOVER:    // Error recovery.
                    default:
                        {
#ifdef PHTRACE
                            os_ << "DISCARDING TOKEN: ";
                            lexer_.DumpLexeme(os_);
                            os_ << std::endl;
#endif

                            if(ulPhase == PHASE_ERROR)
                            {
                                // Pop states from stack to return to error state
                                while(ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN) < 0)
                                    Pop(1);

                                // Shift the error virtual token
                                ulShiftError = ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN);
                                stack_.push_back(StackEntry(ulShiftError));
#ifdef PHTRACE
                                os_ << "RETURNED TO ERROR STATE WHILE RECOVERING" << std::endl;
#endif
                            }
                            else
                            {
                                // Pop states from stack to try to find error state (a state from which the
                                // error virtual token can be shifted onto the stack).
                                while(stack_.size() > 0 && ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN) < 0)
                                    Pop(1);

                                if(ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN) < 0)
                                {
                                    // Error state not found: Fatal error.
                                    OnFatalError(FALSE);
                                    bContinueParsing = FALSE;
                                }
                                else
                                {
                                    //// Found error state

                                    // Signal beginning of recovery phase.
                                    OnEnterErrorRecovery(ulPhase == PHASE_ERROR);

                                    // Enter error recovery phase.
                                    ulPhase = PHASE_ERROR;

                                    // Shift the error virtual token.
                                    ulShiftError = ActionGoto(stack_.back().ulSymbol_, ERROR_TOKEN);
                                    stack_.push_back(StackEntry(ulShiftError));
#ifdef PHTRACE
                                    os_ << "SHIFT ERROR TOKEN FROM ERROR STATE" << std::endl;
#endif
                                }
                            }
                            bNeedNextToken = TRUE;
                        }
                        break;
                }

                break;

            case ACCEPT:
                if(ulPhase == PHASE_ERROR)
                    ExitErrorRecovery();

#ifdef PHTRACE
                os_ << "@@@ ACCEPT" << std::endl;
#endif
                OnAccept();
                bContinueParsing = FALSE;
                break;

            default:    // Shift or reduce
                if(ulPhase == PHASE_ERROR)
                    ExitErrorRecovery();

                if(lAction > 0) // SHIFT
                {
                    stack_.push_back(StackEntry(lAction, tokenAttributes));

#ifdef PHTRACE
                    os_ << "@@@ SHIFT " << lAction << "(";
                    lexer_.DumpLexeme(os_);
                    os_ << ")" << std::endl;
#endif
                    bNeedNextToken = TRUE;
                }
                else           // REDUCE
                {
                    lProduction = -lAction;
                    lPlen = ProdLength[lProduction];

                    // Define an entry pointing one below the leftmost symbol's
                    // entry in the stack.
                    StackEntry* pProdAttr = &stack_[stack_.size() - lPlen - 1];

#ifdef PHTRACE
                    os_ << "@@@ REDUCE " << lProduction << std::endl;
#endif

                    // Perform the semantic action associated with the production.
                    bContinueParsing = PerformSemanticAction(avl, lProduction, leftSymbolAttributes, pProdAttr, ulPhase);

                    // Did the semantic action ask the parser to stop?
                    if(!bContinueParsing)
                    {
                        if(!bError)
                            OnStopParser();
                        else
                            OnFatalError(ulPhase == PHASE_ERROR);
                        break;
                    }

                    // Pop the production's symbols off the stack.
                    Pop(lPlen);

                    // Push the nonterminal on the left side of the production on the stack.
                    lState = ActionGoto(stack_.back().ulSymbol_, LeftSide[lProduction]);
                    stack_.push_back(StackEntry(lState, leftSymbolAttributes));
                    bNeedNextToken = FALSE;
                }
                break;
        }
    } while(bContinueParsing);

    return !bError;
}

template <typename U>
inline BOOL PhParser<U>::Parse(CTag*& pRoot, unsigned long* pulErrorLine, unsigned long* pulErrorCol)
{
    BOOL bRet;
    AddressQueryList aql;
    AddressValueList avl;
    if((bRet = Parse(aql, avl, NULL, NULL, TRUE)))
        pRoot = pRoot_;
    return bRet;
}

// Called after the parser reaches a syntax error.
// lLine, lCol: the line & column of the beginning of the token where the error appeared.
// bDuringRecovery: TRUE error was detected while attempting to recover from a previous
// error.
// Return:  RAISE_FATAL_ERROR:  Emit a fatal error
//          TRY_TO_RECOVER:     Attempt to recover from error
//          STOP_PARSER         Stop the parser go to the accepting state
//                                  (successful termination).
template <typename U>
inline PhParser<U>::AfterErrorAction PhParser<U>::OnError(long lLine, long lCol, BOOL /*bDuringRecovery*/)
{
#ifdef PHTRACE
    os_ << "Syntax error line " << lLine << ", column " << lCol << std::endl;
#else
    lLine; lCol;    // To avoid "unreferenced formal parameter" warning.
#endif
    return TRY_TO_RECOVER;
}

// Called on an unrecovered syntax error.
// bDuringRecovery: TRUE if function was called while attempting to recover
// from a previous error.
template <typename U>
inline void PhParser<U>::OnFatalError(BOOL /*bDuringRecovery*/)
{
#ifdef PHTRACE
    os_ << " FATAL ERROR" << std::endl;
#endif
}

template <typename U>
inline void PhParser<U>::Pop(unsigned long ulLen)
{
    while(ulLen)
    {
        stack_.pop_back();
        ulLen--;
    }
}

template <typename U>
void PhParser<U>::OnEnterErrorRecovery(BOOL /*bDuringRecovery*/)
{
#ifdef PHTRACE
    os_ << "ENTERING ERROR RECOVERY PHASE..." << std::endl;
#endif
}

template <typename U>
void PhParser<U>::OnErrorRecovered()
{
#ifdef PHTRACE
    os_ << "RECOVERED FROM SYNTAX ERROR" << std::endl;
#endif
}

template <typename U>
inline
BOOL PhParser<U>::PerformSemanticAction(
    AddressValueList& avl, long lProduction,
    CSymbolAttributes<U>& leftSymbolAttributes,
    StackEntry* pProdAttr, ParsingPhase& ulPhase)
{
     BOOL bRet = TRUE;           // By default, continue parsing.
     switch(lProduction)
     {
         #include "wsparac.h"    // Phoenix - generated parser actions.
         default:
             // By default, the left nonterminal's attributes are copied from the
             // leftmost symbol's in the right side of the production.
             leftSymbolAttributes = GetATT(1, pProdAttr);
             break;
     }
     return bRet;
}

template <typename U>
inline
void PhParser<U>::ProcessOpeningTag(CSymbolAttributes<U>& tokenAttributes)
{
    Tags eNewTag = tokenAttributes.eTag_;

    // Ignore unknown opening tags!!
    if(eNewTag == TAG_NONE)
    {
        bDisable_ = TRUE;
        return;
    }
    else
        bDisable_ = FALSE;

    // Manage implicit opening tags!
    if(!bHasHTMLTag_)
    {
        if(eNewTag != TAG_HTML)
        {
            // Implicit <HTML> tag!
            CSymbolAttributes<U> HTMLAttributes;
            HTMLAttributes.eTag_ = TAG_HTML;
            ProcessOpeningTag(HTMLAttributes);
        }
        else
            bHasHTMLTag_ = TRUE;
    }
    if(!bHasHeadTag_)
    {
        switch(eNewTag)
        {
            case TAG_BASE:
            case TAG_BASEFONT:
            case TAG_BGSOUND:
            case TAG_LINK:
            case TAG_META:
            case TAG_SCRIPT:
            case TAG_STYLE:
            case TAG_TITLE:
            case TAG_OBJECT:
                // Implicit <HEAD> tag!
                {
                    CSymbolAttributes<U> headAttributes;
                    headAttributes.eTag_ = TAG_HEAD;
                    ProcessOpeningTag(headAttributes);
                }
                break;
            case TAG_HEAD:
                bHasHeadTag_ = TRUE;
                break;
        }
    }
    if(!bHasBodyTag_)
    {
        switch(eNewTag)
        {
            case TAG_SCRIPT:
            case TAG_INS:
            case TAG_DEL:
            // block
            case TAG_P:
            case TAG_H1:
            case TAG_H2:
            case TAG_H3:
            case TAG_H4:
            case TAG_H5:
            case TAG_H6:
            case TAG_UL:
            case TAG_OL:
            case TAG_PRE:
            case TAG_DL:
            case TAG_DIV:
            case TAG_NOSCRIPT:
            case TAG_BLOCKQUOTE:
            case TAG_FORM:
            case TAG_HR:
            case TAG_TABLE:
            case TAG_FIELDSET:
            case TAG_ADDRESS:
            case TAG_MARQUEE:
            // inline (transitional DTD).
            case TAG_A:
            case TAG_TT:
            case TAG_I:
            case TAG_B:
            case TAG_BIG:
            case TAG_SMALL:
            case TAG_EM:
            case TAG_STRONG:
            case TAG_DFN:
            case TAG_CODE:
            case TAG_SAMP:
            case TAG_KBD:
            case TAG_VAR:
            case TAG_OBJECT:
            case TAG_CITE:
            case TAG_ABBR:
            case TAG_ACRONYM:
            case TAG_IMG:
            case TAG_BR:
            case TAG_MAP:
            case TAG_Q:
            case TAG_SUB:
            case TAG_SUP:
            case TAG_SPAN:
            case TAG_BDO:
            case TAG_INPUT:
            case TAG_SELECT:
            case TAG_TEXTAREA:
            case TAG_LABEL:
            case TAG_BUTTON:
            case TAG_FONT:
            case TAG_BASEFONT:
                // Implicit <BODY> tag!
                {
                    if(bHeadClosed_ && (eNewTag == TAG_SCRIPT || eNewTag == TAG_OBJECT)
                        || (eNewTag != TAG_SCRIPT && eNewTag != TAG_OBJECT))
                    {
                        CSymbolAttributes<U> bodyAttributes;
                        bodyAttributes.eTag_ = TAG_BODY;
                        ProcessOpeningTag(bodyAttributes);
                    }
                }
                break;
            case TAG_BODY:
            case TAG_FRAMESET:
                bHasBodyTag_ = TRUE;
                break;
        }
    }
    if(bInTable_)
    {
        if(eNewTag == TAG_THEAD || eNewTag == TAG_TFOOT)
            bTableHeadOrFoot_ = TRUE;
        else if(eNewTag == TAG_TR && !bTableHeadOrFoot_)
        {
            if(!bHasTbody_)
            {
                // Implicit <TBODY> tag!
                CSymbolAttributes<U> tbodyAttributes;
                tbodyAttributes.eTag_ = TAG_TBODY;
                ProcessOpeningTag(tbodyAttributes);
                bHasTbody_ = TRUE;
            }
        }
        else if(eNewTag == TAG_TBODY)
            bHasTbody_ = TRUE;
    }

    if(eNewTag == TAG_TABLE)
        bInTable_ = TRUE;

    int iChild = -1;
    TagHierarchy& th = currentAddress_.tagHierarchy_;
    if(th.size() != 0)
    {
        // Get the current tag.
        Tags eTag = th.back().eTag_;

        // Get the current tag's behavior.
        TagCouple tc = {eTag, eNewTag};
        TagToBehavior::const_iterator it = tagToBehavior.find(tc);
        TagBehavior tagBeh;
        if(it == tagToBehavior.end())
            tagBeh = tagToDefaultBehavior[eTag];
        else
            tagBeh = it->second;

        // Update the current address accordingly.
        if(tagBeh == TAG_CLOSES)
        {
            // The current tag closes. It must be popped from the current address,
            // and the new opening tag must be pushed into the address.

            // Derive child number from sibling.
            iChild = th.back().iChild_;

            // Pop the current tag from the address.
            if(eTag == TAG_HEAD)
                bHeadClosed_ = TRUE;
            else if(eTag == TAG_TABLE)
                bInTable_ = bHasTbody_ = FALSE;
            else if(eTag == TAG_THEAD || eTag == TAG_TFOOT)
                bTableHeadOrFoot_ = FALSE;
            else if(eTag == TAG_TBODY)
                bHasTbody_ = FALSE;
            th.pop_back();

            if(bProduceTree_)
                pCurrentTag_ = pCurrentTag_->pParent_;
        }
        else
        {
            // The current tag creates a child.
            // Derive child number from address level child number.
            iChild = currentAddress_.iCurChild_;
        }

        // Clear the child and attribute info.
        currentAddress_.iCurChild_ = -1;
        currentAddress_.iCurChildAttribute_ = -1;
        currentAddress_.strAtt_.erase();
    }
    else
    {
        // Root node!
    }

    // Push the new tag onto the current address stack.
    th.push_back(AddressNode(++iChild, eNewTag));

    if(bProduceTree_)
    {
        CTag* pNewTag = new CTag;
        pNewTag->eTagType_ = eNewTag;
        pNewTag->strLabel_ = tagToName[eNewTag];
        pNewTag->pParent_  = pCurrentTag_;
        if(pCurrentTag_ == NULL)
        {
            // Make the root.
            pRoot_ = pCurrentTag_ = pNewTag;
        }
        else
        {
            pCurrentTag_->sons_.push_back(pNewTag);
            pCurrentTag_ = pNewTag;
        }
    }

#ifdef _DEBUG
#ifdef PHTRACE
    os_ << "Current address: ";
    currentAddress_.Dump(os_);
    os_ << endl;
#endif
#endif
}

template <typename U>
inline
void PhParser<U>::ProcessClosingTag(CSymbolAttributes<U>& tokenAttributes)
{
    // Explicit tag close.
    // Look for the matching opening tag, and if it exists, then close all
    // the tags contained in it.
    // If it was not found, then ignore the current closing tag.
    Tags eTag = tokenAttributes.eTag_;

    // Ignore unknown closing tags!!
    if(eTag == TAG_NONE)
    {
        bDisable_ = TRUE;
        return;
    }
    else
        bDisable_ = FALSE;

    TagHierarchy& th = currentAddress_.tagHierarchy_;
    int iAddrSize = th.size();
    BOOL bCloseHead = FALSE;
    for(int i = iAddrSize - 1; i >= 0; i--)
    {
        Tags eTmpTag = th[i].eTag_;
        if(eTmpTag == TAG_HEAD)
            bCloseHead = TRUE;
        else if(eTmpTag == TAG_TABLE)
            bInTable_ = bHasTbody_ = FALSE;
        else if(eTmpTag == TAG_THEAD || eTag == TAG_TFOOT)
            bTableHeadOrFoot_ = FALSE;
        else if(eTmpTag == TAG_TBODY)
            bHasTbody_ = FALSE;
        if(eTmpTag == eTag) // Found matching opening tag!
            break;
    }
    if(i >= 0)
    {
        // Set the address level child number to that of the last tag to be popped,
        // so that the next possible opening tag, sibling of this one, can take the
        // following child number.
        currentAddress_.iCurChild_ = th[i].iChild_;

        if(bProduceTree_)
        {
            for(int j = iAddrSize - 1; j >= i; j--)
                pCurrentTag_ = pCurrentTag_->pParent_;
        }

        // Pop up to and including the matching opening tag.
        if(bCloseHead)
            bHeadClosed_ = TRUE;
        th.erase(th.begin() + i, th.end());
    }

#ifdef _DEBUG
#ifdef PHTRACE
    os_ << "Current address: ";
    currentAddress_.Dump(os_);
    os_ << endl;
#endif
#endif
}

template <typename U>
inline
BOOL PhParser<U>::ProcessAttributeValue(
    AddressValueList& avl,
    CSymbolAttributes<U>& attrValue)
{
    BOOL bRes = FALSE;

    // If inside an unknown tag, ingnore.
    // We do not process attributes with an empty address.
    if(!bDisable_ && !currentAddress_.tagHierarchy_.empty())
    {
        currentAddress_.addrType_ = Address::ADDRESS_ANONYMOUS_ATTRIBUTE;
        currentAddress_.iCurChildAttribute_++;
        bRes = AddressCheck(avl, attrValue);

        if(bProduceTree_)
        {
            CTag::CAttribute att;
            att.strAddress_ = currentAddress_.ToString();
            att.strText_    = MakeString(attrValue.label_.pcBegin_, attrValue.label_.ulSize_);
            pCurrentTag_->attributes_.push_back(att);
        }
    }

    return bRes;
}

template <typename U>
inline
BOOL PhParser<U>::ProcessAttributeValue(
    AddressValueList& avl,
    CSymbolAttributes<U>& attrName,
    CSymbolAttributes<U>& attrValue)
{
    BOOL bRes = FALSE;

    // If inside an unknown tag, ingnore.
    // We do not process attributes with an empty address.
    if(!bDisable_ && !currentAddress_.tagHierarchy_.empty())
    {
        currentAddress_.addrType_ = Address::ADDRESS_ATTRIBUTE;
        currentAddress_.iCurChildAttribute_++;
        currentAddress_.strAtt_ = MakeString(attrName.label_.pcBegin_, attrName.label_.ulSize_);
        bRes = AddressCheck(avl, attrValue);

        if(bProduceTree_)
        {
            CTag::CAttribute att;
            att.strAddress_ = currentAddress_.ToString();
            att.strText_    = MakeString(attrValue.label_.pcBegin_, attrValue.label_.ulSize_);
            att.strLabel_   = MakeString(attrName.label_.pcBegin_, attrName.label_.ulSize_);
            pCurrentTag_->attributes_.push_back(att);
        }
    }

    return bRes;
}

template <typename U>
inline
BOOL PhParser<U>::ProcessText(AddressValueList& avl, CSymbolAttributes<U>& text)
{
    BOOL bRes = FALSE;

    // We do not process texts with an empty address.
    if(!currentAddress_.tagHierarchy_.empty())
    {
        currentAddress_.addrType_ = Address::ADDRESS_TEXT;
        currentAddress_.iCurChild_++;
        currentAddress_.strAtt_.erase();
        bRes = AddressCheck(avl, text);

        if(bProduceTree_)
        {
            CTag::CNode node;
            node.strAddress_ = currentAddress_.ToString();
            node.strText_    = MakeString(text.label_.pcBegin_, text.label_.ulSize_);
            pCurrentTag_->texts_.push_back(node);
        }
    }

    return bRes;
}

template <typename U>
inline
BOOL PhParser<U>::ProcessScript(AddressValueList& avl, CSymbolAttributes<U>& script)
{
    BOOL bRes = FALSE;

    // We do not process scripts with an empty address.
    if(!currentAddress_.tagHierarchy_.empty())
    {
        if(script.label_.pcBegin_ != NULL)
        {
            currentAddress_.addrType_ = Address::ADDRESS_SCRIPT;
            currentAddress_.iCurChild_++;
            currentAddress_.strAtt_.erase();
            bRes = AddressCheck(avl, script);

            if(bProduceTree_)
            {
                CTag::CNode node;
                node.strAddress_ = currentAddress_.ToString();
                node.strText_    = MakeString(script.label_.pcBegin_, script.label_.ulSize_);
                pCurrentTag_->scripts_.push_back(node);
            }
        }
    }

    return bRes;
}

template <typename U>
inline
BOOL PhParser<U>::ProcessComment(AddressValueList& avl, CSymbolAttributes<U>& comment)
{
    BOOL bRes = FALSE;

    // We do not process comments with an empty address.
    if(!currentAddress_.tagHierarchy_.empty())
    {
        currentAddress_.addrType_ = Address::ADDRESS_COMMENT;
        currentAddress_.iCurChild_++;
        currentAddress_.strAtt_.erase();
        bRes = AddressCheck(avl, comment);

        if(bProduceTree_)
        {
            CTag::CNode node;
            node.strAddress_ = currentAddress_.ToString();
            node.strText_    = MakeString(comment.label_.pcBegin_, comment.label_.ulSize_);
            pCurrentTag_->comments_.push_back(node);
        }
    }

    return bRes;
}

template <typename U>
inline
BOOL PhParser<U>::AddressCheck(AddressValueList& avl, CSymbolAttributes<U>& valueAttributes)
{
    // Loop on all addresses of the input address list, and check for a match.
    // Resolve if match.
    // If all addresses are resolved, stop parsing and return success code.

#ifdef _DEBUG
#ifdef PHTRACE
    os_ << "Current address: ";
    currentAddress_.Dump(os_);
    os_ << endl;
#endif
#endif

    BOOL bAllAddressesResolved = TRUE;
    for(AddressValueList::iterator it = avl.begin(); it != avl.end(); it++)
    {
        if(it->pcBegin_ == NULL)    // Unresolved address.
        {
#ifdef _DEBUG
#ifdef PHTRACE
    os_ << "Input address: ";
    it->pAq_->address_.Dump(os_);
    os_ << endl;
#endif
#endif
            if(it->pAq_->address_ == currentAddress_)
            {
                //// Found a match for an unresolved address: resolve it!
                BOOL bOK = TRUE;

                // Process the offset.
                const AddressQuery::ob* pOb = &it->pAq_->offset_;
                U pcBegin = valueAttributes.label_.pcBegin_;
                U pcEnd   = pcBegin + valueAttributes.label_.ulSize_;
                unsigned long ulLength = valueAttributes.label_.ulSize_;
                if(pOb->rep != AddressQuery::IS_UNUSED)
                {
                    if(pOb->rep == AddressQuery::IS_NUMBER)
                    {
                        pcBegin += pOb->iNumber;
                        ulLength -= pOb->iNumber;
                        if(!pOb->bIncluded)
                        {
                            ++pcBegin;
                            --ulLength;
                        }
                    }
                    else
                    {
                        // Find the offset string.
                        U pcAfterOffset;
                        pcBegin = FindString(pcBegin, pcEnd, ulLength,
                            pOb->str.data(), pOb->str.size(),
                            pOb->bCaseSensitive, pcAfterOffset);
                        if(pcBegin == NULL)
                            bOK = FALSE;
                        else
                        {
                            if(!pOb->bIncluded)
                                pcBegin = pcAfterOffset;
                            ulLength = pcEnd - pcBegin;
                        }
                    }
                }

                // pcBegin now points to the logical beginning of the HTML element's string,
                // i.e., the zero position. And ulLength is the maximum length of the result.
                // Process the first boundary.
                // The goal is to adjust pcBegin to point to the beginning of the result.
                if(bOK)
                {
                    pOb = &it->pAq_->begin_;
                    if(pOb->rep != AddressQuery::IS_UNUSED)
                    {
                        if(pOb->rep == AddressQuery::IS_NUMBER)
                        {
                            pcBegin += pOb->iNumber;
                            ulLength -= pOb->iNumber;
                            if(!pOb->bIncluded)
                            {
                                ++pcBegin;
                                --ulLength;
                            }
                        }
                        else
                        {
                            // Find the first boundary string.
                            U pcAfterBoundary;
                            pcBegin = FindString(pcBegin, pcEnd, ulLength,
                                pOb->str.data(), pOb->str.size(),
                                pOb->bCaseSensitive, pcAfterBoundary);
                            if(pcBegin == NULL)
                                bOK = FALSE;
                            else
                            {
                                if(!pOb->bIncluded)
                                    pcBegin = pcAfterBoundary;
                                ulLength = pcEnd - pcBegin;
                            }
                        }
                    }
                }

                // pcBegin now points to the actual beginning of the result, and
                // ulLength is the maximum length of the result.
                // Process the second boundary.
                // The goal is to determine the length of the result, and update
                // ulLength accordingly.
                if(bOK)
                {
                    pOb = &it->pAq_->end_;
                    if(pOb->rep != AddressQuery::IS_UNUSED)
                    {
                        if(pOb->rep == AddressQuery::IS_NUMBER)
                        {
                            if(pOb->iNumber > 0 && (unsigned long)pOb->iNumber < ulLength)
                                ulLength = pOb->iNumber;

                            // Exclusion is irrelevant.
                        }
                        else
                        {
                            // Find the second boundary string.
                            U pcAfterBoundary;
                            U pcBound = FindString(pcBegin, pcEnd, ulLength,
                                pOb->str.data(), pOb->str.size(),
                                pOb->bCaseSensitive, pcAfterBoundary);
                            if(pcBound != NULL)
                            {
                                if(pOb->bIncluded)
                                    ulLength = pcAfterBoundary - pcBegin;
                                else
                                    ulLength = pcBound - pcBegin;
                            }

                            // If second boundary not found, expand to the end of string.
                        }
                    }

                    // Now, extract the result!
                    it->pcBegin_ = pcBegin;
                    it->ulSize_  = ulLength;
                }
                else
                    bAllAddressesResolved = FALSE;
            }
            else
                bAllAddressesResolved = FALSE;
        }
    }

    return bAllAddressesResolved && bStopWhenAllAddressesResolved_;
}
