/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/

//==================================================================< Pragmas >
#ifdef  WIN32
#  pragma warning(disable : 4710)
#endif  // WIN32


//=================================================================< Includes >
#ifndef __ADMINISTRATION_LISTEN_CLIENT__
#  include "adminlc.hpp"
#endif  // __ADMINISTRATION_LISTEN_CLIENT__


//====================================================================< Class >

///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////// Class CGAdministrationListen //
///////////////////////////////////////////////////////////////////////////////

//============================================< Constructor(s) and destructor >
//-----------------------------------------------------------------------------
// CGAdministrationListen::CGAdministrationListen
//
// Constructor.
//
// Arguments : task : task associeted.
// Return    : none.
//-----------------------------------------------------------------------------
CGAdministrationListen::CGAdministrationListen (CGTask& task)
        : task_ (task)
{
    // Set the task data.
    task_.setData (this) ;

    // Save this pointer.
    CTCurrentThread::setData (this) ;

    // Set the abort routine.
    CTCurrentThread::setAbortRoutine (abort) ;
} // CGAdministrationListen::CGAdministrationListen

//-----------------------------------------------------------------------------
// CGAdministrationListen::~CGAdministrationListen
//
// Destructor.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
CGAdministrationListen::~CGAdministrationListen ()
{
    if (GpProcessMain->getTraceMode () )
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        GpProcessMain->getLogFileList () << task_.getTaskName () << ": listener <" << listen_.getInetAddress ().getHostAddress () << ":" << listen_.getPort () << "> closed." << endl ;
    }
    else
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        GpProcessMain->getLogFileList () << "Close listener on <" << listen_.getInetAddress ().getHostAddress () << ":" << listen_.getPort () << ">." << endl ;
    }

    // Close the listen socket.
    listen_.close () ;
} // CGAdministrationListen::~CGAdministrationListen

//=========================================================< Control methodes >
void CGAdministrationListen::init (void)
{
    if (GpProcessMain->getTraceMode () )
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        GpProcessMain->getLogFileList () << task_.getTaskName () << ": try to start listener on port " << *( (string*) task_.getInitInfo () ) << "." << endl ;
    }

    port_t  tAdminPort ;

    // Get the administration port.
    sscanf ( ( (string*) (task_.getInitInfo () ) )->data (), "%hu", &tAdminPort) ;

    // Specify if reuse of address is allow.
    listen_.reuseAddress (GpProcessMain->canReuseAddress () ) ;

    // Binding the listen socket to the administration port.
    listen_.bindToPort (tAdminPort) ;

    // Mark the listen socket as a listen socket.
    listen_.listen () ;

    if (GpProcessMain->getTraceMode () )
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        GpProcessMain->getLogFileList () << task_.getTaskName () << ": <gateway listener      <- administration client>" << endl ;
        char szAddress [22] ;
        char szLine [80] ;
        sprintf (szAddress, "%s:%hu", listen_.getInetAddress ().getHostAddress ().c_str (), listen_.getPort () ) ;
        sprintf (szLine, "<%-21.21s <- aaa.bbb.ccc.ddd:xxxxx>", szAddress) ;
        GpProcessMain->getLogFileList () << task_.getTaskName () << ": " << szLine << endl ;
    }
    else
    {
        CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
        GpProcessMain->getLogFileList () << "Listening on <" << listen_.getInetAddress ().getHostAddress () << ":" << listen_.getPort () << ">." << endl ;
    }

    // Tell the control task that initialization is terminated.
    task_.initializationTerminated () ;
} // CGAdministrationListen::init

//-----------------------------------------------------------------------------
// CGAdministrationListen::listenForConnections
//
// Listen for the administration clients to connect.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGAdministrationListen::listenForConnections (void)
{
    try
    {
        CTSocket::CTSelection event ;

        do
        {
            if (task_.hasToBeReinitialized () )
            {
                task_.reinitialized () ;
            }

            // Listen for the administration clients to connect.
            event = listen_.poll (_S_RECEIVE_ | _S_EXCEPTION_) ;
            switch (event)
            {
            case CTSocket::RECEIVE :
                // A client wants to connect.
                newClient () ;
                break ;
            case CTSocket::TIMEDOUT :
                // No client is waiting.
                break ;
            case CTSocket::EXCEPTION :
                // Error on socket listen_.
                throw CTSocketException ("listenForConnections: Exception event on socket listen_.", __FILE__, __LINE__) ;
                break ;
            default :
                // Unexpected event on socket listen_.
                throw CTSocketException ("listenForConnections: Unexpected event on socket listen_.", __FILE__, __LINE__) ;
            }
            CTCurrentThread::sleep (1) ;
        }
        while ( (event == CTSocket::TIMEDOUT) || (event == CTSocket::RECEIVE) ) ;
    }
    catch (CTSocketException& e)
    {
#ifdef  WIN32
        if (e.getNumber () != WSAEINTR)
#endif  // WIN32
#ifdef  UNIX
            if (e.getNumber () != EINTR)
#endif  // UNIX
            {
                throw e ;
            }
    }
} // CGAdministrationListen::listenForConnections

//-----------------------------------------------------------------------------
// CGAdministrationListen::newClient
//
// Accept a new client connection.
//
// Arguments : none.
// Return    : none.
//-----------------------------------------------------------------------------
void CGAdministrationListen::newClient (void)
{
    try
    {
        // Create an administration client task.
        task_.getTasksList ().newTask (CGTask::TASK_ADMINISTRATION_CLIENT, (void*) &listen_, task_) ;
    }
    catch (CGTaskException&)
    {}

} // CGAdministrationListen::newClient

//-----------------------------------------------------------------------------
// CGAdministrationListen::abort
//
// Abort Routine.
//
// Arguments : pStatus : pointer to the exit status.
// Return    : none.
//-----------------------------------------------------------------------------
void CGAdministrationListen::abort (void*)
{
    // Delete the CGAdministrationListen pointer.
    ( (CGAdministrationListen*) CTCurrentThread::getData () )->~CGAdministrationListen () ;
} // CGAdministrationListen::abort

///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////// End of Class CGAdministrationListen //
///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////// Globals Functions //
///////////////////////////////////////////////////////////////////////////////

void administrationListenTask (CGTask& task)
{
    try
    {
        if (GpProcessMain->getTraceMode () )
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": new administration listen task." << endl ;
        }

        // Destruction of this object must occur before killing the task.
        CGAdministrationListen  administrationListen (task) ;
        administrationListen.init () ;
        administrationListen.listenForConnections () ;

        if (GpProcessMain->getTraceMode () )
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": administration listen task terminated." << endl ;
        }
    }
    catch (CTException& e)
    {
        if (GpProcessMain->getTraceMode () )
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": !! " << e.what () << endl ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": administration listen task terminated." << endl ;
        }
        else
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << "!! " << e.what () << endl ;
        }
    }
    catch (...)
    {
        if (GpProcessMain->getTraceMode () )
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": !! unknown exception." << endl ;
            GpProcessMain->getLogFileList () << task.getTaskName () << ": administration listen task terminated." << endl ;
        }
        else
        {
            CTLockGuard<CTMutex>    guard (GpProcessMain->getLogFileList ().getMutex () ) ;
            GpProcessMain->getLogFileList () << "!! unknown exception." << endl ;
        }
    }
    if (task.initializationIsTerminated () )
    {   // Remove the task from the list.
        task.invalidate () ;
        for (int iIndex=0; iIndex < (int)task.getTasksList ().getTasksList ().size(); iIndex++)
        {
            if (task.getTasksList ().getTasksList ()[iIndex] == &task)
            {
                task.getTasksList ().killTask (iIndex) ;
                break ;
            }
        }
    }
    else
    {   // Tell the control task that initialization is terminated.
        task.initializationTerminated (false) ;
    }
} // administrationListenTask


///////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////// End of Globals Functions //
///////////////////////////////////////////////////////////////////////////////
