/* Copyright (C) 2000 Cyrano SA
 
   This file is part of OpenSTA
 
   OpenSTA is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free
   Software Foundation; either version 2, or (at your option) any later
   version.
 
   OpenSTA is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   for more details.
 
   You should have received a copy of the GNU General Public License along
   with OpenSTA; see the file COPYING.  If not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
 
   For more information about the GNU Public license see:
   http://www.gnu.org/copyleft/gpl.html
 
   OpenSTA is at:
   http://www.opensta.org/
   http://opensta.sourceforge.net/
*/
#include "TestExecuter_i.h"
#include <CECInetInfo.h>
#include <mutex.hpp>
#include "process.h"
#include <iostream>
#include "CECSimpleParse.h"
#include "NTPerfMain.h"
#include "CECTestOutput.h"
#include "ctime.hpp"
#include "ntp.h"
#include "OnExecuterStop.hpp"
#include "OstaTraceDef.h"
using namespace std;

extern executerEventList gExecuterEventList ;
extern CTMutex	gEventListMutex ;
extern CTCondition eventStopCondition ;

void AllQueryThread(TestExecuterImpl* pTE)
{
	LOG_MESSAGE(TRCCAT_STARTUP,TRCLEV_LOW,"Query thread started");
	pTE->RunAllQueries();
}

void RunThread(TestExecuterImpl* args)
{
	TestExecuterImpl* pTE=reinterpret_cast<TestExecuterImpl*>(args);

	TaskExecInfo task = pTE->m_execInf.seqTasks[0];

	string strProfile = task.strScriptName;
	strProfile += ".NTP";
	if (!pTE->m_pControl->GetProfileFile(strProfile.c_str()))
	{

	}

	// Parse profile
	string strPath = getenv("OPENSTA_HOME");
	strPath+="Engines\\Temp\\" + strProfile;
	if (!pTE->ParseProfile(strPath))
	{
		LOG_MESSAGE(TRCCAT_STARTUP,TRCLEV_LOW,"Profile does not contain any valid queries. Stopping...");
		pTE->StopTaskGroup(pTE->m_execInf.lTaskGroupId);
		return;
	}

	// Write Headers
	LOG_MESSAGE(TRCCAT_STARTUP,TRCLEV_LOW,"Writing Query Headers");
	for (unsigned int i=0; i<pTE->m_queries.size(); i++)
	{

	    CECNTPHeader header(pTE->m_offset, pTE->m_execInf.lTaskGroupId, i, pTE->m_queries[i].strHeading.c_str() );
		pTE->m_pControl->AddTestOutput(CECNTPHeader::OUTPUT_TYPE, header.ToString().c_str());
	}

	pTE->m_bRunning = true;

	// One thread for all the queries
	CTThread thr (makeFunctorA1(AllQueryThread, TestExecuterImpl*, pTE)) ;

	if (pTE->m_execInf.bIsDuration)
	{
		long nDur = pTE->m_execInf.lTaskGroupDuration;
        char cval[32];
        sprintf(cval,"%d",nDur);
		LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Setting duration = "+string(cval));
		
		while (pTE->m_bRunning)
		{
			if (nDur == 0)
			{
				LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Timed duration complete. Stopping TaskGroup.");
				pTE->StopTaskGroup(pTE->m_execInf.lTaskGroupId);
				pTE->m_bRunning = false;
			}
			else
			{
				::Sleep(1000);
				nDur--;
			}
		}
	}
}



TestExecuterImpl::TestExecuterImpl(CORBAServer* corba, string strTestRunString, string strTestName, string strReposHost)
{
    m_corba=corba;
	m_bRunning = false;
	m_nQueriesCompleted = 0;

	LOG_MESSAGE(TRCCAT_STARTUP,TRCLEV_LOW,"Registering TestExecuterImpl");

	// Register the interface with the name servive
    char strReg[64];
    CECInetInfo inet;
    sprintf(strReg, "%s_%8.8lX", inet.GetIPAddressAsString().c_str(), getpid());

    m_registered_name=strReg;
    m_corba->ActivateObject(this);    
    m_corba->NameObject( this->_this(),TestExecuter::CLASS_NAME,m_registered_name.c_str());

	LOG_MESSAGE(TRCCAT_STARTUP,TRCLEV_LOW,"Connecting to InjectorControl");
	CORBA::Object_ptr obj = m_corba->FindObjectWithName(InjectorControl::CLASS_NAME, inet.GetIPAddressAsString().c_str());
	m_pControl = InjectorControl::_narrow(obj);

	LOG_MESSAGE(TRCCAT_STARTUP,TRCLEV_LOW,"Finding time offset");
	m_offset = m_pControl->GetTimeOffset();

	CTTime tm;
	tm.now();
	m_nStartTime = tm.getSeconds();

    char strOffset[24];
    sprintf(strOffset, "Time Offset = %d", m_offset);
	LOG_MESSAGE(TRCCAT_STARTUP,TRCLEV_LOW,string(strOffset));
}

TestExecuterImpl::~TestExecuterImpl()
{
}

void TestExecuterImpl::Shutdown()
{
	m_bRunning = false;
	if (m_pRunThread)
	{
		CTCurrentThread::join(*m_pRunThread);
		delete m_pRunThread;
		m_pRunThread = NULL;
	}
	// De-register the interface
    m_corba->UnNameObject( TestExecuter::CLASS_NAME,m_registered_name.c_str());
}


CORBA::Boolean TestExecuterImpl::StartTaskGroup(TaskGroupExecInfo& tgExec)
{
    char strTG[4];
    sprintf(strTG, "%d", tgExec.lTaskGroupId);

	LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Starting Task Group "+string(strTG));
	m_execInf = tgExec;
	m_pRunThread = new CTThread(makeFunctorA1(RunThread, TestExecuterImpl*, this));
	LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Task Group "+string(strTG)+" is running");
	return true;
}

CORBA::Boolean TestExecuterImpl::StopTaskGroup(long lTaskGroupId)
{
    char strTG[4];
    sprintf(strTG, "%d", lTaskGroupId);
	LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Stopping Task Group "+string(strTG));

	CTLockGuard<CTMutex> guard(gEventListMutex) ;
	CExecuterEvent ev(EEE_ONSTOPTASKGROUP, lTaskGroupId) ;
	gExecuterEventList.push_back(ev);
	LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"All task groups complete... stopping.");
	CExecuterEvent ev1(EEE_ONSTOPEXECUTER) ;
	CExecuterEvent ev2(EEE_FINISHED) ;
	gExecuterEventList.push_back(ev1);
	gExecuterEventList.push_back(ev2);

	LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Sending STOP EVENT for "+string(strTG));
	eventStopCondition.signal();



	return true;
}

CORBA::Boolean TestExecuterImpl::Stop()
{
	LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Stop invoked on Test Executer");

	CTLockGuard<CTMutex> guard(gEventListMutex) ;
	CExecuterEvent ev1(EEE_ONSTOPEXECUTER) ;
	CExecuterEvent ev2(EEE_FINISHED) ;
	gExecuterEventList.push_back(ev1);
	gExecuterEventList.push_back(ev2);

	LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Send Stop Signal");
	eventStopCondition.signal();

	return true;
}

CORBA::Boolean TestExecuterImpl::GetStatsSnapshot(TestExecuterSnapshot& teSnapshot)
{
	teSnapshot.strTestExecuter = m_registered_name.c_str();
	teSnapshot.lNumberOfActiveVUs = 0;

	TaskSnap task;
	task.lActiveUsers = 0;
	task.lTaskId = 1;

	CTTime tm;
	tm.now();

	TaskGroupSnap tgSnap;
	tgSnap.lRunTime = tm.getSeconds() - m_nStartTime;
	tgSnap.lTaskGroupId = m_execInf.lTaskGroupId;
	tgSnap.lCompletedIterations = m_nQueriesCompleted;
    tgSnap.lActiveUsers = 0;
	tgSnap.seqTaskSnaps.length(1);
	tgSnap.seqTaskSnaps[0] = task;
	tgSnap.dBytes = 0.;
	tgSnap.lRequests = 0;


	teSnapshot.seqTaskGroupSnaps.length(1);
	teSnapshot.seqTaskGroupSnaps[0] = tgSnap;

	return true;
}

bool TestExecuterImpl::ParseProfile(string strFilePath)
{
	LOG_MESSAGE(TRCCAT_STARTUP,TRCLEV_LOW,"Parsing Profile: "+strFilePath);
	ifstream ifs(strFilePath.c_str());
	int nIndex = 0;
	while (ifs.good() && !ifs.eof())
	{
		string strline;
		getline(ifs, strline);
		CECSimpleParse parse(strline);
		
		QueryInfo query;
		query.nIndex = nIndex;
		query.strHeading = parse.getNext(',');
		query.strQuery = parse.getNext(',');
		string strtmp = parse.getNext(',');
		if (atoi(strtmp.c_str()) == 1)
		{
			query.bDelta = true;
		}
		else
		{
			query.bDelta = false;
		}
		strtmp = parse.getNext(',');
		query.nFreq = atoi(strtmp.c_str());
		query.iNbSleep = 0 ;
		query.bInitOK = false;
		// Next field is if the query is enabled
		strtmp = parse.getNext(',');
		if (atoi(strtmp.c_str()) == 1)
		{
			// before adding the query to the vector validate the fields
			if (query.strHeading.size() > 0 && query.strQuery.size() > 0)
			{
				LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Adding query "+query.strHeading);
				m_queries.push_back(query);
				nIndex++;
			}
			else
			{
				if (query.strHeading.size() <= 0)
					LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Invalid query: The heading is empty.");
				if (query.strQuery.size() <= 0)
					LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Invalid query: The query string is empty.");
			}
		}
	}

	if (m_queries.size() > 0)
	{
	    char strQ[4];
	    sprintf(strQ, "%d", m_queries.size());
		LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,string(strQ)+" enabled queries found.");
		return true;
	}
	return false;
}

void TestExecuterImpl::RunAllQueries()
{
#ifdef _DEBUG
_asm { int 3 };   // ### L.M. LM TO REMOVED!!
#endif
	LOG_MESSAGE(TRCCAT_RUNNING,TRCLEV_LOW,"Starting query execution for queries.");
	OtaNTPerform ntp;
	size_t i=0; 

	for (i=0; i<m_queries.size(); i++)
	{
		m_queries[i].bInitOK = ntp.InitializeQuery (m_queries[i].strQuery) ;
	}
	
	long lResult = 0 ;
	long lLastVal = 0 ;
	while (m_bRunning)
	{

		CTTime tmStart;
		CTTime tmEnd;

		bool bExecuted = false;
		for (i=0; i<m_queries.size(); i++)
		{
			if (m_queries[i].bInitOK && m_queries[i].iNbSleep == m_queries[i].nFreq)
			{
				// If it is time for a request to get a value, we get the value
				// only once.
				// the other values (for the other requests) are get by the call
				if (!bExecuted)
				{
					tmStart.now();
					ntp.ExecQueries() ;
					tmEnd.now();
					bExecuted = true ;
				}
				if (m_queries[i].bDelta)
					lLastVal = ntp.GetLastVal(i);

				m_queries[i].iNbSleep = 0;
				lResult = 0 ;
				ntp.GetValue(i, lResult);

				if (m_queries[i].bDelta)
				{	
					// the first value for delta case must not be taken
					// into account
					if (lLastVal == -1)
						continue ;
					lResult = lResult - lLastVal;
				}
				CECNTPQuery result( m_offset, m_execInf.lTaskGroupId, m_queries[i].nIndex, tmStart.getSeconds(), tmEnd.getSeconds(), lResult );
				m_pControl->AddTestOutput(CECNTPQuery::OUTPUT_TYPE, result.ToString().c_str());
			}
		}

		::Sleep(1000);
		for (i=0; i<m_queries.size(); i++)
		{
			m_queries[i].iNbSleep++;
		}
	}
}
